/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup Beamsearch Beamsearch - A Module for incremental beam search
 * @author Niels Beuck
 * @date 2.8.10
 *
 * This module implements an incremental beam search.
 *
 *
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <string.h>
#include <blah.h>
#include "cdg.h"
#include "beamsearch.h"
#include "increment.h"
#include "constraintnet.h"
#include "hook.h"
#include "set.h"
#include "virtualnode.h"
#include "frobbing.h"
#include "input.h"

/* -- MACROS --------------------------------------------------------------- */

/* -- TYPE DEFINITIONS ----------------------------------------------------- */

/* -- VARIABLES ------------------------------------------------------------ */

int bsMethod = 0; // method for lexem selection, 0:= per frobbing, 1:= per frobbing but print out suitable rules, 2:= per rules
int bsBeamWidth = 10;
int bsBottomUpPredictionDepth =2;
float bsBeamToleranceFactor = 0.3;
int bsVerbosity = 3; // 0= no intermediate output, 1= step end output, 2= entry end output, 3= as 2 plus warnings 4= all output
Boolean bsLockAllLevels = TRUE;
Boolean bsUnlockUnusedVNs = FALSE; // used for frobbing based top down prediction, slow! alternative to bsTopDownPrediction
Boolean bsTopDownPrediction = TRUE; // will use top down predictors defined in the grammar

/* -- FUNCTIONS ------------------------------------------------------------ */


/* -- IMPLEMENTATION ------------------------------------------------------- */

/*
 * registers the set variables for this strategy
 */
void bsInitialize() {
	setRegister("bsBeamWidth",               SET_INT,   &bsBeamWidth,               NULL, NULL, NULL, NULL);
	setRegister("bsBottomUpPredictionDepth", SET_INT,   &bsBottomUpPredictionDepth, NULL, NULL, NULL, NULL);
	setRegister("bsBeamToleranceFactor",     SET_FLOAT, &bsBeamToleranceFactor,     NULL, NULL, NULL, NULL);
	setRegister("bsVerbosity", 	             SET_INT,   &bsVerbosity,               NULL, NULL, NULL, NULL);
	setRegister("bsLockAllLevels",           SET_BOOL,  &bsLockAllLevels,           NULL, NULL, NULL, NULL);
	setRegister("bsUnlockUnusedVNs",         SET_BOOL,  &bsUnlockUnusedVNs,         NULL, NULL, NULL, NULL);
	setRegister("bsTopDownPrediction",       SET_BOOL,  &bsTopDownPrediction,       NULL, NULL, NULL, NULL);
	setRegister("bsMethod",                  SET_INT,   &bsMethod,                  NULL, NULL, NULL, NULL);
}


/*
 * returns a beam entry for a given analysis
 *
 */
BeamEntry beFromAnalysis(Analysis a, BeamEntry previous, String method) {
	int i;
	LexemNode ln;

	BeamEntry newBe = (BeamEntry) memMalloc(sizeof(BeamEntryStruct));
	newBe->best = a;
	newBe->p = analysis2parse(a);
	newBe->p->searchStrategy = strRegister("BeamSearch");
	newBe->virtualNodeUsed = NULL;      // we add it later
	newBe->virtualNodeAvailable = NULL; // same here
	newBe->previous = previous;
	newBe->method = method;
	newBe->rank = 0; // initialize unranked
	newBe->slots = vectorClone(a->slots);

	newBe->lexemeSelection = vectorNew(vectorSize(a->P->lg->graphemnodes));
	// read selected lexemes from analysis
	for(i = 0; i < vectorSize(a->P->lg->nodes); i++) {
		if(bvElement(a->isUnselectedLexeme, i)) {
			continue;
		}
		ln = vectorElement(a->P->lg->nodes, i);
		vectorSetElement(newBe->lexemeSelection, ln, ln->arc->from);
	}

	return newBe;
}

/*
 * free the memory of the beam entry
 * including the analysis and the parse, if there is one
 */
void beDelete(BeamEntry be) {
	be->best = NULL;

	if(be->p) {
		parseDelete(be->p);
		be->p = NULL;
	}

	if(be->virtualNodeAvailable) {
		be->virtualNodeAvailable = NULL;
	}

	if(be->virtualNodeUsed) {
		be->virtualNodeUsed = NULL;
	}

	be->lexemeSelection = NULL;

	be->slots = NULL;

	// don't delete the previous BE, as we don't own it
	be->previous= NULL;

	be->method= NULL;
	be->rank=0;
	memFree(be);
}

/*
 * get the score of the first entry in the beam
 * make sure the entries are sorted
 *
 * if the beam is empty, return 0.0
 */
Number beamGetHighestScore(List beam) {
	if(!beam) {
		return 0.0;
	}

	BeamEntry be = beam->item;
	Badness b = be->best->b;

	if(b->hard)
		return 0.0;

	return b->soft;
}

/*
 * get the score of the last entry in the beam
 * make sure the entries are sorted
 *
 * if the beam is empty, return 0.0
 */
Number beamGetLowestScore(List beam, int beamWidth) {
	if(!beam || listSize(beam) < beamWidth) {
		return 0.0;
	}

	BeamEntry be = listLastElement(beam);
	Badness b = be->best->b;

	if(b->hard)
		return 0.0;

	return b->soft;
}


/* ----------------------------------------------------------------------
   Print a parse that has no words or lexemgraph set to stdout
 */
void printAbstractParse(Parse parse, LexemGraph lg) {
	int i, j;
	String label;
	Arc arc;
	int nrLevels = inputCurrentGrammar->noOfLevels;
	Level level;

	for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
		level = listNthElement(inputCurrentGrammar->levels, (i % nrLevels) + 1);
		if (!level || !level->showflag)
			continue;

		label = vectorElement(parse->verticesLabels, i);

		// skip uninteresting values of other levels and for virtual nodes
		if (!strcmp(label, "") &&
				(int) vectorElement(parse->verticesStructure, i) == -1 &&
				(level->no > 0 ||  (lg->ve && i/nrLevels >= lg->ve->vStartID) ) ) {
			continue;
		}

		if (lg->ve && i/nrLevels >= lg->ve->vStartID){
			cdgPrintf(CDG_DEFAULT, "   %03d %10s: %s --%s-->",
				i , level->id,
				"[virtual]", label);
		} else {
			arc = listNthElement(lg->lattice->arcs, i / nrLevels +1);

			cdgPrintf(CDG_DEFAULT, "   %03d %10s: %s --%s-->",
				arc->from * nrLevels + level->no, level->id,
				arc->word, label);
		}

		j = (int) vectorElement(parse->verticesStructure, i);
		if (j == -1) {
			cdgPrintf(CDG_DEFAULT, "NIL\n");
		} else if (j == -2) {
			cdgPrintf(CDG_DEFAULT, "NONSPEC\n");
		} else if (lg->ve && j >= lg->ve->vStartID){
			cdgPrintf(CDG_DEFAULT, "[virtual]\n");
		} else {
			arc = listNthElement(lg->lattice->arcs, j+1);
			cdgPrintf(CDG_DEFAULT, "%s\n", arc->word);
		}
	}

	cdgPrintf(CDG_DEFAULT, "\n");
}


/*
 * returns the best LevelValue from the given list,
 *
 */
LevelValue bestScoringLVfromList(List lvs) {
	List l= lvs;
	LevelValue best  = NULL;
	Number bestScore = 0.0;
	LevelValue lv;

	while (l) {
		lv= l->item;
		l=l->next;
		if(lv->score > bestScore) {
			best= lv;
			bestScore = lv->score;
		}
	}

	return best;
}

/**
 * returns a bytevector that says which lexemes are available in the given BeamEntry
 *
 * assumes:
 * - that only the first level is restrictive regarding lexemes
 * - that the net was already extended by one arc compared to the parse in the BeamEntry
 *
 */
ByteVector getAvailableLexemes(BeamEntry be, ConstraintNet net) {
	ByteVector ret, tmp;
	int i, j;
	int dep;
	int reg;
	String label;
	List lvs, l, l2;
	int noOfLevls= inputCurrentGrammar->noOfLevels;
	GraphemNode gn;
	LevelValue lv;
	Boolean b;

	ret= bvNew(bvSize(net->lexemgraph->isDeletedNode));
	bvSetAllElements(ret, TRUE);

	for(i=0; i < vectorSize(be->p->verticesStructure); i++ ) {
		if(i%noOfLevls)
			continue;

		dep= i/noOfLevls;
		reg= (int)vectorElement(be->p->verticesStructure, i);
		label= vectorElement(be->p->verticesLabels, i);

		// because we assume the net was already extended,
		// we have to do some stunts regarding virtual node indices
		if(dep >= net->lexemgraph->ve->vStartID -1) {
			dep+= 1;
		}
		if(reg >= net->lexemgraph->ve->vStartID -1) {
			reg+= 1;
		}

		// prepare a ByteVector for the local availability ...
		tmp = bvNew(bvSize(ret));
		bvSetAllElements(tmp,TRUE);

		// ... first mark all possible dependent lexemes as false,
		// so we can mark the actually occurring ones as true again
		gn = vectorElement(net->lexemgraph->graphemnodes, dep);
		l = gn->lexemes;
		while(l) {
			bvSetElement(tmp, FALSE, ((LexemNode)l->item)->no);
			l=l->next;
		}

		// same for the regents
		if(reg >= 0) {
			gn = vectorElement(net->lexemgraph->graphemnodes, reg);
			l = gn->lexemes;
			while(l) {
				bvSetElement(tmp, FALSE, ((LexemNode)l->item)->no);
				l=l->next;
			}
		}

		lvs= cnGetAccordingLVs(net,0,dep,label,reg);

		// now let's look at the actually occurring lexemes
		l = lvs;
		while(l) {
			lv = l->item;
			l2= lv->dependents;
			while(l2) {
				bvSetElement(tmp, TRUE, ((LexemNode)l2->item)->no);
				l2=l2->next;
			}

			l2= lv->regents;
			while(l2) {
				bvSetElement(tmp, TRUE, ((LexemNode)l2->item)->no);
				l2=l2->next;
			}

			l=l->next;
		}

		// ... and multiply it with the global vector
		for(j=0; j < bvSize(ret); j++) {
			b= bvElement(ret,j) && bvElement(tmp, j);
			bvSetElement(ret, b, j);
		}

		lvs = NULL;
	}


	return ret;
}

/*
 * returns TRUE, if at least one of the given level values has a score above the threshold
 */
Boolean compatibleLvExists(ConstraintNet net, int lvlid, int dep, String label, int reg, Number threshold, ByteVector lexemAvailable) {

	List lvs;
	LevelValue lv;
	List l, l2;

	lvs = cnGetAccordingLVs(net, lvlid, dep, label, reg);

	// for all according LVs
	l   = lvs;
	while (l) {
		lv= l->item;
		if( significantlyGreater(lv->score, threshold) ) {
			// score is OK, next question: are the lexemes compatible?
			if(reg >= 0) {
				l2= lv->regents;
				while(l2) {
					if(bvElement(lexemAvailable, ((LexemNode)l2->item)->no)) {
						lvs = NULL;
						return TRUE;
					}
					l2= l2->next;
				}
			} else {
				// regent is not not an actual graphemnode, so the is no issue of lexem compatibility
				lvs = NULL;
				return TRUE;
			}
		}
		l=l->next;
	}

	lvs = NULL;
	return FALSE;
}

/*
 * calculates a score estimation by multiplying the best score for every slot
 */
Number getScoreEstimation(Parse p, ConstraintNet net, Boolean virtOffset) {
	Number score = 1.0;
	int sid_i, wid_i;
	int wid_r;
	int noOfLevels = inputCurrentGrammar->noOfLevels;
	String label;
	int lvlno;
	List lvs;
	LevelValue lv;

	for(sid_i=0; sid_i< vectorSize(p->verticesStructure); sid_i++) {
		wid_i = sid_i / noOfLevels;
		wid_r = (int)vectorElement(p->verticesStructure, sid_i);
		label =      vectorElement(p->verticesLabels   , sid_i);
		lvlno = sid_i % noOfLevels;

		if(virtOffset && net->lexemgraph->ve && wid_i >= net->lexemgraph->ve->vStartID-1)
			wid_i +=1;

		if(virtOffset && net->lexemgraph->ve && wid_r >= net->lexemgraph->ve->vStartID-1)
			wid_r +=1;

		lvs   = cnGetAccordingLVs(net, lvlno, wid_i, label, wid_r);
		lv    = bestScoringLVfromList(lvs);

		lvs = NULL;
		if(lv) {
			score = score * lv->score;
		} else {
			return 0.0;
		}

	}

	return score;
}

/**
 * compares the given beamEntry with all entries in the given beams for whether they represent the same parse
 */
Boolean beClash(BeamEntry be, List beam) {
	List l;
	BeamEntry be2;

	for(l= beam; l; l=l->next) {
		be2 = l->item;
		if(parseSameVertices(be->p, be2->p)) {
			return TRUE;
		}
	}

	return FALSE;
}

/*
 * compares two beam entries by the score of their respective analysis
 */
Boolean beCompare(BeamEntry a, BeamEntry b) {
	return bCompare(a->best->b, b->best->b);
}

/*
 * sort the beam entries by score and return a list containing the best n,
 * where n is the given beam size
 *
 * if a factor > 0 is given, there is an additional restriction:
 * no entries may be worse than the best score times the given factor
 *
 * deletes the given beam
 */
List pruneBeam(List oldBeam, int beamSize, Number f) {
	int i;
	List l;
	List ret = listNew();
	Number threshold;
	BeamEntry be;

	l= listSort(oldBeam, beCompare);

	// calculate the threshold from the best score
	be= l->item;
	threshold= be->best->b->soft * f;

	i= 0;
	for( ; l && i< beamSize; l=l->next) {
		be= l->item;
		if(be->best->b->soft < threshold) {
			break;
		}
		if(!beClash(be, ret)) {
			ret= listAppendElement(ret, be);
			be->rank = i+1;
			i++;
		} else {
			beDelete(be);
		}
	}

	// now delete all remaining entries
	while(l) {
		beDelete(l->item);
		l= l->next;
	}

	return ret;
}

/*
 * undeletes all LVs with virtual nodes as dependent, that are unused in the given parse
 */
void unlockUnusedVNs(ConstraintNet net, Parse p) {
	LexemGraph lg = net->lexemgraph;
	int i, j, k;
	int reg;
	int noOfLevels = inputCurrentGrammar->noOfLevels;
	String label;
	LevelValue lv;
	Vector lvs;

	if(!lg->ve)  {
		// no VNs, no problem
		return;
	}

	for(i=lg->ve->vStartID; i < lg->max; i++) {
		reg  = (int)vectorElement(p->verticesStructure, i* noOfLevels);
		label=      vectorElement(p->verticesLabels   , i* noOfLevels);

		if(reg==-1 && strcmp(label,"")==0) {
			// is unused, reactivate all lvs
			for(k=0; k<noOfLevels; k++) {
				lvs =((ConstraintNode)vectorElement(net->nodes, i*noOfLevels +k))->values;
				for(j=0; j< vectorSize(lvs); j++) {
					lv = vectorElement(lvs, j);
					if(lv->score > 0.0) {
						lv->isDeleted = FALSE;
						lv->limit = lv->score;
					}
				}
			}
		}
	}

	for(i=0; i< vectorSize(net->nodes); i++) {
		incrUpdateNumberCaches(vectorElement(net->nodes, i));
	}
}

/**
 * Searches a suitable LV for the given dep --label--> reg and lvl combination,
 * and adds that to the analysis
 */
void attachInAnalysis(Analysis a, GraphemNode dep, String label, GraphemNode reg, int lvl) {
	LevelValue lv;
	int slot = dep->no * inputCurrentGrammar->noOfLevels + lvl;

	List lvs;

	for(lvs = a->P->domains[slot]; lvs; lvs = lvs->next) {
		lv= lvs->item;
		if(lvGetRegentGN(lv) != reg || strcmp(lv->label, label) != 0 || lv->isDeleted) {
			continue;
		} else {
			aInsert(a, lv);
			break;
		}
	}
}

/**
 * Searches all suitable LV for the given dep --label--> reg and lvl combination,
 * and returns the corresponding analyses
 */
List attachAllInAnalysis(Analysis a, GraphemNode dep, String label, GraphemNode reg, int lvl) {
	LevelValue lv;
	LexemNode curReg;
	int slot = dep->no * inputCurrentGrammar->noOfLevels + lvl;
	Analysis a2;
	List results = listNew();

	List l;
	lv =vectorElement(a->slots, reg->no * inputCurrentGrammar->noOfLevels + lvl );
	for(l = lv->dependents; l; l= l->next) {
		curReg = l->item;
		if(!bvElement(a->isUnselectedLexeme, curReg->no)) {
			break;
		}
	}

	List lvs;
	for(lvs = a->P->domains[slot]; lvs; lvs = lvs->next) {
		lv= lvs->item;
		if(lvGetRegentGN(lv) != reg || strcmp(lv->label, label) != 0 || lv->isDeleted || !listContains(lv->regents, curReg)) {
			continue;
		} else {
			a2 = aClone(a);
			aInsert(a2, lv);
			results = listAppendElement(results, a2);
		}
	}
	return results;
}

/**
 * returns a list of all variants of the given analysis, where dep is attached to reg
 * generates all label variants but only the best lexeme selection
 */
List generatePossibleAttachments(Analysis a, GraphemNode reg, GraphemNode dep, List labels, Parse origP, Number threshold) {
	List results = listNew();
	List l;
	String label;
	Analysis a2;
	ConstraintNet net = a->P->net;
	Problem P;
	Boolean aboveTh = FALSE;

	// for all possible labels
	for(l= labels;l ; l= l->next) {
		label= l->item;

		a2= aClone(a);

		if(bsMethod == 0 || bsMethod == 1) {
			// delete and undelete LVs to restrict the search to the lexeme variants
			cnResetToThreshold(net,threshold);
			if(bsLockAllLevels) {
				cnLockToParse(net, origP, NULL);
			} else {
				cnLockToParse(net, origP, inputCurrentGrammar->levels->item);
			}
			cnLockToNewAttachment(net, 0, dep, label, reg);

			//cdgPrintf(CDG_WARNING, ">>>>>>>>>>>>>>> Possible attachments: \n");
			//cnPrintPossibleAttachments(net, CDG_WARNING);

			// manipulate the analysis to contain according lv
			attachInAnalysis(a2, dep, label, reg, 0);

			// reuse the old problem, but reset the relevant parts
			P= a->P;
			P->current = a2;
			P->best = aClone(a2);
			P->globalMaximum = 1.0;
			P->globalMinimum = 0.0;
			P->unremovable = listNew();
			combined(P);

			a2= P->best;

			// evaluate the resulting analysis and add it to the list
			aboveTh= aJudgeWithThreshold(a2, threshold);
			if(aboveTh && !a2->b->hard && a2->b->soft >= threshold) {
				results = listAppendElement(results, a2);
			}
		} else { // no frobbing
			List a2s = attachAllInAnalysis(a2, dep, label, reg, 0);
			for(; a2s; a2s=a2s->next) {
				a2 = a2s->item;
				aboveTh= aJudgeWithThreshold(a2, threshold);
				if(aboveTh && !a2->b->hard && a2->b->soft >= threshold) {
					results = listAppendElement(results, a2);
				}
			}
		}
	}
	return results;
}

/*
 * returns a list of analyses
 * the given Analysis is guaranteed to be among them
 * then the analysis is checked for conflicts in the topDownPredictors list
 * and variants are generated accordingly
 */
List predictTopDown(Analysis a, Parse origP, Number threshold) {
	List l, m;
	TopDownPredictor tdp;
	ConstraintViolation cv;
	String s;
	GraphemNode gn, gn2;
	LevelValue lv;
	VirtualNodeSpec spec;
	VirtualLGExtension ve = a->P->lg->ve;

	List res = listAppendElement(listNew(), a);

	// look for conflicts that are td-predictors
	for(l=a->conflicts; l; l=l->next) {
		cv = l->item;
		tdp = inputGetTopDownPredictor(inputCurrentGrammar, cv->constraint->id);
		if(!tdp) {
			continue;
		}

		lv= NULL;
		if(tdp->edge == 1) {
			lv= cv->lv1;
		} else if(tdp->edge == 2) {
			lv= cv->lv2;
		} else if(tdp->edge == 3) {
			lv= cv->lv3;
		}

		if(!lv) {
			continue;
		}

		if(tdp->bottom) {
			gn = lvGetDependentGN(lv);
		} else {
			gn = lvGetRegentGN(lv);
		}

		// check all suitable specs
		for(m = tdp->specs; m; m=m->next) {
			s = m->item;
			spec= veGetSpecByName(ve, s);

			if(!spec) {
				cdgPrintf(CDG_WARNING, "WARNING: could not find vn-spec of name '%s' for tdp via '%s'!\n", s, tdp->constraintName);
				continue;
			}

			// findUnused vn for the spec
			gn2 = vnsGetUnusedInstance(spec, a, ve)->gn;
			if(!gn2) {
				continue;
			}

			// then attach that vn to the gn with all possible labels
			res= listAppendList(res,
					generatePossibleAttachments(a, gn, gn2, tdp->labels, origP, threshold));
		}
	}

	return res;
}

/*
 * returns a list with all LVs for the given slot, that are compatible with the given parse
 */
List getCompatibleLVsfromSlot(Parse p, int slot, ConstraintNet net, Number threshold) {
	ConstraintNode cnode = vectorElement(net->nodes, slot);
	Boolean unusedVirtual =
			cnode->level->no == 0 && // we only skip for the first level, the rest is done by lexeme compatibility
			((GraphemNode)vectorElement(net->lexemgraph->graphemnodes, slot/inputCurrentGrammar->noOfLevels))->isVirtual &&
			strcmp(vectorElement(p->verticesLabels, slot),"") ==0 &&
			(int)vectorElement(p->verticesStructure, slot) == -1;
	List ret = listNew();
	int i;
	for(i = 0; i < vectorSize(cnode->values); i++) {
		LevelValue lv = vectorElement(cnode->values, i);
		if(lv->score ==0 || lv->score < threshold) {
			continue;
		}
		if(pIsCompatibleLV(p, lv)) {
			ret= listAppendElement(ret, lv);
			if(unusedVirtual) {
				// we don't need more than one lv for unused virtual nodes
				break;
			}
		}
	}

	return ret;
}

/**
 * returns a vector with all those given LVs that are compatible with the given parse
 */
Vector filterLVsByParse(Parse p, Vector lvs) {
	int i;
	Vector ret = vectorNew(vectorSize(lvs));

	for(i= 0; i < vectorSize(lvs); i++) {
		LevelValue lv = vectorElement(lvs, i);
		if(pIsCompatibleLV(p, lv)) {
			vectorAddElement(ret, lv);
		}
	}

	return ret;
}

/**
 * returns whether the given LVs could possible be part of an analysis at the same time,
 * based on the lexemes they bind
 */
Boolean isLexemeCompatible(LevelValue lv, List lvs, LexemGraph lg) {
	List l, m;
	List lexemes;
	List tmp;
	LevelValue lv2;

	int dep = lvGetDependentGN(lv)->no;
	lexemes = listClone(lv->dependents);
	// for all given LVs ...
	for (l= lvs; l; l=l->next) {
		lv2 = l->item;
		if(lvGetDependentGN(lv2)->no == dep) {
			tmp = listNew();
			for(m = lexemes; m; m=m->next) {
				if(listContains(lv2->dependents, m->item)) {
					tmp = listAppendElement(tmp, m->item);
				}
			}
			if(!tmp) {
				// no lexemes left, the LVs are incompatible
				return FALSE;
			}
			lexemes = tmp;
		}
		if(lvGetRegentGN(lv2) && lvGetRegentGN(lv2)->no == dep) {
			tmp = listNew();
			for(m = lexemes; m; m=m->next) {
				if(listContains(lv2->regents, m->item)) {
					tmp = listAppendElement(tmp, m->item);
				}
			}
			if(!tmp) {
				// no lexemes left, the LVs are incompatible
				return FALSE;
			}
			lexemes = tmp;
		}
	}

	if(lvGetRegentGN(lv)) {
		int reg = lvGetRegentGN(lv)->no;
		lexemes = listClone(lv->regents);
		// for all given LVs ...
		for (l= lvs; l; l=l->next) {
			lv2 = l->item;
			if(lvGetDependentGN(lv2)->no == reg) {
				tmp = listNew();
				for(m = lexemes; m; m=m->next) {
					if(listContains(lv2->dependents, m->item)) {
						tmp = listAppendElement(tmp, m->item);
					}
				}
				if(!tmp) {
					// no lexemes left, the LVs are incompatible
					return FALSE;
				}
				lexemes = tmp;
			}
			if(lvGetRegentGN(lv2) && lvGetRegentGN(lv2)->no == reg) {
				tmp = listNew();
				for(m = lexemes; m; m=m->next) {
					if(listContains(lv2->regents, m->item)) {
						tmp = listAppendElement(tmp, m->item);
					}
				}
				if(!tmp) {
					// no lexemes left, the LVs are incompatible
					return FALSE;
				}
				lexemes = tmp;
			}
		}
	}

	return TRUE;
}


/*
 * Build an Analysis given a list of lvs
 * and without a problem
 */
Analysis aFromSlots(Problem P, Vector slots, ConstraintNet net) {
	Analysis a = aNew(P);
	int i;
	LevelValue lv;

	aBuildPath(a);
	for(i=0; i < vectorSize(slots); i++) {
		lv = vectorElement(slots, i);
		aInsert(a, lv);
	}
	return a;
}



/**
 * prints all constraint violations that are in 'a' but not in 'b'
 * and saves them in the profiling objects
 */
void removedCVs(List a, List b, ProfilingObject prof) {
	List l, m;
	ConstraintViolation cv1,cv2;
	Boolean found;

	cdgPrintf(CDG_WARNING, "Conflicts: %d vs %d\n", listSize(a), listSize(b));

	for (l= a; l; l= l->next) {
		cv1 = l->item;
		found = FALSE;
		for (m= b; m; m= m->next) {
			cv2 = m->item;
			if(cv1->constraint == cv2->constraint &&
					cv1->nodeBindingIndex1 == cv2->nodeBindingIndex1 &&
					cv1->nodeBindingIndex2 == cv2->nodeBindingIndex2 &&
					cv1->nodeBindingIndex3 == cv2->nodeBindingIndex3)
			{
				found = TRUE;
				break;
			}
		}

		if(!found) {
			// cdgPrintf(CDG_WARNING, "Rule candidate: %s\n", cv1->constraint->id);
			if(hashContainsKey(prof->contraintCount, cv1->constraint->id)) {
				hashSet(prof->contraintCount, cv1->constraint->id,
						((void*)((int)hashGet(prof->contraintCount, cv1->constraint->id)) +1));
			} else {
				hashSet(prof->contraintCount, cv1->constraint->id, (void *)1);
			}
		}
	}
}

/**
 * For a BeamEntry we need a score, for that we need an analysis and for that we need a lexeme selection
 * here the lexeme selection is done by frobbing
 *
 * Also see addParseToBeam
 */
List addParseToBeamByFrobbing(Parse p, int beamSize, Number f, List oldBeam,
		ConstraintNet net, BeamEntry oldBE, String method, ProfilingObject prof)
{
	List beam = oldBeam;
	BeamEntry newBE;
	Problem P;
	Number threshold;
	List results, l;
	Analysis a;
	String s;

	Timer timer = timerNew();
	ProfilingObject prof2 = proNew();

	threshold=  max(
			beamGetHighestScore(beam) * bsBeamToleranceFactor,
			beamGetLowestScore(beam, bsBeamWidth));

	if(bsVerbosity>=4) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		printAbstractParse(p, net->lexemgraph);
		cdgPrintf(CDG_INFO, "\n");
	}

	if(bsVerbosity < 4) {
		if(bsVerbosity < 3) {
			hkDeactivate(CDG_WARNING);
		}
		hkDeactivate(CDG_INFO);
	}
	proMeasureTime(timer, prof2, "  add prepare");

	// we have to limit the search space, only lexeme variations are allowed
	cnResetToThreshold(net,threshold);
	if(bsLockAllLevels) {
		cnLockToParse(net, p, NULL);
	} else {
		cnLockToParse(net, p, inputCurrentGrammar->levels->item);
	}
	if(bsUnlockUnusedVNs) {
		unlockUnusedVNs(net, p);
	}
	proMeasureTime(timer, prof2, "  add lock net");

	// To make an initial analysis from a parse, we would need a decorated path,
	// and for decorating a path we would need words with descriptions
	// instead we will not give a parse for now and hope for the locking to do the job
	//P = net2problem(net, 0.0, NULL);
	P = recycleProblem(net, 0.0, NULL, prof);

	proMeasureTime(timer, prof2, "  add n2p");

	// let frobbing find the best lexeme selection
	if(P) {
		combined(P);

		proMeasureTime(timer, prof2, "  add frob");

		if(bsTopDownPrediction) {
			results = predictTopDown(P->best, p, threshold);
			proMeasureTime(timer, prof2, "  add tdp");
		} else {
			results = listAppendElement(listNew(), P->best);
		}

		s= method;
		for(l= results; l; l= l->next) {
			a = l->item;

			hkRestoreVerbosity();
			if(bsVerbosity>=4)
				cdgPrintf(CDG_INFO, "Score: %.5f\n", a->b->hard?0.0:a->b->soft);
			if(a->b->hard || significantlyGreater(threshold, a->b->soft)) {
				// this variant's score is too bad, delete it
				a = NULL;

			} else {
				newBE = beFromAnalysis(a, oldBE, s);
				beam = listAppendElement(beam, newBE);

				beam = pruneBeam(beam, beamSize, f);
			}
			s= "tdp"; // if there is more than one analysis in the list, those are from tdp
		}

		// either way, we don't need the problem anymore, so:
		P->current = NULL;
		P = NULL;
	} else {
		// this parse did not work,
		// probably as the lvs compatible with one of the new edges
		// are not compatible lexeme wise with the rest of the parse
		hkRestoreVerbosity();
		if(bsVerbosity>=3)
			cdgPrintf(CDG_WARNING, "Invalid parse, skipping \n");
	}
	proMeasureTime(timer, prof2, "  add postpare");
	proMerge(prof, prof2);

	return beam;
}


/**
 * For a BeamEntry we need a score, for that we need an analysis and for that we need a lexeme selection
 *
 * This is a hybrid of addParseToBeamByFrobbing and addParseToBeamByRules
 * This method investigates which rules are needed by the latter
 * by looking at which constraints associated to the lexeme selection
 * are repaired by frobbing
 */
List addParseToBeamCollectingNeededRules(Parse p, int beamSize, Number f, List oldBeam, ConstraintNet net, BeamEntry oldBE,
		String method, List positions, ProfilingObject prof)
{
	List beam = oldBeam;
	BeamEntry newBE;
	Problem P;
	Number threshold;
	List results, l, m, m2;
	Analysis a;
	String s;
	Vector slots;

	Timer timer = timerNew();
	ProfilingObject prof2 = proNew();

	threshold=  max(
			beamGetHighestScore(beam) * bsBeamToleranceFactor,
			beamGetLowestScore(beam, bsBeamWidth));

	if(bsVerbosity>=4) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		printAbstractParse(p, net->lexemgraph);
		cdgPrintf(CDG_INFO, "\n");
	}

	if(bsVerbosity < 4) {
		if(bsVerbosity < 3) {
			hkDeactivate(CDG_WARNING);
		}
		hkDeactivate(CDG_INFO);
	}
	proMeasureTime(timer, prof2, "  add prepare");

	List lvCombinations = listAppendElement(listNew(), listNew()); // list of lists
	List tmp;
	List compatibleLVs;
	int slot;
	for (l= positions; l; l=l->next) {
		slot = ((int)l->item) * inputCurrentGrammar->noOfLevels + 0;
		compatibleLVs = getCompatibleLVsfromSlot(p, slot, net, threshold);
		tmp = listNew();
		for(m = lvCombinations; m; m=m->next) {
			for(m2 = compatibleLVs; m2; m2=m2->next) {
				if(isLexemeCompatible(m2->item, m->item, net->lexemgraph)) {
					tmp = listAppendElement(tmp,
							listAppendElement(listClone(m->item), m2->item));
				}
			}
		}
		lvCombinations = tmp;
	}

	// slots might have changed in the new parse due to insertion,
	// so we have to remove incompatible LVs from the slots list
	Vector parseSlots = filterLVsByParse(p, oldBE->slots);
	LevelValue lv;

	// cdgPrintf(CDG_WARNING, "Testing %d LVs \n", listSize(compatibleLVs));

	for(l = lvCombinations; l; l = l->next) {

		if(bsVerbosity < 4) {
			if(bsVerbosity < 3) {
				hkDeactivate(CDG_WARNING);
			}
			hkDeactivate(CDG_INFO);
		}

		compatibleLVs = l->item;

		// we have to limit the search space, only lexeme variations are allowed
		cnResetToThreshold(net,threshold);
		if(bsLockAllLevels) {
			cnLockToParse(net, p, NULL);
		} else {
			cnLockToParse(net, p, inputCurrentGrammar->levels->item);
		}
		if(bsUnlockUnusedVNs) {
			unlockUnusedVNs(net, p);
		}
		for(m= compatibleLVs; m; m=m->next) {
			lv = (LevelValue) m->item;
			cnLockToLV(net, lv);
		}

		proMeasureTime(timer, prof2, "  add lock net");

		slots = vectorClone(parseSlots);
		for(m= compatibleLVs; m; m=m->next) {
			lv = (LevelValue) m->item;
			vectorAddElement(slots, lv);
		}

		P = recycleProblem(net, 0.0, slots, prof);

		proMeasureTime(timer, prof2, "  add n2p");

		// let frobbing find the best lexeme selection
		if(P) {
			List initialCVs =  listNew();
			for(m= P->best->conflicts; m; m=m->next) {
				initialCVs = listAppendElement(initialCVs, cvClone((ConstraintViolation)m->item));
			}

			combined(P);

			List remainingCVs = listNew();
			for(m= P->best->conflicts; m; m=m->next) {
				remainingCVs = listAppendElement(remainingCVs, cvClone((ConstraintViolation)m->item));
			}

			removedCVs(initialCVs, remainingCVs, prof);

			proMeasureTime(timer, prof2, "  add frob");

			if(bsTopDownPrediction) {
				results = predictTopDown(P->best, p, threshold);
				proMeasureTime(timer, prof2, "  add tdp");
			} else {
				results = listAppendElement(listNew(), P->best);
			}

			s= method;
			for(m= results; m; m= m->next) {
				a = m->item;

				hkRestoreVerbosity();
				if(bsVerbosity>=4)
					cdgPrintf(CDG_INFO, "Score: %.5f\n", a->b->hard?0.0:a->b->soft);
				if(a->b->hard || significantlyGreater(threshold, a->b->soft)) {
					// this variant's score is too bad, delete it
					a = NULL;

				} else {
					newBE = beFromAnalysis(a, oldBE, s);
					beam = listAppendElement(beam, newBE);

					beam = pruneBeam(beam, beamSize, f);
				}
				s= "tdp"; // if there is more than one analysis in the list, those are from tdp
			}

		} else {
			// this parse did not work,
			// probably as the lvs compatible with one of the new edges
			// are not compatible lexeme wise with the rest of the parse
			hkRestoreVerbosity();
			if(bsVerbosity>=3)
				cdgPrintf(CDG_WARNING, "Invalid parse, skipping \n");
		}
	}

	proMeasureTime(timer, prof2, "  add postpare");
	proMerge(prof, prof2);

	return beam;
}

/**
 * For a BeamEntry we need a score, for that we need an analysis and for that we need a lexeme selection
 * Here lexeme selection is done by testing all combinations for the affected edges and repairing remaining
 * conflicts with a given set of rules
 */
List addParseToBeamByRules(Parse p, int beamSize, Number f, List oldBeam, ConstraintNet net, BeamEntry oldBE,
		String method, List positions, ProfilingObject prof)
{
	List beam = oldBeam;
	BeamEntry newBE;
	Number threshold, localThreshold;
	List results, l, m, m2;
	Analysis a, best;
	String s;
	Vector slots;
	Boolean tooBad, aboveTh;

	Timer timer = timerNew();
	Timer timer2;
	ProfilingObject prof2 = proNew();
	ProfilingObject prof3;

	List tmp;
	List compatibleLVs;
	int slot;
	int i;
	Problem P = makeZombieProblem(net);

	threshold=  max(
			beamGetHighestScore(beam) * bsBeamToleranceFactor,
			beamGetLowestScore(beam, bsBeamWidth));
	cdgPrintf(CDG_INFO, "Threshold: %f\n", threshold);

	// we are only interested in the best lexeme selection,
	// therefore we use the maximum of the beam threshold
	// and the score of the currently best lexeme selection
	// as local threshold
	localThreshold = threshold;

	if(bsVerbosity>=4) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		printAbstractParse(p, net->lexemgraph);
		cdgPrintf(CDG_INFO, "\n");
	}

	// slots might have changed in the new parse due to insertion,
	// so we have to remove incompatible LVs from the slots list
	Vector parseSlots = filterLVsByParse(p, oldBE->slots);
	LevelValue lv;

	proMeasureTime(timer, prof2, "  add: prepare");

	List lvCombinations = listAppendElement(listNew(), listNew()); // list of lists
	// for all positions where a change occurred
	// (i.e. the new word, newly introduced VNs and all changes from inserting a new word for a VN)
	for (l= positions; l; l=l->next) {
		for(i=0; i < P->nrLevels; i++) {
			slot = ((int)l->item) * inputCurrentGrammar->noOfLevels + i;
			prof3 = proNew();
			timer2 = timerNew();
			compatibleLVs = getCompatibleLVsfromSlot(p, slot, net, threshold);
			proMeasureTime(timer2, prof3, "   mc: getCompatibleLVs");

			tmp = listNew();
			// for every lv combination so far ...
			for(m = lvCombinations; m; m=m->next) {
				// ... check for every potential lv for the current slot ...
				for(m2 = compatibleLVs; m2; m2=m2->next) {
					// ... whether it is compatible lexeme wise with the LVs already in the list ...
					if(isLexemeCompatible(m2->item, m->item, net->lexemgraph)) {
						// ... if yes add the extended list as a new combination
						tmp = listAppendElement(tmp,
								listAppendElement(listClone(m->item), m2->item));
					}
				}
			}
			lvCombinations = tmp;

			proMeasureTime(timer2, prof3, "   mc: make lists");
			proMerge(prof2, prof3);
		}
	}
	proMeasureTime(timer, prof2, "  add: make combinations");

	cdgPrintf(CDG_INFO, "%s:%2d->%4d", method, listSize(positions), listSize(lvCombinations));

	best = NULL;
	// now judge all combinations to get the best one
	// and add it to the beam if it is above the threshold
	for(l = lvCombinations; l; l = l->next) {
		compatibleLVs = l->item;
		prof3 = proNew();
		slots = vectorClone(parseSlots);
		tooBad= FALSE;

		// first compare individual LVs to the local threshold
		for(m= compatibleLVs; m; m=m->next) {
			lv = (LevelValue) m->item;
			if(significantlyGreater(localThreshold, lv->score)) {
				// can't get better than the currently best analysis, no need to try this lexeme variant
				tooBad = TRUE;
				break;
			}
			vectorAddElement(slots, lv);
		}
		if(tooBad) {
			continue;
		}
		proMeasureTime(timer, prof3, "  add: prepare slots");

		// make an analysis just by adding up all the LVs from all the slots
		a= aFromSlots(P, slots, net);
		proMeasureTime(timer, prof3, "  add: make analysis");

		aboveTh = aJudgeWithThreshold(a, localThreshold);
		//aCheck(a,3);
		//aPrint(CDG_INFO, a);

		if(!aboveTh || a->b->hard || significantlyGreater(localThreshold, a->b->soft )) {
			a = NULL;
			continue;
		}

		if(!a->b->hard && a->b->soft * bsBeamToleranceFactor > threshold) {
			threshold = a->b->soft * bsBeamToleranceFactor;
		}

		localThreshold = a->b->soft;
		best= a;

		proMeasureTime(timer, prof3, "  add: judge analysis");
	}

	if(best) {
		cdgPrintf(CDG_INFO, " => %f\n", localThreshold);

		if(bsTopDownPrediction) {
			results = predictTopDown(best, p, threshold);
			proMeasureTime(timer, prof2, "  add tdp");
		} else {
			results = listAppendElement(listNew(), best);
		}
		proMeasureTime(timer, prof3, "  add: apply tdp rules");

		s= method;
		for(m= results; m; m= m->next) {
			a = m->item;

			if(bsVerbosity>=4)
				cdgPrintf(CDG_INFO, "Score: %.5f\n", a->b->hard?0.0:a->b->soft);
			if(a->b->hard || significantlyGreater(threshold, a->b->soft)) {
				// this variant's score is too bad, delete it
				a = NULL;

			} else {
				newBE = beFromAnalysis(a, oldBE, s);
				//parsePrint(newBE->p);
				beam = listAppendElement(beam, newBE);

				beam = pruneBeam(beam, beamSize, f);
			}
			s= "tdp"; // if there is more than one analysis in the list, those are from tdp
		}

		proMeasureTime(timer, prof3, "  add: make beam");
		proMerge(prof2, prof3);
	} else {
		cdgPrintf(CDG_INFO, " => fail\n");
	}

	proMerge(prof, prof2);

	return beam;
}

/*
 * takes a list of parses and generates a list of beam entries of the given beam size
 */
List addParseToBeam(Parse p, int beamSize, Number f, List oldBeam, ConstraintNet net, BeamEntry oldBE,
		String method, List positions, ProfilingObject prof)
{
	if(bsMethod == 0) {
		return addParseToBeamByFrobbing(p, beamSize, f, oldBeam, net, oldBE, method, prof);
	} else if(bsMethod ==1) {
		return addParseToBeamCollectingNeededRules(p, beamSize, f, oldBeam, net, oldBE, method, positions, prof);
	} else if(bsMethod ==2) {
		return addParseToBeamByRules(p, beamSize, f, oldBeam, net, oldBE, method, positions, prof);
	}
	cdgPrintf(CDG_ERROR, "ERROR: unknown bsMethod %d !\n", bsMethod);
	abort();
}

/* ---------------------------------------------------------------------------
 * Make variants of the given parse by attaching the new arc
 * - to every existing node
 * - with every existing label
 * in the first level, in all other levels it will be attached to nil
 *
 * only choose combinations with a according LV in the net,
 * that has a score above the threshold
 *
 * @param arc	the arc to be attached
 * @param p 	the parse to be varied, may differ from the one in the BE,
 *                                      this is needed for recursive calls for virtual nodes
 * @param be	beam entry, contains the net and virtual node information
 * @param insert, whether the arc has to be inserted or is already part of the parse
 * @param childTracking, list containing transitive children for circle detection
 * @param bss   beamsearchstate, used to add profiling information to it
 *
 * @returns a list of variants of the given parse
 */
List appendArcToParse(Arc arc, Parse p, ConstraintNet net, BeamEntry be, List oldBeam,
		Boolean insert, List childTracking, ProfilingObject prof)
{
	List results = oldBeam;
	Number threshold = 0.0;
	int i, j, pos, slot, r, wid_r;
	int noOfLvls = inputCurrentGrammar->noOfLevels;
	List lvs;
	String label;
	Parse p2;
	Vector tracker; // vector of lists, indexed by slot ids
	LevelValue lv;
	Boolean vReg;
	GraphemNode gn;

	Timer timer = timerNew();
	ProfilingObject prof2= proNew();

	if(listSize(childTracking) > bsBottomUpPredictionDepth) {
		return results;
		// more than n steps in the unknown are cut off
	}

	Number estimatedScore = getScoreEstimation(p, net, insert);
	if(estimatedScore == 0.0) {
		// the given slots already prevent a soft solution, so we can stop right here
		// then again, why is this parse in the beam? probably a score fluctuation effect
		return results;
	}

	ByteVector availableLexemes = getAvailableLexemes(be, net);

	tracker = vectorNew(vectorSize(net->nodes) +2); // +2 for nil and nonspec
	for(i=0; i < vectorSize(net->nodes); i++) {
		vectorSetElement(tracker, listNew(), i);
	}

	pos= arc->from;
	slot= pos* noOfLvls;

	lvs= vectorToList(((ConstraintNode)vectorElement(net->nodes, slot))->values);

	threshold=  max(
			beamGetHighestScore(results) * bsBeamToleranceFactor,
			beamGetLowestScore(results, bsBeamWidth));

	// try all levelValues
	for( ; lvs; lvs= lvs->next) {

		lv= lvs->item;

		if(lv->score * estimatedScore < threshold) {
			// no need to check this levelValue, as it is too bad,
			// as the lvs are sorted by score, there is also no need to go any further
			break;
		}

		label= lv->label;
		if(root(lv->regents)) {
			wid_r= -1;
			vReg= FALSE;
		} else if(nonspec(lv->regents)) {
			wid_r= -2;
			vReg= FALSE;
		} else {
			gn= lvGetRegentGN(lv);
			wid_r = gn->no;
			vReg= gn->isVirtual;
		}

		if(!insert && childTracking && listContains(childTracking, (void *)wid_r) ) {
			// circle detection!
			// the word we are about to append is an virtual one
			// and i currently points to one of the reasons we need the new virtual word
			// so skip the word i to prevent a circle
			continue;
		}

		if(listContains(vectorElement(tracker,wid_r +2), label)) {
			// we already had that combination
			continue;
		}

		// add combination to tracker
		vectorSetElement(tracker,
				listAppendElement(
						vectorElement(tracker, wid_r +2),
						label),
				wid_r +2);

		if(!insert && wid_r==-1 && strcmp(label,"")==0) {
			// We are currently introducing a virtual word because it is needed.
			// A dependency to NIL with the empty label would result in a
			// partially used virtual word, that is forbidden.
			// To save time it is better to stop it here,
			// instead of letting the grammar handle it.
			continue;
		}

		// be->virtualNodeAvailable uses old indices
		if(vReg && !bvElement(be->virtualNodeAvailable, wid_r-1)) {
			// that virtual word is not for use in this entry, skip it
			// there should be another free virtual word of the same kind
			continue;
		}

		if(compatibleLvExists(net, 0, pos, label, wid_r, threshold/estimatedScore, availableLexemes) ) {
			// found a LevelValue above the threshold,

			// so add the current combination to the copy of the parse
			p2= parseClone(p, FALSE); // No need to decorate the parse
			if(insert) {
				vectorInsertElement(p2->verticesLabels, label, slot);
				vectorInsertElement(p2->verticesStructure, (void *)wid_r, slot);
				for(j= 1; j< noOfLvls; j++) {
					// add default values for other parses
					vectorInsertElement(p2->verticesLabels, "", slot+j);
					vectorInsertElement(p2->verticesStructure, (void *)-1, slot+j);
				}
				if(useVirtualNodes) {
					// push references to virtual nodes
					for(j= 0; j< vectorSize(p2->verticesStructure); j++) {
						if(j== slot) // skip the word itself
							continue;
						r = (int) vectorElement(p2->verticesStructure, j);
						if(r >= pos) {
							vectorSetElement(p2->verticesStructure, (void *)r+1, j);
						}
					}
				}
			} else {
				vectorSetElement(p2->verticesLabels, label, slot);
				vectorSetElement(p2->verticesStructure, (void *)wid_r, slot);
			}

			if(vReg && !bvElement(be->virtualNodeUsed, wid_r-1)) {
				results = appendArcToParse(((GraphemNode)vectorElement(net->lexemgraph->graphemnodes,wid_r))->arc,
						p2, net, be, results, FALSE, listAppendElement(listClone(childTracking), (void *)pos), prof );
			} else {
				// now go from parse to an actual analysis, that we can evaluate, and add that to the beam
				results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net, be, "app",
						listAppendElement(listClone(childTracking), (void*) pos), prof);
			}
			parseDelete(p2);

			// update threshold
			threshold=  max(
					beamGetHighestScore(results) * bsBeamToleranceFactor,
					beamGetLowestScore(results, bsBeamWidth));
		}
	} // end for all lvs

	tracker = NULL;
	childTracking = NULL;

	if(insert) {
		proMeasureTime(timer, prof2, " append");
		proMerge(prof, prof2);
	}

	return results;
}


/* -------------------------------------------------------------------------
 * Make variants of the given parse by replacing the virtual nodes with the given arc.
 * If there are no virtual nodes, the list is empty.
 *
 * @returns a list of variants of the given parse
 */
List insertArcIntoParse(Arc arc, BeamEntry be, ConstraintNet net, List oldBeam, ProfilingObject prof) {
	List results = oldBeam;
	int wid_v, wid_n, old_wid_v; // word IDs
	int sid_i;                   // slot IDs
	int i, r, reg;
	Parse p = be->p;
	Parse p2;
	List lvs;
	Boolean toobad = FALSE;
	LevelValue lv;
	Number score;
	Number threshold;

	ProfilingObject prof2 = proNew();
	Timer timer= timerNew();

	int noOfLvls= inputCurrentGrammar->noOfLevels;

	wid_n= arc->from;

	List changedPositions;

	// for all virtual nodes
	for(wid_v= net->lexemgraph->ve->vStartID; wid_v< net->lexemgraph->max ; wid_v++) {

		// the virtual nodes in the LexemGraph are already pushed back, but we still need the old index at some points
		old_wid_v = wid_v -1;

		// the new word and the replaced virtual node will definitely be affected by insertion,
		// so we add them to the list
		changedPositions = listAppendElement(listAppendElement(listNew(), (void*)wid_n), (void*)wid_v);

		if(! bvElement(be->virtualNodeAvailable, old_wid_v) ||
				! bvElement(be->virtualNodeUsed, old_wid_v)) {
			// the node is not available or not used, so there is not point in replacing it with the new word
			continue;
		}

		threshold=  max(
				beamGetHighestScore(results) * bsBeamToleranceFactor,
				beamGetLowestScore(results, bsBeamWidth));

		p2= parseClone(p, FALSE); // no need to decorate it

		// push references to virtual nodes
		for(i= 0; i< vectorSize(p2->verticesStructure) ; i++) {
			r = (int) vectorElement(p2->verticesStructure, i);
			if(r >= wid_n) {
				vectorSetElement(p2->verticesStructure, (void *)r+1, i);
			}
		}

		toobad=FALSE;
		score= 1.0;

		// we have to
		// 1. replace the regent of all edges going to the node to be replaced
		// 2. replace the edge going out from the virtual node with a parking edge
		// 3. add an edge going out from the new node

		// we do step 1 and 2 in one for loop
		// for all old slots
		for(sid_i=0; sid_i< vectorSize(p->verticesStructure); sid_i++) {
			int lvl_i = sid_i % noOfLvls;
			int wid_i = sid_i/noOfLvls;

			if( ((int)vectorElement(p->verticesStructure, sid_i)) == old_wid_v ) {
				// step 1:
				// This edge points towards the virtual node to-be-replaced.
				// Change its head so it points to the new word instead.
				vectorSetElement(p2->verticesStructure, (void *)wid_n,
						sid_i); // virtual nodes are not pushed back yet in this vector, so we can still use the old id

				// keep the label the same, nothing to do there

				// add it to the list of changed words
				changedPositions = listAppendElement(changedPositions,
						(void*) (wid_i+1 < net->lexemgraph->ve->vStartID?wid_i:(wid_i+1)) /* we need the new id, the parse has the old one*/
						);

				// check max score
				// parse->verticesStructure starts at 0, as do the GraphemNode numbers, no need to add or subtract something
				lvs = cnGetAccordingLVs( net, // net has already the new IDs,
						lvl_i,
						(wid_i >= wid_n)?(wid_i+1):(wid_i) , 	// dependent (wID), we get the new id for virtual nodes by adding 1
						vectorElement(p->verticesLabels, sid_i),	// label
						wid_n);					// regent (wID)

				lv= bestScoringLVfromList(lvs);
				lvs = NULL;
				// we don't know the best possible combination of lv's,
				// but we can compare the ideally best combination to the threshold
				score= score * (lv?lv->score:0.0);
				if(score < threshold || score == 0.0) {
					// local score already too bad, abort this variant
					parseDelete(p2);
					toobad=TRUE;
					break; // for all old slots, we break for all virtual nodes later via 'toobad'
				}
			} else if (wid_i == old_wid_v) { // virtual nodes are not pushed back yet in this vector, so have to use the old id
				// step 2:
				// This edge's dependent is the virtual node to-be-replaced.
				// We replace this edge, so that the virtual node is now unused,
				// i.e., the regent becomes NIL and the label is empty.
				vectorSetElement(p2->verticesStructure, (void *) -1 , sid_i); // the NIL regent, its wID is -1 here
				vectorSetElement(p2->verticesLabels   , "", sid_i);			  // empty label
			}
		}

		if(toobad) {
			// the multiplied best scores are below the threshold, skip to next virtual word
			continue;
		}

		// step 3:
		// finally we insert the new edges for the new word

		reg = (int) vectorElement(p->verticesStructure, old_wid_v*noOfLvls +0);
		if(reg >= net->lexemgraph->ve->vStartID -1) {
			reg += 1;
		}

		// first check max score, we only check level 0 here
		lvs = cnGetAccordingLVs( net, // net has already the new IDs,
				0, 		// level
				wid_n, 	// dependent (wID),
				vectorElement(p->verticesLabels,    old_wid_v*noOfLvls +0),	// label
				reg
				);	// regent (wID)
		lv= bestScoringLVfromList(lvs);
		lvs = NULL;
		if(!lv || lv->score * score < threshold) {
			continue;
		}

		// for all levels
		for(i=0; i < noOfLvls; i++) {
			reg = (int) vectorElement(p->verticesStructure, old_wid_v*noOfLvls +i);
			if(reg >= net->lexemgraph->ve->vStartID -1) {
				reg += 1;
			}
			vectorInsertElement(
				p2->verticesStructure,
				(void*)reg,
				wid_n* noOfLvls + i);
			vectorInsertElement(
				p2->verticesLabels,
				vectorElement(p->verticesLabels, old_wid_v*noOfLvls +i),
				wid_n* noOfLvls + i);
		}

		// OK, we got a valid parse, add it to the beam
		results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net, be, "ins",
				changedPositions, prof);

//		cdgPrintf(CDG_INFO, "This parse was generated by insertArcIntoParse\n");
//		printAbstractParse(p2, net->lexemgraph);
//		cdgPrintf(CDG_INFO, "------------------------------------\n");

		parseDelete(p2);

	} // end for all virtual nodes

	proMeasureTime(timer, prof2, " insert");
	proMerge(prof, prof2);

	return results;
}




///**
// * ads an arc to the given analysis in all possible ways
// *
// */
//List/*<BeamEntry>*/ appendArcToAnalysis(Arc arc, Analysis a, ConstraintNet net, BeamEntry be, List oldBeam, ProfilingObject prof) {
//	List results = oldBeam;
//	Number threshold = 0.0;
//	int i, j, pos, slot, r, wid_r;
//	int noOfLvls = inputCurrentGrammar->noOfLevels;
//	List lvs;
//	String label;
//	Parse p2;
//	Vector tracker; // vector of lists, indexed by slot ids
//	LevelValue lv;
//	Boolean vReg;
//	GraphemNode gn;
//
//	Timer timer = timerNew();
//	ProfilingObject prof2= proNew();
//
//	ByteVector availableLexemes = getAvailableLexemes(be, net);
//
//	tracker = vectorNew(vectorSize(net->nodes) +2); // +2 for nil and nonspec
//	for(i=0; i < vectorSize(net->nodes); i++) {
//		vectorSetElement(tracker, listNew(), i);
//	}
//
//	pos= arc->from;
//	slot= pos* noOfLvls;
//
//	lvs= vectorToList(((ConstraintNode)vectorElement(net->nodes, slot))->values);
//
//	threshold=  max(
//			beamGetHighestScore(results) * bsBeamToleranceFactor,
//			beamGetLowestScore(results, bsBeamWidth));
//
//	// try all levelValues
//	for( ; lvs; lvs= lvs->next) {
//
//		lv= lvs->item;
//
//		if(lv->score * estimatedScore < threshold) {
//			// no need to check this levelValue, as it is too bad,
//			// as the lvs are sorted by score, there is also no need to go any further
//			break;
//		}
//
//		label= lv->label;
//		if(root(lv->regents)) {
//			wid_r= -1;
//			vReg= FALSE;
//		} else if(nonspec(lv->regents)) {
//			wid_r= -2;
//			vReg= FALSE;
//		} else {
//			gn= lvGetRegentGN(lv);
//			wid_r = gn->no;
//			vReg= gn->isVirtual;
//		}
//
//		if(!insert && childTracking && listContains(childTracking, (void *)wid_r) ) {
//			// circle detection!
//			// the word we are about to append is an virtual one
//			// and i currently points to one of the reasons we need the new virtual word
//			// so skip the word i to prevent a circle
//			continue;
//		}
//
//		if(listContains(vectorElement(tracker,wid_r +2), label)) {
//			// we already had that combination
//			continue;
//		}
//
//		// add combination to tracker
//		vectorSetElement(tracker,
//				listAppendElement(
//						vectorElement(tracker, wid_r +2),
//						label),
//				wid_r +2);
//
//		if(!insert && wid_r==-1 && strcmp(label,"")==0) {
//			// We are currently introducing a virtual word because it is needed.
//			// A dependency to NIL with the empty label would result in a
//			// partially used virtual word, that is forbidden.
//			// To save time it is better to stop it here,
//			// instead of letting the grammar handle it.
//			continue;
//		}
//
//		// be->virtualNodeAvailable uses old indices
//		if(vReg && !bvElement(be->virtualNodeAvailable, wid_r-1)) {
//			// that virtual word is not for use in this entry, skip it
//			// there should be another free virtual word of the same kind
//			continue;
//		}
//
//		if(compatibleLvExists(net, 0, pos, label, wid_r, threshold/estimatedScore, availableLexemes) ) {
//			// found a LevelValue above the threshold,
//
//			// so add the current combination to the copy of the parse
//			p2= parseClone(p, FALSE); // No need to decorate the parse
//			if(insert) {
//				vectorInsertElement(p2->verticesLabels, label, slot);
//				vectorInsertElement(p2->verticesStructure, (void *)wid_r, slot);
//				for(j= 1; j< noOfLvls; j++) {
//					// add default values for other parses
//					vectorInsertElement(p2->verticesLabels, "", slot+j);
//					vectorInsertElement(p2->verticesStructure, (void *)-1, slot+j);
//				}
//				if(useVirtualNodes) {
//					// push references to virtual nodes
//					for(j= 0; j< vectorSize(p2->verticesStructure); j++) {
//						if(j== slot) // skip the word itself
//							continue;
//						r = (int) vectorElement(p2->verticesStructure, j);
//						if(r >= pos) {
//							vectorSetElement(p2->verticesStructure, (void *)r+1, j);
//						}
//					}
//				}
//			} else {
//				vectorSetElement(p2->verticesLabels, label, slot);
//				vectorSetElement(p2->verticesStructure, (void *)wid_r, slot);
//			}
//
//			if(vReg && !bvElement(be->virtualNodeUsed, wid_r-1)) {
//				results = appendArcToParse(((GraphemNode)vectorElement(net->lexemgraph->graphemnodes,wid_r))->arc,
//						p2, net, be, results, FALSE, listAppendElement(listClone(childTracking), (void *)pos), prof );
//			} else {
//				// now go from parse to an actual analysis, that we can evaluate, and add that to the beam
//				results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net, be, "app",
//						listAppendElement(listClone(childTracking), (void*) pos), prof);
//			}
//			parseDelete(p2);
//
//			// update threshold
//			threshold=  max(
//					beamGetHighestScore(results) * bsBeamToleranceFactor,
//					beamGetLowestScore(results, bsBeamWidth));
//		}
//	} // end for all lvs
//
//	for(i=0; i< vectorSize(tracker); i++) {
//		listDelete(vectorElement(tracker, i));
//	}
//	vectorDelete(tracker);
//	listDelete(childTracking);
//
//	if(insert) {
//		proMeasureTime(timer, prof2, " append");
//		proMerge(prof, prof2);
//	}
//
//	return results;
//}


/* ---------------------------------
 * Make variants of the given parse.
 *
 * @param be    the beamentry containing the parse to be varied
 * @param arc	the arc to be appended
 * @param net   the constraint net used through out the search
 * @param oldBeam the beam so far
 * @param bss   the beamsearch state, just used for profiling
 * @returns a list of variants of the given beamEntry
 */
List beAddArc(BeamEntry be, Arc arc, ConstraintNet net, List oldBeam, ProfilingObject prof) {
	List results = oldBeam;
//	Analysis a = aClone(be->best);
//	// TODO calculate score estimation without context
//
//	extendAnalysis(a, a->P, 1, FALSE, NULL, prof);
//
//	if(useVirtualNodes) {
//		results = insertArcIntoAnalysis(arc, aClone(a), be, net, results, prof);
//		results = appendArcToAnalysis(arc, aClone(a), net, be, results, TRUE, NULL, prof);
//	} else {
//		results = appendArcToAnalysis(arc, aClone(a), net, be, results, TRUE, NULL, prof);
//	}


	if(useVirtualNodes) {
		results = insertArcIntoParse(arc, be, net, results, prof);
		results = appendArcToParse(arc, be->p, net, be, results, TRUE, NULL, prof);

	} else {
		results = appendArcToParse(arc, be->p, net, be, results, TRUE, NULL, prof);
	}

	return results;
}


/*
 * returns a new beam search state
 * the constraintNet has to be set separately
 * virtual node specifications are also not set here
 */
BeamSearchState bssNew() {
	BeamSearchState bss = memMalloc( sizeof(BeamSearchStateStruct));
	bss->entries = vectorNew(bsBeamWidth);
	bss->net = NULL;
	bss->virtualNodeTypes = vectorNew(10);
	bss->profiling = proNew();
	bss->previous = NULL;
	bss->newWord= "";
	return bss;
}

/**
 * print details for the given bs state to the console
 */
void bssPrintDetails(BeamSearchState bss) {
	List l;

	if(bss->previous) {
		bssPrintDetails(bss->previous);
	}

	cdgPrintf(CDG_INFO, "\n%s\n-------------------------------------------\n", bss->newWord);

	for(l= bss->profiling->measurePoints; l; l= l->next) {
		ProfilingMeasurePoint mp = l->item;
		cdgPrintf(CDG_INFO, "%30s | %8.3f sec\n", mp->name, mp->time/1000.0 );
	}

	cdgPrintf(CDG_INFO, "\nRule Candidates:\n");
	for(l = hashListOfKeys(bss->profiling->contraintCount); l ; l=l->next) {
		cdgPrintf(CDG_INFO, "%15s : %4d\n", (String)l->item, (int)hashGet(bss->profiling->contraintCount, l->item));
	}
}

/*
 * This method marks as available
 * - all used entries
 * - and the first unused occurrence
 * for each virtual node specification
 *
 */
void markAvailableVirtualNodes(ConstraintNet net, BeamEntry be) {
	Boolean foundUnusedOne;
	List l, m;
	VirtualLGExtension ve= net->lexemgraph->ve;
	List specs = ve->specs;
	GraphemNode gn;

	be->virtualNodeAvailable = bvNew(net->lexemgraph->max);
	bvSetAllElements(be->virtualNodeAvailable, TRUE);

	// for all specs
	for(l = specs; l; l=l->next ) {
		VirtualNodeSpec spec = l->item;
		foundUnusedOne= FALSE;
		List ocs = veOccurences(ve, spec);

		// for all vns using this spec
		for(m = ocs; m; m=m->next) {
			VirtualNode vn = m->item;
			gn= vn->gn;
			if(!bvElement(be->virtualNodeUsed, gn->no)) {
				if(!foundUnusedOne) {
					foundUnusedOne= TRUE;
					// this last one is still available
					bvSetElement(be->virtualNodeAvailable, TRUE, gn->no);
				} else {
					// we don't need any more of those!
					bvSetElement(be->virtualNodeAvailable, FALSE, gn->no);
				}
			} else {
				bvSetElement(be->virtualNodeAvailable, TRUE, gn->no);
			}
		}
	}
}


/* -------------------------------------------------------------------------
 * Execute one step of beam search in the following way:
 * For every parse in the beam, all possible continuations are generated and evaluated via the grammar
 * the best n are used for the next beam front
 *
 * configure n via bsBeamWidth
 *
 */
BeamSearchState beamSearchStep(BeamSearchState prevBss, Arc newArc, Boolean complete) {
	if(bsVerbosity>=1)
		cdgPrintf(CDG_INFO, "---------------------------------------------------------\n");

	List newEntries = listNew();
	int i,j;
	BeamEntry be;
	List l;

	Timer timer1 = timerNew();
	Timer timer2 = timerNew();

	BeamSearchState bss = bssNew();
	//BeamSearchState bss = (BeamSearchState) memMalloc(sizeof(BeamSearchStateStruct));

	bss->virtualNodeTypes = vectorClone(prevBss->virtualNodeTypes);
	bss->net = prevBss->net;
	bss->previous = prevBss;
	bss->newWord = newArc->word;

	// handle virtual nodes in the net and in the beam
	if(useVirtualNodes) {

		// first detect, whether we need more virtual node, depending on the beam entries
//		for(i=0; i< vectorSize(prevBss->entries); i++) {
//			be = vectorElement(prevBss->entries, i);
//			cnManageVirtualNodes(bss->net, be->best, NULL, NULL);
//		}

		// then determine which ones are used in which beam entry
		for(i=0; i< vectorSize(prevBss->entries); i++) {
			GraphemNode gn;
			be = vectorElement(prevBss->entries, i);
			be->virtualNodeUsed = bvNew(bss->net->lexemgraph->max);
			// we start at zero, but it doesn't really matter what value we enter for non-virtual words
			for(j=0; j < vectorSize(bss->net->lexemgraph->graphemnodes); j++) {
				gn = vectorElement(bss->net->lexemgraph->graphemnodes, j);
				if(gnIsUnusedVirtualNode(gn,be->best)) {
					bvSetElement(be->virtualNodeUsed, FALSE, j);
				} else {
					bvSetElement(be->virtualNodeUsed, TRUE, j);
				}
			}
		}

		// and we mark which VNs should be available for each beam entry, as they share the net but might not need
		// the additional entries
		for(i=0; i< vectorSize(prevBss->entries); i++) {
			be = vectorElement(prevBss->entries, i);
			markAvailableVirtualNodes(bss->net, be);
		}
	}

	if(complete) {
		bss->net->lexemgraph->complete = TRUE;
	}
	// now we add the new arc to the net
	bss->net = extendNet(prevBss->net, newArc, NULL);
	bss->net = postExtendNetCleanup(bss->net, NULL, proNew());
	// ATTENTION: expanding the net made some indices in the old parses invalid,
	// namely the indices of virtual nodes, as the new arc is inserted in front of them

	proMeasureTime(timer2, bss->profiling, " prepare");

	// for all entries in the beam ...
	for(i=0; i< vectorSize(prevBss->entries); i++) {
		if(bsVerbosity>=2) {
			cdgPrintf(CDG_INFO, "----------------------\n");
			cdgPrintf(CDG_INFO, "Beam entry #%d\n", i+1);
		}

		be = vectorElement(prevBss->entries, i);

		// ... find all possible ways to add the new arc to the current parse.
		newEntries = beAddArc(be, newArc, bss->net, newEntries, bss->profiling);
	}

	proMeasureTime(timer2, bss->profiling, " add");

	bss->entries = listToVector(newEntries);

	if(!bss->entries) {
		cdgPrintf(CDG_WARNING,"WARNING: empty beam \n");
		bss->entries= vectorNew(10);
		return bss;
	}

	l= newEntries;

	if(bsVerbosity>=1) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		while(l) {
			be = l->item;
			while(be) {
				cdgPrintf(CDG_INFO, "%d <--(%s)-- ", be->rank, be->method);
				be = be->previous;
			}
			cdgPrintf(CDG_INFO, "\n");

			be = l->item;
			parsePrint2( be->p, TRUE, frobIgnoreThreshold);
			cdgExecHook(HOOK_PARTIALRESULT, be->p);
			l=l->next;
		}
		int n = vectorSize(bss->entries);
		cdgPrintf(CDG_INFO,"INFO: There %s %d entr%s in the beam \n", n==1?"is":"are", n, n==1?"y":"ies");
		if(newEntries)
			cdgPrintf(CDG_INFO," Best  score: %.5f \n", bScore(((BeamEntry)newEntries->item)->best->b));
		cdgPrintf(CDG_INFO," Worst score: %.5f \n", bScore(((BeamEntry)listLastElement(newEntries))->best->b));
	}
	newEntries = NULL;

	proMeasureTime(timer2, bss->profiling, " postpare");
	proMeasureTime(timer1, bss->profiling, "overall");

	return bss;
}

/**
 * same as beamSearchStep but without a previous state
 */
BeamSearchState beamSearchInitialStep(Arc arc, Boolean complete) {
	if(bsVerbosity>=1)
		cdgPrintf(CDG_INFO, "---------------------------------------------------------\n");

	List l, entries;
	int i, j, k;
	BeamSearchState bss;
	ConstraintNet cn;
	Lattice lat;
	Parse p;
	VirtualNodeSpec spec;
	LevelValue lv;
	int slot;
	ConstraintNode cnode;

	// make initial state
	bss = bssNew();
	bss->newWord = arc->word;

	Timer timer1 = timerNew();
	Timer timer2 = timerNew();

	// select the virtual nodes to be used
	if(useVirtualNodes) {
		l= inputCurrentGrammar->vnspecs;
		while(l) {
			spec = l->item;
			for(i= 0; i < spec->copies; i++) {
				vectorAddElement(bss->virtualNodeTypes, spec);
			}
			l= l->next;
		}
	}

	// make lattice
	lat = (Lattice) memMalloc(sizeof (LatticeStruct));
	lat->id = strPrintf("wordgraph0");
	lat->filename = strRegister("*BS*");
	lat->lineNo = 0;
	lat->lookahead = listNew();
	lat->arcs = listAppendElement(listNew(), arc);
	inputScanLattice(lat, complete);

	if(bsVerbosity>=3)
		cdgPrintf(CDG_INFO,"Building initial CNet\n");
	cn = cnBuild(lat, TRUE, complete);
	if(useVirtualNodes) {
		cnInitVirtualNodes(cn, inputCurrentGrammar->vnspecs, NULL, NULL);
	}
	cn = postExtendNetCleanup(cn, NULL, proNew());
	bss->net= cn;

	BeamEntry initialEntry = memMalloc(sizeof(BeamEntryStruct));
	initialEntry->best=NULL;
	initialEntry->virtualNodeAvailable = bvNew(listSize(inputCurrentGrammar->vnspecs));
	bvSetAllElements(initialEntry->virtualNodeAvailable, TRUE);
	initialEntry->virtualNodeUsed = bvNew(listSize(inputCurrentGrammar->vnspecs));
	bvSetAllElements(initialEntry->virtualNodeUsed, FALSE);
	initialEntry->previous= NULL;
	initialEntry->rank= 0;
	initialEntry->method = "initial";
	initialEntry->slots = vectorNew(10);
	initialEntry->lexemeSelection = vectorNew(10);

	p= parseNew(NULL);
	p->verticesLabels= vectorNew(10);
	p->verticesStructure= vectorNew(10);

	p->words= vectorNew(0);
//	w = newWord();
//	w->from = ln->arc->from;
//	w->to = ln->arc->to;
//	w->word = ln->lexem->word;
//	w->description = strRegister(ln->lexem->description);
//	vectorAddElement(p->words, w);

	initialEntry->p=p;

	if(useVirtualNodes) {
		// add edges for the virtual nodes to the parse
		for(i=0; i< vectorSize(bss->virtualNodeTypes); i++) {
			for(j=0; j< inputCurrentGrammar->noOfLevels; j++) {
				slot = (i +1 /*cn already extended*/ )*inputCurrentGrammar->noOfLevels +j;
				vectorAddElement(p->verticesLabels, "");
				vectorAddElement(p->verticesStructure, (void *)-1);

				// get the first compatible LV and add it to the list of slot-fillers
				cnode = vectorElement(cn->nodes, slot);
				for(k = 0; k < vectorSize(cnode->values); k++) {
					lv = vectorElement(cnode->values, k);
					if(lv->score ==0) {
						continue;
					}

					// we search a neutral attachment, i.e., one to nil via the empty label
					if(lv->regents == NULL && strcmp(lv->label, "") == 0) {
						break;
					}
					if(k== vectorSize(cnode->values)-1) {
						cdgPrintf(CDG_ERROR, "ERROR: no value for virtual node found\n");
						abort();
					}
				}
				vectorAddElement(initialEntry->slots, lv);
			}
		}
	}

	proMeasureTime(timer2, bss->profiling, "prepare");

	entries = appendArcToParse(arc, p, cn, initialEntry, listNew(), TRUE, NULL, bss->profiling);

	proMeasureTime(timer2, bss->profiling, "append");

	bss->entries = listToVector(entries);
	if(!bss->entries) {
		cdgPrintf(CDG_WARNING,"WARNING: empty beam \n");
		bss->entries= vectorNew(10);
	}

	l= entries;
	while(l) {
		BeamEntry be = l->item;
		parsePrint2(be->p, TRUE, frobIgnoreThreshold);
		cdgExecHook(HOOK_PARTIALRESULT, be->p);
		l=l->next;
	}

	if(bsVerbosity>=1) {
		int n = vectorSize(bss->entries);
		cdgPrintf(CDG_INFO,"INFO: There %s %d entr%s in the beam \n", n==1?"is":"are", n, n==1?"y":"ies");
		if(entries)
			cdgPrintf(CDG_INFO," Best  score: %.5f \n", ((BeamEntry)entries->item)->best->b->hard?0.0:((BeamEntry)entries->item)->best->b->soft);
		cdgPrintf(CDG_INFO," Worst score: %.5f \n", beamGetLowestScore(entries, 0));
	}
	entries = NULL;

	proMeasureTime(timer2, bss->profiling, "postpare");
	proMeasureTime(timer1, bss->profiling, "overall");

	return bss;
}


/*
 * this function implements the incremental beam search strategy
 */
BeamSearchState beamSearch(Lattice lat) {
	List arcs = lat->arcs;
	BeamSearchState bss;

	bss  = beamSearchInitialStep(arcs->item, (arcs->next || !incrAutoFinalize)?FALSE:TRUE);

	arcs = arcs->next;
	while( arcs ) {
		Arc arc = arcs->item;

		bss = beamSearchStep(bss ,arc, (arcs->next || !incrAutoFinalize)?FALSE:TRUE);

		arcs= arcs->next;
	}

	bssPrintDetails(bss);

	return bss;
}


/*
 * returns a string containing a sequence of 1s and 0s
 */
void bvPrint(ByteVector bv) {
	int i;
	for(i=0;i < bvSize(bv); i++) {
		cdgPrintf(CDG_INFO, "%s",bvElement(bv,i)?"1":"0");
	}
	cdgPrintf(CDG_INFO, "\n");
}

/**
 *
 */
TopDownPredictor tdpNew(String name, int edge, Boolean bottom, List labels, List specs) {
	TopDownPredictor tdp = memMalloc(sizeof(TopDownPredictorStruct));
	tdp->constraintName = name;
	tdp->edge = edge;
//	if(strcmp(bottom, "@")==0 || strcmp(bottom, "true")==0) {
//		tdp->bottom = TRUE;
//	} else {
//		tdp->bottom = FALSE;
//	}
	tdp->bottom = bottom;
	tdp->labels = labels;
	tdp->specs = specs;

	return tdp;
}

/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
