/*
 * Copyright (C) 1997-2011 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Michael Schulz
 * Birth: 1.9.98
 *
 */

#ifndef INCREMENT_H
#define INCREMENT_H

/* -------------------------------------------------------------------------
 * @addtogroup ISearch
 *
 * @{ */

#include "input.h"
#include "constraintnet.h"
#include "transform.h"
#include "parse.h"
#include "profiling.h"

extern ConstraintNet incSearch(Lattice, int, Number, Number);
extern List incGetNonSpecIncompatibles(void);


/* ---------------------------------------------------------------------- */

#define INCR_WRITE_TOKEN (1 << 0)
#define INCR_WRITE_TIME  (1 << 1)
#define INCR_WRITE_SCORE (1 << 2)
#define INCR_WRITE_UNREM (1 << 3)
#define INCR_WRITE_PATCHATTEMPTS (1 << 4)
#define INCR_WRITE_REMAINING (1 << 5)
#define INCR_WRITE_REAPPEARING_LVALS (1 << 6)
#define INCR_WRITE_REPAIRATTEMPTS (1 << 7)
#define INCR_WRITE_REAPPEARING_LEXEMES (1 << 8)
#define INCR_WRITE_STABLE_SUBTREES (1 << 9)
#define INCR_WRITE_EVALUATION_COUNT (1 << 10)
#define INCR_WRITE_AMOUNTS (1 << 11)


typedef enum { // FreezingLevelType
  FLTnone, // no freezing based processing whatsoever
  FLTsimulate, // it is just recorded, which elemets would be frozen
  FLTfull // frozen elements are initialized as deleted each step
} FreezingLevelType;

typedef enum { // FreezingHeuristic
  FHconflictFreeProjective, // a heuristic where only finished parts are frozen, but also for incoming edges
  FHradical, // everything is frozen, but only for outgoing edges
  FHnotSoRadical, // as radical, but not for the rightmost subtree
  FHlookahead // all except the last n domains will be frozen
} FreezingHeuristic;

struct IncrementalStepStruct {
	// ID
	String word; // the token added this step
	String mode; // which mode was used e.g full incremental or prefix parsing
	int iteration; // num of the iteration

	// core
	Problem problem; // associated problem for the work in progress, might be outdated later on
	Parse parse; // the resulting parse
	struct IncrementalStepStruct *previousStep; // the step before this one, NULL if this one is the first

	// heuristics
	List stableStructures; // list of incr subtrees
	List frozenLVs; // list of lvs concurrating to those in a stable substruct
	Vector frozenLvCount; // vector containing number of lvs frozen for different resons
		// 0 ^= same domain, past
		// 1 ^= target domain, past
		// 2 ^= same domain, not even built
		// 3 ^= target domain, not even built
	List frozenLexemes; // list of lexemes concurrating to those in a stable substruct

	// infos
	Number score; // final score of the solution
	Number goal; // score goal for this step
	unsigned long timeNeeded; // time in millis for the whole step
	Boolean reanalysed; // was frobbing started several times for this step?
	Boolean virtualNodeMatch; // if the new node matched on a virtual node
	Boolean killedByTimeLimit; // if the step was terminated by the time limit
	int frobbingSteps; // steps in frobbing
	int patchAttempts; // how often did frobbing attack a conflict for this step
	int successfulPatches; // how often was a analysis improved by a patch
	long evalUnary; // grabbed from problem->net
	long evalBinary; // see eval unary
	int lexemesTotal; // amount of Lexemes appearing in the corresponding problem
	int lvalsTotal; // amount of LevelValues appearing in the corresponding problem
	int nrTokens; // tokens in the problem so far

	int lvCountTotal;			/**< Counting all LVs being built */
	int lvCountBuild;			/**< Counting LVs being built and remaining in the net */
	int lvCountDeleted;       /**< Counting LVs being immediately deleted*/
	int lvCountPostponed;     /**< Counting LVs being postponed when built*/
	List repairAttempts; // some data logged during frobbing
	List unremovedConflicts; // conflicts still on the unremovable list at the end, in string form as Constraint Violations are not persistent
	List unremovedEstablishedConflicts; // unremoved conflicts, not in the rightmost subtree and accumulated from previous steps
	List deletedLVs;  // LVs deleted in a previous step
	List returnedLVs; // LVs previously deleted but used in this step
	List deletedLexemes; // lexemes previously deleted
	List returnedLexemes; // lexemes previously deleted but used in this step
	List unfrozenLVs; // list of LVs being used after being frozen
	List unfrozenLexemes; // list of Lexemes being used after being frozen
	List selectedTags; // stores the result of this steps tagging run;
	ByteVector domainFrozen; // index as in Problem->domains, frozen domain will not get added LVs
	ByteVector slotFrozen; // index as in Problem->domains, frozen slots will not get new LVs pointing towards it
			// slotFrozen is needed in addition to domain frozen, since the root node in a frozen subtree will have it's slot frozen, but not it's domain

	ProfilingObject profiling; // some timing info

};
typedef struct IncrementalStepStruct IncrementalStepStruct, *IncrementalStep;

struct IncrSubtreeStruct{
	LevelValue root;
	Boolean conflictFree;
	struct IncrSubtreeStruct *regent;
	Boolean isHead;
	Boolean acceptable;
	List underlings;
	List conflicts;
	int rightBranches;
};
typedef struct IncrSubtreeStruct IncrSubtreeStruct, *IncrSubtree;

struct ProtoEdgeStruct {
	Level level;
	GraphemNode modifier;
	GraphemNode modifiee;
	Boolean allLabels;
	List labels;
};
typedef struct ProtoEdgeStruct ProtoEdgeStruct, *ProtoEdge;

struct ChunkIncrementStruct {
	String string;
	List words;
	int length;
};
typedef struct ChunkIncrementStruct ChunkIncrementStruct, * ChunkIncrement;

/* --- Procedures -----------  */
extern IncrementalStep incrementalParsing(Lattice lat, Boolean finalize, Boolean onechunk);
extern IncrementalStep incrContinueStep(IncrementalStep step, Vector words, Boolean final);
extern void incrPrintStep(IncrementalStep step, Boolean recursively);
extern unsigned long incrPrintTimeNeeded(IncrementalStep step, Boolean printHeader);
extern void incrPrintDetails(IncrementalStep step, int flags);
extern void incrPrintFrozenStuff(IncrementalStep step);
extern void incrInitialize();
extern IncrementalStep incrNonIncremental(Lattice wholeLat, int frobMethod);
extern void incrWriteProfilingToFile(IncrementalStep step, String name, int flags);
extern void incrWriteAnnotationToFile(IncrementalStep step, Boolean recursive, String name);
extern IncrementalStep incrPrefixParsing(Lattice lat);
extern IncrementalStep incrPseudoIncrementalParsing(Lattice lat);
extern ProtoEdge incrProtoEdgeNew();
extern Boolean incrConstraintIsFluctuating(Constraint c, Boolean isNonspec);
extern void incrBuildLvsTowardsNewNode(ConstraintNet net, GraphemNode newNode, ByteVector exceptions, Vector freezeCount);
extern void pushArc(Arc arc, LexemGraph lg);
extern void extendCNIsBuild(ConstraintNet net);
extern Analysis extendAnalysis(Analysis A, Problem P, int chunkLength, Boolean match, IncrementalStep step, ProfilingObject prof);
extern void incrUpdateNumberCaches(ConstraintNode cNode);
extern ConstraintNet extendNet(ConstraintNet net, Arc arc, IncrementalStep step);
extern ConstraintNet postExtendNetCleanup(ConstraintNet net, IncrementalStep step, ProfilingObject prof);
extern void undeleteLexemes(LexemGraph lg, Boolean freeze);
extern IncrementalStep incrPurelyVirtualInitialStep();

/* --- global variables -------------------------*/

/* NULL if no Step has been created, or a pointer to the last created one */
extern IncrementalStep incrMostRecentlyCreatedIncrementalStep;
extern FreezingLevelType incrFreezeLevel;

extern List incrTemporarilyAllowedConstraints;
extern List incrTempPenalizedConstraints;
extern List incrVolatileConstraints;
extern List incrPersistentConstraints;
extern Boolean incrScoreFluctuation;
extern Boolean incrAutoFinalize;
extern Boolean incrUseFrontierPredicate;
extern Boolean incrEdgeReactivation;
extern Boolean alwaysShowAllSolutions;
extern Boolean initVNsUnused;

extern Number incrPressureRedux;
extern String labelExceptionProperty;
extern String lexemeAttributeExceptionProperty;

/* -- FLAGS ------------------------------------*/
#define INCR_SUBSTRUC_CONFLICTS   (1 << 0)
#define INCR_SUBSTRUC_NONSPEC     (1 << 1)
#define INCR_SUBSTRUC_NIL         (1 << 2)
#define INCR_SUBSTRUC_RIGHT_BRANCHES  (1 << 3)
#define INCR_SUBSTRUC_LEAVES      (1 << 4)


/** @} */
#endif /* don't insert anything after this #endif */
