/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 */

/*
 * File: parse.h
 * Author: Jens Grabarske
 * Purpose: data structures for the solution trees
 * Birth: 9/8/99
 *
 */

#ifndef PARSE_H
#define PARSE_H

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>
#include <time.h>
#include <blah.h>
#include "constraintnet.h"
#include "cdg.h"
#include "eval.h"
#include "levelvalue.h"

/* -- MACROS ------------------------------------------------------------ */
#ifndef SWIG

#define PC_SILENT     0     /* do not produce any messages      */
#define PC_SUMMARY    1     /* produce only summary report      */
#define PC_MEASURES   2     /* produce only measure information */
#define PC_ALL        3     /* produce detailed information     */

#define PC_STRICT     0     /* words not found at end of tested word list are errors */
#define PC_NONSTRICT  1     /* words not found at end of tested word list are not errors */

#endif

/* -- TYPE DEFINITIONS -------------------------------------------------- */
struct ParseStruct {
  String  id;                /* id of the parse in the list of global parses */

  String  latticeId;         /* Name of the word graph      */
  String  userName;          /* Who did the run?            */
  time_t  dateOfCreation;    /* When?                       */
  List    grammarFiles;      /* With which grammars?        */
  String  searchStrategy;    /* With which search strategy? */
  String  comment;           /* comment about the solution  */

  int     solutionNo;        /* This is solution number X   */
  int     noOfSolutions;     /* of Y                        */

  Number  score;             /* What score did it reach?    */
  Boolean optimal;           /* Is it optimal?              */
  List    violatedConstraints; /* List of violated constraints (Strings) */

  Vector  words;             /* Vector of all word nodes        */
  List    levels;            /* List of the used levels (Strings) */
  List    labels;            /* List of the used labels (Strings) */

  Vector verticesStructure;  /* holds, as an int, the elder node for each node i in level j */
  Vector    verticesLabels;  /* holds the label for every vertex (i, j) */

  long           evalUnary;       /* Number of unary evaluations         */
  long           evalBinary;      /* Number of binary evaluations        */
  unsigned long numberOfHits;     /* Score cache statistic               */
  unsigned long evaluationTime;   /* How long did the evaluation take?   */
  String        statisticComment; /* Comment about the statistic data    */

  /* The following parts are not required, but useful for many operations.
     They cannot be assumed to exist, and they are never saved or restored.
  */

  Vector LVs;                 /* holds the constituting LVs */
  LexemGraph lg;              /* lexeme graph used by the LVs */
  List violations;            /* List of full ConstraintViolation structures */
  Badness badness;            /* generalized score */
  Vector map;                 /* maps vertex no -> LV index */
  Vector rmap;                /* maps LV index -> vertex no */
};
//typedef ParseStruct *Parse;


/* The flag optimal may be set to indicate that this analysis was
   proven to be the optimal analysis as defined by the grammar at that
   time.

   The field verticesStructure holds the structural information of
   the analysis. It contains one int for each dependency edge
   (assumed to be ordered by their natural order, as defined by
   lvCompareByNaturalOrder()), with~-1 representing a NIL
   subordination, 0 representing subordination under the first word, etc.
   Note that this way of addressing a word differs from just stating its
   time point.

   The Vector verticesLabels holds the labels of all dependency
   edges, sorted by their natural order.

   The fields LVs, lg, and violations hold, respectively, all LVs that
   constitute the analysis, the underlying lexeme graph, and the actual
   ConstraintViolation structures. These fields contradict the design goal
   of being independent of internal cdg structures, but are present
   nevertheless to make life easier for those applications that do know
   them. For instance, it is straightforward to re-apply the current
   constraint grammar to a Parse if its LVs are known, but rather hairy
   otherwise. However, these fields are strictly optional, i.e. no
   function may blindly assume that it they are non-NULL unless it has
   created them itself, and their contents are never stored or retrieved
   from disk. A Parse which has these fields set is called `decorated'
   (antonym: `abstract'). Also, the optional fields in a decorated
   Parse are totally owned by the Parse, even the LVs and the LexicalEntry
   structures within them. This is to make the Parse independent of the
   lifetime of any structures that may have produced it, such as a
   ConstraintNet. A Parse may thus live as long as the grammar that
   produced it is loaded.

   Here is how to produce a decorated Parse properly:

   - Call parseNew().
   - Specify the structure of the Parse by setting the
     corresponding fields. These are the fields
     words, levels, labels, verticesLabels
     and verticesStructure. An easy way to do this is to install
     all LVs that describe the Parse in a Vector in their
     natural order, and then to call parseComputeStructure().
   - Call parseDecorate() with the lattice that ultimately
     underlies the Parse. This will fill the optional fields with
     newly allocated structures owned by the Parse and no one
     else.
   - Set other fields as desired. Functions such as
     parseEval() can be helpful here.

*/


typedef struct
{
  int     from;           /* Interval in the                       */
  int     to;             /* lattice graph                         */

  String  word;           /* Which word does this node represent?  */
  String  description;    /* == description of the currently corresponding LexiconItem */
} WordStruct;
typedef WordStruct *Word;

typedef struct {
  int errorextrawordref;  /* nr of errors: extra word in reference */
  int errorextrawordtest; /* nr of errors: extra word in test */
  int errorwordlex;       /* nr of errors: words have different lexical readings */
  int errorgovword;       /* nr of errors: governors are different words */
  int errorgovlex;        /* nr of errors: governors have different lexical readings */
  int erroredgelabel;     /* nr of errors: edges have different label */
  int matchedgebase;      /* nr of matches: edges are identical (not counting gov's) */
  int matchedgegov;       /* nr of matches: edges are identical (same gov's words) */
  int matchedgetot;       /* nr of matches: edges are identical (same gov's read.) */
  int totalextrawordref;  /* nr of checks: extra word in reference */
  int totalextrawordtest; /* nr of checks: extra word in test */
  int totalwordlex;       /* nr of checks: words have different lexical readings */
  int totalgovword;       /* nr of checks: governors are different words */
  int totalgovlex;        /* nr of checks: governors have different lexical readings */
  int totaledgelabel;     /* nr of checks: edges have different label */
  int totaledgebase;      /* nr of checks: edges are identical (not counting gov's) */
  int totaledgegov;       /* nr of checks: edges are identical (same gov's words) */
  int totaledgetot;       /* nr of checks: edges are identical (same gov's read.) */
  int totaledgetest;      /* total number of edges in test      */
  int totaledgeref;       /* total number of edges in reference */
  float accuracy;
  float recall;
  float fmeasure;
} ParseCmpStatStruct;
typedef ParseCmpStatStruct *ParseCmpStat;

typedef struct {
  Parse parse;		     /* the parse */
  AnnoEntry annoEntry;	     /* the annotation */
  Hashtable tries;	     /* maps levels to tries */
  Hashtable structHits;	     /* maps levels to strucural hits */
  Hashtable labelledHits;    /* maps levels to labelled hits */
  Hashtable lexicalHits;     /* maps levels to lexical hits */
  Hashtable strictHits;	     /* maps levels to strict hits */
  List errors;		     /* list of ParseError objects */

} ParseVerificationStruct;
typedef ParseVerificationStruct *ParseVerification;

typedef enum {
  PEWord,		     /* word missing */
  PELevel,		     /* level missing */
  PEStruct,		     /* structural error */
  PELabel,		     /* label error */
  PELexem,		     /* lexem error, i.e. a wrong or missing
			      * feature */
} ParseErrorType;

typedef struct {
  ParseErrorType type;	     /* type of error */
  union {
    String level;	     /* missing level */
    String word;	     /* missing word */
    struct {
      String level;	     /* level where the error occured */
      int modifier;	     /* position */
      String label;	     /* label */
      int modifiee;	     /* governor */
      String annoLabel;      /* might be wrong too but not reported as label error */
      int annoModifiee;      /* should be this modifiee */
    } structError;	     /* PEStruct */
    struct {
      String level;	     /* level where the error occured */
      int modifier;	     /* position */
      String label;	     /* label */
      int modifiee;	     /* governor */
      String annoLabel;      /* should be this label */
    } labelError;	     /* PELabel */
    struct {
      int modifier;	     /* position */
      String feature;	     /* name of lexical feature */
      String value;	     /* value of feature, possibly NULL if missing */
      String annoValue;      /* should be this value */
    } lexError;		     /* PELexem */
  } data;
} ParseErrorStruct;
typedef ParseErrorStruct *ParseError;

/* -- FUNCTIONS --------------------------------------------------------- */

/* NULL if no parse has been created, or a pointer to the last one */
extern Parse parseMostRecentlyCreatedParse;

extern void parseSwapLabel(Parse parse, int where, String label, Boolean eval);
extern void parseShiftEdge(Parse parse, int where, int to, Boolean eval);
extern void parseSwapWord(Parse parse, int where, String id, Boolean eval);
extern Boolean parseIsAbstract(Parse parse);
extern Parse parseFind(String id);
extern int parseHeight(Parse parse, String levelname);
extern int parseWordHeight(Parse p, String levelname, int wordNo);
extern int parseWordDepth(Parse p, Level level, int wordNo);
extern String parseDate(Parse parse);
extern LevelValue parseGetLV(Parse parse, int where);
extern List parseGetLexemNodes(Parse parse);
extern List parseGetModifiers(Parse parse, Level level, int wordIndex);
extern Boolean parseDecorate(Parse p, LexemGraph lg, AnnoEntry ae);
extern void parseEval(Parse parse);
extern AnnoEntry parse2annotation(Parse parse);
extern Parse parseFromAnno(AnnoEntry ae);
extern void parseRegister(Parse p);
extern void parseMirror(Parse p, Boolean eval);
extern int parseMap(Parse p, int lvindex);
extern int parseRmap(Parse p, int lvindex);
extern ParseVerification parseVerify(Parse p, AnnoEntry ae);
extern void parseVerificationDelete(ParseVerification pv);
extern int parseIndex(Parse p, String levelname, int wordno);
extern LexiconItem anno2le(Annotation an);
extern String parseOptimizeWord(Parse p, int wordno);
extern String parseOptimizeLabel(Parse p, String levelname, int wordno);
extern int parseOptimizeStructure(Parse p, String levelname, int wordno);
extern LexiconItem parseLexiconItem(Parse p, int i);
extern GraphemNode parseGraphemeNode(Parse p, int i);
extern String parseFormatValue(Parse p, int i);

// edge marking - marked edges are considered to be root edges
extern Boolean parseMarkEdge(Parse p, String levelname, int wordno, Boolean mark);
extern Boolean parseIsMarked(Parse p, String levelname, int wordno);
extern Boolean parseMarkOneEdgePerCycle(Parse p, String levelname);
extern void parseClearAllMarks(Parse p, String levelname);
extern void parseMakeMarkedEdgesRootEdges(Parse p, String levelname, Boolean eval);

#ifndef SWIG
extern Parse parseClone(Parse parse, Boolean decorate);
extern void parseReEval(Parse parse, List changes);
extern int parseCompare(Parse ref, Parse test, ParseCmpStat stat, int verbosity, int strict);
       /* for verbosity and strict arguments see macros above */
extern Boolean parseSameVertices(Parse a, Parse b);
extern Boolean parseEquals(Parse a, Parse b);
extern Parse parseNew(String id);
extern void parseDelete(Parse p);
extern Word newWord();
extern void deleteWord(Word w);
extern List parseFindLn(Parse p, Word word, AnnoEntry ae);
extern void parseComputeStructure(Parse p, Vector LVs, ByteVector isDeleted);
extern void parseStrip(Parse p);
extern void parsePrint2(Parse parse, Boolean skipUninteresting, float threshold);
extern void parsePrint(Parse p);
extern void parseFinalize();
extern Boolean parseIsUnderspecified(Parse p);
extern Parse parseFromLVs(ConstraintNet net, Vector lvs, String strategy);
extern int parseMinFrom(Parse p);
extern int parseMaxTo(Parse p);
extern void parseInitialize(void);
extern void parseFinalize(void);
extern ParseVerification parseVerificationNew(Parse p);
extern ParseError parseErrorNew(ParseErrorType);
extern void parseErrorDelete(ParseError);
#endif

extern Boolean pIsCompatibleLV(Parse p, LevelValue lv);

/* -- VARIABLES --------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
#endif
