/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 */

/*
 * File: predicates.c
 * Author: Ingo Schroeder
 * Purpose: CDG predicates
 * Birth: 17/3/97
 *
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "cdg.h"
#include "eval.h"
#include "predicates.h"
#include "increment.h"
#include "hook.h"
#include "chunker.h"
#include "virtualnode.h"

/* -- MACROS ------------------------------------------------------------ */
/* -- PROTOTYPES -------------------------------------------------------- */
/*
Boolean predNthCousin(List args, LexemGraph lg);
Boolean predAncestor(List args, LexemGraph lg);
*/
/* -- VARIABLES --------------------------------------------------------- */
CDGPredicateStruct eval_predicates[] =
  {
    {"start", predStart, "predStart", FALSE, "1",
     "returns TRUE if lattice starts with word"},
    {"stop", predStop, "predStop", FALSE, "1",
     "returns TRUE if lattice ends with word"},
    {"frontier", predFrontier, "predFrontier", FALSE, "2",
     "returns TRUE if the given word (arg #1) is one of the last n (arg #2)"},
    {"root", predRoot, "predRoot", FALSE, "1",
     "returns TRUE if argument is the root"},
    {"spec", predSpec, "predSpec", FALSE, "1",
     "returns TRUE if argument is specific"},
    {"virtual", predVirtual, "predVirtual", FALSE, "2",
     "returns TRUE if argument is virtual, second argument is the reason for this check, so specific checks can be disabled"},
    {"nonspec", predNonSpec, "predNonSpec", FALSE, "1",
     "returns TRUE if argument is nonspecific"},
    {"print", predPrint, "predPrint", FALSE, ">-1",
     "prints string or number argument and returns TRUE"},
    {"subsumes", predSubsumes, "predSubsumes", FALSE, "3",
     "returns TRUE if 2nd arguments subsumes 3rd"},
    {"exists", predExists, "predExists", FALSE, "1",
     "returns TRUE if feature exists in lexical entry"},
    {"compatible", predCompatible, "predCompatible", FALSE, "3",
     "equivalent to `( subsumes(a, b, c) | subsumes(a, c, b) )'"},
    {"chunk_head", predChunkHead, "predChunkHead", FALSE, "1",
     "returns TRUE if lattice is a chunk head"},
    {"occur", predOccur, "predOccur", FALSE, "1",
     "returns TRUE if the specified word form occurs in the sentence somewhere"},
    {"has", predHas, "predHas", TRUE, "2-6",
     "returns TRUE if the LV has a child edge with the specified label"},
    {"is", predIs, "predIs", TRUE, "2-4",
     "returns TRUE if the word is subordinated somewhere with the specified label"},
    {"cyclic", predCycle, "predCycle", TRUE, "1",
     "returns TRUE if the named level is cyclical"},
    {"between", predBetween, "predBetween", FALSE, "3",
     "test if one of the specified punctuation marks separates the two words"},
    {"under", predUnder, "predUnder", TRUE, "2-3",
     "test if word A is indirectly a modifier of word B"},
    {"connected", predConnected, "predConnected", TRUE, "2",
     "test if A and B have the same root"},
    {"partial", predPartial, "predPartial", FALSE, "0",
     "returns TRUE, if the current input is not complete yet, i.e., partial"},
    {"samevnspec", predSameVnSpec, "predSameVnSpec", FALSE, "2",
     "returns TRUE if A and B are both virtual words and originate from the same VN specification"},
    /*
      {"nth_cousin", predNthCousin, "are there common ancestors not more then n levels back?"},
      {"ancestor_at_level", predAncestor, "are there common ancestors at level x or later?"},
    */
    {(String) NULL, (BooleanFunction *) NULL, (String)NULL, FALSE, (String) NULL}
  };

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * return a pointer to the function implementing a functor
 */
CDGPredicate predGetFunction(String functor)
{
  int i;

  for (i = 0; eval_predicates[i].functor != NULL; i++) {
    if (strcmp(functor, eval_predicates[i].functor) == 0) {
      return &eval_predicates[i];
      break;
    }
  }

  return (CDGPredicate)NULL;
}

/* ----------------------------------------------------------------------
 * return the name of the function implementing a functor
 */
String predGetFunctionName(String functor)
{
  int i;

  for (i = 0; eval_predicates[i].functor != NULL; i++) {
    if (strcmp(functor, eval_predicates[i].functor) == 0) {
      return eval_predicates[i].functionName;
      break;
    }
  }

  return NULL;
}

/* ----------------------------------------------------------------------
 * is #1 a start node?
 */
Boolean predStart(List a, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `start' not a lexeme node in cosntraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  if (!spec(val->data.lexemnode)) {
    return FALSE;
  }

  /* detect word that is truly the start */
  if (((LexemNode) val->data.lexemnode)->arc->from == lg->min) {
    return TRUE;
  }

  /* detect word that is the start of an embedded clause */
  if (inputSentenceBoundary(inputLatticePrevArc(lg->lattice,
						((LexemNode) val->data.
						 lexemnode)->arc))) {
    return TRUE;
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
 * is #1 an end node? (not counting virtual nodes)
 */
Boolean predStop(List a, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `stop' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  if(!spec(val->data.lexemnode) || ((LexemNode) val->data.lexemnode)->grapheme->isVirtual ) {
	  return FALSE;
  }

  if(useVirtualNodes && lg->ve ) {
	  return ((LexemNode) val->data.lexemnode)->arc->to == lg->ve->vStartID;  // -1 for last nonvirtual, +1 for to instead of from
  } else {
	  return ((LexemNode) val->data.lexemnode)->arc->to == lg->max;
  }
}

/* ----------------------------------------------------------------------
 * is #1 one of the last #2th nodes (not counting virtual nodes)?
 */
Boolean predFrontier(List a, LexemGraph lg, Vector context)
{
  if(!incrUseFrontierPredicate) {
	  // the idea here is, that virtual nodes need conflicts as hints,
	  // the frontier predicate is used for suppressing them at the right frontier,
	  // so the predicate can be deactivated
	  return FALSE;
  }

  if(lg->complete) {
	  // in complete input, there is no frontier zone where waiting might be the right choice of action
	  return FALSE;
  }

  ValueStruct vas,vas2;
  Value val,val2;
  LexemNode ln;
  int n;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: 1st argument of predicate `frontier' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  val2 = evalTerm((Term) a->next->item, &vas2, lg, context);
  if (val2->type != VTNumber) {
	  cdgPrintf(CDG_WARNING,
			  "WARNING: 2nd argument of predicate `frontier' not a number in constraint `%s'\n",
			  evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
	  return (FALSE);
  }

  ln= val->data.lexemnode;
  n=  floor(val2->data.number);


  if(!spec(ln) || ln->grapheme->isVirtual ) {
	  return FALSE;
  }

  if(useVirtualNodes && lg->ve) {
    return ln->arc->to + n > lg->ve->vStartID;  // -1 for last nonvirtual, +1 for to instead of from
  } else {
    return ln->arc->to + n > lg->max;
  }

  //return ln->arc->to + n > lg->max;
}

/*
 * are #1 and 2# based on the same virtual node specification?
 */
Boolean predSameVnSpec(List args, LexemGraph lg, Vector context) {
	if(!useVirtualNodes) {
		return FALSE;
	}

	Value val, val2;
	ValueStruct vas, vas2;
	LexemNode ln, ln2;

	val = evalTerm((Term) args->item, &vas, lg, context);
	val2= evalTerm((Term) args->next->item, &vas2, lg, context);

	if (val->type != VTLexemNode) {
		cdgPrintf(CDG_WARNING,
			  "WARNING: 1st argument of predicate `sameVnSpec' not a lexeme node in constraint `%s'\n",
			  evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
		return (FALSE);
	}

	if (val2->type != VTLexemNode) {
		cdgPrintf(CDG_WARNING,
			  "WARNING: 2nd argument of predicate `sameVnSpec' not a lexeme node in constraint `%s'\n",
			  evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
		return (FALSE);
	}

	ln = val->data.lexemnode;
	ln2= val2->data.lexemnode;

	if(!ln->grapheme->isVirtual || !ln2->grapheme->isVirtual) {
		return FALSE;
	}

	if(veGetVirtualNode(lg->ve, ln->lexem->word)->spec == veGetVirtualNode(lg->ve, ln2->lexem->word)->spec) {
		return TRUE;
	}

	return FALSE;

}

/* ----------------------------------------------------------------------
 * is #1 a specified node?
 */
Boolean predSpec(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;

  val = evalTerm((Term) a->item, &vas, lg, context);

  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `spec' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  return (spec(val->data.lexemnode));
}

/* ----------------------------------------------------------------------
 * is #1 a specified node?
 * #2 is a tag giving the reason for the virtual check.
 * This way checks with a certain tag can be deactivated.
 * For example virtual nodes can be ordered or unordered,
 * so respective exceptions in the grammar have to be toggled on or off.
 */
Boolean predVirtual(List a, LexemGraph lg, Vector context)
{
  Value val, val2;
  ValueStruct vas, vas2;

  val = evalTerm((Term) a->item      , &vas , lg, context);
  val2= evalTerm((Term) a->next->item, &vas2, lg, context);

  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
    		"WARNING: 1st argument of predicate `virtual' not a lexeme node in constraint `%s'\n",
    		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  if (val2->type != VTString) {
	  cdgPrintf(CDG_WARNING,
			  "WARNING: 2nd argument of predicate `virtual' not a string in constraint `%s'\n",
			  evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
	  return (FALSE);
  }

  LexemNode ln = val->data.lexemnode;
  String s = val2->data.string;
  if(!spec(ln)) {
	  return FALSE;
  }

  if( orderVirtualNodes && strcmp(s, "order") == 0 ) {
	  return FALSE;
  }

  return (gnIsVirtual(ln->grapheme));
}

/* ----------------------------------------------------------------------
 * is #1 an underspecified node?
 */
Boolean predNonSpec(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `nonspec' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  return (nonspec(val->data.lexemnode));
}

/* ----------------------------------------------------------------------
 * is #1 the root node?
 */
Boolean predRoot(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
     cdgPrintf(CDG_WARNING,
        "WARNING: argument of predicate `root' not a lexeme node in constraint `%s'\n",
	evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
     return (FALSE);
  }

  /* tv->lexemnode is NULL if and only if X^id points to the root */
  return (val->data.lexemnode == NULL);
}

/* ----------------------------------------------------------------------
 * is #1 a chunk head?
 */
Boolean predChunkHead(List a, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;
  Chunk chunk;
  LexemNode ln;

  val = evalTerm((Term) a->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `chunkhead' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return FALSE;
  }

  if (!spec(val->data.lexemnode)) {
    return FALSE;
  }

  ln = (LexemNode) val->data.lexemnode;
  chunk = ln->grapheme->chunk;
  if (!chunk) {
    return FALSE;
  } else {
    return (chunk->head == ((LexemNode) val->data.lexemnode)->grapheme);
  }
}

/* ----------------------------------------------------------------------
 * print #1
 */
Boolean predPrint(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;
  LexemNode ln;

  while (a != NULL && a->item != NULL) {
    val = evalTerm((Term) a->item, &vas, lg, context);
    switch (val->type) {
      case VTString:
	cdgPrintf(CDG_DEFAULT, "%s", val->data.string);
	break;
      case VTNumber:
	cdgPrintf(CDG_DEFAULT, "%f", val->data.number);
	break;
      case VTLexemNode:
	ln = (LexemNode) val->data.lexemnode;
	if (!ln) {
	  cdgPrintf(CDG_DEFAULT, "NIL");
	} else if (nonspec(ln)) {
	  cdgPrintf(CDG_DEFAULT, "NONSPEC");
	} else {
	  cdgPrintf(CDG_DEFAULT, "%s@%p",
		    ((LexemNode) val->data.lexemnode)->arc->word,
		    val->data.lexemnode);
	}
	break;
      default:
	cdgPrintf(CDG_WARNING,
		  "WARNING: type mismatch for predicate `print' in constraint `%s'\n",
		  evalCurrentConstraint ? evalCurrentConstraint->
		  id : "*UNKNOWN*");
	return (FALSE);
    }
    a = a->next;
  }

  return (TRUE);
}

/* ----------------------------------------------------------------------
 * Does #2 subsume #3 in hierarchy #1?
 */
Boolean predSubsumes(List l, LexemGraph lg, Vector context)
{
  ValueStruct vsa, vsb, vsh;
  Value va, vb, vh;

  Hierarchy h;
  int a, b;
  String constraint =
    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";

  vh = evalTerm((Term) listNthElement(l, 1), &vsh, lg, context);
  if (vh->type != VTString) {
    if (!evalSloppySubsumesWarnings  && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 1st arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }
  va = evalTerm((Term) listNthElement(l, 2), &vsa, lg, context);
  if (va->type != VTString) {
    if (!evalSloppySubsumesWarnings && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 2nd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }
  vb = evalTerm((Term) listNthElement(l, 3), &vsb, lg, context);
  if (vb->type != VTString) {
    if (!evalSloppySubsumesWarnings && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 3rd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }

  /* get hierarchy */
  if (NULL == (h = findHierarchy(vh->data.string))) {
    return FALSE;
  }

  /* find both sorts */
  if (0 > (a = (int)hashGet(h->types, va->data.string) - 1)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
	      va->data.string, vh->data.string, constraint);

    return (FALSE);
  }
  if (0 > (b = (int)hashGet(h->types, vb->data.string) - 1)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
	      vb->data.string, vh->data.string, constraint);

    return (FALSE);
  }

  /*
   * if ( a == b || (int)arrayElement( h->distance, a, b ) > 0 ) {
   * cdgPrintf(CDG_WARNING, "WARNING: predicate `subsumes' true for %s and
   * %s\n", tva->string, tvb->string); } */

  /* return canned answer */
  return ((int)arrayElement(h->distance, a, b) >= 0);
}

/* ----------------------------------------------------------------------
 * Does #2 subsume #3, or vice versa, in #1?
 */
Boolean predCompatible(List l, LexemGraph lg, Vector context)
{
  ValueStruct vhs, vas, vbs;
  Value vh, va, vb;
  Hierarchy h;
  int a, b;
  String constraint =
    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";

  vh = evalTerm((Term) listNthElement(l, 1), &vhs, lg, context);
  if (vh->type != VTString) {
    if(!evalSloppySubsumesWarnings && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 1st arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }
  va = evalTerm((Term) listNthElement(l, 2), &vas, lg, context);
  if (va->type != VTString) {
    if(!evalSloppySubsumesWarnings && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 2nd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }
  vb = evalTerm((Term) listNthElement(l, 3), &vbs, lg, context);
  if (vb->type != VTString) {
    if(!evalSloppySubsumesWarnings && !evalSuppressWarnings) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid 3rd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n", constraint);
      evalInvalidArgEncountered = TRUE;
    }
    return FALSE;
  }

  h = findHierarchy(vh->data.string);
  if (!h) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: hierarchy `%s' not found\n", vh->data.string);
    return (FALSE);
  }

  /* find both sorts */
  if (0 > (a = (int)hashGet(h->types, va->data.string) - 1)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
	      va->data.string, vh->data.string,
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }
  if (0 > (b = (int)hashGet(h->types, vb->data.string) - 1)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
	      vb->data.string, vh->data.string,
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  return ((int)arrayElement(h->distance, a, b) >= 0 ||
	  (int)arrayElement(h->distance, b, a) >= 0);
}

/* ----------------------------------------------------------------------
 * Is #1 defined?
 */
Boolean predExists(List l, LexemGraph lg, Vector context)
{
  Term t;
  Value val;
  ValueStruct vas;

  t = (Term) l->item;
  if (t->type != TTTopPeek && t->type != TTBottomPeek) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: predicate `exists' needs a lexical access in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return (FALSE);
  }

  val = evalTerm(t, &vas, lg, context);

  return (val->type != VTError);
}

/* ----------------------------------------------------------------------
   Is the edge LV a direct or indirect descendant of LN?

   Used by predHas() in recursive mode.
*/
Boolean descendant(LevelValue lv, LexemNode ln,
                   Vector context, int counter,
                   Hierarchy h, String s) {

  Boolean result = FALSE;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  LevelValue parent;
  /* break out of loops */
  if(counter > ln->lexemgraph->max) {
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "l");
#endif
    return FALSE;
  }

  /* answer FALSE for NIL LVs */
  if(!spec(lv->regents)) {
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "n");
#endif
    return FALSE;
  }

  /* reject wrong labels */
  if(counter && s && !inputSubsumes(h, s, lv->label)) {
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "s`%s'", lv->label);
#endif
    return FALSE;
  }

  /* detect trivial success */
  if(((LexemNode)(lv->regents->item))->arc->from == ln->arc->from) {
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "y");
#endif
    return TRUE;
  }

  /* maybe call self */
  parent =
    vectorElement(context,
                    nrlevels * ((LexemNode)(lv->regents->item))->arc->from + lv->level->no);
  if(parent) {
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "c");
#endif
    result = descendant(parent, ln, context, counter+1, h, s);
  }

  return result;

}
/* ----------------------------------------------------------------------
   Does the LV #1 have a daughter edge with label #2?

   `has' checks whether one of the edges below the specified lexeme node
   has particular properties.

   If there are two arguments, the first is the lexeme node and the second
   is a label that must be present.

   If the second argument is the name of a unary constraint, its
   formula will be used for matching instead of the edges' labels.

   If there is a third argument, the third is a hierarchy which should be
   used for label matching. Thus, a constraint can say `has(X@id,
   OBJA_OBJC, Labels)', and (assuming a fitting `Labels' hierarchy) either
   OBJA or OBJC will satisfy the predicate.

   If there is a fourth argument, the search is conducted recursively. The
   value of the argument is a node in the previously specified hierarchy.
   This node subsumes all the labels which are *allowed* on the path
   through the tree. So to express that anything which has a relative
   pronoun in its scope is a relative clause, you can say

     X@cat = vfin & has(X@id, find_prel, Labels, AUX)
     ->
     X.label = REL

   This constraint forces a verb to be a relative clause if one word in its
   VP has a relative pronoun as a subordinate, but not if a verb in a
   subordinated clause has a relative pronoun, because then at least one
   label other than `AUX' will intervene.

   If there is a fifth and sixth argument, they must be numbers which then
   constrain the range of time points to search. All words that start
   outside that exclusive range are not even considered. A value of -1
   indicates `no restriction' for the corresponding side.

*/
Boolean predHas(List l, LexemGraph lg, Vector context) {
  String label = NULL;
  String hname = NULL;
  String labels = NULL;
  Hierarchy h = NULL;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  LexemNode ln;
  ValueStruct vas;
  Value val;
  int cache_index = -1;
  List args;
  Constraint c = NULL;
  Boolean recursive = FALSE;
  int levelno;
  Boolean result = FALSE;
  int min = lg->min, max = lg->max;
  void *implArgs[11];

  /*
    The fifth and sixth parameter limit the range of the has() loop.
    Since they can be specified through terms such as X@from, which
    vary between invocations of the same has() formula *even when the
    tree does not change*, we cannot cache the result of such
    invocations.
  */
  Boolean use_cache = lock_counter > 0 && 5 > listSize(l);

  /* if there is no context at all, we can only assume that
     the edge might be found if we had it. */
  if(!context) {
    return TRUE;
  }

  args = l;

  Term t = (Term) args->item;
  if(TTTopPeek != t->type && TTBottomPeek != t->type) {
    cdgPrintf(CDG_ERROR, "ERROR: 1st argument of predicate `has' is not a lexeme node access!\n");
    return TRUE;
  }
  VarInfo vi = t->data.peek.varinfo;
  levelno = vi->level->no;

  val = evalTerm((Term) args->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `has' not a lexeme node in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if(!spec(ln)) {
      if(!(NONSPEC == ln && skipNonspecFailure)) {
          cdgPrintf(CDG_ERROR, "WARNING: lexeme node specified by term %s in constraint `%s' not specified when applied to edge ",
                    formatTerm((Term) args->item, FALSE), evalCurrentConstraint->id);
          lvPrint(CDG_ERROR, (LevelValue)vi->levelvalue, NULL, 0);
          cdgPrintf(CDG_ERROR, "\n");
      }
    return FALSE;
  }

  /* read cache */
  if(use_cache) {
    int result;
    cache_index = evalCurrentFormula->index * lock_width + ln->arc->from;
    result = has_cache[cache_index];
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "has_cache[%d] == %d\n", cache_index, result);
#endif
    if(2 == result) {
      return TRUE;
    }
    if(1 == result) {
      return FALSE;
    }
  }

  /* read second parameter */
  args = args->next;
  val = evalTerm((Term) args->item, &vas, lg, context);
  if(val->type == VTString) {
    label = val->data.string;
  }
  /* complain */
  else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2rd arg) for `has' in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }

#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, "DEBUG: has(`%s', `%s'", ln->lexem->word, label);
#endif

  /* if the string is the name of a unary constraint,
     use this constraint instead of label match  */
  c = hashGet(inputCurrentGrammar->constraints, label);
  if(c && 1 != listSize(c->vars)) {
    c = NULL;
  }

  /* read third parameter and find Hierarchy to use */
  if(NULL != (args = args->next)) {

    val = evalTerm((Term) args->item, &vas, lg, context);
    if(val->type == VTString) {
      hname = val->data.string;
    } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (3rd arg) for `has' in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }
    if(!(h = findHierarchy(hname))) {
      cdgPrintf(CDG_WARNING,
                "WARNING: hierarchy `%s' not found in constraint `%s'\n", hname,
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return FALSE;
    }
#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ", `%s'", hname);
#endif
  }


  /* read fourth parameter */
  if(args && (args = args->next)) {
    recursive = TRUE;
    val = evalTerm((Term) args->item, &vas, lg, context);
    if(val->type == VTString) {
      labels = val->data.string;
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (4th arg) for `has' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }

#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ", `%s'", labels);
#endif

  }

  /* read range parameters */
  if(args && (args = args->next)) {
    val = evalTerm((Term) args->item, &vas, lg, context);
    if(val->type == VTNumber) {
      if(val->data.number > 0) {
        min = val->data.number;
      }
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (5th arg) for `has' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return FALSE;
    }

#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, ", %.0f", val->data.number);
#endif
  }

  if(args && (args = args->next)) {
    val = evalTerm((Term) args->item, &vas, lg, context);
    if(val->type == VTNumber) {
      if(-1 != val->data.number &&
         val->data.number < lg->max) {
        max = val->data.number;
      }
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (5th arg) for `has' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return FALSE;
    }

#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, ", %.0f", val->data.number);
#endif
  }


#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ")? ");
#endif

  /* look for a suitable child edge */
  implArgs[0] = ln;
  implArgs[1] = label;
  implArgs[2] = c;
  implArgs[3] = h;
  implArgs[4] = labels;
  implArgs[5] = &recursive;
  implArgs[6] = &min;
  implArgs[7] = &max;
  implArgs[8] = context;
  implArgs[9] = &nrlevels;
  implArgs[10] = &levelno;

  result = predHasImpl(implArgs);

  /* write cache */
  if(use_cache) {

#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "has_cache[%d] = %d\n",
              cache_index, 1 + (result == TRUE));
#endif

    has_cache[cache_index] = 1 + (result == TRUE);
  }

  return result;

}

Boolean predHasImpl(void **args) {
  LexemNode ln = (LexemNode)args[0];
  String label = (String)args[1];
  Constraint c = (Constraint)args[2];
  Hierarchy h = (Hierarchy)args[3];
  String labels = (String)args[4];
  Boolean recursive = *(Boolean*)args[5];
  int min = *(int*)args[6];
  int max = *(int*)args[7];
  Vector context = (Vector)args[8];
  int nrlevels = *(int*)args[9];
  int levelno = *(int*)args[10];

  Boolean result = FALSE;
  int i;
  LevelValue lv;

#ifdef DEBUGHAS
  if (evalCurrentConstraint->no == 545) { /* constraint nummer anpassen!! */
    if ((evalEvaluationMethod == EMTInterpreted)) {
      cdgPrintf(CDG_DEBUG, "predHasImpl() called in interpreted mode in constraint %s\n", evalCurrentConstraint->id);
    } else if ((evalEvaluationMethod == EMTCompiled)) {
      cdgPrintf(CDG_DEBUG, "predHasImpl() called in compiled mode in constraint %s\n", evalCurrentConstraint->id);
    }
    cdgPrintf(CDG_DEBUG, "arguments are:\n");
    cdgPrintf(CDG_DEBUG, "ln = %d\n", ln->no);
    cdgPrintf(CDG_DEBUG, "label = %s\n", label);
    cdgPrintf(CDG_DEBUG, "c = %s\n", c ? c->id : "NULL");
    cdgPrintf(CDG_DEBUG, "h = %s\n", h ? h->id : "NULL");
    cdgPrintf(CDG_DEBUG, "labels = %s\n", labels);
    cdgPrintf(CDG_DEBUG, "recursive = %s\n", recursive ? "TRUE" : "FALSE");
    cdgPrintf(CDG_DEBUG, "min = %d\n", min);
    cdgPrintf(CDG_DEBUG, "max = %d\n", max);
    cdgPrintf(CDG_DEBUG, "context = %s\n", context ? "SET" : "UNSET");
    cdgPrintf(CDG_DEBUG, "nrlevels = %d\n", nrlevels);
    cdgPrintf(CDG_DEBUG, "levelno = %d\n", levelno);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif


  for(i = min; i <= max; i++) {
    int index = nrlevels * i + levelno;
    if(index >= vectorSize(context)) {
      break;
    }
    lv = vectorElement(context, index);

#ifdef DEBUGHAS
    if(lv) cdgPrintf(CDG_DEBUG, "@%d ", lvIndex(lv));
#endif

    if(!lv || !spec(lv->regents)) continue;
    if((!recursive && ((LexemNode)(lv->regents->item))->arc->from == ln->arc->from) ||
       (recursive && descendant(lv, ln, context, 0, h, labels))) {

      if(c) {

        /* call constraint */
        Constraint thisConstraint = evalCurrentConstraint;
        if(evalUnaryConstraint(c, NULL, context, lv)) {
          evalCurrentConstraint = thisConstraint;
          result = TRUE;
#ifdef DEBUGHAS
          cdgPrintf(CDG_DEBUG, " ...yes (`%s').\n",
                    ((LexemNode)lv->dependents->item)->arc->word);
#endif
          break;
        }
        evalCurrentConstraint = thisConstraint;

      } else {

        /* compare label */
        if((!h && lv->label == label)
           ||
           (h && inputSubsumes(h, label, lv->label))) {
          result = TRUE;
#ifdef DEBUGHAS
          cdgPrintf(CDG_DEBUG, " ...yes (`%s').\n",
                    ((LexemNode)lv->dependents->item)->arc->word);
#endif
            break;
        }
      }
    }
  }
#ifdef DEBUGHAS
  if(!result) {
    cdgPrintf(CDG_DEBUG, " ...no.\n");
  }
#endif
  return result;
}

/* ----------------------------------------------------------------------
   Is the word #1 subordinated with label #2?

   `is' checks whether the specified word is subordinated with the
   specified label.

   The first parameter is a lexeme node in the graph, and
   the second parameter is the label that it should carry.

   If the second argument is the name of a unary constraint, its
   formula will be used for matching instead of the edges' labels.

   If there is a third argument, the third is a hierarchy which should be
   used for label matching. Thus, a constraint can say `is(X@id,
   OBJC_NEB, Labels)', and (assuming a fitting `Labels' hierarchy) either
   OBJC or NEB will satisfy the predicate.

   If there is a fourth argument, the search is conducted recursively. The
   value of the argument is a node in the previously specified hierarchy.
   All the labels which are subsumed by this node are skipped during the
   search (unless a NIL edge interferes). In other words, specifying four
   parameters effectively asks, `Is word #1 *ultimately* labelled #2?',
   where the fourth parameter specifies which intermediate dependencies
   should be considered irrelevant.
*/
Boolean predIs(List l, LexemGraph lg, Vector context) {

  Boolean result;
  LevelValue lv;
  String label = NULL;
  Constraint c;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int i;
  int levelno = -1;
  LexemNode ln;
  ValueStruct vas;
  Value val;
  String hname;
  Hierarchy h = NULL;
  String labels = NULL;
  int counter;

  /* if there is no context at all, we can only assume that
     the edge might be found if we had it. */
  if(!context) {
    return TRUE;
  }


  /* use the level mentioned in the variable declaration */
  Term t = (Term) l->item;
  if(TTTopPeek != t->type && TTBottomPeek != t->type) {
    cdgPrintf(CDG_ERROR, "ERROR: 1st argument of predicate `is' is not a lexeme node access!\n");
    return TRUE;
  }
  VarInfo vi = t->data.peek.varinfo;
  levelno = vi->level->no;

  /* read lexeme node */
  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `is' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNSET*");
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!spec(ln)) {
    return FALSE;
  }
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, "DEBUG: is(%s(%d),", ln->arc->word, ln->arc->from);
#endif
  /* read label */
  l = l->next;
  val = evalTerm((Term) l->item, &vas, lg, context);
  if(val->type == VTString) {
    label = val->data.string;
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, " %s", label);
#endif
  } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2rd arg) for `is' in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }

  /* if the string is the name of a unary constraint,
     use this constraint instead of label match  */
  c = hashGet(inputCurrentGrammar->constraints, label);
  if(c && 1 != listSize(c->vars)) {
    c = NULL;
  }

  /* maybe read hierarchy */
  if(l->next) {
    l = l->next;
    val = evalTerm((Term) l->item, &vas, lg, context);
    if(val->type == VTString) {
      hname = val->data.string;
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, ", %s", hname);
#endif
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (3rd arg) for `is' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }
    if(!(h = findHierarchy(hname))) {
      cdgPrintf(CDG_WARNING,
                "WARNING: hierarchy `%s' not found in constraint `%s'.\n", hname,
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return FALSE;
    }
  }

  /* maybe read second label name */
  if(l->next) {
    l = l->next;
    val = evalTerm((Term) l->item, &vas, lg, context);
    if(val->type == VTString) {
      labels = val->data.string;
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, ", %s", labels);
#endif
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (4rd arg) for `is' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }
  }
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, ")? ");
#endif

  /* Find the LV to check. If there is no fourth argument, this is the LV
     above the specified word. */
  i = ln->arc->from;
  lv = vectorElement(context, nrlevels * i + levelno);
  if(!lv) {
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, " No LV found.\n");
#endif
    return FALSE;
  }

  /*  Otherwise, we have to ascend across irrelevant labels. */
  counter = 0;
  while(labels &&
        counter <= lg->max &&
        spec(lv->regents) &&
        inputSubsumes(h, labels, lv->label)) {
    counter++;
    ln = lv->regents->item;
    i = ln->arc->from;
    lv = vectorElement(context, nrlevels * i + levelno);
    if(!lv) {
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, " No LV found.\n");
#endif
      return FALSE;
    }
  }

  if(counter > lg->max) {
#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, " Cycle.\n");
#endif
    return FALSE;
  }

  if(c) {

    /* If a constraint was specified, eval it... */
    Constraint thisConstraint = evalCurrentConstraint;
#ifdef DEBUGIS
    cdgPrintf(CDG_DEBUG, "\n<call: %s>\n", label);
    cdgPrintf(CDG_DEBUG, "lv == ");
    lvPrint(CDG_DEBUG, lv, NULL, 0);
    cdgPrintf(CDG_DEBUG, "\n");
#endif
    result = evalUnaryConstraint(c, NULL, context, lv);
#ifdef DEBUGIS
    cdgPrintf(CDG_DEBUG, "</call>\n", label);
#endif
    evalCurrentConstraint = thisConstraint;

#ifdef DEBUGIS
    if(result) {
      cdgPrintf(CDG_DEBUG, " ...yes (`%s').\n",
                ((LexemNode)lv->dependents->item)->arc->word);
    } else {
      cdgPrintf(CDG_DEBUG, " ...no (`%s').\n",
                ((LexemNode)lv->dependents->item)->arc->word);
    }
#endif

  } else {

    /* ...otherwise compare the label. */
    result = FALSE;
    if((!h && lv->label == label)
       ||
       (h && inputSubsumes(h, label, lv->label))) {
      result = TRUE;
#ifdef DEBUGIS
      cdgPrintf(CDG_DEBUG, " ...yes (`%s').\n",
                ((LexemNode)lv->dependents->item)->arc->word);
#endif
    }
  }

#ifdef DEBUGIS
  cdgPrintf(CDG_DEBUG, " %s.\n", result ? "Yes" : "No");
#endif

  return result;
}

/* ----------------------------------------------------------------------
 *  Does word #1 occur somewhere in the sentence?
 */
Boolean predOccur(List l, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value val;


  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `occur' in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }

  return lgContains(lg, val->data.string);
}

/* ----------------------------------------------------------------------
 *  Is the current input sequence marked as finished or is it partial?
 */
Boolean predPartial(List l, LexemGraph lg, Vector context) {
  return !lg->complete;
}

/* ----------------------------------------------------------------------
   Does a punctuation mark contained in S occur between two words?

   The first and second parameters are lexeme nodes, and the third parameter
   is a string. TRUE is returned if one of the words between the two lexeme
   nodes is one character long and consists of one of the characters in the
   string.

   The first and second parameters may also be numbers, which stand for the
   corresponding time points in the lexeme graph. The sentinel value -1
   means lg->min and lg->max, respectively.

   Either of the lexeme nodes may be replaced by a Number, which is then
   used as an index into the lexeme graph instead.
 */
Boolean predBetween(List l, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;
  int a, b;
  LexemNode ln;
  String s;

  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type == VTLexemNode) {
    ln = (LexemNode) val->data.lexemnode;
    if (ln == NULL) {
      return FALSE;
    }
    if (nonspec(ln)) {
      a = lg->max + 1;
    } else {
      a = ln->arc->from;
    }
  } else if (val->type == VTNumber) {
    a = val->data.number;
    if (-1 == a) {
      a = lg->min;
    }
  } else {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `between' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }

  l = l->next;
  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type == VTLexemNode) {
    ln = (LexemNode) val->data.lexemnode;
    if (ln == NULL) {
      return FALSE;
    }
    if (nonspec(ln)) {
      b = lg->max + 1;
    } else {
      b = ln->arc->from;
    }
  } else if (val->type == VTNumber) {
    b = val->data.number;
    if (-1 == b) {
      b = lg->max;
    }
  } else {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (2nd arg) for `between' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }

  if (a > b) {
    int c = a;

    a = b, b = c;
  }

  l = l->next;
  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (3rd arg) for `between' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }
  s = val->data.string;

  return predBetweenImpl(&a, &b, s, lg, context);
}

Boolean predBetweenImpl(Pointer arg1, Pointer arg2, Pointer arg3, LexemGraph lg,
			Vector context)
{
  int a = *((int *)arg1), b = *((int *)arg2);
  String s = arg3;
  List l;

#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, "between(%d, %d, %s)?", a, b, s);
#endif
  for (l = lg->lattice->arcs; l != NULL; l = l->next) {
    Arc arc = l->item;

    if (arc->from < a || arc->from > b)
      continue;
#ifdef DEBUGBETWEEN
    cdgPrintf(CDG_DEBUG, " %s(%d)", arc->word, arc->from);
#endif
    if (strlen(arc->word) != strspn(arc->word, s))
      continue;
#ifdef DEBUGBETWEEN
    cdgPrintf(CDG_DEBUG, " yes.\n");
#endif
    return TRUE;
  }
#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, " no.\n");
#endif
  return FALSE;
}

/* ----------------------------------------------------------------------
 *  Is A indirectly a modifier of B?
 */
Boolean predUnder(List l, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;
  LexemNode ln1, ln2;
	Level level;
	VarInfo vi = ((Term)l->item)->data.peek.varinfo;
  int levelno = vi->level->no;

  Term t = (Term) l->item;
  val = evalTerm(t, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `under' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  ln1 = (LexemNode) val->data.lexemnode;
  if (!spec(ln1)) {
    return FALSE;
  }
  val = evalTerm((Term) l->next->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (2nd arg) for `under' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  ln2 = (LexemNode) val->data.lexemnode;
  if (!spec(ln2)) {
    return FALSE;
  }

if(l->next->next)
{
  val = evalTerm((Term) l->next->next->item, &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (3rd arg) for `under' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);

    return FALSE;
  }

  /* find level */
  level = findLevel(val->data.string);
  if (!level) {
    cdgPrintf(CDG_ERROR, "ERROR: no level `%s' in constraint `%s'\n", val->data.string,
	evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return TRUE;
  }

	levelno = level->no;

}

  return predUnderImpl(ln1, ln2, lg, context, levelno);
}

Boolean predUnderImpl(Pointer arg1, Pointer arg2, LexemGraph lg, Vector context, int levelno) {
  LexemNode ln1 = (LexemNode) arg1;
  LexemNode ln2 = (LexemNode) arg2;
  int a = ln1->arc->from;
  int b = ln2->arc->from;
  int j = 0;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  LevelValue lv;

  //int levelno = ((VarInfo)evalCurrentConstraint->vars->item)->level->no;

#ifdef DEBUGUNDER
  cdgPrintf(CDG_DEBUG, "under(%s(%d), %s(%d))?", ln1->arc->word, a, ln2->arc->word, b);
#endif

  do {
    lv = vectorElement(context, nrlevels * a + levelno);
    if(!lv || !spec(lv->regents)) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " no.\n");
#endif
      /* failure */
      return FALSE;
    }
    if(((LexemNode)(lv->regents->item))->arc->from == b) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " yes.\n");
#endif
      /* success */
      return TRUE;
    }

    if(j > lg->max) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " no.\n");
#endif
      /* cycle */
      return FALSE;
    }
    a = ((LexemNode)(lv->regents->item))->arc->from;
    j++;
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " %d", a);
#endif
  } while(lv);

  return FALSE;
}

/* ----------------------------------------------------------------------
   Are A and B part of the same tree?
 */
Boolean predConnected(List l, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;
  int a, b, roota = -1, rootb = -1;
  LexemNode lna, lnb;
  LevelValue lv;
  VarInfo vi = evalCurrentConstraint->vars->item;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int max = lg->max - lg->min;
  int count;

  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `connected' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  lna = (LexemNode) val->data.lexemnode;
  if (!spec(lna)) {
    return FALSE;
  }
  a = lna->arc->from;
  l = l->next;

  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `connected' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  lnb = (LexemNode) val->data.lexemnode;
  if (!spec(lnb)) {
    return FALSE;
  }
  b = lnb->arc->from;

  count = 0;
  lv = vectorElement(context, nrlevels * a + vi->level->no);
  while (lv) {
    if (!spec(lv->regents)) {
      roota = ((LexemNode) (lv->dependents->item))->arc->from;
      break;
    } else if (nonspec(lv->regents)) {
      return FALSE;
    } else if (count++ > max) {
      roota = a;
      break;
    } else {
      lv =
	vectorElement(context,
		      nrlevels * ((LexemNode) (lv->regents->item))->arc->from +
		      vi->level->no);
    }
  }

  count = 0;
  lv = vectorElement(context, nrlevels * b + vi->level->no);
  while (lv) {
    if (!spec(lv->regents)) {
      rootb = ((LexemNode) (lv->dependents->item))->arc->from;
      break;
    } else if (nonspec(lv->regents)) {
      return FALSE;
    } else if (count++ > max) {
      rootb = b;
      break;
    } else {
      lv =
	vectorElement(context,
		      nrlevels * ((LexemNode) (lv->regents->item))->arc->from +
		      vi->level->no);
    }
  }

  return roota == rootb;
}

/* ----------------------------------------------------------------------
 * Does the level contain a cycle?
 */
Boolean predCycle(List l, LexemGraph lg, Vector context)
{
  ByteVector cleared;
  ByteVector seen;
  Level level;
  Value val;
  ValueStruct vas;
  int i;

  /* if there is no context at all, we cannot check this condition, so we
   * must assume that it is satisfied. */
  if (!context) {
    return FALSE;
  }

  val = evalTerm((Term) l->item, &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `cycle' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);

    return FALSE;
  }

  /* find level */
  level = findLevel(val->data.string);
  if (!level) {
    cdgPrintf(CDG_ERROR, "ERROR: no level `%s' in constraint `%s'\n", val->data.string,
	evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return TRUE;
  }

  /* check all edges */
  seen = bvNew(lg->max);
  bvSetElements(seen, FALSE, 0, lg->max);
  cleared = bvNew(lg->max);
  bvSetElements(cleared, FALSE, 0, lg->max);

  for (i = 0; i < vectorSize(context); i++) {
    LevelValue lv = vectorElement(context, i);

    if (!lv) {
      continue;
    }

    /* skip LVs of other levels */
    if (i % inputCurrentGrammar->noOfLevels != level->no) {
      continue;
    }
#ifdef DEBUGPREDCYCLE
    cdgPrintf(CDG_ERROR, "DEBUG: checking time point %d\n",
	      ((LexemNode) (lv->dependents->item))->arc->from);
#endif
    /* search connection to NIL */
    while (lv) {
      int index;

#ifdef DEBUGPREDCYCLE
      cdgPrintf(CDG_ERROR, "DEBUG: loop reaches time point %d.\n",
		((LexemNode) (lv->dependents->item))->arc->from);
#endif
      /* detect loop */
      if (bvElement(seen, ((LexemNode) (lv->dependents->item))->arc->from)) {
	seen = NULL;
	cleared = NULL;
#ifdef DEBUGPREDCYCLE
	cdgPrintf(CDG_ERROR, "DEBUG: loop detected\n",
		  ((LexemNode) (lv->dependents->item))->arc->from);
#endif
	return TRUE;
      }
      bvSetElement(seen, TRUE, ((LexemNode) (lv->dependents->item))->arc->from);

      /* find already cleared LV */
      if (!spec(lv->regents) ||
	  bvElement(cleared, ((LexemNode) (lv->dependents->item))->arc->from)) {

	/* set seen LVs to cleared also */
	int j;

	for (j = 0; j < bvSize(seen); j++) {
	  if (bvElement(seen, j)) {
	    bvSetElement(cleared, TRUE, j);
	  }
	}
#ifdef DEBUGPREDCYCLE
	cdgPrintf(CDG_ERROR, "DEBUG: time point %d is ok.\n",
		  ((LexemNode) (lv->dependents->item))->arc->from);
#endif
	bvSetElements(seen, FALSE, 0, lg->max);
	break;
      }

      /* step upward */
      index = inputCurrentGrammar->noOfLevels *
	((LexemNode) (lv->regents->item))->arc->from + lv->level->no;
      lv = vectorElement(context, index);
    }
  }

  seen = NULL;
  cleared = NULL;
  return FALSE;
}

/* ----------------------------------------------------------------------
 * In hierarchy #1, is there a common ancestor of sorts #2 and #3
 * that is not more than #4 levels away?
 */
/*
Boolean predNthCousin(List l, LexemGraph lg, Vector context) {
   TermValueUnion tvuh, tvua, tvub, tvun;
   TermValue tvh = &tvuh;
   TermValue tva = &tvua;
   TermValue tvb = &tvub;
   TermValue tvn = &tvun;
   EvalTermType tth,tta,ttb,ttn;
   Hierarchy h;
   Sort s;
   int i, a, b, n;
   int levela=0, levelb=0, leveli;

   if (listSize(l) != 4) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: predicate `nth_cousin' needs four arguments\n");
     return (FALSE);
   }
   tth = evalTerm((Term) listNthElement(l, 1), tvh, lg, Vector context);
   if (tth != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (1st arg) for `nth_cousin' in constraint `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   tta = evalTerm((Term) listNthElement(l, 2), tva, lg, Vector context);
   if (tta != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (2nd arg) for `nth_cousin' in constraint `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttb = evalTerm((Term) listNthElement(l, 3), tvb, lg, Vector context);
   if (ttb != ETTString) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (3rd arg) for `nth_cousin' in constraint `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttn = evalTerm((Term) listNthElement(l, 4), tvn, lg, Vector context);
   if (ttn != ETTNumber) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (4rd arg) for `nth_cousin' in constraint `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   n = tvun.number;

   for (l = inputCurrentGrammar->hierarchies; l != NULL; l = l->next)
      if (strcmp(tvh->string, ((Hierarchy) l->item)->id) == 0)
	 break;
   if (l == NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: hierarchy `%s' not found\n", tvh->string);
      return (FALSE);
   }
   h = (Hierarchy) l->item;

   a = b = -1;
   for (i = 0; i < vectorSize(h->sorts); i++) {
      s = (Sort) vectorElement(h->sorts, i);
      if (strcmp(s->id, tva->string) == 0) {
	a = i;
	levela = s->depth;
      }
      if (strcmp(s->id, tvb->string) == 0) {
	 b = i;
	 levelb = s->depth;
      }
   }
   if (a < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tva->string, tvh->string);
      return (FALSE);
   }
   if (b < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tvb->string, tvh->string);
      return (FALSE);
   }

   if(abs(levela-levelb) > n)
   return(FALSE);


   for (i = 0; i < vectorSize(h->sorts); i++) {
     s = (Sort) vectorElement(h->sorts, i);
     leveli = s->depth;

     if ( levela-leveli <= n &&
	  levelb-leveli <= n &&
	 (Boolean) arrayElement(h->subsumption, i, a) &&
	 (Boolean) arrayElement(h->subsumption, i, b)
	  )
       return (TRUE);

   }
   return(FALSE);
}
*/
/* ----------------------------------------------------------------------
 * In hierarchy #1, is there a common ancestor of sorts #2 and #3
 * at level #4 or lower?
 */
/*
Boolean predAncestor(List l, LexemGraph lg, Vector context) {
   TermValueUnion tvuh, tvua, tvub, tvun;
   TermValue tvh = &tvuh;
   TermValue tva = &tvua;
   TermValue tvb = &tvub;
   TermValue tvn = &tvun;
   EvalTermType tth,tta,ttb,ttn;
   Hierarchy h;
   Sort s;
   int i, a, b, n;
   int levela=0, levelb=0, leveli;

   if (listSize(l) != 4) {
     cdgPrintf(CDG_WARNING,
       "WARNING: predicate `ancestor_at_level' needs four arguments\n");
     return (FALSE);
   }
   tth = evalTerm((Term) listNthElement(l, 1), tvh, lg, Vector context);
   if (tth != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (1st arg) for `ancestor_at_level' in constraint `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   tta = evalTerm((Term) listNthElement(l, 2), tva, lg, Vector context);
   if (tta != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (2nd arg) for `ancestor_at_level' in constraint `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttb = evalTerm((Term) listNthElement(l, 3), tvb, lg, Vector context);
   if (ttb != ETTString) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (3rd arg) for `ancestor_at_level' in constraint `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttn = evalTerm((Term) listNthElement(l, 4), tvn, lg, Vector context);
   if (ttn != ETTNumber) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (4rd arg) for `ancestor_at_level' in constraint `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   n = tvun.number;

   for (l = inputCurrentGrammar->hierarchies; l != NULL; l = l->next)
      if (strcmp(tvh->string, ((Hierarchy) l->item)->id) == 0)
	 break;
   if (l == NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: hierarchy `%s' not found\n", tvh->string);
      return (FALSE);
   }
   h = (Hierarchy) l->item;

   a = b = -1;
   for (i = 0; i < vectorSize(h->sorts); i++) {
      s = (Sort) vectorElement(h->sorts, i);
      if (strcmp(s->id, tva->string) == 0) {
	a = i;
	levela = s->depth;
      }
      if (strcmp(s->id, tvb->string) == 0) {
	 b = i;
	 levelb = s->depth;
      }
   }
   if (a < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tva->string, tvh->string);
      return (FALSE);
   }
   if (b < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tvb->string, tvh->string);
      return (FALSE);
   }


 if(n > levela)
   return(FALSE);

 if(n > levelb)
   return(FALSE);

   for (i = 0; i < vectorSize(h->sorts); i++) {
     s = (Sort) vectorElement(h->sorts, i);
     leveli = s->depth;

     if ( leveli >= n &&
	 (Boolean) arrayElement(h->subsumption, i, a) &&
	 (Boolean) arrayElement(h->subsumption, i, b))
       return (TRUE);

   }
   return(FALSE);
}
*/

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */

