/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*
 * File: set.c
 * Author: Horia F. Pop
 * Purpose: management of cdg set variables
 * Birth: March 21, 2001
 *
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "set.h"
#include "cdg.h"
#include "hook.h"

/* -- MACROS ------------------------------------------------------------ */

/* -- TYPE DEFINITIONS -------------------------------------------------- */

typedef union {
	Boolean b;
	int i;
	unsigned long ul;
	float f;
	String s;
} SetValueStruct;

typedef struct {
	String param;
	SetValueStruct value;
} ParameterStruct, *Parameter;

typedef struct {
	Variable var;
	unsigned long bitmask; /* this is dirty */
	BooleanFunction *funcConvert;
	List aliases;
	List values;
} SetVarStruct, *SetVar;

/* -- VARIABLES --------------------------------------------------------- */

static List setList = NULL;

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * setIteratorNew
 */
void setIteratorNew(SetIterator *vip) {
	*vip = (SetIterator) setList;
}

/* ----------------------------------------------------------------------
 * setIteratorNextName
 */
String setIteratorNextName(SetIterator *vip) {
	String s = NULL;
	List l = (List) *vip;

	if (l) {
		s = varName(((SetVar) l->item)->var);
		*vip = (SetIterator) l->next;
	}

	return s;
}

/* ----------------------------------------------------------------------
 * setIteratorFree
 */
void setIteratorFree(SetIterator *vip) {
	*vip = NULL;
}

/* ----------------------------------------------------------------------
 * setReturnVariable: returns the variable with the given name or alias, 
 * or NULL if none found
 */
SetVar setReturnVariable(String name) {
	List l, m;

	for (l = setList; l; l = l->next) {
		SetVar i = (SetVar) l->item;
		if (strcmp(varName(i->var), name) == 0) {
			return i;
		}
		for (m = i->aliases; m; m = m->next) {
			if (strcmp((String) m->item, name) == 0) {
				return i;
			}
		}
	}

	return NULL;
}

/* ----------------------------------------------------------------------
 * Sets default boolean special arguments
 */
List setRegisterBooleanSpecials() {
	String optTRUE[] = { "on", "true", "yes", "1", NULL };
	String optFALSE[] = { "off", "false", "no", "0", NULL };

	Parameter paramPtr;
	List l = NULL;
	int i;

	for (i = 0; optTRUE[i] != NULL; i++) {
		paramPtr = (Parameter) memMalloc(sizeof(ParameterStruct));
		paramPtr->param = strRegister(optTRUE[i]);
		paramPtr->value.b = TRUE;
		l = listAppendElement(l, paramPtr);
	}
	for (i = 0; optFALSE[i] != NULL; i++) {
		paramPtr = (Parameter) memMalloc(sizeof(ParameterStruct));
		paramPtr->param = strRegister(optFALSE[i]);
		paramPtr->value.b = FALSE;
		l = listAppendElement(l, paramPtr);
	}

	return l;
}

/* ----------------------------------------------------------------------
 * Registers a variable. The arguments are the name, type, address, and 
 * address of validation function, callback function and special arguments
 */
Boolean setRegister(String name, int type, Pointer ptr,
		BooleanFunction *funcConvert, BooleanFunction *funcValidate,
		VoidFunction *funcCallback, ...) {
	SetVar v;
	Parameter paramPtr;
	String param;
	va_list ap;

	if (setReturnVariable(name) != NULL) {
		cdgPrintf(CDG_ERROR, "ERROR: variable `%s' already registered\n", name);
		return FALSE;
	}

	v = (SetVar) memMalloc(sizeof(SetVarStruct));
	v->var = varAllocate(name, type, ptr, funcValidate, funcCallback, NULL);
	v->funcConvert = funcConvert;
	v->bitmask = 0;
	v->aliases = NULL;
	v->values = NULL;
	setList = listAppendElement(setList, v);

	/* potential problem - checks pairs of arguments only */
	va_start(ap, funcCallback);
	if (varType(v->var) == SET_BIT) {
		v->bitmask = va_arg(ap, unsigned long);
	}
	while ((param = va_arg(ap, String)) != NULL) {
		paramPtr = (Parameter) memMalloc(sizeof(ParameterStruct));
		paramPtr->param = strRegister(param);
		switch (varType(v->var)) {
		case SET_BOOL:
		case SET_BIT: {
			Boolean value = va_arg(ap, int);
			paramPtr->value.b = value;
		}
			break;
		case SET_INT:
		case SET_ENUM: {
			int value = va_arg(ap, int);
			paramPtr->value.i = value;
		}
			break;
		case SET_FLOAT: {
			double value = va_arg(ap, double);
			paramPtr->value.f = value;
		}
			break;
		case SET_STRING: {
			String value = va_arg(ap, String);
			paramPtr->value.s = strRegister(value);
		}
			break;
		case SET_UNSLONG: {
			unsigned long value = va_arg(ap, unsigned long);
			paramPtr->value.ul = value;
		}
			break;
		default:
			cdgPrintf(CDG_ERROR, "ERROR: illegal setRegister type `%d'\n",
					varType(v->var));
			return FALSE;
		}
		v->values = listAppendElement(v->values, paramPtr);
	}
	va_end(ap);

	if ((varType(v->var) == SET_BOOL || varType(v->var) == SET_BIT)
			&& (v->values == NULL)) {
		v->values = setRegisterBooleanSpecials();
	}
	return TRUE;
}

/* ----------------------------------------------------------------------
 * Sets an alias to an already registered variable
 */
Boolean setAlias(String alias, String name) {
	SetVar v = setReturnVariable(name);

	if (v) {
		v->aliases = listAppendElement(v->aliases, alias);
		return TRUE;
	} else {
		cdgPrintf(CDG_ERROR, "ERROR: variable `%s' not registered\n", name);
		return FALSE;
	}
}

/* ----------------------------------------------------------------------
 * Sets the value of a previously registered variable.
 * The second argument must be of the type associated to the given name.
 * Returns TRUE if the named variable does exist and if the value is valid.
 * It is setValue's responsibility to set the value; the validating function
 * only validates it.
 */
Boolean setValue(String name, String value) {
	List l;
	Boolean success = FALSE;
	SetVar v = setReturnVariable(name);
	Parameter par = NULL;

	if (v == NULL) {
		cdgPrintf(CDG_ERROR, "ERROR: variable `%s' not registered\n", name);
		return FALSE;
	}

	/* find the right parameter */
	for (l = v->values; l; l = l->next) {
		Parameter p = (Parameter) l->item;
		if (strcmp(value, p->param) == 0) {
			par = p;
		}
	}

	switch (varType(v->var)) {
	case SET_BOOL: {
		int var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.b;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = sscanf(value, "%d", &cnv);
				if (cnv < 0 || cnv > 1) {
					success = FALSE;
				}
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			*((Boolean *) varPtr(v->var)) = cnv;
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_INT: {
		int var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.i;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = sscanf(value, "%d", &cnv);
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			*((int *) varPtr(v->var)) = cnv;
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_ENUM: {
		int var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.i;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = FALSE;
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			*((int *) varPtr(v->var)) = cnv;
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_FLOAT: {
		float var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.f;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = sscanf(value, "%f", &cnv);
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			*((float *) varPtr(v->var)) = cnv;
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_STRING: {
		String var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.s;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			success = TRUE;
			if (par) {
				cnv = var;
			} else {
				cnv = value;
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			cdgFreeString(*((String *) varPtr(v->var)));
			*((String *) varPtr(v->var)) = strRegister(cnv);
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_UNSLONG: {
		unsigned long var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.ul;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = sscanf(value, "%ld", &cnv);
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			*((unsigned long *) varPtr(v->var)) = cnv;
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	case SET_BIT: {
		Boolean var, cnv;

		/* the value from specials */
		if (par) {
			var = par->value.b;
		}

		/* use conversion function or try a simple conversion */
		if (v->funcConvert != NULL) {
			if (par) {
				success = v->funcConvert(name, value, &var, &cnv);
			} else {
				success = v->funcConvert(name, value, NULL, &cnv);
			}
		} else {
			if (par) {
				success = TRUE;
				cnv = var;
			} else {
				success = sscanf(value, "%d", (int *) &cnv);
			}
		}

		/* use validation function */
		if (success) {
			if (varValidate(v->var) != NULL) {
				success = varValidate(v->var)(name, value, &cnv);
			}
		}

		/* final update */
		if (success) {
			if (cnv) {
				*((unsigned long *) varPtr(v->var)) |= v->bitmask;
			} else {
				*((unsigned long *) varPtr(v->var)) &= ~v->bitmask;
			}
			if (varCallback(v->var)) {
				varCallback(v->var)(name, varPtr(v->var));
			}
			return TRUE;
		}
		break;
	}
	}

	/* error message */
	cdgPrintf(CDG_ERROR, "ERROR: variable `%s' can only be set to ", name);
	for (l = v->values; l; l = l->next) {
		Parameter p = (Parameter) l->item;
		cdgPrintf(CDG_ERROR, "`%s'", p->param);
		if (varType(v->var) == SET_ENUM && v->funcConvert == NULL && l->next
				&& l->next->next == NULL) {
			cdgPrintf(CDG_ERROR, " or ");
		} else if (l->next) {
			cdgPrintf(CDG_ERROR, ", ");
		}
	}

	if (varType(v->var) != SET_ENUM || v->funcConvert != NULL) {
		if (v->values) {
			cdgPrintf(CDG_ERROR, " or ");
		}
		cdgPrintf(CDG_ERROR, "a valid ");
		switch (varType(v->var)) {
		case SET_BOOL:
			cdgPrintf(CDG_ERROR, "boolean value");
			break;
		case SET_INT:
			cdgPrintf(CDG_ERROR, "integer value");
			break;
		case SET_ENUM:
			cdgPrintf(CDG_ERROR, "enumeration value");
			break;
		case SET_FLOAT:
			cdgPrintf(CDG_ERROR, "float value");
			break;
		case SET_UNSLONG:
			cdgPrintf(CDG_ERROR, "unsigned long value");
			break;
		case SET_STRING:
			cdgPrintf(CDG_ERROR, "string");
			break;
		case SET_BIT:
			cdgPrintf(CDG_ERROR, "bitset value");
			break;
		}
	}

	cdgPrintf(CDG_ERROR, "\n");
	return FALSE;
}

/* ----------------------------------------------------------------------
 * Prints the value of a previously registered variable.
 * The second argument must be an address to the type associated with the
 * given name. Returns TRUE if the named variable does exist. On the exit 
 * of this function, the second argument will hold the value of that 
 * variable.
 */
Boolean setPrint(unsigned long mode, String name) {
	SetVar v = setReturnVariable(name);

	if (v == NULL) {
		cdgPrintf(CDG_ERROR, "ERROR: variable `%s' not registered\n", name);
		return FALSE;
	}

	/* to be updated so that it prints significant names
	 instead of numbers, whenever possible */
	switch (varType(v->var)) {
	case SET_BOOL: {
		Boolean b;
		List l;
		Parameter param = NULL;
		Boolean found = FALSE;

		b = *(Boolean *) varPtr(v->var);
		for (l = v->values; l; l = l->next) {
			param = l->item;
			if (param->value.i == b) {
				found = TRUE;
				break;
			}
		}
		if (param && found) {
			cdgPrintf(mode, "%s = %s", varName(v->var), param->param);
		} else {
			cdgPrintf(mode, "%s = %s", varName(v->var), (b ? "true" : "false"));
		}
		break;
	}
	case SET_INT: {
		int i = *((int *) varPtr(v->var));
		cdgPrintf(mode, "%s = %d", varName(v->var), i);
		break;
	}
	case SET_UNSLONG: {
		unsigned long ul = *((unsigned long *) varPtr(v->var));
		cdgPrintf(mode, "%s = %ld", varName(v->var), ul);
		break;
	}
	case SET_FLOAT: {
		float f = *((float *) varPtr(v->var));
		cdgPrintf(mode, "%s = %f", varName(v->var), f);
		break;
	}
	case SET_STRING: {
		String s = *((String *) varPtr(v->var));
		cdgPrintf(mode, "%s = %s", varName(v->var), s);
		break;
	}
	case SET_BIT: {
		cdgPrintf(mode, "%s = %s", varName(v->var),
				(*((unsigned long *) varPtr(v->var)) & v->bitmask) ? "on"
						: "off");
		break;
	}
	case SET_ENUM: {
		int paramNr = *((int *) varPtr(v->var));
		List l;
		Parameter param = NULL;
		Boolean found = FALSE;

		for (l = v->values; l; l = l->next) {
			param = l->item;
			if (param->value.i == paramNr) {
				found = TRUE;
				break;
			}
		}
		if (param && found) {
			cdgPrintf(mode, "%s = %s", varName(v->var), param->param);
		} else {
			cdgPrintf(mode, "%s = %d", varName(v->var), paramNr);
		}
		break;
	}
	default:
		cdgPrintf(CDG_ERROR, "ERROR: illegal variable type `%d'\n", varType(
				v->var));
		return FALSE;
	}

	/* show aliases */
	if (v->aliases) {
		List l;
		cdgPrintf(mode, " (");
		for (l = v->aliases; l; l = l->next) {
			String alias = l->item;
			cdgPrintf(mode, "%s%s", alias, l->next ? " " : "");
		}
		cdgPrintf(mode, ")");
	}
	cdgPrintf(mode, "\n");

	return TRUE;
}

/* ----------------------------------------------------------------------
 * Returns a list of Strings holding the names of all variables.
 */
List setReturnListOfNames() {
	List l, rez = NULL;

	Boolean cmpString(String s1, String s2) {
		return (Boolean) (strcmp(s1, s2) < 0);
	}

	for (l = setList; l; l = l->next) {
		SetVar v = (SetVar) l->item;
		String s = varName(v->var);
		rez = listInsertSorted(rez, s, cmpString);
#if 0
		for (m=v->aliases; m; m=m->next) {
			rez = listInsertSorted(rez, (String)m->item, cmpString);
		}
#endif
	}
	return rez;
}

/* ----------------------------------------------------------------------
 * Returns the type of the given variable, or SET_NONE otherwise.
 */
int setReturnType(String name) {
	SetVar v = setReturnVariable(name);

	if (v) {
		return varType(v->var);
	} else {
		cdgPrintf(CDG_ERROR, "ERROR: variable `%s' not registered\n", name);
		return SET_NONE;
	}
}

/* -----------------------------------------------------------------------
 * Maybe combine two paths if a leading or trailing colon is given.
 */
Boolean setJoinPaths(String name, String value, String *var, String *cnv) {

	/* find old value */
	SetVar a = setReturnVariable(name);
	String s = *((String *) varPtr(a->var));

	/* if the new value ends in colon, prepend it */
	if (strIsSuffix(value, ":")) {
		*cnv = strAppend(value, s, NULL);
	}

	/* if the new value starts with a comma, prepend it */
	else if (strIsPrefix(value, ":")) {
		*cnv = strAppend(s, value, NULL);
	}

	/* otherwise, just overwrite */
	else {
		*cnv = strCopy(value);
	}

	return TRUE;
}

/* ----------------------------------------------------------------------
 * Deletes the list of registered variables
 */
void setFinalize() {
	List l, m;

	for (l = setList; l; l = l->next) {
		SetVar v = (SetVar) l->item;
		cdgFreeString(varName(v->var));
		memFree(v->var);
		for (m = v->aliases; m; m = m->next) {
			cdgFreeString((String) m->item);
		}
		v->aliases = NULL;
		for (m = v->values; m; m = m->next) {
			memFree(m->item);
		}
		v->values = NULL;
		memFree(v);
	}
	setList = NULL;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
