/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "cdg.h"
#include "hook.h"
#include "timer.h"
#include "transform.h"
#include "increment.h"
#include "profiling.h"
#include "virtualnode.h"

/* How many items should the local search agenda hold? */
int localSize = 3000;

/* How many of the Patches generated in a step should actually be tried out? */
int beamWidth = INT_MAX;

/* counter to generate unique problem ids */
int problem_counter = 0;

/* will partial solutions be printed in xcdg? */
Boolean suppressPartialSolutions = FALSE;


/* ----------------------------------------------------------------------
 Mark an Analysis as covering only part of an utterance.
 */
void aRestrict(Analysis a, int from, int to) {
	a->subStart = max(from, a->P->minTimepoint);
	a->subStop = min(to, a->P->maxTimepoint);

	/* This has almost certainly changed the score of the structure. */
	aJudge(a);
}

/* ----------------------------------------------------------------------
 Mark an Analysis as covering the entire utterance.
 */
void aUnrestrict(Analysis a) {
	a->subStart = a->P->minTimepoint;
	a->subStop = a->P->maxTimepoint;
	aJudge(a);
}

/* ---------------------------------------------------------------------
 Print the currently selected lexeme nodes.
 */
void printSelectedPath(unsigned long mode, Analysis a) {

	int i;

	for (i = 0; i < vectorSize(a->P->lg->nodes); i++) {
		if (bvElement(a->isUnselectedLexeme, i)) {
			continue;
		}
		lgPrintNode(mode, vectorElement(a->P->lg->nodes, i));
		cdgPrintf(mode, "\n");
	}
	cdgPrintf(mode, "\n");
}

/* ----------------------------------------------------------------------
 Return the lexeme node from the list that is selected in a, or NULL.
 */
LexemNode selectedLexeme(Analysis a, List lexemes) {

	LexemNode result = NULL;
	List l;

	if (!a) {
		return NULL;
	}

	for (l = lexemes; l; l = l->next) {
		result = (LexemNode) l->item;
		if (!bvElement(a->isUnselectedLexeme, result->no)) {
			return result;
		}
	}

	return NULL;

}

/* ----------------------------------------------------------------------
 Return the lexeme node that is selected in a at time point WHERE.

 The return value is never NULL. If an existing time point is unbound in
 am Analysis, that is an internal error.
 */
LexemNode selectedAt(Analysis a, int where) {

	int i;
	LexemNode ln;

	/* Column #n of the vector lexemes contains a full listing
	 of all lexemes that span timepoint #n.
	 Therefore we need only look at that row. */
	for (i = 0; i < a->P->lgWidth; i++) {
		ln = vectorElement(a->P->lexemes, a->P->lgWidth * where + i);
		if (ln && !bvElement(a->isUnselectedLexeme, ln->no)) {
			return ln;
		}
	}

	/* This can't happen */
	printSelectedPath(CDG_ERROR, a);
	cdgPrintf(CDG_ERROR, "PANIC: unbound time interval %d\n", where);
	abort();

	return NULL;

}

/* Functions dealing with LevelValues
 ==================================

 This module takes great care not to clone LVs gratuitously, since great
 numbers of LVs are modified during the repair process, and much more
 during lookahead to see which repair step should be taken.

 Instead it uses the precomputed LVs in the ConstraintNet wherever
 possible. This means that wherever an LV with indexWRTNet x
 appears in this module, it is not only equal but _identical_ to the
 corresponding LV in the constraint net.

 Conversely, any LV that carries the indexWRTNet -1 is guaranteed to
 have the score 0.0 (otherwise it would have been found in the
 ConstraintNet). Such impossible LVs can only be introduced in two
 ways: either as a side effect of selecting another path through the
 lexeme graph, or by manipulating individual features of an LV during
 manual frobbing. In either case a totally new LV is cloned from the
 previously selected LV and changed as requested. The LVs contained in
 the ConstraintNodes of the net passed as a parameter to frobbing()
 are never changed at all, except to lower their limits. */

/* ----------------------------------------------------------------------
 Translates the level use flags into our own array for fast access.
 */
int gatherLevels(Problem P) {

	Level level;
	List l;

	P->nrLevels = inputCurrentGrammar->noOfLevels;
	P->levelActive = bvNew(P->nrLevels);

	P->nrActiveLevels = 0;
	for (l = inputCurrentGrammar->levels; l != NULL; l = l->next) {
		level = l->item;
		if (level->useflag) {
			P->nrActiveLevels++;
			bvSetElement(P->levelActive, TRUE, level->no);
		} else {
			bvSetElement(P->levelActive, FALSE, level->no);
		}
	}
	return P->nrActiveLevels;
}

/* ----------------------------------------------------------------------
 Should we bind slot x? Returns FALSE if its level is inactive,
 or the timepoint is not used by the lexeme graph, or
 if we are currently solving only a subproblem.
 */
Boolean slotActive(Analysis a, int x) {
	if (x / a->P->nrLevels < a->subStart) {
		return FALSE;
	}
	if (x / a->P->nrLevels >= a->subStop) {
		return FALSE;
	}
	if (!bvElement(a->P->levelActive, x % a->P->nrLevels)) {
		return FALSE;
	}
	return TRUE;
}

/* ----------------------------------------------------------------------
 Do lva and lvb cause the binary conflict cv?

 This function checks whether the two LVs violate the constraint recorded
 in CV. This checks both instantiation variants, so the order of the LVs
 is irrelevant. If the constraint has a variable penalty, TRUE is
 returned only if lva and lvb make the constraint fail with the same
 penalty as recorded in the conflict.

 */
Boolean lvsCauseConflict(Analysis a, LevelValue lva, LevelValue lvb,
		ConstraintViolation cv) {

	if (matchBinarySignature(lva, lvb, cv->constraint->sig)
			&& !evalBinaryConstraint(cv->constraint, NULL, a->slots, lva, lvb)
			&& cv->constraint->penalty == cv->penalty) {
		return TRUE;
	}

	else if (matchBinarySignature(lvb, lva, cv->constraint->sig)
			&& !evalBinaryConstraint(cv->constraint, NULL, a->slots, lvb, lva)
			&& cv->constraint->penalty == cv->penalty) {
		return TRUE;
	}

	return FALSE;

}
#ifdef TERNARY
/* ----------------------------------------------------------------------
 Do lva, lvb and lvc cause the ternary conflict cv?

 This function checks whether the three LVs violate the constraint recorded
 in CV. This checks both instantiation variants, so the order of the LVs
 is irrelevant. If the constraint has a variable penalty, TRUE is
 returned only if lva, lvb and lvc make the constraint fail with the same
 penalty as recorded in the conflict.

 */
Boolean threelvsCauseConflict(Analysis a, LevelValue lva, LevelValue lvb,
		LevelValue lvc, ConstraintViolation cv) {

	if (matchTernarySignature(lva, lvb, lvc, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lva, lvb,
					lvc) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	}

	else if (matchTernarySignature(lva, lvc, lvb, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lva, lvc,
					lvb) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	} else if (matchTernarySignature(lvb, lva, lvc, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lvb, lva,
					lvc) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	} else if (matchTernarySignature(lvb, lvc, lva, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lvb, lvc,
					lva) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	} else if (matchTernarySignature(lvc, lva, lvb, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lvc, lva,
					lvb) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	} else if (matchTernarySignature(lvc, lvb, lva, cv->constraint->sig)
			&& !evalTernaryConstraint(cv->constraint, NULL, a->slots, lvc, lvb,
					lva) && cv->constraint->penalty == cv->penalty) {
		return TRUE;
	}

	return FALSE;

}
#endif
/* ----------------------------------------------------------------------
 Is this LV valid in a?

 Returns TRUE if LV binds a set of lexemes that are valid in the current
 lexical path, i.e. at least one modifier and at least one modifiee (if
 there are any) must currently be selected.

 If STRICT is FALSE, then lexeme nodes outside the current subrange of
 the Analysis are always considered valid no matter what the path
 selection is. Usually this should be TRUE. It only makes sense to say
 FALSE if you are sure you will take care of the path criterion yourself
 before expanding the subrange of the Analysis again.
 */
Boolean isValid(Analysis a, LevelValue lv, Boolean strict) {

	if (spec(lv->regents)
			&& (strict || (((LexemNode) lv->regents->item)->arc->from
					>= a->subStart
					&& ((LexemNode) lv->regents->item)->arc->from < a->subStop))
			&& !listContains(lv->regents, selectedAt(a,
					((LexemNode) lv->regents->item)->arc->from))) {
		return FALSE;
	}

	if ((strict || (((LexemNode) lv->dependents->item)->arc->from
			>= a->subStart && ((LexemNode) lv->dependents->item)->arc->from
			< a->subStop)) && !listContains(lv->dependents, selectedAt(a,
			((LexemNode) lv->dependents->item)->arc->from))) {
		return FALSE;
	}

	return TRUE;

}

/* ----------------------------------------------------------------------
 Insert SOURCE int A.

 Does not respect the path criterion; if isValid(SOURCE, A) does not
 hold, A may become path inconsistent afterwards. You must use
 pLexicalize() before this function.
 */
void selectLV(Analysis a, LevelValue source) {

	int where = lvIndex(source);
	LevelValue target = vectorElement(a->slots, where);

	/* add to history */
	if (!listContains(a->P->history, (Pointer) where)) {
		a->P->history = listAppendElement(a->P->history, (Pointer) where);
	}

	/* self-replacement? */
	if (source == target) {
		return;
	}

	/* delete temporaries */
	if (target && target->indexWRTNet == -1) {
		target = NULL;
	}

	/* insert */
	vectorSetElement(a->slots, source, where);

}

/* ----------------------------------------------------------------------
 Return TRUE if LV would bind slot WHICH if it were selected.
 */
Boolean lvBindsSlot(Problem P, LevelValue lv, int where) {

	/* must bind same level */
	if (lv->level->no != where % P->nrLevels) {
		return FALSE;
	}

	/* must cover timespan */
	if (((LexemNode) lv->dependents->item)->arc->from > where / P->nrLevels) {
		return FALSE;
	}

	if (((LexemNode) lv->dependents->item)->arc->to <= where / P->nrLevels) {
		return FALSE;
	}

	return TRUE;

}

/* ----------------------------------------------------------------------
 Return the LV that binds slot WHERE in A.

 Note that the slot WHERE itself may be empty; in that case
 an LV in an earlier slot must exist that binds WHERE.
 */
LevelValue slotFiller(Analysis a, int where) {

	LevelValue result = NULL;
	int i = where;

	while (i >= 0 && !result) {
		result = vectorElement(a->slots, i);
		i -= a->P->nrLevels;
	}

	/* if the Analysis is incomplete (under construction), then the
	 latest previous LV may not actually bind that slot. In that case
	 no LV binds it. */
	if (result && !lvBindsSlot(a->P, result, where)) {
		result = NULL;
	}

	return result;

}

/*
 * returns the Edge directly above the given edge,
 * i.e. the LV that binds the slot the given slot has as regent
 * NULL if regent of given edge is nil or nonspec
 */
LevelValue edgeAbove(Analysis a, LevelValue lv) {
	if (!lv->regents || lv->regents == NONSPEC) {
		return NULL;
	}
	int upperTimeSlot = ((LexemNode) lv->regents->item)->grapheme->no;
	int slot = upperTimeSlot * a->P->nrLevels + lv->level->no;
	return slotFiller(a, slot);
}

/* ----------------------------------------------------------------------
 Propagate a limit to an LV.

 This can only happen for two reasons: either LV binds a set of lexeme
 nodes (as dependents or as regents) that all have a common limit not
 higher than F, or there is, for a certain set of domains, no combined
 assignment including LV with a limit greater than F.

 Accordingly, the only calls to downgrade() are during the routines
 updateLimit() and dSearch().
 */
void downgrade(Problem P, LevelValue lv, Number f) {

	/* downgrade */
	lv->limit = f;

	/* downgrading an LV past the global record means deleting */
	if (significantlyGreater(P->globalMinimum, f) && !lv->isDeleted) {
		lv->isDeleted = TRUE;
		P->nrUndeletedLVs -= (((LexemNode) lv->dependents->item)->arc->to
				- ((LexemNode) lv->dependents->item)->arc->from);
	}
}

/* ----------------------------------------------------------------------
 Recompute the limit of an LV based on the limits of the lexemes it binds.
 */
void updateLimit(Problem P, LevelValue lv) {

	Number f;
	List l;
	LexemNode lnb;

	/* deleted? */
	if (lv->isDeleted) {
		downgrade(P, lv, 0.0);
	}

	/* dependents limited? */
	f = 0.0;
	for (l = lv->dependents; l; l = l->next) {
		lnb = l->item;
		if (lnb->limit > f) {
			f = lnb->limit;
		}
	}
	if (f < lv->limit) {
		downgrade(P, lv, f);
	}

	/* regents limited? */
	if (spec(lv->regents)) {

		f = 0.0;
		for (l = lv->regents; l; l = l->next) {
			lnb = l->item;
			if (lnb->limit > f) {
				f = lnb->limit;
			}
		}
		if (f < lv->limit)
			downgrade(P, lv, f);
	}
}

/* ----------------------------------------------------------------------
 Return the limit of an LV.

 This function first updates the field ->limit of its argument, then
 returns that. This is the first of the two possibilities for a limit to
 be reduced. The decision to test the limits of an LV's lexemes in this
 routine is somewhat arbitrary but it guarantees that lexically motivated
 limits are propagated rather often.
 */
Number getLimit(Problem P, LevelValue lv) {
	updateLimit(P, lv);
	return lv->limit;

}

/* -----------------------------------------------------------------------
 Given LV, find an LV in a->P that has the same properties.

 Searches through the corresponding domains of a->P for an LV that

 - has the same label as LV
 - has the same structure as LV
 - is currently lexically valid.

 If one is found, it is returned, otherwise NULL is returned.

 This function is used to avoid unnecessary LV cloning. It judges LVs by
 comparing their lexeme nodes for identity, so it will only work
 correctly if P and lv are ultimately derived from the same lexeme graph.
 It will fail if P derives from one constraint net and lv is taken from
 another, or from a Parse structure -- use translateLV() for that.

 */
LevelValue findExistingLV(Analysis a, LevelValue lv) {

	List l;
	LevelValue model;

	/* if the LV is physically inconsistent,
	 don't even bother searching the domain */
	if (spec(lv->regents) && !lgCompatibleNodes(a->P->lg, lv->dependents->item,
			lv->regents->item)) {
#ifdef DEBUGFINDMODEL
		cdgPrintf(CDG_DEBUG, "DEBUG: findExistingLV() fails immediately\n");
#endif
		return NULL;
	}

	/* otherwise, inspect each LV in the domain closely */
	for (l = a->P->domains[lvIndex(lv)]; l; l = l->next) {
		model = (LevelValue) l->item;
#ifdef DEBUGFINDMODEL
		cdgPrintf(CDG_DEBUG, "DEBUG:  trying #%d\n", model->indexWRTNet);
		lvPrint(CDG_DEBUG, model, NULL, 2);
		cdgPrintf(CDG_DEBUG, "\n");
#endif
		if (strcmp(model->label, lv->label)) {
#ifdef DEBUGFINDMODEL
			cdgPrintf(CDG_DEBUG, "DEBUG: wrong label\n");
#endif
			continue;
		}
		if (((LexemNode) model->dependents->item)->arc->from
				!= ((LexemNode) lv->dependents->item)->arc->from
				|| ((LexemNode) model->dependents->item)->arc->to
						!= ((LexemNode) lv->dependents->item)->arc->to) {
#ifdef DEBUGFINDMODEL
			cdgPrintf(CDG_DEBUG, "DEBUG: wrong modifier\n");
#endif
			continue;
		}
		if (spec(lv->regents) && spec(model->regents)) {
			if ((((LexemNode) model->regents->item)->arc->from
					!= ((LexemNode) lv->regents->item)->arc->from
					|| ((LexemNode) model->regents->item)->arc->to
							!= ((LexemNode) lv->regents->item)->arc->to)) {
#ifdef DEBUGFINDMODEL
				cdgPrintf(CDG_DEBUG, "DEBUG: wrong modifiee\n");
#endif
				continue;
			}
		} else {
			if (lv->regents != model->regents) {
#ifdef DEBUGFINDMODEL
				cdgPrintf(CDG_DEBUG, "DEBUG: wrong specification\n");
#endif
				continue;
			}
		}
		if (!isValid(a, model, TRUE)) {
#ifdef DEBUGFINDMODEL
			cdgPrintf(CDG_DEBUG, "DEBUG: not lexically valid\n");
#endif
			continue;
		}

#ifdef DEBUGFINDMODEL
		cdgPrintf(CDG_DEBUG, "DEBUG: returning #%d\n", model->indexWRTNet);
#endif
		return model;
	}

#ifdef DEBUGFINDMODEL
	cdgPrintf(CDG_DEBUG, "DEBUG: returning NULL\n");
#endif
	return NULL;

}

/* ----------------------------------------------------------------------
 Find the lexeme node in a P's lexeme graph that corresponds to the
 first node in LNS.
 */
LexemNode translateLn(Problem P, List lns) {

	int i;
	LexemNode ln;
	if (!spec(lns)) {
		return NULL;
	}

	ln = lns->item;
#ifdef DEBUGTRANSLATELN
	cdgPrintf(CDG_DEBUG, "DEBUG: translateLn(%s)\n", ln->lexem->description);
#endif

	for (i = 0; i < vectorSize(P->lg->nodes); i++) {
		LexemNode lnNew = vectorElement(P->lg->nodes, i);
#ifdef DEBUGTRANSLATELN
		cdgPrintf(CDG_DEBUG, "DEBUG:   trying `%s'... ", lnNew->lexem->description);
#endif
		if (lnNew->arc->from != ln->arc->from) {
#ifdef DEBUGTRANSLATELN
			cdgPrintf(CDG_DEBUG, "wrong ->to\n");
#endif
			continue;
		}
		if (lnNew->arc->to != ln->arc->to) {
#ifdef DEBUGTRANSLATELN
			cdgPrintf(CDG_DEBUG, "wrong ->from\n");
#endif
			continue;
		}
		if (lnNew->lexem->description != ln->lexem->description) {
#ifdef DEBUGTRANSLATELN
			cdgPrintf(CDG_DEBUG, "wrong ->description\n");
#endif
			continue;
		}

#ifdef DEBUGTRANSLATELN
		cdgPrintf(CDG_DEBUG, "ok.\n");
#endif
		return lnNew;
	}

#ifdef DEBUGTRANSLATELN
	cdgPrintf(CDG_ERROR, "ERROR: translateLn() failed!\n");
#endif
	return NULL;

}

/* ----------------------------------------------------------------------
 Like findExistingLV(), but for `foreign' LVs, that is, LVs that
 do not belong to the same net and therefore have their own lexical
 entries. This means that it does not suffice to call listContains() &
 friends, which makes this function slower than findExistingLV. Indeed,
 this function may have to construct a totally new LV with index -1.
 */
LevelValue translateLV(Problem P, LevelValue lv) {

	List l, m;
	LevelValue model;
	LexiconItem le;
	int from, to;
	Boolean found;
#ifdef DEBUGTRANSLATELV
	cdgPrintf(CDG_DEBUG, "\nDEBUG: translating: ");
	lvPrint(CDG_DEBUG, lv, NULL, 2);
	cdgPrintf(CDG_DEBUG, "\n");
#endif

	/* otherwise, inspect each LV in the domain closely */
	for (l = P->domains[lvIndex(lv)]; l; l = l->next) {
		model = (LevelValue) l->item;

#ifdef DEBUGTRANSLATELV
		cdgPrintf(CDG_DEBUG, "  comparing to: ");
		lvPrint(CDG_DEBUG, model, NULL, 2);
		cdgPrintf(CDG_DEBUG, "\n");
#endif
		if (strcmp(model->label, lv->label)) {
#ifdef DEBUGTRANSLATELV
			cdgPrintf(CDG_INFO, "   ...wrong label.\n");
#endif
			continue;
		}

		le = ((LexemNode) lv->dependents->item)->lexem;
		from = ((LexemNode) lv->dependents->item)->arc->from;
		to = ((LexemNode) lv->dependents->item)->arc->to;
		found = FALSE;
		for (m = model->dependents; m != NULL; m = m->next) {
			LexemNode ln = m->item;
			if (ln->lexem->description == le->description && ln->arc->from
					== from && ln->arc->to == to) {
				found = TRUE;
				break;
			}
		}
		if (!found) {
#ifdef DEBUGTRANSLATELV
			cdgPrintf(CDG_INFO, "   ...wrong modifier.\n");
#endif
			continue;
		}

		if (lv->regents != NONSPEC && model->regents == NONSPEC)
			continue;
		if (lv->regents == NONSPEC && model->regents != NONSPEC)
			continue;
		if (lv->regents == NULL && model->regents != NULL) {
#ifdef DEBUGTRANSLATELV
			cdgPrintf(CDG_INFO, "   ...wrong modifiee.\n");
#endif
			continue;
		}

		if (spec(lv->regents)) {
			le = ((LexemNode) lv->regents->item)->lexem;
			from = ((LexemNode) lv->regents->item)->arc->from;
			to = ((LexemNode) lv->regents->item)->arc->to;
			found = FALSE;
			for (m = model->regents; m != NULL; m = m->next) {
				LexemNode ln = m->item;
				if (ln->lexem->description == le->description && ln->arc->from
						== from && ln->arc->to == to) {
					found = TRUE;
					break;
				}
			}
			if (!found) {
#ifdef DEBUGTRANSLATELV
				cdgPrintf(CDG_INFO, "   ...wrong modifiee.\n");
#endif
				continue;
			}
		}
		if (found) {
#ifdef DEBUGTRANSLATELV
			cdgPrintf(CDG_INFO, "   ...OK.\n");
#endif
			return model;
		}
	}

	/* The LV specified by the annotation is invalid, i.e. it violates a hard
	 unary constraint. This can happen if the grammar is bad or has changed
	 since the analysis was made. Ever helpful, we build a new LV in this
	 case. */
	{
		LexemNode ln1 = translateLn(P, lv->dependents);
		LexemNode ln2 = translateLn(P, lv->regents);
		List dependents = listPrependElement(NULL, ln1);
		List regents = NULL;
		LevelValue result;
		if (!ln1 || (lv->regents && !ln2)) {
			return NULL;
		}
		dependents = listPrependElement(NULL, ln1);
		if (ln2) {
			regents = listPrependElement(NULL, ln2);
		}
		result = lvNew(dependents, lv->level, lv->label, regents);
		result->indexWRTNet = -1;
		return result;
	}
}

/* ----------------------------------------------------------------------
 Construct an LV that can bind slot where in A.

 This is the exact opposite of a lexeme swap: rather than change the
 Analysis to suit an LV we would like to try, we invent an LV to suit the
 Analysis. This way of ensuring the path criterion is only used to build
 the initial Analysis.
 */
LevelValue inventLV(Analysis a, int slot) {

	LevelValue original, result;
	int from = slot / a->P->nrLevels;
	int levelno = slot % a->P->nrLevels;
	Level level;
	LexemNode dependent;
	List dependents;
	String label;

#ifdef DEBUGINVENTLV
	cdgPrintf(CDG_DEBUG, "DEBUG: inventLV(%p, $d) called\n", a, slot);
#endif

	/* If there is at least one LV in the corresponding domain,
	 we can steal and copy it. */
	if (a->P->domains[slot]) {
		result = lvClone(a->P->domains[slot]->item);
#ifdef DEBUGINVENTLV
		cdgPrintf(CDG_DEBUG, "DEBUG: using existing lv:\n  ");
		lvPrint(CDG_DEBUG, result, NULL, 2);
		cdgPrintf(CDG_DEBUG, "\n");
#endif

		result->indexWRTNet = -1;
		result->no = -1;
		result->dependents = NULL;
		dependent = selectedAt(a, from);
		result->dependents = listPrependElement(NULL, dependent);

		if (spec(result->regents)) {
			int from = ((LexemNode) result->regents->item)->arc->from;
			result->regents = listPrependElement(NULL, selectedAt(a, from));
		}

		if ((original = findExistingLV(a, result)) != NULL) {
			result = original;
		}
	}
	/* Otherwise we have to work really hard to invent one out of thin air. */
	else {
		level = listNthElement(inputCurrentGrammar->levels, 1 + levelno);
		label = level->labels->item;
		dependent = selectedAt(a, from);
		dependents = listPrependElement(NULL, dependent);
		result = lvNew(dependents, level, label, NULL);
#ifdef DEBUGINVENTLV
		cdgPrintf(CDG_DEBUG, "DEBUG: invented new lv:\n  ");
		lvPrint(CDG_DEBUG, result, NULL, 2);
		cdgPrintf(CDG_DEBUG, "\n");
#endif
	}

	return result;

}

/* ----------------------------------------------------------------------
 Does LV violate something in A?
 */
Boolean causesConflict(Analysis a, LevelValue lv) {

	List l;
	ConstraintViolation cv;
	int index = lvIndex(lv);

	for (l = a->conflicts; l; l = l->next) {
		cv = l->item;
		if (index == cv->nodeBindingIndex1)
			return TRUE;
		if (index == cv->nodeBindingIndex2)
			return TRUE;
	}

	return FALSE;

}

/* ----------------------------------------------------------------------
 Should LV be displayed in a solution?

 The reasoning here is that most slots in an analysis are useless
 (such as the OBL slots of function words) and do not contribute to
 the understanding of a solution, but merely take up space.
 Therefore we want to suppress displaying them.

 An lv is considered boring if it modifies NIL, and all other
 bindings of the same lexeme node also modify NIL. That way, if
 there are the LVs

 #200 = OBL1:ich(2,3)------>NIL  and
 #300 = OBL1:versichern(0,4)----->wir(5,6)

 then #200 will be boring and #300 will be interesting.

 However, FALSE is always returned if LV causes at least one conflict,
 so that conflicts will not refer to invisible LVs.
 */
Boolean isBoring(Analysis a, LevelValue lv) {

	List l;
	int where = lvIndex(lv);
	LevelValue lvb;

	if (causesConflict(a, lv)) {
		return FALSE;
	}

	for (l = a->P->domains[where]; l; l = l->next) {
		lvb = l->item;
		if (spec(lvb->regents)
				&& ((LexemNode) lvb->dependents->item)->arc->from
						== ((LexemNode) lv->dependents->item)->arc->from
				&& ((LexemNode) lvb->dependents->item)->arc->to
						== ((LexemNode) lv->dependents->item)->arc->to) {
			return FALSE;
		}
	}

	return TRUE;

}

/* ----------------------------------------------------------------------
 Fix self-modifying LVs.

 When swapping in a lexeme that is so long that it overlaps both the
 modifier and the modifiee of an LV, an LV may end up with a lexeme
 node that modifies itself.

 Since this is unlikely to be helpful, we point those LVs to NIL
 immediately. Perhaps it would be smarter to point them wherever is
 best at the moment, but that is a problem for another day.

 Note that since this LV is invalid, it cannot be the member of any
 domain. So we don't have to check its index to avoid changing the
 original LVs.
 */
void breakCircle(LevelValue lv) {

	if (spec(lv->regents) && lgOverlap(lv->dependents->item, lv->regents->item)) {
#ifdef DEBUGLVBREAKCIRCLE
		cdgPrintf(CDG_WARNING,
				"WARNING: #%d modifies self at slot @%d\n",
				lv->indexWRTNet, lvIndex(lv));
#endif
		lv->regents = NULL;
	}
}

/* ----------------------------------------------------------------------
 Modify Analysis a so that the LV at slot #where points to the
 lexeme at time point direction. If no such LV exists, a temporary
 LV is allocated for the purpose.
 */
void aRedirect(Analysis a, int where, int direction) {
	LevelValue lv = (LevelValue) vectorElement(a->slots, where);
	LevelValue original = NULL;
	LevelValue tmpLV;
	LexemNode modifiee;
	while (!lv) {
		where--;
		lv = (LevelValue) vectorElement(a->slots, where);
	}
	tmpLV = lvClone(lv);
	if (lv->indexWRTNet == -1) {
		lv = NULL;
	}

	/* re-orientate */
	if (direction < 0) {
		tmpLV->regents = NULL;
	} else if (direction >= a->P->maxTimepoint) {
		tmpLV->regents = NONSPEC;
	} else {
		modifiee = selectedAt(a, direction);
		tmpLV->regents = listPrependElement(NULL, modifiee);
	}

	/* update info */
	tmpLV->indexWRTNet = -1;
	tmpLV->no = -1;

	/* If the swap produced an LV that already exists,
	 take this one and delete the clone. */
	if ((original = findExistingLV(a, tmpLV)) != NULL) {
		tmpLV = original;
	}

	/* Otherwise we retain the new LV.
	 (It will be deallocated the next time this domain is changed.) */

	/* Finally, substitute the LV into the Analysis. */
	vectorSetElement(a->slots, tmpLV, where);

}

/* ----------------------------------------------------------------------
 Modify Analysis a so that the LV at slot #where carries the passed
 label. If no such LV exists, a temporary LV is allocated for the
 purpose.
 */
Boolean aRelabel(Analysis a, int where, String label) {
	LevelValue lv = (LevelValue) vectorElement(a->slots, where);
	LevelValue original = NULL;
	LevelValue tmpLV;

	while (!lv) {
		where--;
		lv = (LevelValue) vectorElement(a->slots, where);
	}

	/* sanity check */
	if (!listContains(lv->level->labels, label)) {
		cdgPrintf(CDG_WARNING, "WARNING: Level `%s' has no label `%s'.\n",
				lv->level->id, label);
		return FALSE;
	}

	tmpLV = lvClone(lv);

	/* swap */
	tmpLV->label = label;

	/* update info */
	tmpLV->indexWRTNet = -1;
	tmpLV->no = -1;

	/* If the swap produced an LV that already exists,
	 take this one and delete the clone. */
	if ((original = findExistingLV(a, tmpLV)) != NULL) {
		tmpLV = original;
	}
	/* Otherwise we retain the new LV.
	 (It will be deallocated the next time this domain is changed.) */

	/* Finally, substitute the LV into the Analysis. */
	vectorSetElement(a->slots, tmpLV, where);
	return TRUE;

}

/* ----------------------------------------------------------------------
 Add LVs of non-main levels for time point WHERE.

 For each slot of A that is not on the main level, an LV is selected and
 inserted. If several LVs are possible (= already present in the net
 ConstraintNet and the Problem), the one that fits the present LVs best
 is chosen. However, this optimization is done strictly the unary way. If
 several auxiliary levels must cooperate to achieve a high score this may
 be overlooked.

 If REPLACE is set, even slots that are already filled are re-optimized.
 Otherwise only empty slots are filled.

 If OVERRIDE is set, then this function may change the lexical selection
 of existing LVs in the Analysis.

 */
void aComputeAux(Analysis a, int where, Boolean replace, Boolean override) {
	Problem P = a->P;
	Patch p;
	int i;
	Level m = inputGetMainlevel(inputCurrentGrammar);
	int flags = override ? OD_INT | OD_ALLOWALL | OD_PREDICT | OD_DELETED
			: OD_INT | OD_ORIENT | OD_LABEL | OD_PREDICT | OD_DELETED;

	for (i = 0; i < P->nrLevels; i++) {
		int index = P->nrLevels * where + i;
		LevelValue lv = vectorElement(a->slots, index);
		if (lv && !replace)
			continue;
		if (i == m->no)
			continue;
		if (!slotActive(a, index))
			continue;

		p = optimizeDomains(a, (Pointer) index, flags);
		if (p) {
			apply(a, p);
		}
	}
}

/* Functions dealing with Analyses
 =============================== */

/* --------------------------------------------------------------------
 Calculate all conflicts in an analysis.

 Sets the fields ->b and ->conflicts. If these fields were non-empty,
 the contents are deallocated.

 if the score falls below the given threshold, cancel evaluation and return false
 else true is returned
 */
Boolean aJudgeWithThreshold(Analysis a, Number threshold) {

    LevelValue lva, lvb;
    int i, j;

	if (!a) {
		return TRUE;
	}

	/* remove old conflict set */
	if (a->conflicts) {
		a->conflicts = NULL;
	}
	/* start out with the perfect badness */
	a->b->no = 0;
	a->b->hard = 0;
	a->b->soft = 1.0;

	lock_tree(a->P->maxTimepoint);

	/* seek unary conflicts */
	for (i = 0; i < vectorSize(a->slots); i++) {
		if (!slotActive(a, i))
			continue;
		lva = (LevelValue) vectorElement(a->slots, i);
		if (!lva)
			continue;
		(void) evalUnary(lva, a->P->net, a->slots, FALSE, a->b, &a->conflicts,
				FALSE);

		if(threshold>0.0 && (a->b->hard || significantlyGreater(threshold, a->b->soft) ) ) {
			unlock_tree();
			return FALSE;
		}

		/* seek binary conflicts */
		for (j = i + 1; j < vectorSize(a->slots); j++) {
			if (!slotActive(a, j))
				continue;
			lvb = (LevelValue) vectorElement(a->slots, j);
			if (!lvb)
				continue;

      (void) evalBinary(lva, lvb, a->P->net, a->slots, FALSE, a->b, &a->conflicts);
#ifdef TERNARY
            if(inputCurrentGrammar->no_ternary) {
			/* seek ternary conflicts */
      int k;
			for (k = j + 1; k < vectorSize(a->slots); k++) {
				if (!slotActive(a, k)) {
					continue;
				}
          LevelValue lvc = (LevelValue) vectorElement(a->slots, k);
				if (!lvc) {
					continue;
				}

				(void) evalTernary(lva, lvb, lvc, a->P->net, a->slots, FALSE,
						a->b, &a->conflicts);
			}
            }
#endif
            if(threshold>0.0 && (a->b->hard || significantlyGreater(threshold, a->b->soft) ) ) {
            	unlock_tree();
				return FALSE;
			}
		}
	}

	unlock_tree();

	/* sort conflicts */
	if (a->conflicts) {
		a->conflicts = listSort(a->conflicts, cvCompare);
	}

	return TRUE;
}

/* --------------------------------------------------------------------
 Calculate all conflicts in an analysis.

 Sets the fields ->b and ->conflicts. If these fields were non-empty,
 the contents are deallocated.
 */
void aJudge(Analysis a) {
	aJudgeWithThreshold(a, 0.0);
}

/* ----------------------------------------------------------------------
 Re-compute conflicts after applying a patch.

 After applying a Patch to an Analysis, it is faster (~20% less time needed)
 not to evaluate the grammar on the entire new Analysis, but simply to
 re-evaluate it merely on the changed bits. This is slightly
 complicated by the fact that the LVs listed in the Patch are not
 the only ones that might have changed. If the Patch also changed
 the path though the lexeme graph, other LVs might have changed to
 follow suit.

 Basically this procedure first inspects every LV in a to see if it
 has changed. Then all conflicts caused by the changed LVs are
 erased, because those are the ones that may have been resolved.
 Finally the grammar is evaluated on the changed LVs.
 */
void aRejudge(Analysis a, Patch p) {

	ConstraintViolation cv;
    LevelValue lv, lvb;
	LexemNode ln;
	ByteVector changedLV = bvNew(a->P->nrSlots);
	ByteVector changedInterval = bvNew(a->P->maxTimepoint);

    List l, m;
	List lvList;
	List old = NULL, new = NULL;
    int i, j;

	Vector partialslots = vectorClone(a->slots);

	/* 1. Note down all LVs that were changed by the patch. */
	bvSetAllElements(changedLV, FALSE);

	/* explicitly changed LVs */
	for (i = 0; i < vectorSize(p->selection); i++) {
		lv = vectorElement(p->selection, i);
		if (!lv)
			continue;
		bvSetElement(changedLV, TRUE, lvIndex(lv));
	}

	/* implicitly changed LVs:
	 to find those, we note down all changed time intervals */
	if (p->newLexemes) {

		/* make vector of changed time intervals */
		bvSetAllElements(changedInterval, FALSE);
		for (l = p->newLexemes; l; l = l->next) {
			ln = l->item;
			bvSetElements(changedInterval, TRUE, ln->arc->from, ln->arc->to);
		}

		/* Mark those LVs that modify one of the new lexemes... */
		for (i = 0; i < vectorSize(a->slots); i++) {
			lv = vectorElement(a->slots, i);
			if (!lv)
				continue;
			if (!spec(lv->regents))
				continue;
			for (j = ((LexemNode) lv->regents->item)->arc->from; j
					< ((LexemNode) lv->regents->item)->arc->to; j++) {
				if (bvElement(changedInterval, j)) {
					bvSetElement(changedLV, TRUE, i);
					break;
				}
			}
		}

		/* ...or are modified by one of them */
		for (i = 0; i < vectorSize(a->slots); i++) {
			lv = vectorElement(a->slots, i);
			if (!lv)
				continue;
			for (j = ((LexemNode) lv->dependents->item)->arc->from; j
					< ((LexemNode) lv->dependents->item)->arc->to; j++) {
				if (bvElement(changedInterval, j)) {
					bvSetElement(changedLV, TRUE, i);
					break;
				}
			}
		}
	}

	/* 2. Remove conflicts that may have disappeared. */
	for (l = a->conflicts; l; l = l->next) {
		cv = (ConstraintViolation) l->item;

		/* caused by a changed lv? */
		if (bvElement(changedLV, lvIndex(cv->lv1)) || (cv->lv2 && bvElement(
				changedLV, lvIndex(cv->lv2))) || (cv->lv3 && bvElement(
				changedLV, lvIndex(cv->lv3))) ||

		/* context-sensitive conflict? This can disappear
		 whenever anything change at all. */
		cv->constraint->is_context_sensitive ||
		/* caused by an lv that is unchanged but has disappeared?
		 (This can only happen during initialization.) */
		vectorElement(a->slots, lvIndex(cv->lv1)) == NULL || (cv->lv2
				&& vectorElement(a->slots, lvIndex(cv->lv2)) == NULL)
				|| (cv->lv3 && vectorElement(a->slots, lvIndex(cv->lv3))
						== NULL)) {

			/* remove */

			/* Since we are going to remove this conflict,
			 we must remove its penalty as well. */
			if (cdgMultiply == cdgAggregation) {
				bSubtract(a->b, cv->penalty);
			}

			/* de-allocate */
			cv = NULL;

		} else {

			/* retain */
			old = listAppendElement(old, cv);
		}

	}

	/* de-allocate old list, don't free items:
	 each item was either freed above, or is needed in the new list. */
	a->conflicts = NULL;

	/* Under the minimum strategy, Badnesses cannot be computed in reverse.
	 Therefore, recompute the partial Badness from scratch. */
	if (cdgMinimum == cdgAggregation) {
		a->b->no = 0;
		a->b->hard = 0;
		a->b->soft = 1.0;
		for (l = old; l != NULL; l = l->next) {
			cv = l->item;
			bAdd(a->b, cv->penalty);
		}
	}

	/* 3. Calculate newly introduced conflicts. */

	/* make list of new LVs */
	lvList = NULL;
	for (i = 0; i < vectorSize(a->slots); i++) {
		if (!slotActive(a, i)) {
			vectorSetElement(partialslots, NULL, i);
			continue;
		}
		lv = vectorElement(a->slots, i);
		if (!lv)
			continue;
		if (bvElement(changedLV, i)) {
			lvList = listAppendElement(lvList, lv);
		}
	}

	lock_tree(a->P->maxTimepoint);

	/* seek new unary conflicts */
	for (l = lvList; l; l = l->next) {
		lv = l->item;
		(void) evalUnary(lv, a->P->net, NULL, FALSE, a->b, &new, FALSE);

		/* seek new binary conflicts between changed & unchanged LVs */
		for (i = 0; i < vectorSize(a->slots); i++) {
			if (!slotActive(a, i))
				continue;
			lvb = (LevelValue) vectorElement(a->slots, i);
			if (!lvb)
				continue;

			/* don't evaluate against another member of lvList */
			if (bvElement(changedLV, i))
				continue;

			(void) evalBinary(lv, lvb, a->P->net, NULL, FALSE, a->b, &new);
#ifdef TERNARY
            if(inputCurrentGrammar->no_ternary) {
			/* seek new ternary conflicts between one changed LV & two unchanged LVs */
            int k;
			for (k = i + 1; k < vectorSize(a->slots); k++) {

				if (!slotActive(a, k)) {
					continue;
				}
                LevelValue lvc = (LevelValue) vectorElement(a->slots, k);
				if (!lvc) {
					continue;
				}

				/* don't evaluate against another member of lvList */
				if (bvElement(changedLV, k)) {
					continue;
				}

                (void) evalTernary(lv, lvb, lvc, a->P->net, NULL, FALSE, a->b, &new);

			}
            }
#endif
		}

		/* seek new binary conflicts between changed LVs */
		for (m = l->next; m; m = m->next) {
			lvb = m->item;
			(void) evalBinary(lv, lvb, a->P->net, NULL, FALSE, a->b, &new);
#ifdef TERNARY
            if(inputCurrentGrammar->no_ternary) {
			/* seek new ternary conflicts between two changed LVs & one unchanged LV */
            int k;
			for (k = 0; k < vectorSize(a->slots); k++) {

				if (!slotActive(a, k)) {
					continue;
				}
                LevelValue lvc = (LevelValue) vectorElement(a->slots, k);
				if (!lvc) {
					continue;
				}

				/* don't evaluate against another member of lvList */
				if (bvElement(changedLV, k)) {
					continue;
				}

                (void) evalTernary(lv, lvb, lvc, a->P->net, NULL, FALSE, a->b, &new);

			}

			/* seek new ternary conflicts between three changed LVs */
            List n;
			for (n = m->next; n; n = n->next) {
                LevelValue lvc = n->item;
                (void) evalTernary(lv, lvb, lvc, a->P->net, NULL, FALSE, a->b, &new);
			}
            }
#endif
		}
	}

	/* find context-sensitive conflicts */
	evalInContext(partialslots, a->slots, a->b, &new);
	partialslots = NULL;

	unlock_tree();

	/* sort new conflicts */
	if (new) {
		l = new;
		new = listSort(new, cvCompare);
		l = NULL;
	}

	/* ensure that new conflicts precede old conflicts */
	a->conflicts = listAppendList(new, old);

	lvList = NULL;

	changedLV = NULL;
	changedInterval = NULL;

}

/* ----------------------------------------------------------------------
 Allocate an Analysis.

 This function allocates a new Analysis for the Problem P. It does not
 allocate all fields of the structure, since the first Analysis in a
 Problem will be completed by net2problem(), and further structures are
 produced by aClone().

 */
Analysis aNew(Problem P) {

	Analysis result = memMalloc(sizeof (AnalysisStruct));
	result->P = P;
	result->slots = vectorNew(P->nrSlots);
	result->isUnselectedLexeme = bvNew(P->nrLexemes);
	result->b = bestBadness();
	result->conflicts = NULL;
	result->subStart = P->minTimepoint;
	result->subStop = P->maxTimepoint;

	vectorSetAllElements(result->slots, NULL);
	bvSetAllElements(result->isUnselectedLexeme, TRUE);

	return result;
}

/* ----------------------------------------------------------------------
 Clone an Analysis.

 This function creates a deep copy of A, except for the constituting LVs,
 which are usually simply copied rather than cloned. LVs are only cloned
 if they are not owned by the underlying constraint net (i.e., have
 indexWRTNet == -1).
 */
Analysis aClone(Analysis a) {
	int i;
	LevelValue lv;
	List l;
	ConstraintViolation cv;

	Analysis result = memMalloc(sizeof (AnalysisStruct));
	result->P = a->P;

	/* mostly shallow copy of LVs */
	result->slots = vectorClone(a->slots);
	for (i = 0; i < vectorSize(result->slots); i++) {
		lv = vectorElement(a->slots, i);
		if (!lv) {
			continue;
		}
		if (lv->indexWRTNet == -1) {
			lv = lvClone(lv);
			vectorSetElement(result->slots, lv, i);
		}
	}

	/* deep copy of everything else */
	result->isUnselectedLexeme = bvClone(a->isUnselectedLexeme);
	result->conflicts = NULL;
	for (l = a->conflicts; l; l = l->next) {
		cv = cvClone((ConstraintViolation) l->item);
		result->conflicts = listAppendElement(result->conflicts, cv);
	}
	result->b = bClone(a->b);
	result->subStart = a->subStart;
	result->subStop = a->subStop;

	return result;

}

/* ----------------------------------------------------------------------
 Return TRUE if A has no hard violations.
 */
Boolean aValid(Analysis a) {

	return (a->b->hard == 0);

}

/* ----------------------------------------------------------------------
 Re-calculate the soft score of an Analysis by multiplying the
 scores of all soft conflicts.

 This is to prevent rounding errors after repeated division and
 multiplication during the repair process.
 */
void aCalibrate(Analysis a) {

	List l;
	ConstraintViolation cv;
	Number score = 1;

	if (cdgMinimum == cdgAggregation)
		return;

	for (l = a->conflicts; l; l = l->next) {
		cv = (ConstraintViolation) l->item;
		if (cv->penalty)
			score *= cv->penalty;
	}

	if (score != a->b->soft) {

#ifdef DEBUGACALIBRATE
		cdgPrintf(CDG_DEBUG, "adjusted by %4.3e)\n", score - a->b->soft);
#endif

		a->b->soft = score;

	}
}

/* ----------------------------------------------------------------------
 Check that an LV is consistent internally _and_ with the
 containing Analysis.
 */
void checkValidity(Analysis a, LevelValue lv) {

	if (isValid(a, lv, TRUE)) {
		return;
	}

	if (spec(lv->regents) && !selectedLexeme(a, lv->regents)) {
		cdgPrintf(CDG_ERROR, "\nPANIC: regents of #%d are all unselected:\n",
				lv->indexWRTNet);
		lvPrint(CDG_ERROR, lv, NULL, 2);
		cdgPrintf(CDG_ERROR, "\n");
		printSelectedPath(CDG_ERROR, a);
		abort();
	}

	if (!selectedLexeme(a, lv->dependents)) {
		cdgPrintf(CDG_ERROR,
				"\nPANIC: dependents of #%d are all unselected:\n",
				lv->indexWRTNet);
		lvPrint(CDG_ERROR, lv, NULL, 2);
		cdgPrintf(CDG_ERROR, "\n");
		printSelectedPath(CDG_ERROR, a);
		abort();
	}
}

/* ----------------------------------------------------------------------
 Show which LVs are not the unary best ones.

 Displays all domains that do not have their optimal element selected
 (i.e. the one with the highest limit).

 */
void showUnary(Analysis a) {

	int i;
	LevelValue lv, optimal;
	Number x, y;

	for (i = 0; i < vectorSize(a->slots); i++) {
		lv = vectorElement(a->slots, i);
		if (!lv)
			continue;

		optimal = a->P->domains[i]->item;

		x = getLimit(a->P, lv);
		y = getLimit(a->P, optimal);

		if (x < y) {
			cdgPrintf(CDG_INFO, "@%d: #%d limit %4.3e, #%d limit %4.3e\n", i,
					lv->indexWRTNet, x, optimal->indexWRTNet, y);
		}
	}

	cdgPrintf(CDG_INFO, "\n\n");

}

/* ----------------------------------------------------------------------
 Selects list of lexemes and switches all LVs correspondingly.

 Substitutes all lexeme nodes in WHICH into A. All previously selected
 LVs are changed just enough so that they comply with the new path
 selected. Naturally, this may make part of the structure and the labels
 of the Analysis totally meaningless if the new path differs
 substantially from the old one.

 If forced is set, new LevelValues might be created
 if no existing one fits the new path.

 Returns false, if path selection fails due to missing LV
 and forced = false. In this case the resulting analysis might be corrupt!
 In every other case true is returned
 */
Boolean selectPath(Analysis a, List which, Boolean forced) {
	LevelValue oldLV, newLV, original;
	LexemNode ln, lnold, lnnew;
	List l;
	int index, i, j;
	ByteVector changed;

	/* It is admissible to call selectPath() vacuously. */
	if (!which) {
		return TRUE;
	}

	changed = bvNew(a->P->maxTimepoint);
	bvSetAllElements(changed, FALSE);

#ifdef DEBUGSELECTPATH
	cdgPrintf(CDG_DEBUG, "old path:\n");
	for (i = 0; i < a->P->nrLexemes; i++) {
		ln = vectorElement(a->P->lg->nodes, i);
		if (bvElement(a->isUnselectedLexeme, i))
		continue;
		cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
		lgPrintNode(CDG_DEBUG, ln);
		cdgPrintf(CDG_DEBUG, "\n");
	}
	cdgPrintf(CDG_DEBUG, "\nselecting path:\n");
	for (l = which; l; l = l->next) {
		ln = l->item;
		cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
		lgPrintNode(CDG_DEBUG, ln);
		cdgPrintf(CDG_DEBUG, "\n");
	}
#endif

	for (l = which; l; l = l->next) {
		lnnew = l->item;
		if (!bvElement(a->isUnselectedLexeme, lnnew->no)) {
			continue;
		}
		for (i = lnnew->arc->from; i < lnnew->arc->to; i++) {
#ifdef DEBUGSELECTPATH
			cdgPrintf(CDG_DEBUG,
					"  interval %d has changed in word `%s'.\n",
					i, lnnew->lexem->description);
#endif
			bvSetElement(changed, TRUE, i);
		}
	}

	/* unselect all lexemes that cover changed timespans */
	for (i = 0; i < a->P->nrLexemes; i++) {
		lnold = vectorElement(a->P->lg->nodes, i);
		if (!lnold) {
			continue;
		}
		for (j = lnold->arc->from; j < lnold->arc->to; j++) {
			if (bvElement(changed, j)) {
				if (!bvElement(a->isUnselectedLexeme, lnold->no)) {
					bvSetElement(a->isUnselectedLexeme, TRUE, lnold->no);
#ifdef DEBUGSELECTPATH
					cdgPrintf(CDG_DEBUG, "  unselecting ");
					lgPrintNode(CDG_DEBUG, lnold);
					cdgPrintf(CDG_DEBUG, " because of changed time span %d.\n", j);
#endif
					break;
				}
			}
		}
	}

	/* record the selection */
	for (l = which; l; l = l->next) {
		lnnew = l->item;
		bvSetElement(a->isUnselectedLexeme, FALSE, lnnew->no);
	}

#ifdef DEBUGSELECTPATH
	cdgPrintf(CDG_DEBUG, "new path:\n");
	for (i = 0; i < a->P->nrLexemes; i++) {
		ln = vectorElement(a->P->lg->nodes, i);
		if (bvElement(a->isUnselectedLexeme, i))
		continue;
		cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
		lgPrintNode(CDG_DEBUG, ln);
		cdgPrintf(CDG_DEBUG, "\n");
	}
#endif

	/* exchange all LVs that bind unselected lexemes. These are:

	 [A] LVs that use one of them as a modifiee. They are simply
	 switched to the simultaneously selected lexeme. If that results
	 in an LV that already exists, no additional LV must be allocated.

	 To find them, we must iterate over all LVs. */
	for (index = 0; index < vectorSize(a->slots); index++) {
		oldLV = (LevelValue) vectorElement(a->slots, index);

		/* hit? */
		if (!oldLV || !(spec(oldLV->regents)) || selectedLexeme(a,
				oldLV->regents)) {
			continue;
		}

#ifdef DEBUGSELECTPATH
		cdgPrintf(CDG_DEBUG, "[A] have to replace #%d in @%d\n", oldLV->indexWRTNet, index);
#endif

		/* clone the LV (perhaps only temporarily) */
		newLV = lvClone(oldLV);

		/* realize change */
		ln = selectedAt(a, ((LexemNode) oldLV->regents->item)->arc->from);
		newLV->regents = listAppendElement(NULL, ln);
		newLV->indexWRTNet = -1;
		newLV->no = -1;

		/* If the swap produced an LV that already exists,
		 take this one and delete the clone. */
		if ((original = findExistingLV(a, newLV)) != NULL) {
			newLV = original;
		} else {

			if(forced) {
				/* A manipulated LV may even be self-modifying. */
				breakCircle(newLV);
			} else {
				// there is probably a good reason no such LV exists, fail
				return FALSE;
			}

		}

		/* Finally, substitute the new LV into the Analysis. */
		vectorSetElement(a->slots, newLV, index);
		/* This overwrites the old content of the slot,
		 so there is no need to remove the old LV. */
#ifdef DEBUGSELECTPATH
		cdgPrintf(CDG_DEBUG, "replaced with #%d.\n", newLV->indexWRTNet);
#endif

		/* de-allocate temporary */
		if (oldLV->indexWRTNet == -1) {
			oldLV = NULL;
		}
	}

	/* [B] Those LVs that use an unselected lexeme as a modifier.
	 Note that an LV may fall under both category [A] and [B]. */
	for (index = 0; index < vectorSize(a->slots); index++) {
		oldLV = (LevelValue) vectorElement(a->slots, index);

		/* hit? */
		if (!oldLV || selectedLexeme(a, oldLV->dependents)) {
			continue;
		}

		/* since case [B] may replace an LV from domain @X
		 with an lV that goes into domain @Y,
		 we must take care to remove the old LV from its slot. */
		vectorSetElement(a->slots, NULL, index);

#ifdef DEBUGSELECTPATH
		cdgPrintf(CDG_DEBUG, "[B] have to replace #%d in @%d\n", oldLV->indexWRTNet, index);
#endif

		/* Beware of holes in the Analysis!

		 If we replace the lexeme

		 DATUM(0-3)

		 by the lexemes

		 der(0-1) erste(1-2) April(2-3),

		 then the old LV bound three domains. We cannot simply replace
		 this LV with one that binds der(0-1) because that would leave
		 two domains vacant. Instead we inspect all of these domains and
		 make sure that they all are bound. */
		i = ((LexemNode) oldLV->dependents->item)->arc->from;
		while (i < ((LexemNode) oldLV->dependents->item)->arc->to) {

			/* clone the LV (perhaps only temporarily) */
			newLV = lvClone(oldLV);

			/* realize change */
			ln = selectedAt(a, i);
			newLV->dependents = listAppendElement(NULL, ln);
			newLV->indexWRTNet = -1;
			newLV->no = -1;

			/* If the swap produced an LV that already exists, take this one and delete the clone. */
			if ((original = findExistingLV(a, newLV)) != NULL) {
				newLV = original;
			} else {
				if(forced) {
					/* A manipulated LV may even be self-modifying. */
					breakCircle(newLV);
				} else {
					// there is probably a good reason no such LV exists, fail
					return FALSE;
				}
			}

			/* Finally, substitute the LV into the Analysis. */
			vectorSetElement(a->slots, newLV, lvIndex(newLV));
#ifdef DEBUGSELECTPATH
			cdgPrintf(CDG_DEBUG, "replaced with #%d.\n", newLV->indexWRTNet);
#endif

			/* step over this interval */
			i = ln->arc->to;
		}

		/* de-allocate temporary */
		if (oldLV->indexWRTNet == -1) {
			oldLV = NULL;
		}
	}
	changed = NULL;

	return TRUE;
}

/* ----------------------------------------------------------------------
 Let A have an arbitrary, but consistent and complete lexeme selection.
 */
Boolean aBuildPath(Analysis a) {

	List l = lgMakePath(a->P->net->lexemgraph, NULL);
	/* This can only fail if there is no path though the lattice at all.
	 Ideally, CDG should have refused to build a net for such a graph,
	 or the lattice should have been rejected at load time. But just
	 in case an invalid lexeme graph sneaks through here, test it
	 anyway. */
	if (!l) {
		cdgPrintf(CDG_WARNING,
				"\nWARNING: no path through the lexeme graph possible, bailing out...\n");
		return FALSE;
	}

	selectPath(a, l, TRUE);
	l = NULL;
	return TRUE;
}

/* ----------------------------------------------------------------------
 Change A so that it contains LV.

 This function is essentially a wrapper for pNew()+apply(). This means
 that it takes care of the path criterion itself; you can insert any LV
 into any Analysis.
 */
void aInsert(Analysis a, LevelValue lv) {
	Patch p = pNew(a, 1, OD_DELETED);
	Patch q = pNext(p, lv, OD_ALLOWALL | OD_DELETED);
	p = NULL;
	(void) pLexicalize(q, OD_DELETED, FALSE);
	apply(a, q);
}

/* Functions dealing with Patches
 ============================== */

/* ----------------------------------------------------------------------
 Return new Patch.

 Returns a new WIDTH-fold Patch and initializes the internal control
 variables. The result contains no LV yet.

 */
Patch pNew(Analysis a, int width, int flags) {

	Patch result = memMalloc(sizeof (PatchStruct));

	result->a = a;
	result->local = bestBadness();
	result->limit = 1.0;
	result->selection = vectorNew(width);
	result->index = 0;
	result->nrNewLVs = 0;
	result->predicted = NULL;
	result->newLexemes = NULL;
	result->conflicts = NULL;
	result->bound = bvNew(a->P->nrSlots);

	/* normally, disallow deleted lexeme nodes implicitly */
	result->forbiddenLexemes = bvClone(a->P->lg->isDeletedNode);

	/* upon request, allow them all */
	if (flags & OD_DELETED) {
		bvSetAllElements(result->forbiddenLexemes, FALSE);
	}

	bvSetAllElements(result->bound, FALSE);

	return result;
}

/* ----------------------------------------------------------------------
 Clone a Patch.

 Only the PatchStruct itself, the Badness and the Vectors and Lists are
 actually allocated anew, not their contents.
 */
Patch pClone(Patch p) {

	Patch result;

	if (!p) {
		return NULL;
	}

	result = pNew(p->a, vectorSize(p->selection), 0);

	result->index = p->index;
	result-> nrNewLVs = p->nrNewLVs;
	result->limit = p->limit;

	/* deep copy */
	vectorCopy(result->selection, p->selection);
	result->newLexemes = listClone(p->newLexemes);
	result->conflicts = listClone(p->conflicts);

	bvCopy(result->bound, p->bound);
	bvCopy(result->forbiddenLexemes, p->forbiddenLexemes);

	if (p->predicted)
		result->predicted = bClone(p->predicted);

	if (p->local)
		bCopy(result->local, p->local);
	else
		result->local = NULL;

	return result;

}

/* ----------------------------------------------------------------------
 Build a Patch that binds one more LV than the original one.

 Usually, Patches that cannot help in solving the problem are not built,
 and NULL is returned instead. A Patch is considered unhelpful if it
 contains a deleted LV, or has a lower limit than an existing solution,
 or has score 0.

 This is normally a Good Thing, but in some places we expressly want to
 build a Patch even though we know it won't help us, so the pruning can
 be suppressed by setting OD_DELETED.

 In no case will pNext() ever construct a Patch with physically
 incompatible LVs.
 */
Patch pNext(Patch old, LevelValue lv, int flags) {

	LevelValue current;
	int i, j = 0;
	Badness b;
	Boolean prune = ((flags & OD_DELETED) == 0);
	Boolean debug = (flags & OD_DEBUG) != 0;
	Patch new;

	/* reject deleted LVs */
	if (prune && lv->isDeleted) {
		if (debug) {
			cdgPrintf(CDG_DEBUG, "DEBUG: #%d is deleted.\n", lv->indexWRTNet);
		}
		return NULL;
	}

	/* new LV must satisfy the search flags */
	current = slotFiller(old->a, lvIndex(lv));

	if ((flags & OD_ORIENT) == 0) {
		if (current) {
			int new = -1;
			int old = -1;
			if (spec(current->regents)) {
				old = ((LexemNode) current->regents->item)->arc->from;
			}
			if (spec(lv->regents)) {
				new = ((LexemNode) lv->regents->item)->arc->from;
			}
			if (old != new) {
				if (debug) {
					cdgPrintf(CDG_DEBUG,
							"DEBUG: #%d would change the structure\n",
							lv->indexWRTNet);
				}
				return NULL;
			}
		}
	}
	if ((flags & OD_LABEL) == 0) {
		if (current) {
			if (current->label != lv->label) {
				if (debug) {
					cdgPrintf(CDG_DEBUG, "DEBUG: #%d would change the label\n",
							lv->indexWRTNet);
				}
				return NULL;
			}
		}
	}
	if ((flags & OD_LEXICAL) == 0) {
		if (!isValid(old->a, lv, TRUE)) {
			if (debug) {
				cdgPrintf(CDG_DEBUG, "DEBUG: #%d would change the lexemes\n",
						lv->indexWRTNet);
			}
			return NULL;
		}
	}

	/* ...and the path criterion */
	if (!lvVectorCompatible(lv, old->forbiddenLexemes)) {
		if (debug) {
			cdgPrintf(CDG_DEBUG, "DEBUG: #%d is lexically incompatible.\n",
					lv->indexWRTNet);
		}
		return NULL;
	}

	List conflicts = listNew();

	/* ...and the grammaticality constraints */
	b = bestBadness();
	for (i = 0; i < vectorSize(old->selection); i++) {
		LevelValue oldlv = (LevelValue) vectorElement(old->selection, i);
		if (!oldlv)
			continue;
		if (prune
				&& evalBinary(lv, oldlv, old->a->P->net, NULL, FALSE, b, &conflicts)
						== 0.0) {
			if (debug) {
				cdgPrintf(CDG_DEBUG, "DEBUG: pruning Patch with score 0.0\n");
			}
			b = NULL;
			return NULL;
		}
	}

	/* build new Patch */
	new = pClone(old);

	new->conflicts = listAppendList(listClone(old->conflicts), conflicts);

	/* extend by one LV */
	vectorSetElement(new->selection, lv, new->index);
	new->index++;

	/* note down all domains bound by this LV */
	for (i = ((LexemNode) lv->dependents->item)->arc->from; i
			< ((LexemNode) lv->dependents->item)->arc->to; i++) {
		j = old->a->P->nrLevels * i + lv->level->no;
		bvSetElement(new->bound, TRUE, j);
	}

	/* update Badness */
	bAddBadness(new->local, b);
	if (lv->score != 1.0) {
		bAdd(new->local, lv->score);
	}
	/* This is incorrect. A score of f < 1.0 for the new LV may have
	 been caused by several unary faults, so it might have to be
	 represented by (2/0/f) rather than (1/0/f). But we ignore this
	 for the time being. */

	/* update limits */
	if (new->local->soft < new->limit) {
		new->limit = new->local->soft;
	}
	if (lv->limit < new->limit) {
		new->limit = lv->limit;
	}

	/* reject Patches with limit that is too low */
	if (prune) {

		if (significantlyGreater(old->a->P->globalMinimum, new->limit)) {
			if (debug) {
				cdgPrintf(CDG_DEBUG, "DEBUG: limit %4.3e too low\n", new->limit);
			}
			new = NULL;
			b = NULL;
			return NULL;
		}
	}

	/* update compatibility Vector */
	lgRequireLexemes(old->a->P->lg, new->forbiddenLexemes, lv->regents);
	lgRequireLexemes(old->a->P->lg, new->forbiddenLexemes, lv->dependents);

	b = NULL;
	return new;

}

/* ----------------------------------------------------------------------
 Compare Patches, first by limit, then by local Badness.
 */
Boolean pCompare(Patch a, Patch b) {

	if (a->limit > b->limit)
		return TRUE;
	if (a->limit < b->limit)
		return FALSE;

	return (bCompare(a->local, b->local));

}

/* ----------------------------------------------------------------------
 Print exactly what effect P has on A
 */
void pDebug(Analysis a, Patch p) {
	List l;
	Analysis newA = aClone(a);
	apply(newA, pClone(p));
	for (l = newA->conflicts; l != NULL; l = l->next) {
		ConstraintViolation new = l->item;
		if (!cvContains(a->conflicts, new)) {
			cdgPrintf(CDG_DEBUG, "DEBUG: introduces ");
			cvPrint(CDG_DEBUG, new);
			cdgPrintf(CDG_DEBUG, "\n");

		}
	}
	for (l = a->conflicts; l != NULL; l = l->next) {
		ConstraintViolation old = l->item;
		if (!cvContains(newA->conflicts, old)) {
			cdgPrintf(CDG_DEBUG, "DEBUG: removes ");
			cvPrint(CDG_DEBUG, old);
			cdgPrintf(CDG_DEBUG, "\n");
		}
	}

	newA = NULL;

}

/* ----------------------------------------------------------------------
 Display a Patch.

 This function displays a Patch using cdgPrintf().

 If BRIEF is set, self-substitutions are omitted from the output for
 better readability.
 */
void pPrint(unsigned long mode, Patch p, Boolean brief) {

	int i;
	LevelValue lv;
	List l;
	LevelValue original;
	Boolean empty = TRUE;

	/* show new LVs */
	for (i = 0; i < vectorSize(p->selection); i++) {
		lv = (LevelValue) vectorElement(p->selection, i);
		if (!lv)
			continue;

		/* only show those LVs that actually are new */
		original = vectorElement(p->a->slots, lvIndex(lv));
		if (brief && original && original->indexWRTNet == lv->indexWRTNet)
			continue;

		cdgPrintf(mode, "#%d ", lv->indexWRTNet);
		empty = FALSE;
	}

	if (empty) {
		cdgPrintf(mode, "[no change]{%d} ", vectorSize(p->selection));
	}

	cdgPrintf(mode, "/ ");

	/* show new lexeme nodes */
	for (l = p->newLexemes; l; l = l->next) {
		lgPrintNode(mode, l->item);
		cdgPrintf(mode, " ");
	}

	/* show local score */
	cdgPrintf(mode, "%4.3e ", p->local->soft);
	if (p->local->soft > p->limit) {
		cdgPrintf(mode, "limit %4.3e  ", p->limit);
	}

	/* show predicted Badness */
	if (p->predicted) {
		bPrint(mode, p->predicted);
		cdgPrintf(mode, " predicted");
	}
	cdgPrintf(mode, "\n");

}

/**
 * implementation for apply and applyIfPossible
 */
Boolean applyImpl(Analysis a, Patch p, Boolean forced) {
	LevelValue source;
	int i;

	/* swap in all new lexemes */
	if(!selectPath(a, p->newLexemes, forced)) {
		if(forced) {
			cdgPrintf(CDG_ERROR, "ERROR: select path failed even though forced was set!");
		}
		return FALSE;
	}

	/* insert all new LVs */
	for (i = 0; i < vectorSize(p->selection); i++) {
		source = vectorElement(p->selection, i);
		if (!source) {
			continue;
		}
		selectLV(a, source);
	}

	/* rejudge in all the necessary places */
	aRejudge(a, p);
	aCalibrate(a);

	/* cleanup */
	p = NULL;

	return TRUE;
}

/* ----------------------------------------------------------------------
 Apply a Patch to an Analysis. P must already lexicalized.

 After apply(),
 - A contains all Lvs from P
 - all LVs from P are valid in A
 - A is path consistent

 Note that this deletes the Patch when done, so a Patch can only be
 applied once.

 It is not meaningful to apply a Patch to any other Analysis than the
 one it was built for. A different but equivalent Analysis is fine,
 though -- that is in fact what pJudge() does.

 */
void apply(Analysis a, Patch p) {
	applyImpl(a, p, TRUE);
}

/*
 * like apply, but can fail if forced = false
 */
Boolean applyIfPossible(Analysis a, Patch p) {
	return applyImpl(a, p, FALSE);
}


/* ----------------------------------------------------------------------
 Predict accurately the Badness that applying P would yield.

 To this end, the Patch is temporarily applied to its Analysis and the
 resulting Badness is noted.

 if forced is set, levelValues not available in the net might be used,
 if not set, the score would be set to worst instead
 */
void pJudge(Patch p, Boolean forced) {

	Analysis newA = aClone(p->a);

	Boolean applyable;

	if(forced) {
		applyable = TRUE;
		apply(newA, pClone(p));
	} else {
		applyable = applyIfPossible(newA, pClone(p));
	}

	if (p->predicted) {
		p->predicted = NULL;
	}
	if(applyable) {
		p->predicted = bClone(newA->b);
	} else {
		// we have to expect the worst
		p->predicted = bNew(INT_MAX, INT_MAX, 0.0);
	}
	newA = NULL;

}

/* -----------------------------------------------------------------------
 Compute the set of lexemes that have to be substituted into A so
 that P may be applied.

 Basically, this function computes p->newLexemes from
 p->forbiddenLexemes.

 Rationale: adding an LV to a Patch usually decreases the number of
 lexemes that can be used in the new path. But there is no guarantee that
 all LVs requested together *uniquely* define a set of lexemes that has
 to be swapped in. Thus the space of possible lexemes is narrowed down in
 pNext(), and the positive list of lexemes to be used is built here.

 Returns FALSE if no path can be built, TRUE otherwise.

 Usually, this function does not select deleted lexeme nodes. Therefore,
 if a Patch was built with OD_DELETED, the flag must also be given to
 pLexicalize(). This is the only flag that pLexicalize() reads.

 If P was not built with OD_DELETED, this function can only fail because
 the lexeme graph has become altogether disconnected, e.g. there is no
 path at all that spans the entire graph.

 It is a critical error if the lexicalisation fails (this should be
 changed to something more sensible). As it is, the caller is guaranteed
 to get a valid path if one exists, or an assertion otherwise. No
 provision is made to create a path that is as `nearly' valid as
 possible.

 */
Boolean pLexicalize(Patch p, int flags, Boolean debugme) {
	int i, j, k;
	LexemNode ln, orig, found;
	Boolean exactfit;
	Analysis a = p->a;
	LexemGraph lg = a->P->lg;

	/* stores the hypothetically selected lexeme nodes per time point */
	Vector bound = vectorNew(a->P->maxTimepoint);

	/* As of now, all time points are bound. */
	int coverage = a->P->nrTimepoints;

	if (debugme) {
		cdgPrintf(CDG_DEBUG,
				"DEBUG: pLexicalize() called. These are the new LVs:\n");
	}
	for (i = 0; i < vectorSize(p->selection); i++) {
		LevelValue lv = vectorElement(p->selection, i);
		if (lv) {
			if (debugme) {
				lvPrint(CDG_DEBUG, lv, NULL, 2);
				cdgPrintf(CDG_DEBUG, "\n");
			}

		}
	}

	/* if the lexeme graph has become disconnected,
	 pLexicalize() cannot possibly succeed. */
	if (lg->noOfPaths == 0) {
		if (debugme) {
			cdgPrintf(CDG_DEBUG,
					"DEBUG: the graph is disconnected. Giving up.\n\n");
		}
		return FALSE;
	}

	/* Otherwise it must succeed. If it doesn't,
	 that is a programming error. */
	if (debugme) {
		{
			LevelValue lv1, lv2;
			for (i = 0; i < vectorSize(p->selection); i++) {
				lv1 = vectorElement(p->selection, i);
				for (j = 0; j < vectorSize(p->selection); j++) {
					lv2 = vectorElement(p->selection, j);
					if (!lvCompatible(p->a->P->lg, lv1, lv2)) {
						cdgPrintf(CDG_DEBUG,
								"PANIC: pNext is about to build an invalid Patch!\n");
						abort();
					} else {
						cdgPrintf(CDG_DEBUG, "#%d - #%d: compatible.\n",
								lv1->indexWRTNet, lv2->indexWRTNet);
					}
				}
			}
		}
	}

	if (debugme) {
		cdgPrintf(CDG_DEBUG, "checking %d time points.\n", a->P->nrTimepoints);
		cdgPrintf(CDG_DEBUG, "Old path:\n");
		printSelectedPath(CDG_DEBUG, a);
		cdgPrintf(CDG_DEBUG, "Here is the entire lexeme graph:\n");
		lgPrint(CDG_DEBUG, lg);

	}

	i = a->P->minTimepoint;
	while (i < a->P->maxTimepoint) {
		ln = selectedAt(a, i);
		for (j = ln->arc->from; j < ln->arc->to; j++) {
			vectorSetElement(bound, (Pointer) ln, j);
		}
		i = ln->arc->to;
	}

	/* Look for points that must be unbound. */
	i = a->P->minTimepoint;
	while (i < a->P->maxTimepoint) {
		ln = vectorElement(bound, i);

		if (bvElement(p->forbiddenLexemes, ln->no)) {
			coverage -= (ln->arc->to - ln->arc->from);
			for (j = ln->arc->from; j < ln->arc->to; j++) {
				vectorSetElement(bound, NULL, j);
			}
			if (debugme) {
				cdgPrintf(CDG_DEBUG, "unbinding ");
				lgPrintNode(CDG_DEBUG, ln);
				cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
				if (coverage > a->P->nrTimepoints) {
					cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
					abort();
				}
			}
		}

		/* step to next (possibly) unbound time point */
		i = ln->arc->to;
	}

	while (coverage < a->P->nrTimepoints) {
		if (debugme) {
			cdgPrintf(CDG_DEBUG, "%d points bound, still patching...\n",
					coverage);
		}

		/* Find replacements. */
		i = a->P->minTimepoint;
		while (i < a->P->maxTimepoint) {
			if (debugme) {
				cdgPrintf(CDG_DEBUG, "checking point %d...\n", i);
			}
			if (vectorElement(bound, i)) {
				i = ((LexemNode) vectorElement(bound, i))->arc->to;
				continue;
			}
			/* Find a new lexeme binding point #i. */
			orig = selectedAt(a, i);
			if (debugme) {
				cdgPrintf(CDG_DEBUG, "Trying to replace ");
				lgPrintNode(CDG_DEBUG, orig);
				cdgPrintf(CDG_DEBUG, " at point %d\n", i);
			}
			found = NULL;
			exactfit = FALSE;
			for (j = 0; j < a->P->lgWidth; j++) {
				ln = vectorElement(a->P->lexemes, a->P->lgWidth * i + j);
				if (!ln) {
					continue;
				}
				if (debugme) {
					cdgPrintf(CDG_DEBUG, "Trying ");
					lgPrintNode(CDG_DEBUG, ln);
					cdgPrintf(CDG_DEBUG, "... ");
				}
				if (!(flags & OD_DELETED) && bvElement(lg->isDeletedNode,
						ln->no)) {
					if (debugme) {
						cdgPrintf(CDG_DEBUG, "deleted.\n");
					}
					continue;
				}
				if (bvElement(p->forbiddenLexemes, ln->no)) {
					if (debugme) {
						cdgPrintf(CDG_DEBUG, "forbidden.\n");
					}
					continue;
				}
				/* Try very hard to make a good choice. We want to prefer nodes
				 that bind exactly the same time slot as before, because then
				 we don't have to make subsidiary swaps. Secondarily, we want
				 nodes with high limits because those lead to better solutions.
				 */
				if (exactfit && (orig->arc->from != ln->arc->from
						|| orig->arc->from != ln->arc->from)) {
					if (debugme) {
						cdgPrintf(CDG_DEBUG, "not an exact fit.\n");
					}
					continue;
				}
				if (found && found->limit > ln->limit) {
					if (debugme) {
						cdgPrintf(CDG_DEBUG, "got a better one already.\n");
					}
					continue;
				}
				if (debugme) {
					cdgPrintf(CDG_DEBUG, "OK.\n");
				}
				found = ln;
				exactfit = (orig->arc->from == ln->arc->from && orig->arc->from
						== ln->arc->from);

			}

			if (!found) {
				if (debugme) {
					abort();
				} else {
					cdgPrintf(CDG_DEBUG,
							"PANIC: couldn't lexicalize Patch, transcript follows...\n");
					pLexicalize(p, flags, TRUE);
				}
			}

			/* OK, we'll take it. */
			lgRequireLexeme(lg, p->forbiddenLexemes, found);

			/* clear all lexemes that overlap it... */
			j = found->arc->from;
			while (j < found->arc->to) {
				if (NULL != (ln = vectorElement(bound, j))) {
					coverage -= (ln->arc->to - ln->arc->from);
					for (k = ln->arc->from; k < ln->arc->to; k++) {
						vectorSetElement(bound, NULL, k);
					}
					j = ln->arc->to;
					if (debugme) {
						cdgPrintf(CDG_DEBUG, "unbinding ");
						lgPrintNode(CDG_DEBUG, ln);
						cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
						if (coverage > a->P->nrTimepoints) {
							cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
							exit(1);
						}
					}
				} else {
					j++;
				}
			}

			/* ...and insert it. */
			for (j = found->arc->from; j < found->arc->to; j++) {
				vectorSetElement(bound, found, j);
				coverage++;
			}
			if (debugme) {
				cdgPrintf(CDG_DEBUG, "binding ");
				lgPrintNode(CDG_DEBUG, found);
				cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
				if (coverage > a->P->nrTimepoints) {
					cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
					abort();
				}
			}
			i = found->arc->to;
		}
	}

	/* translate array into lexeme List */
	i = a->P->minTimepoint;
	while (i < a->P->maxTimepoint) {
		ln = vectorElement(bound, i);
		if (ln != selectedAt(a, i)) {
			p->newLexemes = listPrependElement(p->newLexemes, ln);
		}
		if (!ln) {
			cdgPrintf(CDG_DEBUG, "PANIC: pLexicalize failed to bind %d!\n", i);
			abort();
		}
		i = ln->arc->to;
	}
	bound = NULL;

	return TRUE;
}

/* ----------------------------------------------------------------------
 Does P imply a lexical change when applied to A?
 */
Boolean pIsLexical(Analysis a, Patch p) {

	int i;
	LevelValue lv;

	for (i = 0; i < vectorSize(p->selection); i++) {
		lv = vectorElement(p->selection, i);
		if (!lv) {
			continue;
		}
		if (!isValid(a, lv, TRUE)) {
			return TRUE;
		}
	}
	return FALSE;

}

/* ----------------------------------------------------------------------
 If P were applied to A, would A then repeat a conflict from
 a->P->removedConflicts?
 */

Boolean detectCircle(Analysis a, Patch p, int flags) {

	Boolean result = FALSE;
	int i;
	ConstraintViolation cv = NULL;
	Analysis newA;
	int arity;

	/* these LVs are actually checked */
	LevelValue lv1 = NULL, lv2 = NULL;
	List l;

	/* We may have to construct an Analysis with a different path
	 selected than a to see whether conflicts would be re-introduced
	 or not. But that is only necessary if p introduces any new
	 lexemes. */
	newA = aClone(a);

	/* preview what the Analysis would look like
	 after swapping in the new lexemes */
	selectPath(newA, p->newLexemes, TRUE);

	/* and the new LVs */
	for (i = 0; i < vectorSize(p->selection); i++) {
		LevelValue lv = vectorElement(p->selection, i);
		if (lv) {
			selectLV(newA, lv);
		}
	}

	/* check all recently removed conflicts */
	for (l = a->P->removedConflicts; l; l = l->next) {
		cv = (ConstraintViolation) l->item;
		arity = listSize(cv->constraint->vars);

		/* find the corresponding LV. */
		lv1 = slotFiller(newA, cv->nodeBindingIndex1);

		/* check it */
		if (arity == 1 && matchUnarySignature(lv1, cv->constraint->sig)
				&& !evalUnaryConstraint(cv->constraint, NULL, newA->slots, lv1)
				&& cv->constraint->penalty == cv->penalty) {
			result = TRUE;
			break;
		}

		/* binary conflict */
		else {

			/* Let lv1 and lv2 be the LVs that may or may not re-introduce cv. */
			lv2 = slotFiller(newA, cv->nodeBindingIndex2);

			if (lv2 && lvsCauseConflict(newA, lv1, lv2, cv)) {
				result = TRUE;
				break;
			}
		}
	}

	if (result && (flags & OD_DEBUG)) {
		pPrint(CDG_DEBUG, p, TRUE);
		cdgPrintf(CDG_DEBUG, " would re-violate ");
		cvPrint(CDG_DEBUG, cv);
		cdgPrintf(CDG_DEBUG, "\n");

	}

	newA = NULL;
	return result;

}

/* Functions dealing with domains
 ============================== */

/* Domains are not a data type in their own. Each domain in a problem
 is simply a List of all LVs that may bind a specific level/time
 point pair. This means that a domain for time point 0 may contain
 LVs with the modifier wer(0,1) and also werden(0,2). Conversely,
 shallow copies of the same LV appear in several domains if its
 modifier spans more than one time interval. */

/* ----------------------------------------------------------------------
 Count remaining LVs in Problem.

 This function computes the field P->nrUndeletedLVs. Note that usually
 this field is maintained by the frobbing routines that may delete LVs.
 This function must be called after a foreign routine has deleted LVs.

 */
void determineAmbiguity(Problem P) {
	int i;
	List l;

	P->nrUndeletedLVs = 0;
	for (i = 0; i < P->nrSlots; i++) {
		for (l = P->domains[i]; l; l = l->next) {
			LevelValue lv = l->item;
			if (lv->level->useflag && !lv->isDeleted) {
				P->nrUndeletedLVs
						+= (((LexemNode) lv->dependents->item)->arc->to
								- ((LexemNode) lv->dependents->item)->arc->from);
			}
		}
	}
}

/* ----------------------------------------------------------------------
 Print all LVs in a domain.
 */
void dPrint(unsigned long mode, Problem P, int where) {

	List m;
	LevelValue lv;
	Number limit = 0.0;
	Number f;

	List l = P->domains[where];

	if (l) {
		lv = (LevelValue) l->item;
		cdgPrintf(mode, "@%d: %d/%s\n", where, where / P->nrLevels,
				lv->level->id);
	} else {
		cdgPrintf(mode, "Empty domain.\n");
	}

	for (m = l; m; m = m->next) {
		lv = m->item;
		cdgPrintf(mode, "#%d: ", lv->indexWRTNet);
		if (lv->isDeleted) {
			cdgPrintf(mode, "[ ");
		}
		lvPrint(mode, lv, NULL, 1);
		cdgPrintf(mode, " (limit %4.3e)", lv->limit);
		if (lv->isDeleted) {
			cdgPrintf(mode, " ]");
		}
		cdgPrintf(mode, "\n");
		f = getLimit(P, lv);
		if (f > limit)
			limit = f;
	}

	cdgPrintf(mode, "\n limit %4.3e\n", limit);
}

/* ----------------------------------------------------------------------
 Re-calculate limits for all LVs in a domain.

 This function propagates limits bidirectionally between LVs and lexeme
 nodes. It is a generalization of cnOptimizeNode(): the former function
 can only delete structures if they are absolutely impossible. This
 function can also detect implausible structures:

 The limit of an LV cannot exceed the limits of all its dependents (or all
 of its regents).

 The limit of a lexeme node cannot exceed the limit of all LVs that bind
 it as a modifier. (Note that a lexeme node can appear in an Analysis
 without ever being a modifiee, so this second condition is not
 symmetrical.)

 Theoretically, the limit of a lexeme node cannot exceed the limits of
 all other lexemes that are needed to connect it to the start and end of
 the word graph. However, because this condition is expensive to compute,
 and is also approximated by the first two conditions, it is not
 enforced.

 !ALERT! This function modifies the domain WHICH. Using it while
 iterating over the same domain will yield undefined behaviour.

 */
void dCheck(Problem P, int which) {

	List l, m;
	Number *localLimit;
	LevelValue lv;
	LexemNode ln;
	int i, no, from;
	Number f;

#ifdef DEBUGDCHECK
	cdgPrintf(CDG_DEBUG, "\nDEBUG: dCheck(@%d)\n", which);
	cdgPrintf(CDG_DEBUG, "DEBUG: GlobalMinimum: %f\n", P->globalMinimum);
#endif
	/* empty domain ? */
	if (!P->domains[which])
		return;

	localLimit = memMalloc(sizeof(Number)*P->nrLexemes);
	from = which / P->nrLevels;

	/* localLimits start out at 0 */
	for (i = 0; i < P->nrLexemes; i++) {
		localLimit[i] = 0.0;
	}

	/* propagate limit of LVs to lexeme nodes */
	for (l = P->domains[which]; l; l = l->next) {
		lv = l->item;
		if(lv->isDeleted) {
			lv->limit=0.0;
			continue;
		}
		f = getLimit(P, lv);

#ifdef DEBUGDCHECK
		cdgPrintf(CDG_DEBUG, "DEBUG: #%d has limit %4.3e\n", lv->indexWRTNet, f);
#endif
		if (significantlyGreater(P->globalMinimum, f) && !lv->isDeleted) {
			lv->isDeleted = TRUE;
			P->nrUndeletedLVs -= (((LexemNode) lv->dependents->item)->arc->to
					- ((LexemNode) lv->dependents->item)->arc->from);
			continue;
		}

		for (m = lv->dependents; m; m = m->next) {
			no = ((LexemNode) m->item)->no;
			if (f > localLimit[no]) {
				localLimit[no] = f;
#ifdef DEBUGDCHECK
				cdgPrintf(CDG_DEBUG, "DEBUG: locallimit[%%%d] = %4.3e\n", no, f);
#endif
			}
		}
	}

	/* lower limits of lexeme nodes that have too little support */
	for (i = 0; i < P->lgWidth; i++) {
		ln = vectorElement(P->lexemes, from * P->lgWidth + i);
		if (!ln) {
			break;
		}
		if (localLimit[ln->no] < ln->limit) {
			ln->limit = localLimit[ln->no];
#ifdef DEBUGDCHECK
			cdgPrintf(CDG_DEBUG, "DEBUG: %%%d->limit  = %4.3e\n", ln->no, localLimit[ln->no]);
#endif
			if (significantlyGreater(P->globalMinimum, ln->limit)) {
#ifdef DEBUGDCHECK
				cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
				lgPrintNode(CDG_DEBUG, ln);
				cdgPrintf(CDG_DEBUG, "\n");
#endif
				lgDeleteNode(P->lg, ln);
			}
		}
	}

	/* propagate limits from lexeme nodes to LV, and re-sort domain */
	l = P->domains[which];
	P->domains[which] = NULL;
	for (m = l; m; m = m->next) {
		updateLimit(P, m->item);
		P->domains[which] = listInsertSorted(P->domains[which], m->item,
				lvCompareByLimit);
	}
	l = NULL;
	memFree(localLimit);
}

/* ----------------------------------------------------------------------
 Optimize several domains after another.

 This function behaves like separate calls to dCheck() with all
 the integers in which as arguments, but is more efficient
 because the cleanup code only runs once.
 */
void dCheckAll(Problem P, List which) {

	List l;

	for (l = which; l != NULL; l = l->next) {
		dCheck(P, (int) l->item);
	}

	which = NULL;

	if (P->net) {
		cnOptimizeNet(P->net);
	}
	/* Since this may have deleted LVs,
	 we must recount the remaining ones. */
	determineAmbiguity(P);

	/* And since deleting lexeme nodes may change the distance between
	 nodes, which may change the score of constraint violations,
	 which may change the score of the current Analysis, we should
	 re-judge it here. */
	aJudge(P->current);
	aJudge(P->best);
	aJudge(P->annotation);
}

/* ----------------------------------------------------------------------
 How many undeleted LVs does the domain WHERE contain?
 */
int dSize(Problem P, int where) {

	int result = 0;
	List l;

	for (l = P->domains[where]; l; l = l->next) {
		if (!((LevelValue) l->item)->isDeleted) {
			result++;
		}
	}

	return result;

}

/**
 * returns a number array the size of the number of levelValues in the problem
 * such an array is cached in the problem, to minimize reallocation
 * it is only reallocated if the number of LVs changed since last allocation
 *
 * The given array cannot be shared by different function calls, so two function that use this
 * buffer may never call each other, be aware!!!
 */
Number* getLimitBuffer(Problem P) {
	int i;

	if(P->nrLVs != P->limitBufferSize) {
		P->limitBuffer = memMalloc(sizeof(Number)*P->nrLVs);
		P->limitBufferSize = P->nrLVs;
	}

	/* table of locally determined limits */
	for (i = 0; i < P->nrLVs; i++) {
		P->limitBuffer[i] = 0.0;
	}

	return P->limitBuffer;
}

/* ----------------------------------------------------------------------
 Search specified domains for good combined assignments.

 The domains searched are the ones whose number is contained in WHICH.
 See below, however, about the OD_INT flag.

 This executes a local search among all LVs of the domains in question.
 Care is taken not to form tuples that are physically incompatible, nor
 is the same LV ever added twice to the same tuple, even if WHICH
 contains duplicates.

 The result is a List of all possible Patches, ordered by pCompare().
 Only the fields ->local and ->limit are computed, not the field
 ->predicted (see optimizeDomains() for a way to do that).

 FLAGS specifies particular kinds of behaviour. The following flags are
 defined:

 OD_INT: the argument WHICH is not a Vector, but an integer
 specifying the one domain to be searched. Not pretty, but it saves
 us the allocation of a Vector for most calls.

 OD_PREDICT: choose the Patch with the best predicted Badness.

 OD_LIMIT: return the Patch with the best limit (default).

 OD_SINGLE: return only unary Patches even when multiple domains are
 specified.

 OD_ORIENT: allow Patches that change the direction of edges.

 OD_LEXICAL: allow Patches that change lexical items.

 OD_LABEL: allow Patches that change labels.

 OD_ALLOWALL: a synonym for specifying all of the three previous flags.
 This ensures that no Patch is rejected for one of its LVs.

 OD_DELETED: allow Patches that contain deleted LVs or lexemes.

 Note that OD_DELETED only re-allows LVs that are present, but were
 marked as deleted during the preceding parsing process. It can not
 magically revive LVs that were rejected when the constraint net was
 first built. Therefore, LVs that violate hard unary constraints are
 usually not found even with OD_DELETE.

 If you specify OD_DELETED to a function in the toolchain, you must also
 pass it to all later functions that operate on the resulting Patches, to
 avoid spooking the internal consistency checks.

 OD_PRUNE: delete LVs that do not appear in any of the possible Patches.

 OD_NARROW: ignore Patches with suboptimal limit.

 OD_INTERACTIVE: ask the user to select a Patch instead of computing the
 result autonomously.

 OD_DEBUG: give overwhelming amount of output.

 OD_CONTEXT: use context-sensitive constraints for evaluating a Patch.
 This will give wrong results unless WHICH actually contains the entire
 context.

 OD_IMMEDIATE: return immediately if a Patch is found that improves A
 beyond a->P->best->b.

 Note that these flags are used by various other functions that ultimately
 call or are called by dSearch(), such as optimizeDomains() and pNext(),
 and not all flags are used by all of them. For instance, OD_IMMEDIATE
 is read only by optimizeDomains() since it implies Patch previewing.

 The set of possible transformations is generated recursively, using an
 Agenda with size localSize. Therefore a simultaneous search in many
 domains will usually be incomplete.

 If the search was not truncated (and OD_PRUNE is set), the result of the
 search is used to propagate limits to LVs from all domains searched.
 Each LV is assigned a limit that is not higher than the best limit of a
 Patch it appears in.

 */
List dSearch(Analysis a, Vector which, int flags) {
	Agenda ag;
	LevelValue lv;
	List l = NULL;
	List result = NULL;
	Number *localLimits = getLimitBuffer(a->P);
	Number bestLimit = 0.0;
	Patch p, newP;
	int i, j, index, width, reg;
	double nrCombinations = 1;

	/* Do not trust the caller; several other invocation options
	 would render pruning invalid, so test them explicitly. */
	Boolean doPruning = (flags & OD_PRUNE) && (flags & OD_SWAP) == 0 && (flags & OD_ALLOWALL)
			== OD_ALLOWALL;
	LevelValue current = NULL;
	Boolean verbose = (flags & OD_INTERACTIVE) != 0;
	Boolean debug = (flags & OD_DEBUG) != 0;

	Vector allowedRegents = NULL;
	//Vector allowedLabels  = NULL;

	Vector context = NULL;
	if (flags & OD_INT) {
		width = 1;
		if (verbose) {
			nrCombinations = dSize(a->P, (int) which);
		}
	} else {
		width = vectorSize(which);

		/* preview how much effort it's going to be */
		for (i = 0; i < vectorSize(which); i++) {

			/* this is the number of the domain */
			index = (int) vectorElement(which, i);
			l = listPrependElement(l, (Pointer) index);

			/* this is the current size of the domain,
			 which gives the actual complexity of searching it now */
			if (verbose) {
				nrCombinations *= dSize(a->P, index);
			}
		}
	}

	if (debug) {
		cdgPrintf(CDG_DEBUG, "\nDEBUG: dSearch(");
		if (flags & OD_INT) {
			cdgPrintf(CDG_DEBUG, "@%d", (int) which);
		} else {
			for (i = 0; i < vectorSize(which); i++) {
				j = (int) vectorElement(which, i);
				cdgPrintf(CDG_DEBUG, "@%d", j);
				if (i < vectorSize(which) - 1) {
					cdgPrintf(CDG_DEBUG, " ");
				}
			}
		}
		cdgPrintf(CDG_DEBUG, ")\n");
	}

	if (verbose) {
		cdgPrintf(CDG_INFO, "searching %.0f combination%s\n", nrCombinations,
				nrCombinations != 1 ? "s" : "");
	}

	if (doPruning) {
		dCheckAll(a->P, l);
	}

	// for swapping, determine the allowed regents
	if(flags & OD_SWAP) {
		allowedRegents = vectorNew( vectorSize(which));
		for(i = 0; i<vectorSize(which); i++) {
			index = (int) vectorElement(which, i);
			lv = vectorElement(a->slots, index);
			reg = lvGetRegentGN(lv)?lvGetRegentGN(lv)->no:-1;
			if(!vectorContains(allowedRegents, (Pointer) reg)) {
				//cdgPrintf(CDG_INFO, "allowed regent: %s \n", lvRegent(lv, FALSE));
				vectorAddElement(allowedRegents, (Pointer) reg);
			}
		}

//		allowedLabels = vectorNew( vectorSize(which));
//		for(i = 0; i<vectorSize(which); i++) {
//			index = (int) vectorElement(which, i);
//			lv = vectorElement(a->slots, index);
//			if(!vectorContains(allowedLabels, strRegister(lv->label))) {
//				cdgPrintf(CDG_INFO, "allowed label: %s \n", lv->label);
//				vectorAddElement(allowedLabels, strRegister(lv->label));
//			}
//		}
	}

	/* conduct local search */
	ag = agNew(localSize, memFreeFunction);
	agSetVerbosity(ag, verbose);
	cdgAgInsert(ag, 1.0, pNew(a, width, flags));

	while (!agIsEmpty(ag)) {

		/* read */
		p = (Patch) agRemoveBest(ag);

		if (debug) {
			cdgPrintf(CDG_DEBUG, "DEBUG: read ");
			pPrint(CDG_DEBUG, p, TRUE);
		}

		/* time limit reached? */
		if (timerExpired) {
			p = NULL;
			continue;
		}

		/* interrupt? */
		if (cdgCtrlCTrapped) {
			p = NULL;
			continue;
		}


		/* Should we delete Patches with suboptimal limits or not?

		 This is one of the truly important parts of the module.
		 Basically, we only want to investigate Patches with the optimal
		 limit, because that is the idea behind having limits. That is
		 the first line of the condition.

		 Also, a Patch with a suboptimal limit may still be useful for
		 assessing the limits of the LVs contained in it. So we only
		 want to suppress Patches if we already know that that stage
		 will be skipped later on -- because the agenda overflowed, or
		 because the user has told us not to. That is the second line of
		 the condition. */
		if (flags & OD_LIMIT && p->limit < bestLimit &&
				(agIsTruncated(ag) || !doPruning)) {

			if (debug) {
				pPrint(CDG_INFO, p, TRUE);
			}
			p = NULL;
			continue;
		}

		/* Since in general word graphs, one lexeme can cover multiple
		 intervals, one LV can bind multiple domains. It is therefore
		 possible that we have already bound the next domain ordered. */

		/* find the next domain to bind */
		while (p->index < width) {
			/* next domain ordered */
			if (flags & OD_INT) {
				i = (int) which;
			} else {
				i = (int) vectorElement(which, p->index);
			}
			/* bound already? */
			if (!bvElement(p->bound, i)) {
				break;
			} else {
				/* skip it */
				vectorSetElement(p->selection, NULL, p->index++);
			}
		}


		/* retain */
		if (p->index == width) {

			if (debug) {
				cdgPrintf(CDG_DEBUG, "DEBUG: retained ");
				pPrint(CDG_DEBUG, p, TRUE);
			}

			/* apply context-sensitive constraints */
			if (flags & OD_CONTEXT) {
				context = vectorNew(a->P->nrSlots);

				for (i = 0; i < vectorSize(p->selection); i++) {
					LevelValue lv = vectorElement(p->selection, i);
					if (lv) {
						vectorSetElement(context, lv, lvIndex(lv));
					}
				}

				evalInContext(p->selection, context, p->local, NULL);
				context = NULL;

				/* maybe update limit */
				if (p->local->soft < p->limit) {
					p->limit = p->local->soft;
				}
				if (p->local->hard) {
					p->limit = 0.0;
				}

			}

			result = listInsertSorted(result, p, pCompare);
			if (p->limit > bestLimit) {
				if (verbose) {
					cdgPrintf(CDG_INFO, "INFO: Patch with limit %4.3e found\n",
							p->limit);
				}
				bestLimit = p->limit;
			}
			continue;
		}

		/* extend */
		if (flags & OD_INT) {
			i = (int) which;
		} else {
			i = (int) vectorElement(which, p->index);
		}
		if (debug) {
			cdgPrintf(CDG_DEBUG, "DEBUG: considering @%d (length %d)\n", i,
					listSize(a->P->domains[i]));
		}

		current = slotFiller(a, i);
		for (l = a->P->domains[i]; l; l = l->next) {
			lv = l->item;

			// skip attachments to blocked virtual nodes
			if(blockVirtualNodes && a->P->lg->ve && lvGetRegentGN(lv) &&
					bvElement(a->P->lg->ve->blocked,lvGetRegentGN(lv)->no) ) {
				continue;
			}

			if (debug) {
				cdgPrintf(CDG_DEBUG, "DEBUG: Trying #%d... ", lv->indexWRTNet);
			}
			/* reject deleted LV */
			if ((flags & OD_DELETED) == 0 && lv->isDeleted) {
				if (debug) {
					cdgPrintf(CDG_DEBUG, "is deleted\n", lv->indexWRTNet);
				}
				continue;
			} else {
				if (debug) {
					cdgPrintf(CDG_DEBUG, "\n");
				}
			}

			// for swapping reject not allowed regents and labels
			if(flags & OD_SWAP) {
				reg = lvGetRegentGN(lv)?lvGetRegentGN(lv)->no:-1;
				if(!vectorContains(allowedRegents, (Pointer) reg)) {
					continue;
				}

//				if(!vectorContains(allowedLabels, strRegister(lv->label))) {
//					continue;
//				}
			}

			if ((newP = pNext(p, lv, flags)) != NULL) {

				/* is it an actual change, or merely a substitution X --> X? */
				if (lv != current) {
					newP->nrNewLVs++;
				}

				/* write */
				cdgAgInsert(ag, newP->limit, newP);

				if (debug) {
					cdgPrintf(CDG_DEBUG, "DEBUG: wrote ");
					pPrint(CDG_DEBUG, newP, TRUE);
				}
			}
		} // end for (l = a->P->domains[i]; l; l = l->next)

		p = NULL;

	} // end while (!agIsEmpty(ag))

	/* Now, propagate limits to all LVs in the searched domains. */
	if (doPruning && !cdgCtrlCTrapped && !agIsTruncated(ag)) {

		if (verbose) {
			cdgPrintf(CDG_INFO, "%d of %.0f combination%s possible\n",
					listSize(result), nrCombinations, nrCombinations > 1 ? "s"
							: "");
		}

		/* update global limit */
		if (a->P->globalMaximum > bestLimit) {
			a->P->globalMaximum = bestLimit;
			cdgPrintf(CDG_INFO, "INFO: it won't get any better than %4.3e.\n",
					a->P->globalMaximum);
		}

		/* if an LV appears in a Patch with limit f,
		 then its limit may be as high as that */
		for (l = result; l; l = l->next) {
			p = (Patch) l->item;
			if (p->index < width)
				continue;
			for (i = 0; i < vectorSize(p->selection); i++) {
				lv = (LevelValue) vectorElement(p->selection, i);
				if (!lv)
					continue;
				index = lv->indexWRTNet;

				if (localLimits[index] < p->limit) {
					if (debug) {
						cdgPrintf(CDG_DEBUG, "DEBUG: #%d can attain %4.3e\n",
								index, p->limit);
					}
					localLimits[index] = p->limit;
				}
			}
		}

		/* degrade LVs that do not cooperate */
		for (i = 0; i < width; i++) {
			if (flags & OD_INT) {
				index = (int) which;
			} else {
				index = (int) vectorElement(which, i);
			}
			for (l = a->P->domains[index]; l; l = l->next) {
				lv = (LevelValue) l->item;
				if (lv->isDeleted)
					continue;

				j = lv->indexWRTNet;

				if (localLimits[j] < lv->limit) {
					if (debug) {
						cdgPrintf(CDG_DEBUG, "DEBUG: limit(#%d) = %4.3e\n",
								lv->indexWRTNet, localLimits[j]);
					}
					downgrade(a->P, lv, localLimits[j]);
				}
			}
		}
	}

	if (verbose) {
		cdgCtrlCTrapped = FALSE;
	}

	agDelete(ag);

	return result;

}

/* ----------------------------------------------------------------------
 Generate a Patch that swaps in LN and changes nothing else.
 */
Patch makeLexicalPatch(Analysis a, Vector lns) {
	Patch result = pNew(a, 0, 0);
	int i;
	result->newLexemes = vectorToList(lns);
	for (i = 0; i < vectorSize(lns); i++) {
		LexemNode ln = vectorElement(lns, i);
		lgRequireLexeme(a->P->lg, result->forbiddenLexemes, ln);
		result->newLexemes = listAppendElement(result->newLexemes, ln);
	}
	return result;

}

/* ----------------------------------------------------------------------
 Optimise n domains in the context of A.

 Performs a dSearch() on the specified domains and applies pJudge() to
 some or all of the resulting patches to decide which should be applied.

 The result is the Patch that is most suitable in the context of the
 current Problem. NULL may be returned if no Patch is possible.

 If the returned Patch is not NULL, it has its field `predicted' set, and
 the value is an accurate prediction of the Badness of A if that Patch
 were applied to it. It has also been successfully been lexicalized and
 can immediately be applied to A.

 Various additional restrictions are obeyed. The variable beamWidth
 limits the number of patches that are considered at all. If a Patch
 fails the detectCircle() test, it is discarded. If OD_IMMEDIATE is set,
 the first Patch that improves on P->best->b stops further previewing. If
 OD_INTERACTIVE is set, the user is asked for a decision via stdout.

 Other flags works as in dSearch().
 */
Patch optimizeDomains(Analysis a, Vector which, int flags) {

	Badness predictedOptimum = worstBadness();
	List l, patches = NULL;
	Patch p, result;

	char buffer[BUFFERSIZE];
	Number bestLimit = 0.0;
	int bestPatch = 0;
	int i, no = 0;

	/* create only single-spot Patches */
	if (flags & OD_SINGLE && !(flags & OD_INT)) {
		for (i = 0; i < vectorSize(which); i++) {
			int where = (int) vectorElement(which, i);
			patches = listAppendList(patches, dSearch(a, (Pointer) where, flags
					| OD_INT));
		}
	}

	/* create Patches by combining LVs from the appropriate domains */
	else {
		patches = dSearch(a, which, flags);
	}

	/* consider all patches... */
	for (l = patches; l; l = l->next) {
		no++;

		/* ...within beam width */
		if (no > beamWidth) {
			break;
		}

		/* react to user break */
		if (cdgCtrlCTrapped) {
			cdgPrintf(CDG_INFO, "\nmaybe that analysis took a little long...\n");
			break;
		}

		p = (Patch) l->item;

		/* if an LV in the Patch was deleted by dSearch itself,
		 disregard the entire Patch */
		for (i = 0; i < vectorSize(p->selection); i++) {
			LevelValue lv = vectorElement(p->selection, i);
			if (lv && lv->isDeleted && (flags & OD_DELETED) == 0) {
				i = -1;
				break;
			}
		}
		if (i == -1) {
			continue;
		}

		/* obey narrowSearch flag */
		if (flags & OD_NARROW) {
			if((flags & OD_LIMIT) && p->limit < bestLimit && p->nrNewLVs > 1) {
				continue;
			}
			if((flags & OD_PREDICT) && p->limit < bScore(predictedOptimum)) {
				continue;
			}
		}

		/* pLexicalize() only fails if the lexeme graph has become
		 disconnected. In this case all further calls to pLexicalize() will
		 also fail, so we can immediately give up. */
		if (!pLexicalize(p, flags, FALSE)) {
			break;
		}

		if (flags & OD_DEBUG) {
			cdgPrintf(CDG_DEBUG, "judging Patch ");
			pPrint(CDG_DEBUG, p, TRUE);
		}

		/* do not consider Patches that would re-introduce conflicts */
		if (detectCircle(a, p, flags)) {
			continue;
		}

		/* update the best limit of patches we have found so far */
		if (p->limit > bestLimit) {
			bestLimit = p->limit;
		}

		pJudge(p, !(flags & OD_NONEWLVS));
		if (flags & OD_INTERACTIVE) {
			cdgPrintf(CDG_INFO, "\n%d) ", no);
			pPrint(CDG_INFO, p, TRUE);
		}

		/* Now for the tricky part: which of the Patches should be
		 returned? It depends on the method. In classical minConflict
		 style, it is the Patch that will result in the best immediate
		 improvement. */
		if (flags & OD_PREDICT) {

			/* the best patch is the one with the best predicted Badness. */
			if (bCompare(p->predicted, predictedOptimum)) {
				*predictedOptimum = *p->predicted;
				bestPatch = no;
			}
		} else {

			/* the best patch is the one with the best limit and the best
			 predicted Badness.

			 But locally non-optimal patches are allowed if they yield an
			 immediate global improvement. */

			if ((p->limit == bestLimit && bCompare(p->predicted,predictedOptimum)) ||
					bCompare(p->predicted, a->P->best->b)) {
				*predictedOptimum = *p->predicted;
				bestPatch = no;
			}
		}

		/* if we have found a method to improve the Analysis globally;
		 don't even bother with further previewing */
		if ((flags & OD_IMMEDIATE) && bCompare(p->predicted, a->P->best->b)) {
			break;
		}
	}

	/* Decide which patch to return */
	result = NULL;
	if (patches) {
		if (0 == (flags & OD_INTERACTIVE)) {
			no = bestPatch;
		} else if (bestPatch > 0) {

			/* get input */
			cdgPrintf(CDG_INFO, "patch (%d): ", bestPatch);
			cdgFlush();
			cdgGetString(buffer, BUFFERSIZE);

			/* remove trailing whitespace */
			while (isspace((int)buffer[strlen(buffer) - 1])) {
				buffer[strlen(buffer) - 1] = '\0';
			}

			no = -1;
			if (!strcmp(buffer, "")) {
				no = bestPatch;
			} else {
				no = 0;
				sscanf(buffer, "%d", &no);
			}
		} else {
			no = 0;
		}
	}

	/* return Patch */
	if (no > 0 && no <= listSize(patches)) {
		result = pClone(listNthElement(patches, no));
	}

	/* free */
	patches = NULL;
	predictedOptimum = NULL;

	return result;

}

/* ----------------------------------------------------------------------
 Optimize N lexical selections in the context of A.

 Of the lexemes in the lexeme graph, those are chosen whose selection
 would collectively lead to the best score.

 The effect of this function is similar to calling optimizeDomains() with
 the corresponding domains numbers, but there are important differences:

 - optimizeDomains() relies on having the domains of P already populated.
 It just combines pre-existing LVs. This function chooses lexemes
 instead, and can produce LVs never before seen in the history of Earth.

 - because of that, optimizeDomains() is generally more efficient for
 wide searches, since it ignores the many idiotic combinations that are
 unarily forbidden (such as `der_ART' as a subject).

 This function is only useful in a problem that was created from a
 decorated Parse, not from a full constraint net.

 */
Patch optimizeLexemes(Analysis a, Vector which) {
	Patch result = NULL;
	List agenda, newagenda;
	int i, j;
	List l;
	LexemGraph lg = a->P->lg;

	agenda = listPrependElement(NULL, vectorNew(5));

	for (i = 0; i < vectorSize(which); i++) {
		int index = (int) vectorElement(which, i);
		newagenda = NULL;
		for (j = 0; j < vectorSize(lg->nodes); j++) {
			LexemNode ln = vectorElement(lg->nodes, j);
			if (ln->arc->from != index)
				continue;
			if (bvElement(lg->isDeletedNode, ln->no))
				continue;
			for (l = agenda; l != NULL; l = l->next) {
				Vector old = l->item;
				Vector new = vectorClone(old);
				vectorAddElement(new, ln);
				newagenda = listPrependElement(newagenda, new);
			}
		}
		agenda = newagenda;
	}

	for (l = agenda; l != NULL; l = l->next) {
		Vector v = l->item;
		Patch p = makeLexicalPatch(a, v);
		v = NULL;
		pJudge(p, TRUE);
		if (!result || bCompare(p->predicted, result->predicted)) {
			if (result) {
				result = NULL;
			}
			result = p;
		} else {
			p = NULL;
		}
	}
	agenda = NULL;
	return result;
}

/* Functions dealing with Problems
 =============================== */

/* ----------------------------------------------------------------------
 Commit to the current value of domain WHERE.

 All other LVs in the domain are marked deleted.
 */
void freezeDomain(Analysis a, int where) {

	LevelValue theOne = slotFiller(a, where);
	List l;
	cdgPrintf(CDG_DEBUG,
			"DEBUG: Freezing @%d, retaining #%d, deleting %d values.\n", where,
			theOne->indexWRTNet, dSize(a->P, where));

	for (l = a->P->domains[where]; l != NULL; l = l->next) {
		LevelValue lv = l->item;
		if (lv != theOne && !lv->isDeleted) {
			lv->isDeleted = TRUE;
			a->P->nrUndeletedLVs
					-= (((LexemNode) lv->dependents->item)->arc->to
							- ((LexemNode) lv->dependents->item)->arc->from);
		}
	}
}

/* --------------------------------------------------------------------
 Detect conflicts introduced by repair.

 A conflict CV is `recent' if there is a `best' Analysis
 and CV is not in it.
 */
Boolean isRecent(Problem P, ConstraintViolation cv) {

	if (!P->best) {
		return FALSE;
	}

	return !cvContains(P->best->conflicts, cv);

}

/* ----------------------------------------------------------------------
 Detect unremovable conflicts.

 This simply returns TRUE if CV is in P->unremovable.
 */
Boolean isUnremovable(Problem P, ConstraintViolation cv) {

	return cvContains(P->unremovable, cv);

}

/* ----------------------------------------------------------------------
 Sort all conflicts in A by recentness and score.

 This is to ensure that conflicts introduced by a repair step will end up
 in the front of the list, and will thus be attacked before other
 conflicts.
 */
void sortConflicts(Analysis a) {

	List old = NULL, new = NULL;
	List l;

	for (l = a->conflicts; l; l = l->next) {
		if (isRecent(a->P, l->item)) {
			new = listInsertSorted(new, l->item, cvCompare);
		} else {
			old = listInsertSorted(old, l->item, cvCompare);
		}
	}
	a->conflicts = listAppendList(new, old);

}

/* ----------------------------------------------------------------------
 List all conflicts in an Analysis.

 cvPrint() is used to display conflicts.
 */
void printConflicts(unsigned long mode, Analysis a) {

	int counter;
	List l;
	ConstraintViolation cv;

	if (a->conflicts)
		cdgPrintf(mode, "\n");
	counter = 0;

	/* show all conflicts */
	for (l = a->conflicts; l; l = l->next) {
		cv = (ConstraintViolation) l->item;
		cdgPrintf(mode, "%3d. ", counter++);
		cvPrint(mode, cv);

		/* mark if this conflict as old or new */
		if (isRecent(a->P, cv)) {
			cdgPrintf(mode, "    (recently introduced)");
		} else if (isUnremovable(a->P, cv)) {
			cdgPrintf(mode, "    (obstinate)");
		}
		cdgPrintf(mode, "\n");

	}

}

/* ----------------------------------------------------------------------
 Reset the transformation history info for a Problem.

 Deallocates the fields P->removedConflicts and P->history as well as
 their elements.

 */
void resetProblem(Problem P) {

	P->removedConflicts = NULL;

	P->history = NULL;

	sortConflicts(P->current);

}

/* ----------------------------------------------------------------------
 Free a remembered Problem state.
 */
void psDelete(ProblemState ps) {
	if (!ps) {
		return;
	}
	memFree(ps->LVLimits);
	ps->deletedLVs = NULL;
	memFree(ps->LnLimits);
	ps->deletedLns = NULL;
	ps->valuesPerNode = NULL;
	memFree(ps);
}

/* ----------------------------------------------------------------------
 Take a snapshot of the Problem at the current time.

 This only remembers the state of deletion of LVs, lexeme nodes, etc.
 The ProblemState will no longer work if you add or delete any
 nodes later.
 */
ProblemState psNew(Problem P) {
	ProblemState result;
	ConstraintNet net = P->net;
	LexemGraph lg;
	int i;

	if (!net) {
		return NULL;
	}

	lg = net->lexemgraph;
	result = memMalloc(sizeof(ProblemStateStruct));
	result->LVLimits = memMalloc(sizeof(Number) * vectorSize(net->values));
	result->deletedLVs = bvNew(vectorSize(net->values));
	result->LnLimits = memMalloc(sizeof(Number) * vectorSize(lg->nodes));
	result->deletedLns = bvNew(vectorSize(lg->nodes));
	result->valuesPerNode = vectorNew(vectorSize(net->nodes));

	/* remember state of LVs */
	for (i = 0; i < vectorSize(net->values); i++) {
		LevelValue lv = vectorElement(net->values, i);
		result->LVLimits[i] = lv->limit;
		bvSetElement(result->deletedLVs, lv->isDeleted, i);
	}

	/* remember state of lexeme nodes*/
	for (i = 0; i < vectorSize(lg->nodes); i++) {
		LexemNode ln = vectorElement(lg->nodes, i);
		result->LnLimits[i] = ln->limit;
		bvCopy(result->deletedLns, lg->isDeletedNode);
	}

	/* remember value count in ConstraintNodes */
	for (i = 0; i < vectorSize(net->nodes); i++) {
		ConstraintNode cn = vectorElement(net->nodes, i);
		vectorSetElement(result->valuesPerNode, (Pointer) cn->noValidValues, i);
	}
	return result;
}

/* ----------------------------------------------------------------------
 Record the current state of the Problem for posterity.
 */
void rememberProblem(Problem P) {
	if (P->savedState) {
		P->savedState = NULL;
	}
	P->savedState = psNew(P);
}

/* ----------------------------------------------------------------------
 Reset Problem to the remembered state.

 This function restores P and its underlying
 constraint net to the state that it had at the time when
 rememberProblem() was last called on it.
 */
void restoreProblem(Problem P, ProblemState ps) {
	cdgPrintf(CDG_INFO, "restoring problem \n"); // NB

	int i;
	if (!ps) {
		return;
	}

	if (!P->net) {
		return;
	}

	/* restore state of LVs */
	for (i = 0; i < vectorSize(P->net->values); i++) {
		LevelValue lv = vectorElement(P->net->values, i);
		lv->limit = ps->LVLimits[i];
		lv->isDeleted = bvElement(ps->deletedLVs, i);
		// if(lv->isFrozen)
		//    	lv->isDeleted=FALSE; //unfreeze LV
	}

	/* restore state of lexeme nodes */
	bvCopy(P->lg->isDeletedNode, ps->deletedLns);
	for (i = 0; i < vectorSize(P->lg->nodes); i++) {
		LexemNode ln = vectorElement(P->lg->nodes, i);
		ln->limit = ps->LnLimits[i];
		// if(bvElement(P->lg->isFrozenNode,i)) // unfreeze lexemes
		//	 bvSetElement(P->lg->isDeletedNode,FALSE,i);
	}
	lgComputeNoOfPaths(P->lg, FALSE);
	lgComputeDistances(P->lg);

	/* restore value counts in ConstraintNodes */
	for (i = 0; i < vectorSize(P->net->nodes); i++) {
		ConstraintNode cn = vectorElement(P->net->nodes, i);
		cn->noValidValues = (int) vectorElement(ps->valuesPerNode, i);
	}
}

/* ----------------------------------------------------------------------
 When the pressure specified by the user turns out too great for the
 problem, halve it and reset the net, because all the deletions done
 so far cannot be trusted any more.

 However, all is not lost; the partial solution found during the
 high-pressure search is likely to be useful for further search. */
Boolean reducePressure(Problem P) {

	resetProblem(P);
	P->globalMaximum = 1.0;

	if (P->globalMinimum < 1e-8) {
		cdgPrintf(CDG_PROFILE, "INFO: cannot reduce pressure further...\n");
		cdgPrintf(CDG_PROFILE, "INFO: perhaps there *is* no solution?\n");
		return FALSE;
	}

	P->globalMinimum /= 10.0;
	cdgPrintf(CDG_INFO, "INFO: Pressure too great, resetting to %4.3e\n",
			P->globalMinimum);
	restoreProblem(P, P->savedState);

	/* By restoring the entire Problem, we have repaired the lexeme
	 graph and reinstated its distance information. Since this information
	 is used by the predicate distance(), each and every constraint formula
	 has to be re-evaluated now, even in unary constraints. */
	aJudge(P->current);
	aJudge(P->best);

	return TRUE;
}

/* ----------------------------------------------------------------------
 Revert to the best known location in the search space;
 that is, make P->current a clone of P->best.
 */
void backTrace(Problem P) {

	if (P->current) {
		P->current = NULL;
	}
	if (P->best) {
		P->current = aClone(P->best);
	}

	resetProblem(P);

}

/* ----------------------------------------------------------------------
 De-allocate a Problem.
 */
void deleteProblem(Problem P) {

	int i;

	if (P->levelActive) {
		P->levelActive = NULL;
	}

	if (P->lexemes) {
		P->lexemes = NULL;
	}

	if (P->domains) {
		for (i = 0; i < P->nrSlots; i++)
			P->domains[i] = NULL;
		memFree(P->domains);
		P->domains = NULL;
	}

	if (P->best) {
		P->best = NULL;
	}

	if (P->current) {
		P->current = NULL;
	}
	if (P->annotation) {
		P->annotation = NULL;
	}

	P->removedConflicts = NULL;
	P->history = NULL;
	P->unremovable = NULL;

	P->savedState = NULL;

	if (P->profileTime) {
		P->profileTime = NULL;
	}

	memFree(P);

}

/* ----------------------------------------------------------------------
 Transform Analysis to Parse.

 This function returns a new decorated Parse built from A.
 It calls parseComputeStructure() to do the real work.

 */
Parse analysis2parse(Analysis a) {
	int i;
	Parse p = parseNew(NULL);
	LevelValue lv;
	Vector LVs;

	/* general info */
	p->latticeId = a->P->lg->lattice->id;
	p->searchStrategy = strRegister("frobbing");
	p->comment = NULL;
	p->solutionNo = 1;
	p->noOfSolutions = 1;
	p->score = a->b->hard ? 0.0 : a->b->soft;
	p->optimal = (!significantlyGreater(a->P->globalMaximum, a->b->soft));

	LVs = vectorNew(0);
	for (i = 0; i < vectorSize(a->slots); i++) {
		lv = vectorElement(a->slots, i);
		if (!lv) {
			continue;
		}
		vectorAddElement(LVs, lv);
	}
	parseComputeStructure(p, LVs, a->isUnselectedLexeme);
	/* can't fail here, the grammar is definitely present */
	(void) parseDecorate(p, a->P->lg, NULL);
	LVs = NULL;

	/* conflicts */
	parseEval(p);

	return p;

}

/* ----------------------------------------------------------------------
 Call parseVerify()
 */
void aVerify(Analysis a) {

	AnnoEntry ae = findAnnoForLattice(a->P->lg->lattice->id, TRUE);
	Parse p;
	if (!ae) {
		return;
	}
	p = analysis2parse(a);
	memFreeFunction(parseVerify(p, ae) );
	parseDelete(p);
}

/* ----------------------------------------------------------------------
 Register a new globally best Analysis.

 This function registers a new globally optimal score found for an
 Analysis. It deallocates P->best and replaces it with a clone of
 a. If the Analysis carries a positive score, this is assigned
 to P->globalMinimum. If a conflict found in P->unremovable does not also
 appear in a->conflicts, it is removed from P->unremovable
 (since it was obviously not unremovable).

 */
void newMaximum(Analysis a) {
	cdgPrintf(CDG_INFO, " newMaximum\n");
	List l, m;
	ConstraintViolation cv;
	int i;
	Parse p;
	Number f = 1.0;
	Problem P = a->P;

	P->lastSolFound = timerElapsed(P->profileTime);
	if (aValid(a) && 0.0 == P->firstSoftFound && a->subStart == P->minTimepoint
			&& a->subStop == P->maxTimepoint) {
		P->firstSoftFound = P->lastSolFound;
	}

	/* register new record for Badness */
	if (P->best) {
		P->best = NULL;
	}
	P->best = aClone(a);

	/* note new record for score */
	if (aValid(P->best)) {

		if(P->best->b->soft >P->globalMinimum) {
			cdgPrintf(CDG_INFO, "globMin: %4.3f --> %4.3f \n", P->globalMinimum,
					P->best->b->soft);
		}

		P->globalMinimum = P->best->b->soft;
		l = NULL;
		for (i = 0; i < P->nrSlots; i++) {
			if (!slotActive(a, i)) {
				continue;
			}
			l = listPrependElement(l, (Pointer) i);
			// just checking for a simple inconsistency regarding the limits of participyting LVs
			LevelValue lv = vectorElement(a->slots, i);
			if (significantlyGreater(P->globalMinimum, lv->score)) {
				cdgPrintf(CDG_WARNING, "WARNING! Score of \n  ");
				lvPrint(CDG_WARNING, lv, NULL, 1);
				cdgPrintf(
						CDG_WARNING,
						"\n  is smaller than Analysis it occurs in (%.3f < %.3f)\n",
						lv->score, P->globalMinimum);
				cdgPrintf(CDG_WARNING, "  constraints it violates \n  ");
				List tmp;
				for (tmp = lv->constraints; tmp; tmp = tmp->next) {
					Constraint c = tmp->item;
					cdgPrintf(CDG_WARNING, "  %s  (%.3f?)\n", c->id,
							c->penaltyTerm ? -1.0 : c->penalty);
				}
			}

		}
		dCheckAll(P, l);
	}

	/* groom the Analysis a bit */
	aCalibrate(P->best);

	/* If an `unremovable' conflict was removed anyway,
	 it was obviously not unremovable.
	 So we retract it from the List.  */
	m = NULL;
	for (l = P->unremovable; l; l = l->next) {
		cv = l->item;
		if (!cvContains(P->best->conflicts, cv)) {
			cdgPrintf(CDG_DEBUG, "DEBUG: Hey, `%s' was removed after all...\n",
					cv->constraint->id);
			f = min(f, cv->penalty);
			cv= NULL;
		}

		/* By the same token, retract all softer conflicts from this list as
		 well.
		 If we were wrong about not being able to repair a hard conflict, we
		 could also have been wrong about not being able to repair a soft
		 conflict. */
		/* Removing the following constraints in the same loop works
		 * because P->unremovable is always sorted in descending order
		 * so it's guaranteed that all less-bad conflicts than the
		 * ones removed above get removed. This ordering stems from
		 * the way frobbing works: it attack the conflicts in
		 * descending order.
		 */
		else if (cv->penalty > f) {
			cdgPrintf(CDG_DEBUG, "DEBUG: I'll take another shot at `%s'.\n",
					cv->constraint->id);
			cv = NULL;
		}

		/* yup, still obstinate */
		else {
			m = listPrependElement(m, cv);
		}
	}

	P->unremovable = m;

	resetProblem(P);

	/* hand the Parse to Tcl for displaying */
	if (!suppressPartialSolutions || alwaysShowAllSolutions) {
		p = analysis2parse(P->best);
		cdgExecHook(HOOK_PARTIALRESULT, p);
		parseRegister(p);
	}

}

/* ----------------------------------------------------------------------
 Show remaining ambiguity of P.

 Prints out the average remaining ambiguity in the
 Problem, i.e. the ratio of remaining LVs to domains.
 */
void printAmbiguity(unsigned long mode, Problem P) {

	if (P->nrUndeletedLVs == P->nrActiveSlots) {
		cdgPrintf(mode, " (unique)");
	} else {
		cdgPrintf(mode, " ambiguity %.1f", (float) P->nrUndeletedLVs
				/ P->nrActiveSlots);
	}

}

/* --------------------------------------------------------------------
 Show what was done recently.

 Displays the conflicts recently removed from the Analysis and the
 domains that were changed for that purpose.
 */
void printHistory(unsigned long mode, Problem P) {

	List l;

	/* show removedConflicts */
	if (P->removedConflicts) {
		cdgPrintf(mode, "\nrecently removed:\n");
		for (l = P->removedConflicts; l; l = l->next) {
			cvPrint(mode, l->item);
			cdgPrintf(mode, "\n");
		}
		cdgPrintf(mode, "\n");
	}

	/* show history */
	if (P->history) {
		for (l = P->history; l; l = l->next) {
			cdgPrintf(mode, "@%d ", (int) l->item);
		}
		cdgPrintf(mode, "\n\n");
	}
}

/* ----------------------------------------------------------------------
 Print all LVs of an Analysis.

 Displays all LVs in A in their natural order.
 */
void aPrint(unsigned long int mode, Analysis a) {
	int i, j, k;

	/* iterate over all values */
	for (i = 0; i < vectorSize(a->slots); i++) {
		LevelValue lv;
		if (!slotActive(a, i)) {
			continue;
		}
		lv = (LevelValue) vectorElement(a->slots, i);

		if (!lv) {
			if (mode != CDG_INFO) {
				cdgPrintf(mode, "@%d: <implicitly bound domain>\n", i);
			}
			continue;
		}

		/* LVs aren't shown if they have no competitors and cause no conflicts. */
		if (mode == CDG_INFO && isBoring(a, lv)) {
			continue;
		}

		/* align number to 6 characters */
		j = lvIndex(lv);
		for (k = 1000; k > j; k /= 10) {
			if (k > 1 || j != 0) {
				cdgPrintf(mode, " ");
			}
		}
		cdgPrintf(mode, "@%d: ", i, lv->indexWRTNet);
		j = lv->indexWRTNet;
		if (j < 0) {
			cdgPrintf(mode, "    ");
		} else {
			for (k = 100000; k > j; k /= 10) {
				if (k > 1 || j != 0) {
					cdgPrintf(mode, " ");
				}
			}
		}
		cdgPrintf(mode, "#%d ", j);
		if (lv->isDeleted) {
			cdgPrintf(mode, "[");
		}
		lvPrint(mode, lv, a->isUnselectedLexeme, 0);
		if (lv->isDeleted) {
			cdgPrintf(mode, "]");
		}
		cdgPrintf(mode, "\n");
	}
}

/*
 * Prints all Levelvalues in a given domain
 */
void printDomain(unsigned long mode, Problem P, int domain) {
	List l = P->domains[domain];
	while (l) {
		LevelValue lv = (LevelValue) l->item;
		l = l->next;
		lvPrint(mode, lv, NULL, 1);
		cdgPrintf(mode, "\n");
	}
}

/* --------------------------------------------------------------------
 Print current state of the problem.

 Displays the current state of the Problem. In particular, prints the
 current Analysis as well as the recent transformation history.
 */
void printProblem(unsigned long mode, Problem P) {

	aCalibrate(P->current);
	cdgPrintf(mode, "\n\n\n");

	aPrint(mode, P->current);

	/* print conflicts */
	printConflicts(mode, P->current);

	/* print Badness */
	cdgPrintf(mode, "\n");
	bPrint(mode, P->current->b);

	/* special case: suboptimal Badness during a repair step */
	if (P->best && bCompare(P->best->b, P->current->b)) {
		cdgPrintf(mode, ", chasing ");
		bPrint(mode, P->best->b);
	}

	/* special case: optimal Badness reached */
	if (P->current->b->hard == 0 && !significantlyGreater(P->globalMaximum,
			P->current->b->soft)) {
		cdgPrintf(mode, " (optimal)");
		resetProblem(P);
	}

	/* some more measures about the current Analysis */
	printAmbiguity(mode, P);
	cdgPrintf(mode, "\n");

	/* print history */
	printHistory(mode, P);

}

/* ----------------------------------------------------------------------
 Check Analysis for internal consistency, and abort with warning if
 an error is found.

 Different levels of checking are necessary so that aCheck() can be used
 during the construction of Analyses, when some of the structures checked
 here are incomplete because they haven't been built yet.
 */
void aCheck(Analysis a, int level) {

	int i, j, k;
	LevelValue lv;
	LexemNode lna, lnb;
	Analysis b;
	LexemGraph lg = a->P->lg;

	ByteVector bound = bvNew(a->P->maxTimepoint);
	ByteVector slots = bvNew(a->P->nrSlots);

	bvSetAllElements(slots, FALSE);

	/* invalid LV in the Analysis? */
	for (i = 0; i < vectorSize(a->slots); i++) {
		lv = vectorElement(a->slots, i);
		if (!lv)
			continue;

		checkValidity(a, lv);
		if (!slotActive(a, i)) {
			continue;
		}
		lv = vectorElement(a->slots, i);
		if (!lv)
			continue;

#if 0
		/* zoning violation? */
		if(spec(lv->regents) &&
				(((LexemNode)lv->regents->item)->arc->from < a->P->subStart ||
						((LexemNode)lv->regents->item)->arc->from >= a->P->subStop)) {
			cdgPrintf(CDG_INFO, "PANIC: illegal LV @%d/#%d points to %d rather than [%d..%d]\n",
					i,
					lv->indexWRTNet,
					((LexemNode)lv->regents->item)->arc->from,
					a->P->subStart, a->P->subStop);
			abort();
		}
#endif

		/* LV residing in wrong domain? */
		if ((j = lvIndex(lv)) != i) {
			cdgPrintf(CDG_ERROR, "PANIC: LV at @%d has lvIndex() of %d:\n\n",
					i, j);
			lvPrint(CDG_ERROR, lv, NULL, 2);
			cdgPrintf(CDG_ERROR, "\n");
			abort();
		}

		for (j = ((LexemNode) lv->dependents->item)->arc->from; j
				< ((LexemNode) lv->dependents->item)->arc->to; j++) {
			k = a->P->nrLevels * j + lv->level->no;
			bvSetElement(slots, TRUE, k);
		}

	}

	/* path criterion */
	bvSetAllElements(bound, FALSE);

	for (i = 0; i < a->P->nrLexemes; i++) {
		lna = (LexemNode) vectorElement(lg->nodes, i);

		if (bvElement(a->isUnselectedLexeme, lna->no))
			continue;

		bvSetElements(bound, TRUE, lna->arc->from, lna->arc->to);

		for (j = i + 1; j < a->P->nrLexemes; j++) {
			lnb = (LexemNode) vectorElement(lg->nodes, j);

			if (bvElement(a->isUnselectedLexeme, lnb->no))
				continue;

			if (lgOverlap(lna, lnb)) {
				cdgPrintf(CDG_ERROR, "PANIC: ");
				lgPrintNode(CDG_ERROR, lna);
				cdgPrintf(CDG_ERROR, " and ");
				lgPrintNode(CDG_ERROR, lnb);
				cdgPrintf(CDG_ERROR, " both selected at time point %d\n",
						lna->arc->from);
				abort();
			}
		}
	}

	/* skip Badness check */
	if (level >= 2) {

		/* wrong Badness calculated? */
		b = aClone(a);
		aJudge(b);
		if (bCompare(a->b, b->b) || bCompare(b->b, a->b)) {

			cdgPrintf(CDG_ERROR, "PANIC: misevaluated Analysis!\n");
			aPrint(CDG_ERROR, b);
			cdgPrintf(CDG_ERROR, "\ncomputed conflicts=");
			bPrint(CDG_ERROR, a->b);
			sortConflicts(a);
			printConflicts(CDG_ERROR, a);
			cdgPrintf(CDG_ERROR, "\nactual conflicts=");
			bPrint(CDG_ERROR, b->b);
			printConflicts(CDG_ERROR, b);
			abort();
		}
		b = NULL;

		/* skip completeness check */
		if (level >= 3) {

			/* path criterion */
			for (i = a->subStart; i < a->subStop; i++) {
				if (!bvElement(bound, i)) {
					cdgPrintf(CDG_ERROR, "PANIC: time point %d not bound\n", i);
					abort();
				}
			}
			/* All domains bound? */
			for (i = 0; i < a->P->nrSlots; i++) {
				if (!slotActive(a, i)) {
					continue;
				}
				if (!bvElement(slots, i)) {
					cdgPrintf(CDG_ERROR, "PANIC: domain @%d not bound\n", i);
					aPrint(CDG_ERROR, a);
					abort();
				}
			}
		}
	}
	bound = NULL;
	slots = NULL;
}

/* ----------------------------------------------------------------------
 Build the fields P->lgWidth and P->lexemes.
 */
void buildLexemeTable(Problem P) {

	LexemNode ln;
	LexemGraph lg = P->lg;
	int i, j, k;

	P->lgWidth = lgWidth(lg);
	P->nrLexemes = vectorSize(lg->nodes);

	if (P->lexemes) {
		P->lexemes = NULL;
	}

	P->lexemes = vectorNew(P->maxTimepoint * P->lgWidth);
	vectorSetAllElements(P->lexemes, NULL);
	for (i = 0; i < P->nrLexemes; i++) {
		ln = (LexemNode) vectorElement(P->lg->nodes, i);
		for (j = ln->arc->from; j < ln->arc->to; j++) {
			k = j * P->lgWidth;
			while (vectorElement(P->lexemes, k) != NULL) {
				k++;
			}
			vectorSetElement(P->lexemes, ln, k);
		}
	}

#ifdef DEBUGBUILDLEXEMETABLE
	cdgPrintf(CDG_DEBUG, "Lexeme table: lgwidth = %d\n", P->lgWidth);
	for(i = 0; i < vectorSize(P->lexemes); i++) {
		ln = vectorElement(P->lexemes, i);
		if (!ln) continue;
		cdgPrintf(CDG_DEBUG, "Cell %d (timeslot %d item %d): ",
				i,
				i / P->lgWidth,
				i % P->lgWidth
		);
		lgPrintNode(CDG_DEBUG, ln);
		cdgPrintf(CDG_DEBUG, "\n");
	}
#endif

}

/* ----------------------------------------------------------------------
 Add LVs to an Analysis so that it binds slots 0 through slot.
 */
void aExtend(Analysis a, int slot) {

	LevelValue lv = NULL;
	List l;

	/* bound already? */
	if (slotFiller(a, slot) != NULL) {
		return;
	}

	/* select the first undeleted LV */
	for (l = a->P->domains[slot]; l; l = l->next) {
		lv = l->item;
		if (!lv->isDeleted && !(initVNsUnused && lvGetDependentGN(lv)->isVirtual && strcmp(lv->label, "")!=0 )) {
			break;
		}
	}

	/* if there is no undeleted LV, take the first one */
	if (lv->isDeleted) {
		lv = (LevelValue) a->P->domains[slot]->item;
	}
	aInsert(a, lv);

#ifdef DEBUGAEXTEND
	{	int j, k;
		lvPrint(CDG_DEBUG, lv, NULL, 0);
		cdgPrintf(CDG_DEBUG, "\n#%d binds", lv->indexWRTNet);
		for (j = ((LexemNode)lv->dependents->item)->arc->from;
				j < ((LexemNode)lv->dependents->item)->arc->to;
				j++) {
			k = a->P->nrLevels * j + lv->level->no;
			cdgPrintf(CDG_DEBUG, " @%d", k);
		}
		cdgPrintf(CDG_DEBUG, "\n");

#ifdef DEBUGAEXTENDBINDINGS
		for (j = 0; j < vectorSize(a->slots); j++) {
			if (j % a->P->nrLevels == 0) {
				cdgPrintf(CDG_DEBUG, "\n");
			}
			if (vectorElement(a->slots, j) != NULL) {
				cdgPrintf(CDG_DEBUG, "X");
			} else if (slotFiller(a,j)) {
				cdgPrintf(CDG_DEBUG, "I");
			} else {
				cdgPrintf(CDG_DEBUG, "0");
			}
		}
		cdgPrintf(CDG_DEBUG, "\n");
#endif
	}
#endif

}

/* ----------------------------------------------------------------------
 Put an LV into all appropriate domains, and count it.

 This function registers lv in all domains of P where it is
 appropriate. It also updates P->nrLVs and
 P->nrUndeletedLVs.
 */
void addLVToProblem(Problem P, LevelValue lv) {

	int k, index;

	/* put this LV into whatever domains need it */
	for (k = ((LexemNode) lv->dependents->item)->arc->from; k
			< ((LexemNode) lv->dependents->item)->arc->to; k++) {
		index = P->nrLevels * k + lv->level->no;
		P->domains[index] = listInsertSorted(P->domains[index], lv, lvCompare);
	}

	/* make sure that
	 (a) LV has a non-negative number
	 (b) P remembers to allocate at least this much space when
	 reasoning about its LVs
	 */
	if (lv->indexWRTNet < 0) {
		lv->indexWRTNet = P->nrLVs;
	}
	P->nrLVs++;

	/* The field nrUndeletedLVs is special because it counts, not
	 the LVs, but the slots that the remaining LVs may bind. That
	 way we can just divide nrUndeletedLVs by the number of all
	 slots and get the exact ambiguity of the Problem.  */
	if (!lv->isDeleted) {
		P->nrUndeletedLVs += (((LexemNode) lv->dependents->item)->arc->to
				- ((LexemNode) lv->dependents->item)->arc->from);
	}
}
/* ----------------------------------------------------------------------
 Transform Parse to Analysis.

 Builds an Analysis from a decorated Parse.
 */
Analysis parse2analysis(Problem P, Parse p) {
	Analysis result;
	LevelValue lv, model;
	int i;
	ByteVector bv;

	/* Parse must be decorated */
	if (!p->LVs) {
		cdgPrintf(CDG_ERROR, "ERROR: parse %s is not decorated!\n", p->id);
		return (NULL);
	}

	result = aNew(P);

	/* Parse must contain the correct number of LVs */
	if (vectorSize(p->LVs) > P->nrSlots) {
		cdgPrintf(
				CDG_ERROR,
				"ERROR: parse %s contains more subordinations than necessary!\n",
				p->id);
		return NULL;
	}

	if (vectorSize(p->LVs) == P->nrSlots) {

		/* copy lexical selection from parse */
		bv = bvClone(P->lg->isDeletedNode);
		if (!lgCopySelection(P->lg, p->lg)) {
			return NULL;
		}
		bvCopy(result->isUnselectedLexeme, P->lg->isDeletedNode);
		bvCopy(P->lg->isDeletedNode, bv);
		bv = NULL;
	} else {
		aBuildPath(result);
	}

	/* copy LVs */
	for (i = 0; i < vectorSize(p->LVs); i++) {
		lv = vectorElement(p->LVs, i);
		if (!lv || !lv->level->useflag) {
			continue;
		}
		model = translateLV(P, lv);
		if (!model) {
			cdgPrintf(CDG_WARNING,
					"WARNING: LV @%d not present in net, winging it\n", i);
			aExtend(result, i);
			model = inventLV(result, i);
		}
		aInsert(result, model);
	}

	/* Again, the Parse may not provide all required LVs,
	 so we make sure to extend the Analysis as needed. */
	for (i = 0; i < vectorSize(result->slots); i++) {
		if (slotActive(result, i) && !slotFiller(result, i)) {
			LevelValue lv = inventLV(result, i);
			cdgPrintf(CDG_WARNING,
					"WARNING: LV @%d not specified in parse, adding it\n", i);
			vectorSetElement(result->slots, lv, i);
		}
	}
	return result;

}

/* ----------------------------------------------------------------------
 Transform AnnoEntry to Analysis.

 Converts AE to a parse, decorates it, and calls parse2analysis().

 Either of these steps may fail so that NULL is returned.
 */
Analysis anno2analysis(Problem P, AnnoEntry ae) {

	Parse p;
	Analysis a = NULL;

	if (!ae) {
		cdgPrintf(CDG_WARNING, "WARNING: No annotation available.\n");
		return NULL;
	}

	/* Luckily, this is a solved problem:
	 we let the parse module do the work for us. */
	if ((p = parseFromAnno(ae)) && parseDecorate(p, P->lg, ae)) {
		a = parse2analysis(P, p);
	}

	if (p) {
		parseDelete(p);
	}
	if (a) {
		return a;
	} else {
		cdgPrintf(CDG_WARNING, "WARNING: Could not build reference analysis!\n");
		return NULL;
	}
}

/* ----------------------------------------------------------------------
 Import LVs from NET into P.

 If INDEX is >0, then only LVs with higher values for ->indexWRTNet
 will be considered.
 */
void importLVs(Problem P, int index) {
	int i, j;

	for (i = 0; i < vectorSize(P->net->nodes); i++) {
		ConstraintNode n = vectorElement(P->net->nodes, i);
		if (!n->level->useflag) {
			continue;
		}
		for (j = 0; j < vectorSize(n->values); j++) {
			LevelValue lv = vectorElement(n->values, j);
			if (lv->indexWRTNet < index)
				continue;
			addLVToProblem(P, lv);
		}
	}
}

/* ----------------------------------------------------------------------
 Allocate a ProblemStruct and fill all information that can be set with
 no context.
 */
Problem newProblem(void) {
	Problem P = memMalloc(sizeof (ProblemStruct));
	P->id = strPrintf("problem%d", problem_counter++);
	P->globalMinimum = 0.0;
	P->judgeCounter = 0;
	P->patchAttempts = 0;
	P->successfulPatches = 0;

	P->firstSoftFound = 0.0;
	P->lastSolFound = 0.0;
	P->killedByTimelimit= FALSE;

	if (0 == gatherLevels(P)) {
		cdgPrintf(CDG_WARNING, "No levels active -- nothing to transform.\n");
		memFree(P);
		return NULL;
	}

	P->globalMaximum = 1.0;
	P->unremovable = NULL;
	P->removedConflicts = NULL;
	P->history = NULL;
	P->lexemes = NULL;
	P->annotation = NULL;
	P->savedState = NULL;
	P->nrLVs = P->nrUndeletedLVs = 0;
	P->net = NULL;
	P->lg = NULL;
	P->best = NULL;
	P->current = NULL;
	P->annotation = NULL;
	P->profileTime = NULL;

	P->limitBuffer     = NULL;
	P->limitBufferSize = 0;
	return P;
}


// generate a problem simply for caching values accessed by analyses
Problem makeZombieProblem(ConstraintNet net) {
	int i;
	Problem P = newProblem();
	LexemGraph lg = net->lexemgraph;
	P->net = net;
	P->lg = net->lexemgraph;
	P->minTimepoint = lg->min;
	P->maxTimepoint = lg->max;
	P->nrTimepoints = lg->max - lg->min;
	P->nrSlots = P->maxTimepoint * P->nrLevels;
	P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;

	buildLexemeTable(P);
	P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
	for (i = 0; i < P->nrSlots; i++) {
		P->domains[i] = NULL;
	}
	importLVs(P, 0);

	return P;
}

/* ----------------------------------------------------------------------
 Build a Problem with an initial Analysis from a net.

 This function allocates a new Problem structure and initializes it. In
 particular, it creates an initial Analysis for the Problem as
 P->current. This Analysis is guaranteed to be complete and acoustically
 consistent, but it may be ungrammatical. (At least it will be
 constructed quickly.) If this fails, NULL is returned.

 If P is not NULL, the initial Analysis the initial Analysis will be
 constructed to match it. Otherwise, if NET already contains a Parse,
 then the initial Analysis will be constructed from it.

 The parameter PRESSURE can be used to exert temporal pressure
 on the processing. It is assigned to the globalMinimum of the
 resulting Problem.

 */
Problem net2problem(ConstraintNet net, Number pressure, Parse p) {

	Analysis A;
	LexemGraph lg = net->lexemgraph;
	Problem P = newProblem();
	Number minimum = 0.0;

	int i;

	if (!P) {
		return NULL;
	}

	P->net = net;
	P->lg = net->lexemgraph;
	P->minTimepoint = lg->min;
	P->maxTimepoint = lg->max;
	P->nrTimepoints = lg->max - lg->min;
	P->nrSlots = P->maxTimepoint * P->nrLevels;
	P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;

	/* translate external time pressure into internal controlling
	 variables. At the moment this is done by using the parameter
	 `pressure' as a non-zero initialization for globalMinimum, but
	 other methods could be invented. */
	minimum = pressure;

	/* build table of lexeme nodes */
	cdgPrintf(CDG_INFO, "Building table of lexemes...");
	cdgFlush();
	buildLexemeTable(P);
	cdgPrintf(CDG_INFO, " done.\n");

	/* build domains */
	cdgPrintf(CDG_INFO, "Building domains...");
	cdgFlush();
	P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
	for (i = 0; i < P->nrSlots; i++) {
		P->domains[i] = NULL;
	}
	importLVs(P, 0);

	P->current = NULL;
	P->best = NULL;

	/* If a domain remains empty, no solution is possible. This is important
	 because transformation methods may assume that each domain has at
	 least one value. */
	for (i = 0; i < P->nrSlots; i++) {
		int tp = i / P->nrLevels;
		if (tp > P->maxTimepoint || tp < P->minTimepoint) {
			continue;
		}
		if (!bvElement(P->levelActive, i % P->nrLevels)) {
			continue;
		}
		dCheck(P, i);
		if (P->domains[i] == NULL) {
			Level level = listNthElement(inputCurrentGrammar->levels, 1 + (i
					% P->nrLevels));
			LexemNode ln = vectorElement(P->lexemes, P->lgWidth * i
					/ P->nrLevels);
			cdgPrintf(
					CDG_WARNING,
					"\nWARNING: empty domain `%s'/%s, problem has no solution.\n",
					ln->arc->word, level->id);
			return NULL;
		}
	}
	cdgPrintf(CDG_INFO, " done.\n");

	/* Since the calls to dCheck() in the previous step may have deleted
	 lexemes, we have to re-optimize the net so that lgMakePath will
	 not select dangling lexemes. */
	if (cnOptimizeNet(net) < 0) {
		deleteProblem(P);
		return NULL;
	};

	/* initial analysis */
	cdgPrintf(CDG_INFO, "Building initial analysis...");
	cdgFlush();
	if (p) {
		A = parse2analysis(P, p);
	} else {
		A = aNew(P);
		aBuildPath(A);
		vectorSetAllElements(A->slots, NULL);
		P->current = A;

		/* iterate over all slots */
		for (i = 0; i < P->nrSlots; i++) {
			if (!slotActive(P->current, i)) {
				continue;
			}
			aExtend(A, i);
		}
	}

	/* parse2analysis() can fail on inadequate input */
	if (NULL == A) {
		return NULL;
	}

	P->current = A;
	P->best = NULL;
	cdgPrintf(CDG_INFO, " done.\n");

	/* compute conflicts */
	cdgPrintf(CDG_INFO, "Computing conflicts...");
	cdgFlush();
	aJudge(P->current);
	aCheck(A, 3);
	cdgPrintf(CDG_INFO, " done.\n");

	/* Now clone the completed Analysis */
	// or not.. we get a new one at newMaxmimum anyway
	//P->best = aClone(P->current);

	/* we *might* just have constructed a valid Analysis by sheer luck... */
	if (aValid(P->current) && P->current->b->soft > P->globalMinimum) {
		P->globalMinimum = P->current->b->soft;
	}

	AnnoEntry ae = findAnnoForLattice(net->lexemgraph->lattice->id, TRUE);
	if(ae) {
		/* build another Analysis that corresponds to an annotation we may
		 have available */
		P->annotation = anno2analysis(P, ae);

		/* Judge the annotation so it won't confuse the optimization
		 with its flashy score of 1.0 */
		aJudge(P->annotation);
	} else {
		P->annotation = NULL;
	}

	/* now apply pressure to the problem */
	P->globalMinimum = max(minimum, P->globalMinimum);

	P->profileTime = timerNew();

	resetProblem(P);
	rememberProblem(P);
	newMaximum(P->current);

	/* register problem */
	hashSet(cdgProblems, P->id, P);

	return P;

}

/* ----------------------------------------------------------------------
 Create a Problem without a constraint net, just from a decorated parse.

 A constraint net will be built, but it will only contain very few values
 and be unsuitable for real parsing; it is just useful for being able to
 use the transform module at all.
 */
Problem parse2problem(Parse p, Number pressure) {

	Analysis temp;
	Problem P;
	Number minimum = 0.0;
	int index = 0;
	int i, j;

	if (parseIsAbstract(p)) {
		cdgPrintf(CDG_ERROR, "ERROR: need a decorated parse.\n");
		return NULL;
	}

	if (NULL == (P = newProblem())) {
		return NULL;
	}

	P->net = NULL;
	P->lg = p->lg;
	P->minTimepoint = P->lg->min;
	P->maxTimepoint = P->lg->max;
	P->nrTimepoints = P->lg->max - P->lg->min;
	P->nrSlots = P->maxTimepoint * P->nrLevels;
	P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;

	/* translate external time pressure into internal controlling
	 variables. At the moment this is done by using the parameter
	 `pressure' as a non-zero initialization for globalMinimum, but
	 other methods could be invented. */
	minimum = pressure;
	if (p->score > minimum) {
		minimum = p->score;
	}

	/* build table of lexeme nodes */
	cdgPrintf(CDG_INFO, "Building table of lexemes...");
	cdgFlush();
	buildLexemeTable(P);
	cdgPrintf(CDG_INFO, " done.\n");

	/* build table of domains */
	cdgPrintf(CDG_INFO, "Building domains...");
	cdgFlush();
	P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
	for (i = 0; i < P->nrSlots; i++) {
		P->domains[i] = NULL;
	}

	cdgPrintf(CDG_INFO, " done.\n");

	P->best = NULL;
	temp = parse2analysis(P, p);
	if (!temp) {
		cdgPrintf(CDG_ERROR, "ERROR: cannot transform `%s' to Analysis\n",
				p->id);
		return NULL;
	}

	/* populate domains */
	P->nrLVs = P->nrUndeletedLVs = 0;
	for (i = 0; i < P->nrSlots; i++) {
		LevelValue lv = slotFiller(temp, i);
		int from = ((LexemNode) lv->dependents->item)->arc->from;

		/* add one LV for each modifier variant */
		for (j = 0; j < P->lgWidth; j++) {
			LexemNode ln = vectorElement(P->lexemes, P->lgWidth * from + j);
			LevelValue alt;
			if (ln) {

				alt = lvClone(lv);
				alt->dependents = listPrependElement(NULL, ln);
				alt->indexWRTNet = index++;
				evalUnary(alt, NULL, NULL, FALSE, NULL, NULL, TRUE);
				alt->limit = alt->score;
				addLVToProblem(P, alt);
			}
		}
	}
	temp = NULL;
	P->current = parse2analysis(P, p);

	/* compute conflicts */
	cdgPrintf(CDG_INFO, "Computing conflicts...");
	cdgFlush();
	aJudge(P->current);
	aCheck(P->current, 3);
	cdgPrintf(CDG_INFO, " done.\n");

	/* Now clone the completed Analysis */
	P->best = aClone(P->current);

	if (aValid(P->current) && P->current->b->soft > P->globalMinimum) {
		P->globalMinimum = P->current->b->soft;
	}

	AnnoEntry ae = findAnnoForLattice(P->lg->lattice->id, TRUE);
	if(ae) {
		/* build another Analysis that corresponds to an annotation we may
		 have available */
		P->annotation = anno2analysis(P, ae);

		/* Judge the annotation so it won't confuse the optimization
		 with its flashy score of 1.0 */
		aJudge(P->annotation);
	} else {
		P->annotation= NULL;
	}

	/* now apply pressure to the problem */
	P->globalMinimum = minimum;

	P->profileTime = timerNew();
	resetProblem(P);
	rememberProblem(P);
	newMaximum(P->current);

	/* register problem */
	hashSet(cdgProblems, P->id, P);

	return P;

}

/*
 * delete the rapairAttempt
 * including the timer and all constraint violations
 */
void repairAttemptDelete(RepairAttempt ra) {
//  Those CV are not owned, at least not all of them,
//  so there is a possible memory leak here
//	List l= ra->otherConflicts;
//	while(l) {
//		cvDelete(l->item);
//		l=l->next;
//	}
	ra->otherConflicts = NULL;
	ra->timer = NULL;
	memFree(ra);
}

/*
 * free the memory used by the problem struct
 * including all cached values
 * excluding the constraintNet, the lexemgraph, all levelvalues
 */
void problemDelete(Problem P) {
	List l;
	int i;

	for(i=0; i<P->nrSlots; i++) {
		l= P->domains[i];
		l = NULL;
		// the domains contain the levelvalues from the constraintnet, so we don't delete them here
	}

	// list of ints, no need to delete elements
	P->history = NULL;


	if(P->levelActive)
		P->levelActive = NULL;

	if(P->lexemes)
		// we delete only the vector, not the lexemes, as the lexemgraph owns it
		P->lexemes = NULL;

	if(P->profileTime)
		P->profileTime = NULL;

	l= P->removedConflicts;
	while(l) {
		l->item = NULL;
		l=l->next;
	}
	P->removedConflicts = NULL;

	l= P->repairAttempts;
	while(l) {
		l->item = NULL;
		l=l->next;
	}
	P->repairAttempts = NULL;

	// those cvs belong to the analysis, TODO confirm
//	l= P->unremovable;
//	while(l) {
//		cvDelete(l->item);
//		l=l->next;
//	}

	P->unremovable = NULL;

	memFree(P);
}


/* ----------------------------------------------------------------------
 Register new Parse (if it is really new).

 Converts P->best to a Parse and adds it to the constraint net underlying
 P. This is only done if the net does not already contain an identical
 Parse.

 */
void addParse(Analysis a) {

	List l;
	Parse parse;
	Problem P = a->P;

	if (!a) {
		return;
	}

	/* only register solutions that aren't already known */
	parse = analysis2parse(a);
	if (P->net) {
		for (l = P->net->parses; l != NULL; l = l->next) {
			if (parseEquals(l->item, parse)) {
				parseDelete(parse);
				return;
			}
		}

		/* add to net */
		P->net->parses = listPrependElement(P->net->parses, parse);
	}

	/* add to global list */
	parseRegister(parse);
	cdgPrintf(CDG_INFO, "INFO: parse %s added to the global parses list.\n",
			parse->id);

}

/* ----------------------------------------------------------------------
 Return a Vector of Badness by domain of A.
 */
Vector aAnalyze(Analysis a) {

	Vector byDomain = vectorNew(a->P->nrSlots);
	int i;
	List l;
	ConstraintViolation cv;
	Badness b;

	for (i = 0; i < a->P->nrSlots; i++) {
		vectorSetElement(byDomain, bestBadness(), i);
	}

	/* count all conflicts */
	for (l = a->conflicts; l; l = l->next) {
		cv = l->item;

		b = vectorElement(byDomain, cv->nodeBindingIndex1);
		bAdd(b, cv->penalty);

		if (cv->lv2) {
			b = vectorElement(byDomain, cv->nodeBindingIndex2);
			bAdd(b, cv->penalty);
		}
	}
	return byDomain;

}

/* ----------------------------------------------------------------------
 Return a domain to change.

 This simply returns any domain that causes at least one conflict.
 */
int selectVariable(Problem P) {

	Vector byDomain = aAnalyze(P->current);
	int i, start;
	int result;

	/* just pick one that violates something */
	do {
		start = result = rand() % P->nrSlots;

		while (((Badness) vectorElement(byDomain, result))->no == 0
				|| listContains(P->history, (Pointer) result)) {
			result = (result + 1) % P->nrSlots;
			if (result == start) {
				result = -1;
				break;
			}
		}
	} while (!slotActive(P->current, start));

	for (i = 0; i < P->nrSlots; i++) {
		memFreeFunction(vectorElement(byDomain, i) );
	}
	byDomain = NULL;
	return result;

}
/* ----------------------------------------------------------------------
 Compare Analyses by Badness.
 */
Boolean aCompare(Analysis a, Analysis b) {
	return bCompare(a->b, b->b);
}

/* ----------------------------------------------------------------------
 Copy of genetic::cross().
 */
Analysis aCross(Analysis a, Analysis b) {

	Analysis result = aClone(a);
	int cut = 0;
	int i = 0;
	Patch p = pNew(result, a->P->nrSlots - cut + 1, 0);
	Patch q;

	if (!a->P->net) {
		return NULL;
	}

	do {
		cut = rand() % a->P->nrSlots;
	} while (!slotActive(a, i));

	for (i = cut; i < a->P->nrSlots; i++) {
		LevelValue lv;
		if (!slotActive(a, i)) {
			continue;
		}
		lv = vectorElement(b->slots, i);
		if (lv->indexWRTNet == -1) {
			lv = lvClone(lv);
		}
		q = pNext(p, lv, OD_ALLOWALL | OD_DELETED);
		p = q;
	}
	(void) pLexicalize(p, OD_DELETED, FALSE);
	apply(result, p);

	/* mutate the result */
	p = pNew(a, 1, OD_DELETED);
	q = pNext(p, vectorElement(a->P->net->values, rand() % vectorSize(
			a->P->net->values)), OD_ALLOWALL | OD_DELETED);
	if (q) {
		(void) pLexicalize(q, OD_DELETED, FALSE);
		apply(result, q);
	}
	p = NULL;

	return result;

}

/*
 * collect the selected lexeme's category for each slot in an analysis
 * virtual nodes are not included
 */
List aGetSelectedCats(Analysis A) {
	List result = listNew();
	int i;
	for(i=0; i< vectorSize(A->slots); i++) {
		LevelValue lv = vectorElement(A->slots,i);
		if(lv->level->no > 0) {
			continue;
		}
		if(lvGetDependentGN(lv)->isVirtual) {
			// don't include tags of virtual nodes
			break;
		}
		LexiconItem li= ((LexemNode)lv->dependents->item)->lexem;
		String sCat= li->values[inputRegisterPath(scanPath("cat"),inputCurrentGrammar)]->data.string;
		result = listAppendElement(result, sCat);
	}

	return result;
}

/* ----------------------------------------------------------------------
 Return LV with indexWRTNet INDEX in P, or NULL.
 */
LevelValue lvByNumber(Problem P, int no) {
	List l;
	int i;

	for (i = 0; i < P->nrSlots; i++) {
		for (l = P->domains[i]; l != NULL; l = l->next) {
			LevelValue lv = l->item;
			if (lv->indexWRTNet == no) {
				return lv;
			}
		}
	}
	return NULL;
}

/* ----------------------------------------------------------------------
 If there is already a Problem associated with NET, return it. If there
 isn't, return NULL.
 */
Problem findProblem(ConstraintNet net) {
	List l, m;

	for (l = m = hashListOfKeys(cdgProblems); l != NULL; l = l->next) {
		Problem P = hashGet(cdgProblems, l->item);
		if (P->net == net) {
			return P;
		}
	}
	return NULL;
}

/*
 *
 */
RepairAttempt repairAttemptNew(ConstraintViolation cv) {
	//cdgPrintf(CDG_INFO, "allocating repairAttempt...\n");
	RepairAttempt ret = (RepairAttempt) memMalloc(sizeof (RepairAttemptStruct));
	ret->otherConflicts = listNew();
	ret->successful = FALSE;
	ret->primaryConflict = cv->constraint;
	ret->time = 0;
	ret->timer = timerNew();
	//cdgPrintf(CDG_INFO, "done\n");
	return ret;
}

/*
 * TODO
 */
Problem recycleProblem(/*Problem P,*/ ConstraintNet net, Number pressure, Vector slots, ProfilingObject prof) {

	Analysis A;
	LexemGraph lg = net->lexemgraph;
	Number minimum = 0.0;
	int i;

	Timer timer = timerNew();
	ProfilingObject prof2 = proNew();

	Problem P = newProblem();
	P->net = net;
	P->lg = net->lexemgraph;
	P->minTimepoint = lg->min;
	P->maxTimepoint = lg->max;
	P->nrTimepoints = lg->max - lg->min;
	P->nrSlots = P->maxTimepoint * P->nrLevels;
	P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;

	if (!P) {
		return NULL;
	}

	minimum = pressure;

	/* build table of lexeme nodes */
	cdgPrintf(CDG_INFO, "Building table of lexemes...");
	cdgFlush();
	buildLexemeTable(P);
	cdgPrintf(CDG_INFO, " done.\n");

	proMeasureTime(timer, prof2, "   rP-0 build-lexs");

	/* build domains */
	cdgPrintf(CDG_INFO, "Building domains...");
	cdgFlush();
	P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
	for (i = 0; i < P->nrSlots; i++) {
		P->domains[i] = NULL;
	}
	importLVs(P, 0);

	P->current = NULL;
	P->best = NULL;

	proMeasureTime(timer, prof2, "   rP-1 build");

	/* If a domain remains empty, no solution is possible. This is important
	 because transformation methods may assume that each domain has at
	 least one value. */
	for (i = 0; i < P->nrSlots; i++) {
		int tp = i / P->nrLevels;
		if (tp > P->maxTimepoint || tp < P->minTimepoint) {
			continue;
		}
		if (!bvElement(P->levelActive, i % P->nrLevels)) {
			continue;
		}
		dCheck(P, i);
		if (P->domains[i] == NULL) {
			Level level = listNthElement(inputCurrentGrammar->levels, 1 + (i % P->nrLevels));
			LexemNode ln = vectorElement(P->lexemes, P->lgWidth * i
					/ P->nrLevels);
			cdgPrintf(
					CDG_WARNING,
					"\nWARNING: empty domain `%s'/%s, problem has no solution.\n",
					ln->arc->word, level->id);
			return NULL;
		}
	}
	proMeasureTime(timer, prof2, "   rP-2 propagate");

	/* Since the calls to dCheck() in the previous step may have deleted
	 lexemes, we have to re-optimize the net so that lgMakePath will
	 not select dangling lexemes. */
	if (cnOptimizeNet(net) < 0) {
		deleteProblem(P);
		return NULL;
	};

	proMeasureTime(timer, prof2, "   rP-3 opti");

	/* initial analysis */
	A = aNew(P);
	aBuildPath(A);
	vectorSetAllElements(A->slots, NULL);
	P->current = A;

	/* iterate over all slots */
	for (i = 0; i < P->nrSlots; i++) {
		if (!slotActive(P->current, i)) {
			continue;
		}
		//if(slots && vectorSize(slots) > i) {
		//	vectorSetElement(A->slots, vectorElement(slots, i), i);
		//} else {
		aExtend(A, i);
		//}
	}

	if(slots) {
		// add the given slots to the analysis
		for (i = 0; i < vectorSize(slots); i++) {
			aInsert(A, vectorElement(slots, i));
		}
	}


	/* parse2analysis() can fail on inadequate input */
	if (NULL == A) {
		return NULL;
	}

	P->current = A;
	P->best = NULL;
	cdgPrintf(CDG_INFO, " done.\n");
	proMeasureTime(timer, prof2, "   rP-4 init");

	/* compute conflicts */
	cdgPrintf(CDG_INFO, "Computing conflicts...");
	cdgFlush();
	aJudge(P->current);
	aCheck(A, 3);
	cdgPrintf(CDG_INFO, " done.\n");

	/* Now clone the completed Analysis */
	// or not.. we get a new one at newMaxmimum anyway
	//P->best = aClone(P->current);

	/* we *might* just have constructed a valid Analysis by sheer luck... */
	if (aValid(P->current) && P->current->b->soft > P->globalMinimum) {
		P->globalMinimum = P->current->b->soft;
	}

	proMeasureTime(timer, prof2, "   rP-5 judge");

	P->annotation = NULL;

	/* now apply pressure to the problem */
	P->globalMinimum = max(minimum, P->globalMinimum);

	P->profileTime = timerNew();

	resetProblem(P);
	proMeasureTime(timer, prof2, "   rP-6 reset");
	rememberProblem(P);
	proMeasureTime(timer, prof2, "   rP-7 remember");

	newMaximum(P->current);
	proMeasureTime(timer, prof2, "   rP-8 newMax");

	proMerge(prof, prof2);

	return P;
}

/* -- ENDOFFILE ------------------------------------------------------------ */

