/*
 * Copyright (C) 1997-2009 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Niels Beuck
 * Birth: 23.10.09
 *
 */


#include <config.h>

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <blah.h>

#include "virtualnode.h"
#include "lexemgraph.h"
#include "hook.h"
#include "transform.h"
#include "parse.h"
#include "cdg.h"
#include "set.h"
#include "increment.h"
#include "input.h"

Boolean useVirtualNodes = FALSE;
Boolean matchVirtualNodes = TRUE;
Boolean orderVirtualNodes = FALSE;
Boolean virtFlexibleLabelMatching = TRUE;
Boolean blockVirtualNodes = TRUE;
float virtMatchingFactor = 0.4;
String LabelGroupsProperty;

/** Initialization ******************************************************************* */


void virtualInitialize() {
	setRegister("useVirtualNodes", SET_BOOL, &useVirtualNodes, NULL, NULL, NULL, NULL);
	setRegister("matchVirtualNodes", SET_BOOL, &matchVirtualNodes, NULL, NULL, NULL, NULL);
	setRegister("orderVirtualNodes", SET_BOOL, &orderVirtualNodes, NULL, NULL, NULL, NULL);
	setRegister("virtMatchingFactor", SET_FLOAT, &virtMatchingFactor, NULL, NULL, NULL, NULL);
	setRegister("virtFlexibleLabelMatching", SET_BOOL, &virtFlexibleLabelMatching, NULL, NULL, NULL, NULL);
	setRegister("blockVirtualNodes", SET_BOOL, &blockVirtualNodes, NULL, NULL, NULL, NULL);
	LabelGroupsProperty = strRegister("LabelGroups");
}


/** VirtualNodes ******************************************************************* */

/**
 *
 */
VirtualNode vnNew(VirtualNodeSpec spec, int n) {
	VirtualNode vn = (VirtualNode) memMalloc(sizeof(VirtualNodeStruct));
	vn->spec = spec;
	vn->id = strPrintf("%s%d", spec->name, n);
	vn->gn = NULL;
	return vn;
}

/** Virtual Extension ******************************************************************* */
VirtualNode veGetVirtualNode(VirtualLGExtension ve, String id) {
	int i;
	for(i=0;i<vectorSize(ve->virtualNodes);i++) {
		VirtualNode vn = vectorElement(ve->virtualNodes,i);
		if(!strcmp(vn->id,id)) {
			return vn;
		}
	}
	return NULL;
}

/**
 * counts how many nodes with this spec are already part of this lexemgraph extension
 */
List veOccurences(VirtualLGExtension ve, VirtualNodeSpec spec) {
	VirtualNode vn;
	List ret = listNew();
	int i;
	for(i=0; i< vectorSize(ve->virtualNodes); i++) {
		vn= vectorElement(ve->virtualNodes, i);
		if(vn->spec== spec) {
			ret= listAppendElement(ret,vn);
		}
	}
	return ret;
}



/**
 * create new Virtual extension for the given lexemgraph
 */
VirtualLGExtension veNew(LexemGraph lg) {
	VirtualLGExtension ve = (VirtualLGExtension) memMalloc(sizeof(VirtualLGExtensionStruct));
	ve->lg = lg;
	ve->virtualNodes= vectorNew(10);
	ve->vStartID = lg->max; // TODO gilt nur, wenn alle arcs die Lnge 1 haben
	ve->specs = NULL;
	ve->blocked = bvNew(vectorSize(lg->graphemnodes));
	bvSetAllElements(ve->blocked, FALSE);
	return ve;
}

/*
 * returns, if the given graphemenodes belongs to the given virtual node specification
 * nonvirtual nodes never belong to a vns
 */
Boolean vnsIsCompatibleGN(VirtualNodeSpec vns, GraphemNode gn) {
	if(!gn->isVirtual)
		return FALSE;

	return strcmp(vns->name,gn->arc->word) == 0;
}

/**
 * create a new ve from an old one
 * the lg will be cloned and set elsewhere
 */
VirtualLGExtension veClone(VirtualLGExtension ve) {
	VirtualLGExtension result = (VirtualLGExtension) memMalloc(sizeof(VirtualLGExtensionStruct));
	result->lg = NULL;
	result->vStartID = ve->vStartID;

	result->virtualNodes= vectorNew(vectorSize(ve->virtualNodes));
	int i;
	for(i=0;i< vectorSize(ve->virtualNodes);i++) {
		VirtualNode oldvn = vectorElement(ve->virtualNodes,i);
		VirtualNode newvn = memMalloc(sizeof(VirtualNodeStruct));
		newvn->id = strRegister(oldvn->id);
		newvn->spec = oldvn->spec;
		newvn->gn = oldvn->gn;

		vectorAddElement(result->virtualNodes,newvn);
	}
	result->blocked = bvClone(ve->blocked);

	return result;
}

/**
 * deallocate a virtual lg extension
 * the virtual nodes will be deleted, too,
 * but not the virtual node specs or the graphemnodes they hold pointers to
 * neither the lexemgraph itself
 */
void veDelete(VirtualLGExtension ve) {
	int i;
	for(i=0; i < vectorSize(ve->virtualNodes); i++) {
		VirtualNode vn = vectorElement(ve->virtualNodes,i);
		vn->gn = NULL;
	}
	ve->virtualNodes = NULL;
}

/**
 * add a virtual grapheme node to the virtual extension
 */
void veAddVirtualNode(VirtualLGExtension ve, GraphemNode gn, VirtualNodeSpec spec, int n) {
	VirtualNode vn= vnNew(spec, n);
	vectorAddElement(ve->virtualNodes,vn);
	vn->gn = gn;

	ve->blocked = bvNew(vectorSize(ve->lg->graphemnodes));
	bvSetAllElements(ve->blocked, FALSE);
}

/**
 * returns the VNspecification of the given name, if any
 */
VirtualNodeSpec veGetSpecByName(VirtualLGExtension ve, String name) {
	List l;
	VirtualNodeSpec spec;
	for(l=ve->specs; l; l=l->next) {
		spec= l->item;
		if(strcmp(spec->name, name)==0) {
			return spec;
		}
	}

	return NULL;
}


/** Arcs ************************************* */

/**
 * a nice new virtual arc not corresponding to a word in the sentence
 */
Arc makeVirtualArc(String name, int tp, int n) {
	String id = strPrintf("%s%d",name,n);
	Arc arc= newArc(tp, tp+1, tp, id, 1.0, NULL, tp , tp+1);
	arc->phraseQuotes=0;
	arc->parens=0;
	arc->quotes=0;
	return arc;
}

/** Virtual Node Specification *********************************************************** */

VirtualNodeSpec vnsNew(String name,Number penalty, List values, int copies) {
	VirtualNodeSpec spec= memMalloc(sizeof(VirtualNodeSpecStruct));
	spec->multiple=FALSE;
	spec->name= name;
	//spec->penalty= penalty;
	spec->values = values;
	spec->copies = copies;
	return spec;
}

/*
 * generates a virtualNodeSpec that has a lexeme corresponding to the annoEntry
 */
VirtualNodeSpec vnsFromAnno(Annotation an) {
	List l;
	List specs = listNew();

	Value val = memMalloc(sizeof(ValueStruct));
	val->type = VTConjunction;
	val->index = 0;
	for(l= an->specs; l; l=l->next) {
		Specification spec = l->item;
		if(spec->type != STTag) {
			break;
		}
		Value valSpec = memMalloc(sizeof(ValueStruct));
		valSpec->type = VTAVNode;
		valSpec->index = 0;
		AVNode avn = memMalloc(sizeof(AVNodeStruct));
		avn->attribute = spec->kind;
		Value valString = memMalloc(sizeof(ValueStruct));
		valString->type = VTString;
		valString->data.string = spec->name;
		valString->index= 0;
		avn->value = valString;
		valSpec->data.avnode = avn;
		specs = listAppendElement(specs, valSpec);
	}
	val->data.list = specs;

	VirtualNodeSpec spec = vnsNew(an->word, 1.0, listAppendElement(listNew(), val), 1);
	return spec;
}

/**
 * looks for occurrences of the vnSpec that are unused in the given analysis
 * returns the first one, if any
 */
VirtualNode vnsGetUnusedInstance(VirtualNodeSpec vns, Analysis a, VirtualLGExtension ve) {
	List l;
	List ocs = veOccurences(ve,vns);
	for(l= ocs; l; l=l->next) {
		VirtualNode vn = l->item;
		if(gnIsUnusedVirtualNode(vn->gn, a)) {
			return vn;
		}
	}
	return NULL;
}

/**
 * build a lexicon item corresponding to the virtual node spec
 */
List makeVirtualLexiconItems(VirtualNodeSpec spec) {
	List ret = listNew();
	List l;
	String word=spec->name;
	String filename="virtual_node";
	int lineNo = 0;
	int i = 0;

	for(l= spec->values;l;l=l->next, i++) {
		Value v= l->item;
		LexiconItem li = newLexiconItem( word, v, filename, lineNo );
		inputCacheLI(inputCurrentGrammar, li);

		String id;
		String sCat= leGetValue(li,inputCurrentGrammar,"cat");
		/*if(peekValue(v,scanPath("case"))!=ErrorValue) {
			String sCase= li->values[inputRegisterPath(scanPath("case"),inputCurrentGrammar)]->data.string;
			id= strPrintf("%s_%s_%s",word,sCat,sCase);
		} else {
			id= strPrintf("%s_%s",word,sCat);
		}*/
		String sNum= "";
		String sPer= "";
		String sBas= "";
		if(peekValue(v,scanPath("number"))!=ErrorValue) {
			sNum= leGetValue(li,inputCurrentGrammar,"number");
		}
		if(peekValue(v,scanPath("person"))!=ErrorValue) {
			sPer= leGetValue(li,inputCurrentGrammar,"person");
		}
		if(peekValue(v,scanPath("base"))!=ErrorValue) {
			sBas= leGetValue(li,inputCurrentGrammar,"base");
			if(strcmp(sBas, "unknown") ==0) {
				sBas= "";
			}
		}
		if(peekValue(v,scanPath("gender"))!=ErrorValue) {
			String sGen= leGetValue(li,inputCurrentGrammar,"gender");
			id= strPrintf("%s-%d_%s_%s_%s_%s_%s",word,i,sCat,sPer,sNum,sGen,sBas);
		} else {
			id= strPrintf("%s-%d_%s_%s_%s_%s",word,i,sCat,sPer,sNum,sBas);
		}
		li->description=id;
		ret= listAppendElement(ret,li);
	}

	return ret;
}


/**
 * Builds Lexemes for a virtual Lexeme Node given a specifiaction
 * tp= timpoint
 * n= number of instance of this spec, just for naming purposes
 */
List vnsMakeLexemNodes(VirtualNodeSpec spec, int tp, int n, GraphemNode gn) {
	List ret = listNew();
	LexemNode ln;
	List l = makeVirtualLexiconItems(spec);

	for(;l;l=l->next) {
		ln= memMalloc(sizeof(LexemNodeStruct));
		ln->arc = gn->arc;
		ln->grapheme = gn;
		ln->lexem = l->item;

		ln->lexemgraph = NULL; //gn->lexemgraph;
		ln->limit = 1.0;//spec->penalty;
		ln->no=0;
		ret= listAppendElement(ret,ln);
	}
	return ret;
}




/** GraphemNodes, procedures related to virtual nodes ************************************* */

/**
 * generate an grapheme node from a virtual Node
 */
GraphemNode gnFromVspec(VirtualNodeSpec spec, int tp, int n) {
	GraphemNode gn = gnNew();
	gn->isVirtual=TRUE;
	gn->arc = makeVirtualArc(spec->name,tp,n);
	gn->lexemes = vnsMakeLexemNodes(spec,tp,n, gn);

	gn->chunk =NULL;
	gn->lexemgraph = NULL; // has to be set later on
	gn->live = TRUE;
	gn->no = tp;
	gn->predictions = NULL; // no need to set it here
	return gn;
}

/**
 * returns if the given graphemenode is virtual,
 * i.e. if it corresponds to a word in the underlying sentence or not
 */
Boolean gnIsVirtual(GraphemNode gn) {
	return gn->isVirtual;
}

/**
 * a virtual graphem node (one that has no corresponding token in the input)
 * is unused if
 * - has NIL as regent on all levels
 * - on all levels it's levelvalue has no label
 * - is no regent for any other gn on any level
 */
Boolean gnIsUnusedVirtualNode(GraphemNode gn, Analysis a) {
	if(!gnIsVirtual(gn))
		return FALSE;

	if(gn->no > a->subStop ) // TODO substop might differ from the last id, but that one is not cached anywhere...
		// the analysis doesn't even contain a slot for the VN so:
		return TRUE;

	int i;
	for(i=0; i< vectorSize(a->slots); i++) {
		LevelValue lv= vectorElement(a->slots, i);
		if(!lv)
			continue;

		if(lvGetRegentGN(lv) == gn) {
			 // gn acts as a regent, so is not unused
			return FALSE;
		}
		if(lvGetDependentGN(lv) == gn && (strcmp(lv->label,"") != 0)) {
			return FALSE;
		}
	}
	// no reason found not to return true;
	return TRUE;
}

/**
 * do lexemes exist for the given graphemnodes that
 * share certain characteristics?
 */
Boolean gnHaveCompatibleLexemes(GraphemNode gna, GraphemNode gnb) {
   return TRUE; // TODO
}


/** Lexemgraph, procedures related to virtual nodes ************************************* */

/*
 * add one graphemnode to a Lexemgraph
 * this precedure is only needed, if virtual nodes are used,
 *   otherwise just add the node at end of the vector
 *
 * if the node is not virtual, it needs to be inserted before virtual ones and push them back
 */
void lgAddGraphemNode(LexemGraph lg, GraphemNode gn) {
	gn->lexemgraph = lg;
	if(!lg->ve || gnIsVirtual(gn)) { // just add it at the end
		gn->no = vectorAddElement(lg->graphemnodes, gn);
	} else { // add it before the virtual nodes
		int pos = lg->ve->vStartID;
		vectorInsertElement(lg->graphemnodes,gn,pos);
		gn->no= pos;
		int i;
		for(i=0; i< vectorSize(lg->ve->virtualNodes); i++){
			VirtualNode vn = vectorElement(lg->ve->virtualNodes,i);
			GraphemNode gn2 = vn->gn;
			gn2->no++;
			pushArc(gn2->arc, lg);
		}
		lg->ve->vStartID++;
	}

	if(lg->ve) {
		lg->ve->blocked = bvNew(vectorSize(lg->graphemnodes));
		bvSetAllElements(lg->ve->blocked, FALSE);
	}
}

/**
 * add a new instance of the virtualNodeSpec to the lexemgraph
 * returns the newly generated graphemnode
 */
GraphemNode lgAddVirtualNode(LexemGraph lg, VirtualNodeSpec spec, int n) {
	GraphemNode gn;
	LexemNode ln;

	lg->max++;

	gn = gnFromVspec(spec, vectorSize(lg->graphemnodes),n);
	lgAddGraphemNode(lg, gn);
	List l= gn->lexemes;
	for(;l;l=l->next) { // lgAddLN done quickly
		ln= l->item;
		ln->no = vectorAddElement(lg->nodes, ln);
		ln->lexemgraph = lg;
		bvSetElement(lg->isDeletedNode, FALSE, ln->no);
		bvSetElement(lg->isFrozenNode, FALSE, ln->no);
	}
	lgComputeDistances(lg);
	lgComputeNoOfPaths(lg, TRUE);

	if (lg->ve)
		veAddVirtualNode(lg->ve,gn,spec, n);

	return gn;
}


/** Constraintnet, procedures related to virtual nodes ************************************* */

/**
 * returns the first index of constraintnodes/slots/domains that belongs to a virtual node
 */
int cnGetFirstVirtualIndex(ConstraintNet net) {
	int ret= 0;
	int i;
	for(i=0;i<vectorSize(net->nodes);i++) {
		ConstraintNode cNode= vectorElement(net->nodes,i);
		if(gnIsVirtual(cNode->gn)) {
			break; // once a virtual node is encountered, there will be no more nonvirtual nodes
		}
		ret++;
	}
	return ret;
}

/**
 * adds a constraint node based on a virtual grapheme node based on the given spec to the constraintnet
 */
void cnAddVirtualNode(ConstraintNet cn, VirtualNodeSpec spec, ByteVector exceptions, Vector freezeCount, int n) {
	GraphemNode vgn = lgAddVirtualNode(cn->lexemgraph,spec,n);
	extendCNIsBuild(cn);
	if(incrFreezeLevel==FLTnone) {
		cnBuildIter(cn, vgn, TRUE);
	} else {
		cdgPrintf(CDG_ERROR,"ERROR: Freezing and virtual nodes are not compatible yet!\n  please disable either freezing or virtual nodes!");
		abort();
	}
//	} else if (incrFreezeLevel== FLTsimulate) {
//		cnBuildIterWithExceptions(cn,vgn, TRUE, exceptions, freezeCount);
//	} else if (incrFreezeLevel== FLTfull) {
//		cnBuildIterWithExceptions(cn,vgn, TRUE, exceptions, NULL);
//	}

	incrBuildLvsTowardsNewNode(cn,vgn,exceptions,freezeCount);
}

/**
 * adds all virtual node specifications as constraint nodes into the net
 *
 * exceptions says weather lvs should be added for each domain
 * freezecount counts lvs not being build because of exceptions
 */
void cnInitVirtualNodes(ConstraintNet cn, List specs, ByteVector exceptions, Vector freezeCount) {
	List l;
	VirtualNodeSpec spec;
	int i;

	if(!cn->lexemgraph->ve) {
		cn->lexemgraph->ve= veNew(cn->lexemgraph);
		cn->lexemgraph->ve->specs = listClone(specs);
	}

	for (l= specs; l; l= l->next) {
		spec= l->item;
		for(i=0; i< spec->copies; i++) {
			if(exceptions) {
				bvExtend(exceptions,inputCurrentGrammar->noOfLevels,FALSE);
			}
			if(freezeCount) {
				vectorInsertElement(freezeCount,(void *)0,vectorSize(freezeCount));
			}
			cnAddVirtualNode(cn,spec,exceptions,freezeCount,i+1);
		}
	}

	for(i=0; i<vectorSize(cn->nodes);i++) {
		ConstraintNode cnode= vectorElement(cn->nodes,i);
		incrUpdateNumberCaches(cnode);
	}
}

/**
 * checks all virtual node specifications and adds corresponding nodes, if needed.
 * prerequisite: net->lexemgraph->ve != NULL
 */
void veBlockVirtualNodes(VirtualLGExtension ve, Analysis a) {
	List l, m;
	GraphemNode gn;
	Boolean firstUnusedFound;

	// first make everything available
	bvSetAllElements(ve->blocked, FALSE);

	for(l= ve->specs; l; l=l->next) {
		VirtualNodeSpec spec = l->item;
		List ocs = veOccurences(ve, spec);
		firstUnusedFound = FALSE;

		// then block all unused instances of a vnSpec, except the first one
		for(m= ocs; m; m=m->next) {
			gn = ((VirtualNode)m->item)->gn;
			if(gnIsUnusedVirtualNode(gn, a)) {
				if(firstUnusedFound) {
					bvSetElement(ve->blocked, TRUE, gn->no);
					cdgPrintf(CDG_DEBUG, "DEBUG: blocked instance '%s' of spec %s\n", gn->arc->word, spec->name);
				} else {
					firstUnusedFound = TRUE;
				}
			}
		}
	}
}


/** Analysis, procedures related to virtual nodes ************************************* */


/**
 * search through a domain for (one of) the levelvalue(s),
 * where the regent is NIL and the label is empty
 */
LevelValue findUnusedLv(Problem P, int i) {
	List l;
	LevelValue lv;
	for(l = P->domains[i]; l; l=l->next) {
		 lv = l->item;
		 if(!lv->regents && !strcmp(lv->label,"")) {
			 return lv;
		 }
	}
	return NULL;
}

/*
 * returns whether the given labels are the same or at least similar
 * similarity is defined by the label hierarchy and the list of label groups
 */
Boolean isCompatibleLabel(String l1, String l2) {
	if(l1 == l2)
		return TRUE;

	//cdgPrintf(CDG_INFO, "INFO: compatible %s %s ?\n", l1, l2);

	Hierarchy h = findHierarchy("Label");
	if(!h) {
		cdgPrintf(CDG_WARNING, "WARNING: Type hierarchy 'Label' not found, can't check label compatibility!\n");
		return FALSE;
	}

	List l;
	String group;

	int p1 = (int)hashGet(h->types, l1) -1;
	if(p1 <= 0 ) {
		cdgPrintf(CDG_WARNING, "WARNING: Label %s not found in hierarchy 'label', can't check label compatibility\n", l1);
		return FALSE;
	}
	int p2 = (int)hashGet(h->types, l2) -1;
	if( p2 <= 0 ) {
		cdgPrintf(CDG_WARNING, "WARNING: Label %s not found in hierarchy 'label', can't check label compatibility\n", l2);
		return FALSE;
	}

	int pg;

	for(l= inputGetProperty(inputCurrentGrammar, LabelGroupsProperty); l; l= l->next) {
		group = l->item;
		pg = (int)hashGet(h->types, group);

		if(pg <= 0) {
			cdgPrintf(CDG_WARNING, "WARNING: Label group %s not found in hierarchy 'label', skipping...\n", group);
				continue;
		}

		if((int)arrayElement(h->distance, pg, p1) >=0 && (int)arrayElement(h->distance, pg, p2) >=0) {
			//cdgPrintf(CDG_INFO, "  Yes\n");
			return TRUE;
		}
	}

	//cdgPrintf(CDG_INFO, "  No\n");
	return FALSE;
}


/*
 * returns all levelValues with same regent and label as lv,
 * but with newNode as dependent
 */
List findSubstitutesDependent(Problem P, LevelValue lv, GraphemNode newNode, ByteVector availableLexemes) {
	List candidates= listNew();
	Boolean available;
	List l;

	List domain = P->domains[newNode->no * P->nrLevels + lv->level->no];
	for( ; domain; domain= domain->next) {
		LevelValue candidate = domain->item;

		if(lvGetRegentGN(lv) != lvGetRegentGN(candidate)) {
			continue;
		}

		if(lv->level->no == 0 && virtFlexibleLabelMatching) {
			if(!isCompatibleLabel(lv->label, candidate->label))
				continue;
		} else {
			if(lv->label != candidate->label)
				continue;
		}

		available = FALSE;
		for(l= candidate->dependents; l; l=l->next ) {
			if(bvElement(availableLexemes, ((LexemNode)l->item)->no) ) {
				available = TRUE;
				break;
			}
		}
		if (!available) {
			// none of the dependent lexemes is available
			continue;
		}

		if(lvGetRegentGN(candidate)) { // only check regent lexemes if the regent is specified
			available = FALSE;
			for(l= candidate->regents; l; l=l->next ) {
				if(bvElement(availableLexemes, ((LexemNode)l->item)->no) ) {
					available = TRUE;
					break;
				}
			}
			if (!available) {
				// none of the regent lexemes is available
				continue;
			}
		}

		candidates= listAppendElement(candidates,candidate);
	}
	return candidates;
}


/*
 * returns all levelvalues with same dependent and label as lv,
 * but with newNode as regent
 */
List findSubstitutesRegent(Problem P, LevelValue lv, GraphemNode newNode, ByteVector availableLexemes) {
	List candidates= listNew();
	Boolean available;
	List l;
	List domain = P->domains[lvGetDependentGN(lv)->no * P->nrLevels + lv->level->no];

	for( ; domain; domain= domain->next) {
		LevelValue candidate = domain->item;

		if(newNode != lvGetRegentGN(candidate)) {
			continue;
		}

		if(lv->level->no == 0 && virtFlexibleLabelMatching) {
			if(!isCompatibleLabel(lv->label, candidate->label))
				continue;
		} else {
			if(lv->label != candidate->label)
				continue;
		}

		available = FALSE;
		for(l= candidate->dependents; l; l=l->next ) {
			if(bvElement(availableLexemes, ((LexemNode)l->item)->no) ) {
				available = TRUE;
				break;
			}
		}
		if (!available) {
			// none of the regent lexemes is available
			continue;
		}

		if(lvGetRegentGN(candidate)) { // only check regent lexemes if the regent is specified
			available = FALSE;
			for(l= candidate->regents; l; l=l->next ) {
				if(bvElement(availableLexemes, ((LexemNode)l->item)->no) ) {
					available = TRUE;
					break;
				}
			}
			if (!available) {
				// none of the regent lexemes is available
				continue;
			}
		}

		candidates= listAppendElement(candidates,candidate);
	}
	return candidates;
}

/**
 * part of aCollapseNodes()
 * subroutine to find all Analyses where a certain edge from the analysis is replaced.
 * If the node to be replaced is neither regent nor dependent of the edge,
 * replace nothing and return a list containing just the old analysis
 */
List aCollapseNodesIter(Analysis a, GraphemNode newNode, GraphemNode virtNode,
		LevelValue lv, int slot, Number threshold) {
	List result= listNew();

	GraphemNode dep = lvGetDependentGN(lv);
	GraphemNode reg = lvGetRegentGN(lv);
	GraphemNode dep2, reg2;
	int i;
	List l;
	LexemNode ln;
	ByteVector available;

	if(dep == virtNode) {
		// first get the available lexemes
		available = bvNew(a->P->nrLexemes);
		bvSetAllElements(available, TRUE);
//		for(i=0; i< vectorSize(a->slots); i++) {
//
//			if (i == slot) {
//				continue;
//			}
		for(i=0; i< slot; i++) {

			LevelValue lv2 = vectorElement(a->slots, i);
			// only the regent graphemNode of lv and the newNode are relevant currently
			// are the ends of lv2 one of them?
			dep2 = lvGetDependentGN(lv2);
			reg2 = lvGetRegentGN(lv2);

			if(dep2 == newNode) {//|| (reg && dep2 == reg)) {
				for(l=dep2->lexemes; l; l=l->next) {
					ln = l->item;
					if(! listContains(lv2->dependents, ln)) {
						bvSetElement(available, FALSE, ln->no);
					}
				}
			}

			//if( (reg && reg2 == reg) || reg2 == newNode ) {
			if( reg2 == newNode ) {
				for(l=reg2->lexemes; l; l=l->next) {
					ln = l->item;
					if(! listContains(lv2->regents, ln)) {
						bvSetElement(available, FALSE, ln->no);
					}
				}
			}
		}

		List candidates= findSubstitutesDependent(a->P,lv,newNode,available);
		for(;candidates;candidates=candidates->next) {
			LevelValue substitute = candidates->item;
			if(threshold > substitute->limit) {
				continue; // levelvalue is too bad, abort here
			}
			Analysis a2= aClone(a);
			//int pos = newNode->no * inputCurrentGrammar->noOfLevels + lv->level->no;
			//vectorSetElement(a2->slots,substitute,pos);
			aInsert(a2,substitute);

			// the replaced node needs a new edge
			LevelValue unused= findUnusedLv(a->P,slot);
			//vectorSetElement(a2->slots,unused,slot);
			aInsert(a2,unused);

			result = listPrependElement(result,a2);
		}

	} else if(reg == virtNode) {
		// first get the available lexemes
		available = bvNew(a->P->nrLexemes);
		bvSetAllElements(available, TRUE);
		for(i=0; i< slot; i++) {

			LevelValue lv2 = vectorElement(a->slots, i);
			// only the dependent graphemNode of lv and the newNode are relevant currently
			// are the ends of lv2 one of them?
			dep2 = lvGetDependentGN(lv2);
			reg2 = lvGetRegentGN(lv2);

			//if(dep2 == dep || dep2 == newNode){
			if(dep2 == newNode){
				for(l=dep2->lexemes; l; l=l->next) {
					ln = l->item;
					if(! listContains(lv2->dependents, ln)) {
						bvSetElement(available, FALSE, ln->no);
					}
				}
			}

			//if(reg2 == dep || reg2 == newNode) {
			if( reg2 == newNode) {
				for(l=reg2->lexemes; l; l=l->next) {
					ln = l->item;
					if(! listContains(lv2->regents, ln)) {
						bvSetElement(available, FALSE, ln->no);
					}
				}
			}
		}

		List candidates= findSubstitutesRegent(a->P,lv,newNode, available);
		for(;candidates;candidates=candidates->next) {
			LevelValue substitute = candidates->item;

			if(threshold > substitute->limit) {
				continue; // levelvalue is too bad, abort here
			}
			Analysis a2= aClone(a);
			//vectorSetElement(a2->slots,substitute,slot);
			// todo is the lv compatible with current lexem choices?
			aInsert(a2,substitute);

			result = listPrependElement(result,a2);
		}
	} else {
		result = listPrependElement(result,a);
	}
	return result;
}

/**
 * returns the best of the Analyses where the virtual node is replaced by the new node.
 * All edges where the virtual node is regent or dependent have to be replaced.
 * There might be more than one way to replace them due to lexem ambiguity,
 * but only the best one is returned.
 * The virtual node itself is changed to unused status (under NIL with empty label)
 */
Analysis aCollapseNodes(Analysis a, GraphemNode newNode, GraphemNode virtNode, Number threshold) {

	if(		gnIsUnusedVirtualNode(virtNode, a) ||
			!gnHaveCompatibleLexemes(newNode, virtNode)) {
		return FALSE;
	}

	cdgPrintf(CDG_INFO, "Replacing %s\n", virtNode->arc->word);

	List current = listNew();
	List previous = listPrependElement(listNew(),a);

	Vector slots= a->slots;
	int i,j, slot;
	List l;

	// for all slots (first levels first, as the main level probably prunes the most paths)
	for(i= 0; i < a->P->nrLevels; i++) {
		for(j=0; j< a->P->nrTimepoints; j++) {
			slot = j*a->P->nrLevels+i;
			LevelValue lv = vectorElement(slots, slot);

			if(!lv || lv->isDeleted) {
				continue;
			}

			if(lvGetDependentGN(lv) == newNode) {
				// if the regent of the new Node is by chance the virtual node,
				// the new node would point to itself when replacing this lv
				// so skip it, the new node will get it's regent soon enough
				continue;
			}

			current= listNew();

			for(l= previous; l; l=l->next) {
				Analysis curA = l->item;
				List tmp = aCollapseNodesIter(curA, newNode, virtNode, lv, slot, threshold);
				current  = listAppendList(current, tmp);
			}

			previous = current;

			cdgPrintf(CDG_INFO, "Number of alternatives after %s (%s): %d \n",
					lvDependent(lv), lv->level->id, listSize(current));

			if(!current) {
				// not replaceable
				return NULL;
			}
		}
	}

	Analysis result= NULL;
	for(l= current; l; l=l->next) {
		Analysis curA = l->item;
		aJudge(curA);
		if(!result || bCompare(curA->b,result->b)) {
			if(result) {
				result = NULL;
			}
			result = curA;
		} else {
			curA = NULL;
		}
	}

	return result;
}

/**
 * Returns the best Analysis from all possible analyses created from the given one
 * by inserting the new node for a virtual one.
 * If none can be found, e.g. because all virtual nodes are unused,
 * or if the best one is lower than the threshold,
 * just return the old analysis
 */
Analysis aBestMatchingReplacement(Analysis a, GraphemNode newNode, Number threshold) {
	Number bestMatch = 0.0;
	Analysis result = NULL;
	Parse p;

	Vector vns = a->P->lg->ve->virtualNodes;
	int i;
	for(i= 0; i< vectorSize(vns); i++) {
		VirtualNode vn = vectorElement(vns,i);
		GraphemNode vgn = vn->gn;
		Analysis current = aCollapseNodes(a, newNode, vgn, threshold);
		if(current) {
			if(!current->b->hard && current->b->soft > bestMatch) {
				result = current;
				bestMatch= result->b->soft;
			}
		}
	}
	if(bestMatch > threshold) {
		if ( alwaysShowAllSolutions) {
			/* hand the Parse to Tcl for displaying */
			p = analysis2parse(result);
			cdgExecHook(HOOK_PARTIALRESULT, p);
			parseRegister(p);
		}
		cdgPrintf(CDG_INFO,"virtual node matching: found matching node!\n");
		return result;
	} else if(bestMatch > 0.0) {
		cdgPrintf(CDG_INFO,"virtual node matching:\n");
		cdgPrintf(CDG_INFO," best match < threshold\n");
		cdgPrintf(CDG_INFO," %1.2f < %1.2f\n", bestMatch, threshold);
		return a;
	} else {
		cdgPrintf(CDG_INFO,"no matching virtual node found\n");
		return a;
	}

}

