/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: blah@nats.informatik.uni-hamburg.de


*/

/* ----------------------------------------------------------------------
 * @defgroup Hashtable Hashtables
 * Implementation of hashtables.
 *
 * A hashtable stores an arbitrary number of objects that are accessed
 * using an arbitrary key. The key is converted to an integer value, so
 * so-called hash value, by the hash function. A hash function should be
 * fast and should map the keys to integers in a highly irregular but
 * consistent way, i.~e., the keys should be distributed evenly over the
 * whole set of integers. Ideally, objects can then be accessed in
 * constant time based on their key.
 * @{
 */

/* ----INCLUDES------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include "blah.h"

#define PRIME_TESTS 10

/* ----------------------------------------------------------------------
 * internal representation of the hash table entry.
 */
struct HashtableEntryStruct {
  Pointer key;                       /**< hash key */
  Pointer value;                     /**< value */
  struct HashtableEntryStruct *next; /**< pointer to next entry */
};

/* ----------------------------------------------------------------------
 * internal representation of the hash table.
 */
struct HashtableStruct {
  int count;                       /**< number of entries in hashtable */
  int threshold;                   /**< limit when table is rehashed */
  int capacity;                    /**< capacity of table */
  double loadFactor;               /**< ratio when to rehash */
  IntFunction *keyEqualFunction;   /**< equality function for key */
  IntFunction *hashFunction;       /**< hash function */
  HashtableEntry *entries;         /**< table of entries */
};

/* ----------------------------------------------------------------------
 * internal representation of the hash iterator.
 */
struct HashIteratorStruct {
  Hashtable hashtable;             /**< Where do we belong to? */
  int index;                       /**< index of entry */
  HashtableEntry entry;            /**< determines _next_ HashtableEntry */
};

/* ----------------------------------------------------------------------
 * rehashes the hashtable.
 *
 * doubles the capacity and this enlarges the space.
 *
 * @param ht the hash table that is to be resized
 * @returns the new resized hashtable.
 */
void rehashHashtable(Hashtable ht)
{
  HashtableEntry *newEntries;
  HashtableEntry entry, old;
  int i, index;
  int newCapacity = primeNext(ht->capacity * 2, PRIME_TESTS);

#ifdef DEBUG
  printf("DEBUG: rehashHashtable %d --> %d...", ht->capacity, newCapacity);
#endif
  newEntries =
    (HashtableEntry *)memMalloc(sizeof(HashtableEntryStruct) * newCapacity);
  for (i = 0; i < newCapacity; i++)
    newEntries[i] = (HashtableEntry)NULL;
  for (i = 0; i < ht->capacity; i++)
    for (old = ht->entries[i]; old != NULL; ) {
      entry = old;
      old = old->next;

      index = (((*ht->hashFunction)(entry->key)) & 0x7fffffff) % newCapacity;

      entry->next = newEntries[index];
      newEntries[index] = entry;
    }

  ht->threshold = (int)(newCapacity * ht->loadFactor);
  ht->capacity = newCapacity;
  memFree(ht->entries);
  ht->entries = newEntries;

#ifdef DEBUG
  printf("done\n");
#endif
}

/* ----------------------------------------------------------------------
 * creates a new hashtable with an initial capacity of c.
 *
 * whenever the number of stored objects exceeds \a loadFactor times the current
 * capacity, the hashtable is automatically resized.
 *
 * @param capacity specifies the capacity of the new hashtable to be created.
 * @param loadFactor it is a number that should be chosen between 0.5 and 0.9.
 * @param hashFunction it is that function which is called with a key as the only parameter and
 * returns the hash value of  * that key.
 * @param keyEqualFunction It is that function which is called with two keys as
 * parameters.It returns 1 if the keys are to be considered equal and 0 otherwise.
 * @returns the new hashtable with a capacity of c.

 */
Hashtable hashNew(int capacity,
		   double loadFactor,
		   IntFunction *hashFunction,
		   IntFunction *keyEqualFunction)
{
  Hashtable ht;
  int i;

  ht = (Hashtable)memMalloc(sizeof(HashtableStruct));
  ht->count = 0;
  if(capacity < 3) {
    capacity = 3;
  }
  ht->capacity = primeNext(capacity, PRIME_TESTS);
  ht->threshold = (int)(ht->capacity * loadFactor);
  ht->loadFactor = loadFactor;
  ht->keyEqualFunction = keyEqualFunction;
  ht->hashFunction = hashFunction;
  ht->entries =
    (HashtableEntry *)memMalloc(sizeof(HashtableEntryStruct)*ht->capacity);
  for (i = 0; i < ht->capacity; i++)
    ht->entries[i] = NULL;

  return(ht);
}

/* ----------------------------------------------------------------------
 * adds the object value with the key \a key in the hashtable.
 *
 * rehashes the hashtable if necessary.
 *
 * @param ht the hashtable in which the object value has to be added with the key.
 * @param key the key value that has to be added with the \a value
 * @param value the object whose value has to be added to the key.
 * @returns the old object that was stored at that key or \a NULL
 */
Pointer hashSet(Hashtable ht,
                Pointer key,
		Pointer value)
{
  HashtableEntry entry;
  int index;
  Pointer p;

  if (ht == NULL || key == NULL || value == NULL) {
    fprintf(stderr, "ERROR: hashSet: arguments must not be NULL\n");
    abort();
  }

  index = (((*ht->hashFunction)(key)) & 0x7fffffff) % ht->capacity;
  for (entry = ht->entries[index]; entry != NULL; entry = entry->next) {
    if ((*ht->keyEqualFunction)(entry->key, key)) {
      p = entry->value;
      entry->key = key;
      entry->value = value;
      return(p);
    }
  }

  if (ht->count >= ht->threshold) {
    /* The number of entries exceeds the thresholds. Rehash the table.
     * TODO:
     * - touching fresh buckets is counted and punished by rehashing
     * - rehashing should depend on collisions not on tabelsize
     */
    rehashHashtable(ht);
    return(hashSet(ht, key, value));
  }

  entry = (HashtableEntry)memMalloc(sizeof(HashtableEntryStruct));
  entry->key = key;
  entry->value = value;
  entry->next = ht->entries[index];
  ht->entries[index] = entry;
  ht->count++;
  return(NULL);
}


/* ----------------------------------------------------------------------
 * retrieves value associated with the key.
 *
 * @param ht the hashtable from which the value at the specified key has to be retrieved.
 * @param key the key at which the value of the hashtable has to be retrieved.
 * @returns the object that(or \a NULL) that is stored under the key in the hashtable.
 */
Pointer hashGet(Hashtable ht,
                Pointer key)
{
  HashtableEntry entry;
  int index;

  if (ht == NULL || key == NULL) {
    fprintf(stderr, "ERROR: hashGet: arguments must not be NULL\n");
    abort();
  }

  index = (((*ht->hashFunction)(key)) & 0x7fffffff) % ht->capacity;
  for (entry = ht->entries[index]; entry != NULL; entry = entry->next) {
    if ((*ht->keyEqualFunction)(entry->key, key)) {
      return(entry->value);
    }
  }

  return(NULL);
}

/* ----------------------------------------------------------------------
 * retrieves value associated with the key
 *
 * @param ht the hashtable from which the value at the specified key has to be retrieved.
 * @param key the key at which the value of the hashtable has to be retrieved.
 * @returns the object that(or \a NULL) that is stored under the key in the hashtable.
 */
Pointer *hashGetPointerToValue(Hashtable ht, Pointer key)
{
  HashtableEntry entry;
  int index;

  if (ht == NULL || key == NULL) {
    fprintf(stderr, "ERROR: hashGet: arguments must not be NULL\n");
    abort();
  }

  index = (((*ht->hashFunction)(key)) & 0x7fffffff) % ht->capacity;
  for (entry = ht->entries[index]; entry != NULL; entry = entry->next) {
    if ((*ht->keyEqualFunction)(entry->key, key)) {
      return(&entry->value);
    }
  }

  return(NULL);
}

/* ----------------------------------------------------------------------
 * removes key/value pair from hashtable
 *
 * @param ht the hashtable from which the key/value pair is to be removed.
 * @param key this shows the value that has to be deleted in the hashtable ht.
 * @returns the object (or \a NULL) that is stored under the key in the hashtable ht.
 */
Pointer hashRemove(Hashtable ht, Pointer key)
{
  HashtableEntry entry, prev;
  int index;
  Pointer value;

  if (ht == NULL || key == NULL) {
    fprintf(stderr, "ERROR: hashRemove: arguments must not be NULL\n");
    abort();
  }

  index = (((*ht->hashFunction)(key)) & 0x7fffffff) % ht->capacity;
  for (entry = ht->entries[index], prev = NULL;
	entry != NULL;
	prev = entry, entry = entry->next) {
    if ((*ht->keyEqualFunction)(entry->key, key)) {
      if (prev != NULL) {
	prev->next = entry->next;
      } else {
	ht->entries[index] = entry->next;
      }
      ht->count--;
      value = entry->value;
      memFree(entry);
      return (value);
    }
  }
  return(NULL);
}

/* ----------------------------------------------------------------------
 * retrieves the size of the hashtable
 *
 * @param ht the hashtable whose size has to be retrieved.
 * @returns the number of objects that are currently stored in the hashtable.
 */
int hashSize(Hashtable ht)
{
  if (ht == NULL) {
    fprintf(stderr, "ERROR: hashSize: argument must not be NULL\n");
    abort();
  }

  return(ht->count);
}

/* ----------------------------------------------------------------------
 * checks if the hashtable is empty.
 *
 * @param ht the hash table whose emptiness is checked.
 * @returns \a TRUE if the hashtable is empty and \a False otherwise.
 */
Boolean hashIsEmpty(Hashtable ht)
{
  if (ht == NULL) {
    fprintf(stderr, "ERROR: hashIsEmpty: argument must not be NULL\n");
    abort();
  }

  return(ht->count == 0 ? TRUE : FALSE);
}

/* ----------------------------------------------------------------------
 * checks if the hashtable contains the specified key.
 *
 * @param ht the hashtable in which the specified key has to be checked.
 * @param key the key whose existance in the hashtable is to be checked.
 * @returns \a TRUE if the an object under the key exists and \a FALSE  otherwise.
 */
Boolean hashContainsKey(Hashtable ht, Pointer key)
{
  HashtableEntry entry;
  int index;

  if (ht == NULL || key == NULL) {
    fprintf(stderr, "ERROR: hashConstainsKey: arguments must not be NULL\n");
    abort();
  }

  index = (((*ht->hashFunction)(key)) & 0x7fffffff) % ht->capacity;
  for (entry = ht->entries[index]; entry != NULL; entry = entry->next)
    if ((*ht->keyEqualFunction)(entry->key, key))
      return(TRUE);

  return(FALSE);
}

/* ----------------------------------------------------------------------
 * checks if the hashtable contains a specific value.
 *
 * Objects are compared by the standard C operator   == and this method is Expensive!!!
 *
 * @param ht the hashtable in which the specific value has to be checked.
 * @param value the value whose existance in the hashtable has to be checked.
 * @returns \a TRUE if the object is contained in the hashtable and \a FALSE otherwise.
 */
Boolean hashContainsValue(Hashtable ht, Pointer value)
{
  HashtableEntry entry;
  int i;

  if (ht == NULL || value == NULL) {
    fprintf(stderr, "ERROR: hashConstainsValue: arguments must not be NULL\n");
    abort();
  }

  for (i = 0; i < ht->count; i++)
    for (entry = ht->entries[i]; entry != NULL; entry = entry->next)
      if (value == entry->value)
	return(TRUE);

  return(FALSE);
}

/* ----------------------------------------------------------------------
 * calls the function `f(key,value)' for each item in the hashtable.
 *
 * @param ht the hashtable in which the function 'f' has to be called.
 * @param f the function that has to be called for every object in the hashtable ht.
 */
void hashForEach(Hashtable ht, VoidFunction *f)
{
   HashtableEntry entry;
   int i;

   for (i = 0; i < ht->capacity; i++) {
     for (entry = ht->entries[i]; entry != NULL; entry = entry->next) {
       (*f)(entry->key, entry->value);
     }
   }
}
/* ----------------------------------------------------------------------
 * calls the function `f(element,data)' for each object in the hashtable.
 *
 * @param ht the hashtable in which the function 'f' has to be called.
 * @param f the function that has to be called in the hashtable ht.
 * @param clientData the data in the function 'f(element,data)'that is called in the hashtable ht.
 */
void hashForEachWithData(Hashtable ht, VoidFunction *f, Pointer clientData)
{
   HashtableEntry entry;
   int i;

   for (i = 0; i < ht->capacity; i++) {
     for (entry = ht->entries[i]; entry != NULL; entry = entry->next) {
       (*f)(entry->key, entry->value, clientData);
     }
   }
}

/* ----------------------------------------------------------------------
 * calls the function `f(key, value)' for each item in the Hashtable.
 *
 * deletes the hashtable and hashtable becomes inaccessible
 *
 * @param ht the hashtable in which the function 'f' has to be called.
 * @param f the function that has to be called for every object in the hashtable ht.
 */
void hashForEachFree(Hashtable ht, VoidFunction *f)
{
   HashtableEntry entry, next;
   int i;

   for (i = 0; i < ht->capacity; i++) {
      for (entry = ht->entries[i]; entry != NULL; entry = next) {
	 next = entry->next;
	 (*f)(entry->key, entry->value);
	 memFree(entry);
      }
   }
   memFree(ht->entries);
   memFree(ht);
}

/* ----------------------------------------------------------------------
   List hashForEachFree(), but frees only the embedded value. F is applied
   to the value only and must be a unariy void function.
*/
void hashForEachFreeValue(Hashtable ht, VoidFunction *f)
{
   HashtableEntry entry, next;
   int i;

   for (i = 0; i < ht->capacity; i++) {
      for (entry = ht->entries[i]; entry != NULL; entry = next) {
	 next = entry->next;
	 (*f)(entry->value);
	 memFree(entry);
      }
   }
   memFree(ht->entries);
   memFree(ht);
}


/* ----------------------------------------------------------------------
 * retrieves a list of keys of all objects in the hashtable.
 *
 * @param ht the hashtable from which the list of keys have to be retrieved.
 * @returns the list of the keys of all objects in the hashtable.
 */
List hashListOfKeys(Hashtable ht)
{
   int i;
   List l = NULL;
   HashtableEntry entry;

   for (i = 0; i < ht->capacity; i++) {
      for (entry = ht->entries[i]; entry != NULL; entry = entry->next) {
	l = listPrependElement(l, entry->key);
      }
   }

   return(l);
}

/* ----------------------------------------------------------------------
 * returns a new hash iterator object.
 *
 * Iterators allow to loop through all the elements of a container.
 * However the behaviour is undefined if the container changes while the iterator is still looping.
 *
 * \code hi = hashIteratorNew(ht);
 *  while (NULL != (key = hashIteratorNextKey(hi))) {
 *    do something with key;
 *  }
 *  hashIteratorDelete(hi); \endcode
 *
 * @param ht the hashtable for which the new iterator object has to be returned.
 * @returns the new hash iterator object for the hashtable ht.
 */
HashIterator hashIteratorNew(Hashtable ht)
{
  HashIterator hi;

  /* Some checks would be nice but I don't have the time right now. */
  hi = (HashIterator)memMalloc(sizeof(HashIteratorStruct));
  hi->hashtable = ht;
  hi->index = -1;
  hi->entry = NULL;
  while (!hi->entry && ++hi->index < ht->capacity) {
    hi->entry=ht->entries[hi->index];
  }

  return(hi);
}

/* ----------------------------------------------------------------------
 * returns the next key of a hash-iterator.
 *
 * Hash iterator points to the following entry afterwards
 *
 * @param hi the hash iterator whose next key has to be returned.
 * @returns the next key of the hash iterator hi
 */
Pointer hashIteratorNextKey(HashIterator hi)
{
  HashtableEntry current=hi->entry;
  Hashtable ht=hi->hashtable;

  if (!current) {
    return(NULL);
  }

  hi->entry = hi->entry->next;
  while (!hi->entry && ++hi->index < ht->capacity) {
    hi->entry=ht->entries[hi->index];
  }

  return(current->key);
}

/* ----------------------------------------------------------------------
 * returns the next value of a hash-iterator.
 *
 * Hash iterator points to the following entry afterwards.
 *
 * @param hi the hash iterator whose next value has to be returned.
 * @returns the next value of a hash-iterator hi.
 */
Pointer hashIteratorNextValue(HashIterator hi)
{
  HashtableEntry current;
  Hashtable ht = hi->hashtable;

  if (NULL == (current = hi->entry)) {
    return(NULL);
  }

  if (NULL == (hi->entry = hi->entry->next)) {
    while (++hi->index < ht->capacity) {
      /*   for (hi->index++; hi->index < ht->capacity; hi->index++) { */
      if (NULL != (hi->entry = ht->entries[hi->index])) {
	return(current->value);
      }
    }
  }

  return(current->value);
}

/* ----------------------------------------------------------------------
 * is an example has function for C strings that can be used in hashNew.
 *
 * A bit rotating function by Knuth is used here.
 * TODO:
 * - try different hash functions
 * - strlen should go out - supply length of key.
 *
 * @param s the string on which the hashStringHashFunction is performed.
 * @returns the integer representation of the string.
 */
int hashStringHashFunction(char *s)
{
   register int len = strlen(s);
   register int value;

#if 1
   /* rotating ala knuth */
   value = len;
   while(len)
      value = (value << 5) ^ (value >> 27) ^ s[--len];

#else
   value = 1;
   while(len)
      value *= s[--len];
      /* value = (value >> 11) + s[--len] */
#endif


   return (value);
}

/* ----------------------------------------------------------------------
 * is an example equality function for C strings that can be used in hashNew.
 *
 * @param s the first string that is used in the string comparison function.
 * @param t the second string that is used in the string comparison function.
 * @returns 0 if they are equal and 1 if not.
 */
int hashStringEqualFunction(char *s, char *t)
{
  return(strcmp(s, t) == 0);
}

/* ---------------------------------------------------------------------- */
/** @} */
