/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
    Author: Ingo Schroeder
    Birth: 23-Apr-97
 
 */

/* ----------------------------------------------------------------------
 * @defgroup ListAgenda ListAgenda
 *
 * This module is an implementation of the agenda interface using a plain list
 * as its the storage medium.
 *
 * This is used by the module netsearch,
 * but has been superseded by the more efficienct module treeagenda.
 * @{
 */

/* -- INCLUDES ------------------------------------------------------- */
#include <stdio.h>
#include <blah.h>

/* -- TYPE DEFINITIONS ------------------------------------------------*/

/* ----------------------------------------------------------------------
 * this type represents an entry of an agenda.
 *
 * Only elements of the same type should be inserted into the same agenda,
 * since there can be only one function (freestate())for deallocating
 * the elements.
 */

struct ListAgendaEntryStruct {
  double score;                   /**< contains the rating of an agenda element.
                                       Large scores are sorted to appear before
				       small scores.
				       */
  Pointer state;                 /**< holds a pointer to the actual element.
                                      Since its type is unknown, the agenda
				      never uses this value except to return it.
				      */
};

/* ----------------------------------------------------------------------
 * quick, should be binary tree.
 */
struct ListAgendaStruct {
  int maxsize;                    /**< Gives the maximal allowed size of the agenda.
                                       If more elements are inserted to it,the agenda
				       will overflow,and the elements with the worst
				       scores will be lost.
				       */
  Boolean verbose;		  /**< If set to true then truncation warnings take effect */
  Boolean truncationWarning;      /**< This is set after an overflow messsage is printed.
                                       It inhibits any further warning messages.
				       */
  int size;                       /**< gives the current size of the agenda. */
  int maxSizeSoFar;               /**< gives the maximal value that the size has reached
                                       so far.
				       */
  double worstScore;              /**< holds the score of the last item of the agenda.
                                       This field helps us to quickly determine whether
				       an insert operation will make the agenda overflow.
				       */
  VoidFunction *freeState;        /**< used on an element when when it is lost through
                                       overflow or when the entire agenda is freed.
				       */
  List entries;                   /**< holds an item in an agenda. The agenda will allocate
                                       and deallocate its own items.The elements,however,
				       will only be deallocated in special cases.
				       */
};

/**< Instantiates the generic agenda iterator.
 *
 * Since the ListAgenda is already sorted by score,
 * iterating does not require any further data structures.
 * The iterator only allocates one List cell, * which is
 * set to the point to whatever item is next.The list
 * entries are unaffected by this
 */

/*--FUNCTIONS----------------------------------------------------------*/

/* ----------------------------------------------------------------------
 * creates a new ListAgenda.
 *
 * @param maxsize the maximum capacity of the new ListAgenda that has to be created.
 * @param f the function that used for deallocating an element.
 * @returns a pointer to a new ListAgenda that can hold upto \a maxsize entries.
 */
ListAgenda laNew( int maxsize, VoidFunction f )
{
  ListAgenda a = (ListAgenda)memMalloc( sizeof(ListAgendaStruct) );
  a->maxsize = maxsize;
  a->verbose = FALSE;
  a->truncationWarning = FALSE;
  a->maxSizeSoFar = 0;
  a->size = 0;
  a->worstScore = 0.0;
  a->freeState = f;
  a->entries = NULL;
  return( a );
}

/* ----------------------------------------------------------------------
 * sets verbosity flag.
 * \param a the ListAgenda whose verbosity property is set
 * \param b a Boolean which we set the verbosity to
 * \returns the old verbosity state
 */
Boolean laSetVerbosity(ListAgenda a, Boolean b)
{
  Boolean old = a->verbose;

  a->verbose = b;
  return (old);
}

/* ----------------------------------------------------------------------
 * gets verbosity flag.
 * \param a the ListAgenda whose verbosity property is set
 * \returns the verbosity state
 */
Boolean laVerbosity(ListAgenda a)
{
  return a->verbose;
}

/* ----------------------------------------------------------------------
 * retrieves size of the agenda.
 *
 * @param a the ListAgenda for which the size has to be retrieved.
 * @returns the current number of entries in the ListAgenda \a a.
 */
int laSize(ListAgenda a)
{
  return (a->size);
}

/* ----------------------------------------------------------------------
 * retrieves the size limit of the agenda.
 *
 * @param a the ListAgenda for which the maximum size has to be determined.
 * @returns the max number of entries that can be held by the ListAgenda a.
 */
int laMaxSize( ListAgenda a )
{
  return(a->maxsize);
}

/* ----------------------------------------------------------------------
 * retrieves the largest attained size of the agenda.
 *
 * @param a the ListAgenda for which the largest size has to be retrieved.
 * @returns the maximum size occupied by the agenda \a a so far.
 */
int laMaxSizeSoFar(ListAgenda a)
{
  return (a->maxSizeSoFar);
}

/* ----------------------------------------------------------------------
 * checks if the specified ListAgenda is empty.
 *
 * @param a the ListAgenda for which emptiness has to be checked.
 * @returns \a TRUE if agenda \a a is empty and \a FALSE otherwise.
 */
Boolean laIsEmpty(ListAgenda a)
{
  return (a->size == 0);
}

/* ----------------------------------------------------------------------
 * checks for the truncation of the ListAgenda \a a.
 *
 * @param a the ListAgenda on which the function \a laIsTruncated is performed.
 * @returns \a TRUE if the agenda is already truncated and \a FALSE otherwise.
 */
Boolean laIsTruncated( ListAgenda a )
{
  return( a->truncationWarning );
}

/* -------------------------------------------------------------------------
 * reset the agenda truncation warning.
 *
 * @param a the ListAgenda whose truncationWarning will be set to FALSE.
 * @returns the old value of the truncationWarning
 */
Boolean laResetTruncated(ListAgenda a)
{
  Boolean oldTruncationWarning = a->truncationWarning;
  a->truncationWarning = FALSE;
  return oldTruncationWarning;
}

/* ----------------------------------------------------------------------
 * inserts a new entry into the agenda.
 *
 * @param a the ListAgenda into which a new element has to be inserted.
 * @param score the \a state is sorted into the list according to the \a score.
 * @param state the item that is inserted into the ListAgenda \a a.
 * @returns \a FALSE if the agenda has been truncated and \a TRUE if we were able
 *        to insert the item without any unpleasant side effects. This is going to
 *        be reported only once.
 *
 * \n If the insertion leads to an overflow, then the worst item from the list is removed.
 * Note that the item to be removed may be the \a state itself.
 */
Boolean laInsert(ListAgenda a, double score, Pointer state)
{
  ListAgendaEntry ae;
  List l, k;
  int pos;

  /* new entry is worse than the worst entry in the agenda */
  if ( a->size >= a->maxsize && a->worstScore > score ) {
    (*a->freeState)( state );
    if ( !a->truncationWarning ) {
      if (a->verbose) {
	//printf("WARNING: truncating agenda, result may be incorrect\n");
      }
      a->truncationWarning = TRUE;
      return FALSE;
    }
    return TRUE;
  }

  /* definitely put it on the agenda */
  ae = (ListAgendaEntry)memMalloc( sizeof(ListAgendaEntryStruct) );
  ae->score = score;
  ae->state = state;

  pos = 1;
  if ( a->entries == NULL || ((ListAgendaEntry)(a->entries->item))->score <= score ) {
    a->entries = listPrependElement( a->entries, ae );
    k = a->entries;
  } else {
    /* find position where to insert item */
    l = a->entries;
    while ( l->next != NULL && ((ListAgendaEntry)(l->next->item))->score > score ) {
      l = l->next;
      pos++;
    }
    /* l is last element that is better than ae */
    k = listAppendElement(NULL, ae);
    k->next = l->next;
    l->next = k;
  }

  /* k is list item that holds ae, pos is position of k */
  a->size++;
  a->worstScore = k->next == NULL ? score : a->worstScore;

  if ( a->size > a->maxsize ) {
    while ( pos < a->maxsize && k->next != NULL ) {
      k = k->next;
      pos++;
    }
    /* delete everything from k->next on */
    for ( l = k->next; l != NULL; l = l->next ) {

      (*a->freeState)(((ListAgendaEntry)(l->item))->state);

      memFree(l->item);
    }
    k->next = NULL;
    a->worstScore = ((ListAgendaEntry)(k->item))->score;
    a->size = pos;
    if ( !a->truncationWarning ) {
      //printf("WARNING: truncating agenda, result may be incorrect\n");
      a->truncationWarning = TRUE;
      a->maxSizeSoFar = max( a->maxSizeSoFar, a->size );
      return FALSE;
    }
  }

  a->maxSizeSoFar = max( a->maxSizeSoFar, a->size );

  return TRUE;
}

/* ----------------------------------------------------------------------
 * returns the best entry (= first) from the agenda.
 *
 * @param a the ListAgenda from which the best item has to be retrieved.
 * @returns the best item in \a a (or NULL if a is empty).The item remains in the agenda.
 */
Pointer laBest(ListAgenda a)
{
  if (a->entries == NULL) {
    return (NULL);
  } else {
    return a->entries->item;
  }
}

/* ----------------------------------------------------------------------
 * removes and returns best entry (= first) from the agenda.
 *
 * @param a the ListAgenda from which the best item has to be retrieved.
 * @returns the best item in \a a.The corresponding entry is removed and deallocated.
 * \n It must not be called on an empty agenda.
 */
Pointer laRemoveBest( ListAgenda a )
{
  Pointer p;
  List l;
  ListAgendaEntry ae;

  l = a->entries;
  a->entries = a->entries->next;
  ae = l->item;
  p = ae->state;

  memFree( ae );
  memFree( l );
  a->size--;

  return( p );
}

/* ----------------------------------------------------------------------
 * deletes the specified agenda.
 *
 * The function deallocates all the items in \a a using *freeState().
 * Then it deallocates all the entries and the agenda itself.
 *
 * @param a the ListAgenda that has to be deleted.
 */
void laDelete( ListAgenda a )
{
  List l;

  l = a->entries;
  while ( l != NULL ) {
    (*a->freeState)(((ListAgendaEntry)(l->item))->state);
    memFree(l->item);
    l = l->next;
  }
  memFree( a );
}

/* ----------------------------------------------------------------------
 * creates a new iterator object.
 *
 * This is actually just a single new list cell that points to the first entry.
 *
 * @param a the ListAgenda for which a new iterator has to be created.
 * @returns a new ListAgendaIterator that will return all items of \a a sorted by score.
 */
ListAgendaIterator laIteratorNew(ListAgenda a)
{
  if ( a->entries != NULL )
    {
      /* This allocates exactly one list cell. */
      return (listPrependElement(a->entries, a->entries->item));
    }
  else
    {
      return (NULL);
    }
}
/* ----------------------------------------------------------------------
 * returns the next item in iterator object.
 *
 * @param ai the iterator of the ListAgenda \a a.
 * @returns the best item in the underlying agenda that was not already returned
 * by the iterator.
 */
Pointer laIteratorNextElement(ListAgendaIterator ai)
{
  Pointer result;

  if (ai->next == NULL) {
    return (NULL);
  }

  result = ((ListAgendaEntry)(ai->item))->state;
  ai->item = ai->next->item;
  ai->next = ai->next->next;

  return (result);
}

/* ----------------------------------------------------------------------
 * deletes the iterator object.
 *
 * The function deallocates the list cell that was used by the iterator.
 *
 * @param ai the iterator object that has to be deleted.
 */
void laIteratorDelete(ListAgendaIterator ai)
{
  memFree( ai );
}

/* ------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------ */
/** @} */

