/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: blah@nats.informatik.uni-hamburg.de


*/

/* ---------------------------------------------------------------------------
 * @defgroup String Strings
 * implementation of strings.
 *
 * Strings in C are represented by arrays of characters.The end of the string
 * is marked with a special character, the null character, which is simply the
 * character with the value 0. Whenever we write a string, enclosed in double
 * quotes, C automatically creates an array of characters for us, containing that
 * string, terminated by the \a NULL character.
 * 
 * FIXME: this comment is plain copy/pasted from cdg.c
 *
 * The symbol table \ref cdgSymbolTable)is used to share strings registered 
 * (cdgRegisterString()) to it. Sharing strings speeds up string comparison 
 * alot as we dont need strcmp for this any more. A pointer comarison suffices.
 * So we store strings in a hash. But
 * be warned: changing a registered string directly will break things seriously.
 * If you need to change a registered string, make a copy of it (strCopy),
 * manipulate it for your needs and then register it once again.
 * @{
 */

/* ------------------------------------------------------------------------ */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <iconv.h>
#include <ctype.h>
#include "blah.h"
//#define DEBUG_DECODING
//#define DEBUG_STRING

/* ---------------------------------------------------------------------------
 * container for shared strings.
 * This global variable stores all registered strings
 */
static Hashtable _strStore = NULL;

/* ---------------------------------------------------------------------------
 * strings with reference counters
 */
typedef struct {
  String data; /**< the workload */
  int counter;    /**< a reference counter */
#ifdef DEBUG_STRING
  int magic; /**< magic value */
#endif
} SharedStringStruct;
typedef SharedStringStruct *SharedString;

static SharedString _strNewSharedString(const String str);
static void _strDeleteSharedString(SharedString);
static void _strDeleteStoreEntry(String, SharedString);
static SharedString _strLookup(String);
static String _strTryRegister(String);
#ifdef DEBUG_STRING
static void _strDebugStoreEntry(String, SharedString);
#endif

/* 
 * Conversion descriptor for String decoding with iconv (3).
 * This is a handler created with iconv_open (3).
 */
static iconv_t _strConversionDescriptor = (iconv_t)-1;

/* ---------------------------------------------------------------------------
 * this performs string copying function.
 *
 * This function constructs a copy of the given source string. The returned string is
 * not shared any more as its source might have been. So in order to manipulate
 * a shared string, first strCopy() it, then alter it and strRegister() it finaly.
 * While copying the string new memory is allocated for you. Take care of it.
 *
 * @param s the string that has to be copied.
 * @returns the new copied string.
 */
String strCopy(const String s)
{
  String t = (String)memMalloc(sizeof(char) * (strlen(s) + 1));
  strcpy(t, s);

  if (t == NULL) {
    fprintf(stderr, "No memory left\n");
    abort();
  }
  return t;
}

/* --------------------------------------------------------------------------
 * returns a formated string.
 * This function is our version of vsprintf(). See strPrintf() for more
 * information.
 */
String strVPrintf(const String fmt, va_list ap)
{
  /* Guess we need no more than 50 bytes. */
  int n, size = 50;
  String p = (String)memMalloc(size), q;

  while (1) {
    /* Try to print in the allocated space. */
    n = vsnprintf(p, size, fmt, ap);
    /* If that worked, return the string. */
    if (n > -1 && n < size) {
      break;
    }
    /* Else try again with more space. */
    if (n > -1)			       /* glibc 2.1 */
      size = n + 1;		       /* precisely what is needed */
    else			       /* glibc 2.0 */
      size *= 2;		       /* twice the old size */
    p = memRealloc(p, size);
  }

  q =  strRegister(p);
  memFree(p);
  return q;
}

/* --------------------------------------------------------------------------
 * returns a formated string.
 * This function basically has been taken from the sprintf() manual page.
 * The differences between sprintf() and strPrintf() are that you don't have
 * to bother about memory allocation. We allocate enuf memory to hold the
 * formated result string. Further more then this string is strRegister()ed
 * for you, so you might get an already shared string returned to you.
 * Use strDelete() to indicate your lake of interest on the result string.
 */
String strPrintf(const String fmt, ...)
{
  String str;
  va_list ap;
  
  va_start(ap, fmt);
  str = strVPrintf(fmt, ap);
  va_end(ap);

  return str;
}

/* --------------------------------------------------------------------------
 * lookup a string in the string store.
 * This function returns a pointer to a SharedString if the given string 
 * argument is already shared, or NULL if this string isn't shared yet.
 */
static SharedString _strLookup(String str)
{
  if (!str) 
    return NULL;
  else
    return (SharedString) hashGet(_strStore, str);
}

/* --------------------------------------------------------------------------
 * try to register a new string.
 * This function only registers new strings. It will not increase the
 * reference counter of an already registered string. In any case it will
 * return a known string.
 */
static String _strTryRegister(String str)
{
  SharedString sstr = _strLookup(str);

  if (!sstr)
    return strRegister(str);
  else
    return str;
}

/* --------------------------------------------------------------------------
 * concatenates two strings.
 *
 * This is our version of the standard unix strcat() with the differences
 * that both arguments are \c const strings. A concatenated shared string of
 * \c a and \c b is returned. Both arguments might be NULL.
 *
 * @returns the target concatenated with the source.
 */
String strCat(const String a, const String b)
{
  if (a) {
    if (b) {
      return strPrintf("%s%s", a, b);
    } else {
      return _strTryRegister(a);
    }
  } else {
    if (b) {
      return _strTryRegister(b);
    } else {
      return strRegister("");
    }
  }
}

/* --------------------------------------------------------------------------
 * concatenates a list of strings.
 *
 * This function takes a list of strings and concatenates them together in
 * a newly allocated string. Be sure that all list elements are realy of
 * type string. We can't grant that here. If the list is NULL or empty
 * NULL is returned to you. The return value is a registered string.
 *
 * @param list of strings
 * @returns the new appended string.
 */
String strFromList(List list)
{
  String str, result;
  List l;
  int len;

  if (!list || listSize(list) == 0) {
#if 0
    return strCopy("");
#else
    return NULL;
#endif
  }

  len = 1;
  for (l = list; l; l = l->next) {
    len += strlen(l->item);
  }

  str = (String)memMalloc(sizeof(String) * len);
  str[0] = '\0';

  for (l = list; l; l = l->next) {
    strcat(str, l->item);
  }

  result = strRegister(str);
  memFree(str);
  return result;
}

/* ---------------------------------------------------------------------------
 * concatenates many strings together.
 *
 * This function allocates the memory for the result string.
 * The argument strings are not modified by the function.
 * The last string in the argument list must be NULL.
 *
 * @param str the head of the string to be produced
 * @param ...  represents the strings to be appended to the head.
 * @returns the new appended string.
 */
String strAppend(const String head, ...)
{
  String str;
  List l = NULL;
  va_list ap;

  va_start(ap, head);
  l = listAppendElement(l, head);
  while(1) {
    str = va_arg(ap, String);
    if (!str)
      break;
    l = listAppendElement(l, str);
  }
  va_end(ap);
  str = strFromList(l);
  l = NULL;

  return str;
}

/* ---------------------------------------------------------------------------
 * allocate a new SharedString.
 * This function constructs a new SharedString. It contains no worload data yet.
 */
static SharedString _strNewSharedString(const String str)
{
  SharedString sstr = memMalloc(sizeof(SharedStringStruct));
  sstr->data = str;
  sstr->counter = 1;
#ifdef DEBUG_STRING
  sstr->magic = 4711;
#endif

  return sstr;
}

#ifdef DEBUG_STRING
/* ---------------------------------------------------------------------------
 * deallocate a key value pair.
 * This is used to deallocate the key and the value of the \c _strStore.
 */
static void _strDebugStoreEntry(String key, SharedString value)
{
  int strcmpResult = strcmp(key, value->data);
  fprintf(stderr, "DEBUG: _strDebugStoreEntry('%s', '%p')\n", key, value);
  assert(strcmpResult == 0);
}
#endif

/* ---------------------------------------------------------------------------
 * deallocate a key value pair.
 * This is used to deallocate the key and the value of the \c _strStore.
 */
static void _strDeleteStoreEntry(String key, SharedString value)
{
  _strDeleteSharedString(value); /* frees the key aswell */
}

/* ---------------------------------------------------------------------------
 * deallocated a SharedString.
 * This function deallocates a SharedString and its workload.
 */
static void _strDeleteSharedString(SharedString sstr)
{
  if(sstr) {
#ifdef DEBUG_STRING
    //fprintf(stderr, "DEBUG: freeing shared string '%s'\n", sstr->data);
    assert(sstr->magic == 4711);
#endif
    memFree(sstr->data);
    memFree(sstr);
  }
}

/* ---------------------------------------------------------------------------
 * unregister a string
 * This function tries deallocate the \c str string when its reference
 * counter licenses it. Note, that the pointer \c str might get invalid
 * or not depending on the reference counter.
 */
void strDelete(String str)
{
  SharedString sstr = _strLookup(str);

  if (sstr && --sstr->counter <= 0) {
    hashRemove(_strStore, str);
    _strDeleteSharedString(sstr);
  }
}

/* ---------------------------------------------------------------------------
 * register a string in symbol table.
 *
 * This function registers a string to be shared. This is done by copying
 * it into the \c _strStore (leaving the argument string pointer untouched).
 * If the string already exists in symbol table the _stored_ string
 * is returned. 
 * If the string doesn't exist then the string is _copied_ and 
 * entered in the symbol table; the new string is returned.
 * In no case the memory of string s is referenced by the symbol table.
 * But the returned string is owned by the symbol table and will be
 * shared by other references later on. So be careful and never change
 * a registered string in place. Use strCopy() first to check out a copy
 * of a shared string. 
 * If \c str is NULL then a registered empty string is returned, that is
 * strRegister(NULL) == strRegister("").
 */
String strRegister(const String str)
{
  SharedString sstr;
  String thisStr = str;

  if (!thisStr) {
    thisStr = "";
  }

  sstr = _strLookup(thisStr);

  if (!sstr) {
    sstr = _strNewSharedString(strCopy(thisStr)); /* diff to _strShare */
    hashSet(_strStore, sstr->data, (Pointer) sstr);
  } else {
    sstr->counter++;
  }
#ifdef DEBUG_STRING
    assert(sstr->magic == 4711);
#endif
  return sstr->data;
}

/* ---------------------------------------------------------------------------
 * return the number of shared strings.
 */
int strStoreSize(void)
{
  return hashSize(_strStore);
}

/* ---------------------------------------------------------------------------
 * module finalization routine.
 * This function is only called by blahFinalize() and should not be
 * used from outside 
 * It basically deallocates the \c _strStore.
 * It also closes the conversionHandler.
 */
void strFinalize(void)
{
  if (_strStore) {
#ifndef DEBUG_STRING
    hashForEachFree((Hashtable)_strStore, _strDeleteStoreEntry);
#else
    List keys = hashListOfKeys(_strStore);
    List l;
    fprintf(stderr, "DEBUG: nr keys=%d size=%d\n", listSize(keys), hashSize(_strStore));
    assert(listSize(keys) == hashSize(_strStore));
    for (l = keys; l; l = l->next) {
      fprintf(stderr, "DEBUG: key='%s'\n", l->item);
    }
    hashForEach((Hashtable)_strStore, _strDeleteStoreEntry);
#endif
  }

  /* close conversion Descriptor */
  if (_strConversionDescriptor != (iconv_t)-1) {
    iconv_close(_strConversionDescriptor);
#ifdef DEBUG_DECODING
    fprintf(stderr, "DEBUG: Conversion descriptor successfully closed\n");
#endif
  }
    
}

/* ----------------------------------------------------------------------
   Comparison function usable for listSort() etc.
*/
Boolean strCompare(String a, String b) {
  return strcmp(a,b) < 0;
}

/* ----------------------------------------------------------------------
   Comparison function for listSort() etc.

   This is the same as strCompare(), but it does the right thing
   when applied to strings with embedded number: each number is
   treated like a single character, and corresponding numbers are
   compared numerically:

   "V1.0" < "V1.1" < "V1.5" < "V1.12" < "V2.0"
   
*/
Boolean strVersionCompare(String a, String b) {
  String pa = a;
  String pb = b;
  int x, y;
  int i;

  while(*pa && *pb) {
    
    /* remove identical non-numeric parts */
    while(pa && pb && pa[0] == pb[0] && !isdigit(pa[0])) {
      pa++;
      pb++;
    }

    /* compare different non-numeric parts alphabetically */
    if(!isdigit(pa[0]) || !isdigit(pb[0])) {
      if(strcmp(pa,pb) < 0) {
        return TRUE;
      } else {
        return FALSE;
      }
    }

    /* compare numeric parts numerically */
    sscanf(pa, "%d", &x);
    sscanf(pb, "%d", &y);
    if(x > y) {
      return FALSE;
    }
    if(x < y) {
      return TRUE;
    }

    i = strspn(pa, "0123456789");
    pa += i;
    i = strspn(pb, "0123456789");
    pb += i;
  }
  return FALSE;

}

/* ----------------------------------------------------------------------------
   Translate a string from unicode UTF-8 to ISO-8859-1.

   The String will be left untouched if there is any problem while decoding.
 */
String strDecode(String word) {
  String result, s, inbuf, outbuf;
  unsigned int inbytesleft, outbytesleft;

  if(!word) {
    return NULL;
  }
  if((iconv_t)-1 == _strConversionDescriptor) {
#ifdef DEBUG_DECODING
    fprintf(stderr, "DEBUG: Conversion Descriptor not available, conversion skipped!\n");
#endif
    return word;
  }
  
  result = memMalloc(strlen(word) +1);
  strcpy(result, "");
  
  inbuf = word; // don't use the original pointers when calling iconv()
  outbuf = result;  
  outbytesleft = inbytesleft = strlen(word) + 1;

  /* convert the string 
   * iconv modifies all it's arguments, so we have to take care, when using
   * them later.
   * 'result' automatically points to the converted string when conversion is done
  */
  if (iconv(_strConversionDescriptor, &inbuf, &inbytesleft, &outbuf, &outbytesleft) ==  -1) {
#ifdef DEBUG_DECODING
    fprintf(stderr, "DEBUG: Conversion from UTF-8 to ISO-8859-1 failed for %s!\n", word);
#endif
    strcpy(result, word); // take the original word as fallback
  } 
  
#ifdef DEBUG_DECODING  
  if(strcmp(word, result)) {
    fprintf(stderr, "DEBUG: normalised `%s' to `%s'\n", word, result);
  }
#endif

  s = strRegister(result);
  memFree(result);
  return s;
}

/* ---------------------------------------------------------------------------
 * module initialization routine.
 * This function is only called by blahInitialize() and should not be
 * used from outside It basically allocates the \c _strStore.
 * It also sets the conversion handler for String decoding
 */
void strInitialize(void)
{
  _strStore =
    hashNew(500, 0.7, hashStringHashFunction, hashStringEqualFunction);

  /* set conversion descriptor: UTF-8 to ISO-8859-1 */
  _strConversionDescriptor = iconv_open("ISO-8859-1", "UTF-8");
#ifdef DEBUG_DECODING
  if (_strConversionDescriptor == (iconv_t) -1) {
    fprintf(stderr, "DEBUG: Opening UTF-8 to ISO-8859-1 conversion handler failed!\n");
  }
#endif

}
 


/* ---------------------------------------------------------------------------
 * explicit declarations
 */
char *strndup(const char *s, size_t n);

/* ---------------------------------------------------------------------------
 * returns the substring from String that starts with the startth character and 
 * being length chars long (not including the final \0)
 * returns NULL if provided with invalid arguments
 */
String strSubStr(String a, int start, int length) {
  
  // Input Filter
  if (a == NULL) return NULL;
  int max = strlen(a);
  if ((start < 0) || (start >= max) || (length < 0)) return NULL;
  String b = (String)((int)a + start);
 
  String t = (String)strndup(b, length);
  if (t == NULL) {
    fprintf(stderr, "No memory left\n");
    abort();
  }
  return t;
}

/* ---------------------------------------------------------------------------
 * checks whether or not a given string needle is a substring of haystack
 * returns the index of the first character of needle when successfull or
 * -1 otherwise
 */
int strFindSubStr(String haystack, String needle) {
  if ((haystack == NULL) || (needle == NULL)) return -1;
  String r = strstr(haystack, needle);
  if (r == NULL) return -1;  
  return (int)r - (int)haystack;
}


/* ---------------------------------------------------------------------------
 * checks whether or not a given string sub is a prefix of String a
 */
Boolean strIsPrefix(String a, String sub) {
  if ((a == NULL) || (sub == NULL)) return FALSE;
  int n = strlen(sub);
  if (n > strlen(a)) return FALSE;
  
  return (strncmp(a, sub, n) == 0);
}

/* ---------------------------------------------------------------------------
 * checks whether or not a given string sub is a suffix of String a
 */
Boolean strIsSuffix(String a, String sub) {
  if ((a == NULL) || (sub == NULL)) return FALSE;
  int n = strlen(sub);
  int alen = strlen(a);
  if (n > alen) return FALSE;
  
  String b = (String)((int)a + alen - n);
  
  return (strcmp(b, sub) == 0);
}


/* ------------------------------------------------------------------------- */
/** @} */
