/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
*/

/* ----------------------------------------------------------------------
 * @defgroup StringTree StringTree
 * A tree-organized set implementation for strings. The strings are ordered
 * in a prefix-based scheme with each word ending in a leaf. (if a word
 * happens to end in an inner node, an empty child-node is automaticially added)
 * @{
 */
 
/*
 * BIG TODO:
 *
 * This still doesn't do exactly what we want ! It is able to represent huge
 * amounts of Strings very memory-efficient (without common prefixes). But it
 * still need an leaf-node (of 12 bytes size) for each String. If not dealing with
 * large Strings this is even more memory-comsuming than before.
 *
 * A possibility for further optimization lies in thinking of the StringTree as an
 * cycle-free finite Automaton. It is because of its lack of cycles it was possible to
 * determinize it that easy. There are alghorithms for minimizing FAs. Some of them
 * can be found @
 * http://odur.let.rug.nl/~vannoord/Fsa/
 * perhaps one of them can even be simplified for cycle-free DFA's.
 *
 */

/* ----INCLUDES------------------------------------------------------- */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "blah.h"

// DEBUG
#define DEBUG

/* ----PREREQUISITES------------------------------------------------------- */

/* ----------------------------------------------------------------------
 *  free a String
 * @param str: the String to be freed
 */
void StrFree(String str) {
  if (str) memFree(str);
}

/* ----------------------------------------------------------------------
 *  clone a String
 * @param str: the String to be cloned
 * @returns: the clone
 *
 * assure: StrEquals(str, res);
 */
String StrClone(String str) {
  if (!str) return NULL;
  int len = strlen(str);
  String new = memMalloc(len+1);
  memcpy(new, str, len+1);
  return new;
}

/* ----------------------------------------------------------------------
 *  concatenate two Strings
 * @param a,b: the Strings to be concatenated
 * @returns: a new String that is prefixed by a and suffixed by b
 */
String StrCat(String a, String b) {
  if (!a) return StrClone(b);
  if (!b) return StrClone(a);
  int len_a = strlen(a);
  int len_b = strlen(b);
  String new = memMalloc(len_a + len_b + 1);
  memcpy(new, a, len_a);
  memcpy(new, b, len_b+1);
  return new;
}


/* ----------------------------------------------------------------------
 *  calculate the Prefix of a String
 * @param str: the Strings whose Prefix is to be calculated
 * @param len: the length of the Prefix to be calculated
 * @returns: a new String that prefixes str and is of length len
 */
String StrPrefix(String str, int len) {
	if (!str) return NULL;
	String new = memMalloc(len+1);
	memcpy(new, str, len);
	new[len] = 0;
	return new;
}

/* ----------------------------------------------------------------------
 *  calculate the Suffix of a String
 * @param str: the Strings whose Suffix is to be calculated
 * @param start: position in which the Suffix starts
 * @returns: a new String suffixing str and starting at str[start]
 */
String StrSuffix(String str, int start) {
	if (!str) return NULL;
	int len = strlen(str);
	if (len < start) return NULL;
	String new = memMalloc(len - start);
	memcpy(new, &str[start], len - start + 1);
	return new;
}

/* ----------------------------------------------------------------------
 *  calculate the common Prefix of two Strings
 * @param a,b: the Strings whose common Prefix is to be calculated
 * @returns: number of charakters that both Strings have in common at their beginning
 * assure: res <= strlen(a)
 * assure: res <= strlen(b)
 */
int StrCommonPrefixLength(String a, String b) {
	if (!a || !b) return 0;
	int i = 0;
	while((a[i] == b[i]) && a[i] && b[i]) i++;
	return i;
}

/* ----------------------------------------------------------------------
 *  returns a Pointer into the String str at Position pos
 * @param str: the Strings to be pointet into
 * @param pos: the index of the Charakter to be pointed at
 * @returns: a String that is NOT newly allocated but points into str
 */
String StrAtPos(String str, int pos) {
	if (!str) return NULL;
	return &str[pos];
}

/* ----------------------------------------------------------------------
 *  length of a String
 * @param a: the Strings whose length is to be estimated
 * @returns: the length of a
 */
int StrLen(String a) {
  if (a) return strlen(a);
  return 0;
}

/* ----------------------------------------------------------------------
 *  Equality Relation on Strings
 * @param a,b: two Strings whose equality is to be tested
 * @returns: TRUE if a equals b, FALSE otherwise
 *
 * @invariant: StrEquals(a,a)
 * @invariant: StrEquals(a,b) == StrEquals(b,a);
 * @invariant: StrEquals(a,b) && StrEquals(b,c) => StrEquals(a,c);
 * @invariant: !StrEquals(a,NULL);
 */
Boolean StrEquals(String a, String b) {
  if (a == b) return TRUE;
  if (!a || !b) return FALSE;
  return strcmp(a,b) == 0;
}

/* ----IMPLEMENTATION------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * get a new StringTree
 *
 * assure: res->value == NULL
 * assure: res->next == NULL
 * assure: res->firstchild == NULL
 *
 * @returns: the new StringTree
 */
StringTree stNew()
{
  StringTree st = memMalloc(sizeof(StringTreeStruct));
  st->value = NULL;
  st->next = NULL;
  st->firstchild = NULL;
  return st;
}

/* ----------------------------------------------------------------------
 *  get a new StringTree and fill its res->value field with value
 *
 * @param value: the String to fill into res->value
 */
StringTree stNewVal(String value)
{
  StringTree st = stNew();
  st->value = value;
  return st;
}

/* ----------------------------------------------------------------------
 *  free a previously allocated StringTree
 *
 * @param this: the StringTree to be freed
 */
void stFree(StringTree this)
{
  if (!this) return;
  
  if (this->firstchild) stFree(this->firstchild);
  
  if (this->next) stFree(this->next);
  
  if (this->value) StrFree(this->value);
  
  memFree(this);
}

/* ----------------------------------------------------------------------
 *  test for equality
 * @param a,b: the two StringTrees to test
 * @returns: TRUE, if the a and b are equal, FALSE otherwise
 *
 * @invariant: a == b => stEquals(a,b)
 * @invariant: [StrEquals(a->value, b->value) &&
 *              stEquals(a->firstchild, b->firstchild) &&
 *              stEquals(a->next, b->next)] <=> stEquals(a,b)
 */
Boolean stEquals(StringTree a, StringTree b) {
  if (a == b) return TRUE;
  if (!a || !b) return FALSE;
  
  return StrEquals(a->value, b->value) &&
   stEquals(a->firstchild, b->firstchild) &&
   stEquals(a->next, b->next);
}

/* ----------------------------------------------------------------------
 *  clone a StringTree
 *
 *  @param this: the StringTree to clone
 *  @returns: the cloned StringTree
 *
 *  assure: stEquals(res, this)
 */
StringTree stClone(StringTree this)
{
  if (!this) return NULL;
  
  StringTree st = stNew();
  st->value = StrClone(this->value);
  st->firstchild = stClone(this->firstchild);
  st->next = stClone(this->next);
  return st;
}

/* ----------------------------------------------------------------------
 *  insert a string into the tree
 *
 * @param this: the StringTree the String should be inserted into
 * @param item: the String to be inserted
 *
 * assure stIsIn(this, item)
 */
void stInsert(StringTree this, String item)
{
  if (!this) return;
  if (!item) return;
  
  // without a value, this tree cannot have any children
  if (!this->value) {
    this->value = StrClone(item);
    return;
  }
  
  int len = StrLen(this->value);
  int item_len = StrLen(item);
  int pre_len = StrCommonPrefixLength(this->value, item);
  if (pre_len <= 0) {
    if ((item_len == 0) && (len == 0)) return;   // the string is already in the tree
    
    // item doesn't prefix st->value -> pass it to the next one or create one
    if (this->next) {
      stInsert(this->next, item);
    } else {
      this->next = stNewVal(StrClone(item));
    }
  } else if (pre_len >= len) {
    // st->value prefixes item -> pass it to the children or create one
    if (this->firstchild) {
      stInsert(this->firstchild, StrAtPos(item, len));
    } else {
      if (item_len == len) return;   // the string is already in the tree
      this->firstchild = stNewVal(StrClone(""));
      this->firstchild->next = stNewVal(StrClone(StrAtPos(item, len)));
    }
  } else {
    // item prefixes st->value -> split
    String old = this->value;
    StringTree st = stNewVal(StrSuffix(old, pre_len));
    st->firstchild = this->firstchild;
    this->value = StrPrefix(old, pre_len);
    this->firstchild = st;
    StrFree(old);
    stInsert(st, StrAtPos(item, pre_len));
  }
}

/* ----------------------------------------------------------------------
 *  remove a string from a tree
 *
 * @param this: the StringTree the String should be removed from
 * @param item: the String to be removed
 *
 * assure !stIsIn(this, item)
 */
void stRemove(StringTree this, String item)
{
  // TODO: implement this
}

/* ----------------------------------------------------------------------
 *  Finds the Position of a String in the StringTree
 * @param this: the StringTree to search in
 * @param item: the String to search te Position for
 * @returns A struct containing:
 *	st: a pointer to the StringTree whose value is prefixed by the last part of item
 *	pos: a pointer to the position in st->value that item ended at
 */
StringTreePos stFind(StringTree this, String item)
{
  if (!this) return NULL;
  if (!item) return NULL;
  
  int len = StrLen(this->value);
  int item_len = StrLen(item);
  int pre_len = StrCommonPrefixLength(this->value, item);
  
  if ((pre_len <= 0) && !((item_len == 0) && (len == 0))) {
    // item doesn't prefix st->value -> pass it to the next one
    return stFind(this->next, item);
  } else if ((pre_len >= len) && (len < item_len)) {
    // st->value prefixes item -> pass it to the children
    return stFind(this->firstchild, StrAtPos(item, len));
  } else {
    if (pre_len < item_len) return NULL;  // if it does not match completely -> it's not in the tree !
    StringTreePos res = malloc(sizeof(StringTreePosStruct));
    res->st = this;
    res->pos = StrAtPos(this->value, pre_len);
    return res;
  }
}

/* ----------------------------------------------------------------------
 *  free a StringTreePos struct allocated by stFind()
 *
 * @param stp: the StringTreePos struct
 */
void stFreeStringTreePos(StringTreePos stp) {
  if (stp) memFree(stp);
}

/* ----------------------------------------------------------------------
 *  tests is a string is in an StringTree - this makes a StringTree
 *  usefull as a Set of Strings
 * 
 * @param this: the Tree to be checked
 * @param item: the Strng to be searched for
 *
 * @returns: TRUE, if the String was inserted into the Tree, FALSE otherwise
 */
Boolean stIsIn(StringTree this, String item)
{
  StringTreePos pos = stFind(this, item);
  Boolean res = pos && pos->st && pos->pos && !pos->st->firstchild && !pos->pos[0];
  stFreeStringTreePos(pos);
  return res;
}

