/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: blah@nats.informatik.uni-hamburg.de
    Author: Ingo Schroeder
    Birth: 22/7/99
   
 */
 
/* ----------------------------------------------------------------------
 * @defgroup TreeAgenda TreeAgenda
 *
 * implemention of an agenda interface using an unbalanced binary search tree.
 *
 * This module exports an agenda as used in the CDG netsearch module.
 * The agenda manages a set of elements annoted with the ratings.
 * Elements can be inserted according to their rating, and the first element
 * can be retrieved.The agenda is not responsible for determining the rating
 * of an element;it can only deal with pairs(subsequently called items) of
 * elements and scores.
 *
 * \todo Actually the b-tree used here can be implemented in a more general
 * way to be more usefull. An agenda is just one way to use a b-treeish 
 * storage organization.
 * @{
 */

/* -- INCLUDES -----------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <blah.h>

/* -- MACROS -------------------------------------------------------------*/
#define TA_DEBUG 0

/* -- TYPE DEFINITIONS ---------------------------------------------------*/

/* -------------------------------------------------------------------------
 * this type represents an entry of an agenda.
 */
struct TANodeStruct {
  int parent;                     /**< index of parent node or -1 if root */
  int better;              /**< index of left child or -1 if non-existent */
  int worse;              /**< index of right child or -1 if non-existent */
  double score;                                 /**< score for this state */
  Pointer state;                                    /**< pointer to state */
};

/* -------------------------------------------------------------------------
 * quick, should be binary tree.
 */
struct TreeAgendaStruct {
  int size;                               /**< total number of used nodes */
  int maxsize;                            /**< number of allocated nodes */
  int maxSizeSoFar;           /**< maximum number of used nodes so far */
  Boolean verbose;		  /**< If set to true then truncation warnings take effect */
  Boolean truncationWarning;  /**< flag whether warning has been emitted */
  int free;           /**< index of first free node or -1 if non-existent */
  int root;                 /**< index of root node or -1 if non-existent */
  int best;                 /**< index of best node or -1 if non-existent */
  int worst;               /**< index of worst node or -1 if non-existent */
  VoidFunction *freeState;                   /**< free function */
  TANodeStruct *nodes;                       /**< array of maxsize nodes */
};

/* -------------------------------------------------------------------------
 * this structure instantiates the generic agenda iterator.
 */

struct TreeAgendaIteratorStruct {
  TreeAgenda treeagenda;           /**< points to the underlying TreeAgenda.*/
  int nextNode;              /**< an index to the next node to be returned.*/
};


/* -- FUNCTIONS ----------------------------------------------------------*/

/* -------------------------------------------------------------------------
 * prints a node.
 *
 * It is used by the taCheckNode().
 *
 * @param a the TreeAgenda whose node has to be printed.
 * @param index the index of the node that has to be printed.
 */
void taPrintNode(TreeAgenda a, int index)
{
  printf("[ %d ", index);

  if (index >= 0) {
    printf("p %d ", a->nodes[index].parent);
    if (a->nodes[index].parent >= 0) {
      printf(" p's b %d p's w %d p's p %d ",
		a->nodes[a->nodes[index].parent].better,
		a->nodes[a->nodes[index].parent].worse,
		a->nodes[a->nodes[index].parent].parent);
    }
    printf("b %d ", a->nodes[index].better);
    if (a->nodes[index].better >= 0) {
      printf(" b's b %d b's w %d b's p %d ",
		a->nodes[a->nodes[index].better].better,
		a->nodes[a->nodes[index].better].worse,
		a->nodes[a->nodes[index].better].parent);
    }
    printf("w %d ", a->nodes[index].worse);
    if (a->nodes[index].worse >= 0) {
      printf(" w's b %d w's w %d w's p %d ",
		a->nodes[a->nodes[index].worse].better,
		a->nodes[a->nodes[index].worse].worse,
		a->nodes[a->nodes[index].worse].parent);
    }
  }
  printf("]");
}

/* -------------------------------------------------------------------------
 * performs internal consistency checks on a TreeAgenda.
 *
 * This is only for debugging purposes.
 *
 * @param a the TreeAgenda that has to be checked.
 * @param index the index of the node in the treeagenda.
 * @param recursive is \a TRUE if repetition occurs and \a FALSE otherwise.
 * @param better is \a TRUE if better consistency and \a FALSE otherwise.
 * @returns the number of nodes that are checked.
 */
int taCheckNode(TreeAgenda a, int index, Boolean recursive, Boolean better)
{
  int count;

  if (index < 0) {
    return (0);
  }

  if (a->best >= 0 && a->nodes[a->best].better >= 0) {
    printf("DEBUG: best inconsistency\n    index ");
    taPrintNode(a, index);
    printf("\n    best ");
    taPrintNode(a, a->best);
    printf("\n");
    abort();
  }

  if (a->worst >= 0 && a->nodes[a->worst].worse >= 0) {
    printf("DEBUG: worst inconsistency\n    index ");
    taPrintNode(a, index);
    printf("\n    worst ");
    taPrintNode(a, a->worst);
    printf("\n");
    abort();
  }

  if ((a->root != index && a->nodes[index].parent < 0)
      || (a->nodes[index].parent >= 0 &&
	  ((better && a->nodes[a->nodes[index].parent].better != index)
	   || (!better && a->nodes[a->nodes[index].parent].worse != index)))) {
    printf("ERROR: %s inconsistency in node ", better &&
	      a->nodes[a->nodes[index].parent].better !=
	      index ? "left" : "right");
    taPrintNode(a, index);
    printf("\n");
    abort();
  }

  count = 1;
  if (recursive) {
    count += taCheckNode(a, a->nodes[index].better, recursive, TRUE);
    count += taCheckNode(a, a->nodes[index].worse, recursive, FALSE);
  }

  return (count);
}

/* -------------------------------------------------------------------------
 * deletes a node.
 *
 * @param a the TreeAgenda in which a node has to be deleted.
 * @param index the index of the node that has to be deleted in TreeAgenda.
 * @param delete \a TRUE if the item itself is also deallocated and \a FALSE otherwise.
  */
int taDeleteNode(TreeAgenda a, int index, Boolean delete)
{
  TANode node;
  int count;

  if (index < 0) {
    return (0);
  }

  node = &a->nodes[index];

  count = 1;
  count += taDeleteNode(a, node->better, delete);
  count += taDeleteNode(a, node->worse, delete);

  if (delete) {
    (*a->freeState) (node->state);
    /* prepend node index to free node list */
    node->parent = a->free;
    a->free = index;
    a->size--;
    if (index == a->root) {
      a->root = -1;
      printf("DEBUG: taDeleteNode, %d was root, resetting\n", index);
    }
  }

  return (count);
}

/* -------------------------------------------------------------------------
 * creates a new agenda.
 *
 * @param maxsize determines the desired maximal size of an agenda.
 * @param f the function for deallocating the element.
 * @returns a pointer to the new structure that can hold upto \a maxsize entries.
 *
 * The agenda cannot determine whether it is safe to call this function on an
 * element or not.Passing the wrong function to a taNew() results undefined behaviour.
 */
TreeAgenda taNew(int maxsize, VoidFunction *f)
{
  int i;
  TreeAgenda a;

  if (maxsize < 1) {
    printf("WARNING: minimum size of agenda is 2, %d is too small\n", maxsize);
    /*
     * The fact that the minimum size is greater than 1 is explored at some points
     * in the code.
     */
    maxsize = 2;
  }

  a = (TreeAgenda) memMalloc(sizeof (TreeAgendaStruct));
  a->size = 0;
  a->maxsize = maxsize;
  a->maxSizeSoFar = 0;
  a->verbose = FALSE;
  a->truncationWarning = FALSE;
  a->root = a->best = a->worst = -1;
  a->freeState = f;
  a->nodes = (TANodeStruct *) memMalloc(sizeof (TANodeStruct) * maxsize);
  memset(a->nodes, 0, sizeof (TANodeStruct) * maxsize);

  for (i = 0; i < maxsize; i++) {
    a->nodes[i].parent = i - 1;
  }

  a->free = maxsize - 1;

  return (a);
}

/* ----------------------------------------------------------------------
 * gets verbosity flag.
 * \param a the TreeAgenda whose verbosity property is set
 * \returns the verbosity state
 */
Boolean taVerbosity(TreeAgenda a)
{
  return a->verbose;
}

/* ----------------------------------------------------------------------
 * sets verbosity flag.
 * \param a the TreeAgenda whose verbosity property is set
 * \param b a Boolean which we set the verbosity to
 * \returns the old verbosity state
 */
Boolean taSetVerbosity(TreeAgenda a, Boolean b)
{
  Boolean old = a->verbose;

  a->verbose = b;
  return (old);
}


/* -------------------------------------------------------------------------
 * retrieves size of the agenda.
 *
 * @param a the TreeAgenda for which the size has to be retrieved.
 * @returns the current number of entries in the TreeAgenda \a a.
 */
int taSize(TreeAgenda a)
{
  return (a->size);
}

/* -------------------------------------------------------------------------
 * retrieves the size limit of the agenda.
 *
 * @param a the TreeAgenda for which the maximum size has to be determined.
 * @returns the max number of entries that can be held by the TreeAgenda a.
 */
int taMaxSize(TreeAgenda a)
{
  return (a->maxsize);
}

/* -------------------------------------------------------------------------
 * retrieves the largest attained size of the agenda.
 *
 * @param a the TreeAgenda for which the largest size has to be retrieved.
 * @returns the maximum size occupied by the agenda \a a so far.
 */
int taMaxSizeSoFar(TreeAgenda a)
{
  return (a->maxSizeSoFar);
}

/* -------------------------------------------------------------------------
 * checks if the specified TreeAgenda is empty.
 *
 * @param a the TreeAgenda for which emptiness has to be checked.
 * @returns \a TRUE if agenda \a a is empty and \a FALSE otherwise.
 */
Boolean taIsEmpty(TreeAgenda a)
{
  return (a->size == 0);
}

/* -------------------------------------------------------------------------
 * checks for the truncation of the TreeAgenda \a a.
 *
 * @param a the TreeAgenda on which the function \a taIsTruncated is performed.
 * @returns \a TRUE if the agenda is already truncated and \a FALSE otherwise.
 */
Boolean taIsTruncated(TreeAgenda a)
{
  return (a->truncationWarning);
}

/* -------------------------------------------------------------------------
 * reset the agenda truncation warning.
 *
 * @param a the TreeAgenda whose truncationWarning will be set to FALSE.
 * @returns the old value of the truncationWarning
 */
Boolean taResetTruncated(TreeAgenda a)
{
  Boolean oldTruncationWarning = a->truncationWarning;
  a->truncationWarning = FALSE;
  return oldTruncationWarning;
}

/* -------------------------------------------------------------------------
 * inserts a new entry into the agenda.
 *
 * @param a the TreeAgenda into which a new element has to be inserted.
 * @param state the item that is inserted into the TreeAgenda \a a.
 * @param score the \a state is sorted into the list according to the \a score.
 * @returns \a FALSE if the agenda has been truncated and \a TRUE if we were able
 *        to insert the item without any unpleasant side effects. This is going to
 *        be reported only once.
 *
 * \n If the insertion leads to an overflow,then one of the elements from the
 * tree is deallocated.Note that the element to be removed may be the \a state itself.
 */
Boolean taInsert(TreeAgenda a, double score, Pointer state)
{
  int i, parent, new, rightmost;
  TANode parentnode = NULL, newnode, rightmostnode;
  Boolean oldTruncationWarning = a->truncationWarning;

#if TA_DEBUG
  if (a->size != taCheckNode(a, a->root, TRUE, TRUE)) {
    abort();
  }
#endif

  /* First, let's see whether the agenda is full */
  if (a->size >= a->maxsize) {
    /* Yep, agenda is full. */
    if (!a->truncationWarning) {
      if (a->verbose) {
	//printf("WARNING: truncating agenda, result may be incorrect\n");
      }
      a->truncationWarning = TRUE;
    }
#if TA_DEBUG
    /* Find worst node so far. */
    for (rightmost = a->root;
	 a->nodes[rightmost].worse >= 0; rightmost = a->nodes[rightmost].worse) {
      /* empty body */
    }

    if (rightmost != a->worst) {
      printf("DEBUG: rightmost inconsistency\n    rightmost ");
      taPrintNode(a, rightmost);
      printf("\n    worst ");
      taPrintNode(a, a->worst);
      printf("\n");
    }
#else
    rightmost = a->worst;
#endif

    /* rightmost is right most node and strictly >= 0 */
    rightmostnode = &a->nodes[rightmost];

    /* Maybe the new state is really bad? */
    if (score <= rightmostnode->score) {
      /* Yes, throw it away. */
      (*a->freeState) (state);
      return (oldTruncationWarning != a->truncationWarning);
    }

    /* At least one node is worse, delete that one. */
#if 0
    printf("WARNING: deleting old state %d with score %4.3f, new one is %4.3f\n",
	      rightmost, rightmostnode->score, score);
#endif
    (*a->freeState) (rightmostnode->state);
    if (rightmostnode->parent >= 0) {
      /* We have a parent, i. e. we aren't the root. */
      a->nodes[rightmostnode->parent].worse = rightmostnode->better;
      if (rightmostnode->better >= 0) {
	a->nodes[rightmostnode->better].parent = rightmostnode->parent;
	for (a->worst = rightmostnode->better;
	     a->worst >= 0 && a->nodes[a->worst].worse >= 0;
	     a->worst = a->nodes[a->worst].worse) {
	  /* empty body */
	}
      } else {
	a->worst = rightmostnode->parent;
      }
    } else {
      /* We are the root */
      if (rightmostnode->better >= 0) {
	a->nodes[rightmostnode->better].parent = -2;
	a->root = rightmostnode->better;
	for (a->worst = rightmostnode->better;
	     a->worst >= 0 && a->nodes[a->worst].worse >= 0;
	     a->worst = a->nodes[a->worst].worse) {
	  /* empty body */
	}
      } else {
	/* Can this happen? */
	a->root = -1;
	a->worst = -1;
      }
    }
    
    /* At this point we explore the fact that the agenda has a minimum
     * size of two. If the agenda is full it cannot be the case that 
     * a->best == a->worst. That means we haven't deleted the best node 
     * and we don't have to fix a->best.
     */

    /* a->free must be -1 at this point, because the agenda is full */
    rightmostnode->parent = -3;
    a->free = rightmost;
    a->size--;
  }

  /* Next, find parent of our new friend. */
  i = a->root;
  parent = -4;
  while (i >= 0) {
    parent = i;
    parentnode = &a->nodes[i];
    if (score >= parentnode->score) {
      i = parentnode->better;
    } else {
      i = parentnode->worse;
    }
  }

  /* 
   * parent is index to parent of new node
   * parent < 0 means agenda is empty, i. e. a->root == -1
   * parentnode is the corresponding node 
   */
  new = a->free;
  newnode = &a->nodes[new];
  a->free = newnode->parent;
  a->size++;
  newnode->better = newnode->worse = -1;
  newnode->parent = parent;
  newnode->score = score;
  newnode->state = state;

  if (a->root < 0) {
    /* Agenda seems to be empty, that's easy. */
    a->root = a->best = a->worst = new;
  } else {
    if (score >= parentnode->score) {
      parentnode->better = new;
      if (a->best == parent) {
	a->best = new;
      }
    } else {
      parentnode->worse = new;
      if (a->worst == parent) {
	a->worst = new;
      }
    }
  }

  if (a->maxSizeSoFar < a->size) {
    a->maxSizeSoFar = a->size;
  }

  return (oldTruncationWarning != a->truncationWarning);
}

/* -------------------------------------------------------------------------
 * returns the best entry (= first) from the agenda.
 *
 * @param a the TreeAgenda from which the best item has to be retrieved.
 * @returns the best item in \a a (or NULL if a is empty).The item remains in the agenda.
 */
Pointer taBest(TreeAgenda a)
{
  if (a->best < 0) {
    return (NULL);
  } else {
    return (a->nodes[a->best].state);
  }
}

/* -------------------------------------------------------------------------
 * removes and returns best entry (= first) from the agenda.
 *
 * @param a the TreeAgenda from which the best item has to be retrieved.
 * @returns the best item in \a a.The corresponding entry is removed and deallocated.
 * \n It must not be called on an empty agenda.
 */
Pointer taRemoveBest(TreeAgenda a)
{
  int leftmost, parent, rightchild;

#if TA_DEBUG
  for (leftmost = a->root;
       leftmost >= 0 && a->nodes[leftmost].better >= 0;
       leftmost = a->nodes[leftmost].better) {
    /* empty body */
  }
  if (leftmost != a->best) {
    printf("DEBUG: leftmost inconsistency\n    leftmost ");
    taPrintNode(a, leftmost);
    printf("\n    best ");
    taPrintNode(a, a->best);
    printf("\n    size %d root %d\n", a->size, a->root);
  }
#else
  leftmost = a->best;
#endif

  if (leftmost < 0) {
    printf("ERROR: taRemoveBest: agenda empty (size is %d)\n", a->size);
    return (NULL);
  }

  /* leftmost is index to left node */
  parent = a->nodes[leftmost].parent;
  rightchild = a->nodes[leftmost].worse;
  if (parent >= 0) {
    /* We have a parent, i. e. we aren't the root. */
    a->nodes[parent].better = rightchild;
    if (rightchild >= 0) {
      a->nodes[rightchild].parent = parent;
      for (a->best = rightchild;
	   a->best >= 0 && a->nodes[a->best].better >= 0;
	   a->best = a->nodes[a->best].better) {
	/* empty body */
      }
    } else {
      a->best = parent;
    }
  } else {
    /* We are the root */
    if (rightchild >= 0) {
      a->nodes[rightchild].parent = -5;
      a->root = rightchild;
      for (a->best = a->root;
	   a->best >= 0 && a->nodes[a->best].better >= 0;
	   a->best = a->nodes[a->best].better) {
	/* empty body */
      }
    } else {
      /* This is the only case where removing the best node
       * also means deleting the worst one. But this one
       * is trivial.
       */
      a->root = a->best = a->worst = -1;
    }
  }

  a->nodes[leftmost].parent = a->free;
  a->free = leftmost;
  a->size--;

  return (a->nodes[leftmost].state);
}

/* -------------------------------------------------------------------------
 * deletes the specified agenda.
 *
 * The function deallocates all the items in \a a using *freeState().
 * Then it deallocates all the entries and the agenda itself.
 *
 * @param a the ListAgenda that has to be deleted.
 */
void taDelete(TreeAgenda a)
{
  int count;
  int oldSize;

  oldSize = a->size;
  count = taDeleteNode(a, a->root, TRUE);
  if (count != oldSize) {
    printf("WARNING: taDelete, mismatch while deleting %d != %d\n", count,
	   oldSize);
  }

  memFree(a->nodes);
  memFree(a);
}

/* -------------------------------------------------------------------------
 * creates a new iterator object.
 *
 * This function creates a new iterator for TreeAgenda \a a.
 *
 * @param a the TreeAgenda for which a new iterator has to be created.
 * @returns a new TreeAgendaIterator that will return all items of \a a sorted by score.
 */
TreeAgendaIterator taIteratorNew(TreeAgenda a)
{
  TreeAgendaIterator ai;

  ai = (TreeAgendaIterator) memMalloc(sizeof (TreeAgendaIteratorStruct));
  ai->treeagenda = a;
  ai->nextNode = a->best;
  return (ai);
}

/* -------------------------------------------------------------------------
 * returns the next item in iterator object.
 *
 * @param ai the iterator of the TreeAgenda \a a.
 * @returns the best item in the underlying agenda that was not already returned
 * by the iterator.
 */
Pointer taIteratorNextElement(TreeAgendaIterator ai)
{
  Pointer state;
  int parent;

  if (ai->nextNode < 0) {
    return (NULL);
  }
  state = ai->treeagenda->nodes[ai->nextNode].state;

  if (ai->treeagenda->nodes[ai->nextNode].worse >= 0) {
    /* Find best node that's worse than the current node. */
    for ( ai->nextNode = ai->treeagenda->nodes[ai->nextNode].worse;
	  ai->treeagenda->nodes[ai->nextNode].better >= 0;
	  ai->nextNode = ai->treeagenda->nodes[ai->nextNode].better ) {
      /* empty body */
    }
  } else {
    parent = ai->treeagenda->nodes[ai->nextNode].parent;
    while (parent >= 0 && ai->nextNode == ai->treeagenda->nodes[parent].worse) {
      ai->nextNode = parent;
      parent = ai->treeagenda->nodes[ai->nextNode].parent;
    }
    ai->nextNode = parent;    
  }

  return (state);
}

/* -------------------------------------------------------------------------
 * deletes the iterator object.
 *
 * The function deallocates the tree element that was used by the iterator.
 *
 * @param ai the iterator object that has to be deleted.
 */
void taIteratorDelete(TreeAgendaIterator ai)
{
  memFree(ai);
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
