/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * Description:
 * make a berkeley database which holds an index of the position
 * of each lexical entry in a cdg lexicon file.
 *
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "db.h"
#include "blah.h"

/* the database handle */
static DB *database;

FILE *stream;
int c;
#define BUFFERSIZE 8000
char key[BUFFERSIZE];
int ptr = 0;

int keysFound = 0;

/* ------------------------------------------------------------------------- 
 * opens a database.
 * this function opens a database, specified by the parameter 'filename'
 * and stores the handle in the static variable 'database'. If there is no
 * database file with this filename, a new database is created.
 * @param filename the filename of the database file
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbOpen(String filename)
{
  int ret = 0;		     // return code for database operations

  // create database handle
  if ((ret = db_create(&database, NULL, 0)) != 0) {
    fprintf(stderr, "ERROR: (db_create) - %s\n", db_strerror(ret));
    return FALSE;
  }
  // setting database to accept more than one data item for one key
  database->set_flags(database, DB_DUP);

  // open database 'filename', if the files already exists, the error EEXIST
  // is thrown by database->open
#if (DB_VERSION_MAJOR < 4 || \
     DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 0)
  if((ret = database->open(database, filename, NULL, DB_BTREE, 
      DB_CREATE | DB_EXCL | DB_TRUNCATE, 0664)) != 0) {
#else /* db 4.1.x or later */
  if((ret = database->open(database, NULL, filename, NULL, DB_BTREE, 
      DB_CREATE | DB_EXCL | DB_TRUNCATE, 0664)) != 0) {
#endif
    fprintf(stderr, "ERROR: (database->open) - %s\n", db_strerror(ret));
    return FALSE;
  } 

  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * close the database
 * write database to disk and close it.
 * @returns TRUE if success, FALSE otherwise
 */

Boolean dbClose(void)
{
  int ret = 0;		     // return code for database operations

  if (database == NULL) {
    fprintf(stderr, "ERROR: No open database found for closing\n");
    return FALSE;
  }

  if ((ret = database->close(database, 0)) != 0) {
    fprintf(stderr, "ERROR: (database->close) - %s\n", db_strerror(ret));
    return FALSE;
  }
  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * adds an entry to the database 
 * adds a key and its position in the lexicon file to the database
 * @param word the word which is the key
 * @param pos element [0]: the beginning, element [1] the end of this entry
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbAdd(String word, int pos[])
{
  int ret = 0;		     // return code for database operations
  DBT key, value;	     // DBT structs for key and value


  
  if (database == NULL) {
    fprintf(stderr, "ERROR: No database opened");
    return FALSE;
  }


  memset(&key, 0, sizeof (key));       // cleaning memory
  memset(&value, 0, sizeof (value));
  key.data = word;
  key.size = strlen(word);
  value.data = pos;
  value.size = sizeof(int) * 2;

  if ((ret = database->put(database, NULL, &key, &value, 0)) == 0) {
#ifdef DEBUG
    printf("added: %s->[%d][%d]\n", (char *)key.data, *((int *)value.data), *((int *)(value.data)+1));
#endif
    keysFound++;
    return TRUE;
  } else {
    fprintf(stderr, "ERROR: (database->put) - %s\n", db_strerror(ret));
    return FALSE;
  }
}


/* return the character after the current character */
char lookahead() {
  char result = (char)getc(stream);
  ungetc(result, stream);
  return result;
}

/* verify that character C follows */
void expect(char c) {
  if(c != (char)getc(stream)) {
    fprintf(stderr, "ERROR: `%c' expected at %ld!\n", c, ftell(stream));
    exit(1);
  }
}

/* iterate over line comment */
void skip_line_comment() {
  while((c = getc(stream)) && !feof(stream)) {
    if('\n' == (char)c) {
      return;
    } else {
      /* do nothing */
    }
  }
  fprintf(stderr, "ERROR: end of file in comment at %ld!\n", ftell(stream));
  exit(1);
}


/* iterate over starred comment */
void skip_star_comment() {
  while((c = getc(stream)) && !feof(stream)) {
    if('*' == (char)c && '/' == lookahead()) {
      getc(stream);
      return;
    } else {
      /* do nothing */
    }
  }
  fprintf(stderr, "ERROR: end of file in comment at %ld!\n", ftell(stream));
  exit(1);
}

/* iterate over space and comments */
void skip_space_and_comments() {

  while((c = getc(stream)) && !feof(stream)) {
    if(isspace(c)) {
      /* do nothing */
    }

    else if('/' == (char)c) {
      c = getc(stream);
      if('*' == (char)c) {
        skip_star_comment();
      }
      else if('/' == (char)c) {
        skip_line_comment();
      }
    }
    else {
      ungetc(c, stream);
      return;
    }
  }
}

/* add character C to the buffer */
void add(char c) {
  if(ptr == BUFFERSIZE) {
    fprintf(stderr, "ERROR: key too long at %ld!\n", ftell(stream));
    exit(1);
  }
  key[ptr] = c;
  ptr++;
}

/* read a valid CDG string, quoted or unquoted, into the buffer `key' */
void read_key() {

  ptr = 0;

  c = getc(stream);
  
  /* unquoted key */
  if(isalnum(c) || (c > 0x80)) {
    while(isalnum(c) || (c > 0x80)) {
      add((char)c);
      c = getc(stream);
    }
    ungetc(c, stream);
    add('\0');
  }
  
  /* double-quoted key */
  else if('"' == (char)c) {
    while(1) {
      c = getc(stream);
      if(feof(stream)) {
        fprintf(stderr,
                "ERROR: end of file in quoted string at %ld!\n",
                ftell(stream));
        exit(1);
      }
      else if('\\' == (char)c) {
        c = getc(stream);
        add((char)c);
      }
      else if('"' == (char)c) {
        add('\0');
        return;
      }
      else {
        add((char)c);
      }
    }
  }

  /* single-quoted key */
  else if('\'' == (char)c) {
    while(1) {
      c = getc(stream);
      if(feof(stream)) {
        fprintf(stderr,
                "ERROR: end of file in quoted string at %ld!\n",
                ftell(stream));
        exit(1);
      }
      else if('\\' == (char)c) {
        c = getc(stream);
        add((char)c);
      }
      else if('\'' == (char)c) {
        add('\0');
        return;
      }
      else {
        add((char)c);
      }
    }
  }
  else {
    fprintf(stderr, "ERROR: invalid start of key at %ld!\n", ftell(stream));
    exit(1);
  }
}

/* read a CDG value into the buffer `value' */
void read_value() {

  int last = 0;

  /* 0: initial
     1: reading single-quoted string
     2: reading double-quoted string */
  int state = 0;

  ptr = 0;
  while((last = c) &&
        (c = getc(stream)) &&
        !feof(stream)) {
#ifdef DEBUG
    printf("DEBUG: `%c' / state %d\n", (char)c, state);
#endif
    if(0 == state) {
      if(';' == (char)c) {
        return;
      }
      else if('\''== (char)c) {
        state = 1;
      }
      else if('"' == (char)c) {
        state = 2;
      }
    }

    else if (1 == state) {
      if('\\' == (char)c) {
        last = c;
        c = getc(stream);
      }
      else if('\''== (char)c && last != '\\') {
        state = 0;
      }
    }

    else if (2 == state) {
      if('\\' == (char)c) {
        last = c;
        c = getc(stream);
      }
      else if('"'== (char) c && last != '\\') {
        state = 0;
      }
    }
  }
}

/* step through file and extract key/value pairs */
void extractKeys() {
  int pos[2];

  while(!feof(stream)) {
    skip_space_and_comments();
    
    pos[0] = ftell(stream);
    read_key();
#ifdef DEBUG
    printf("DEBUG: found key `%s'\n", key);
#endif
    skip_space_and_comments();

    /* ignore template entry */
    if(lookahead() == '=') {
      expect('=');
      expect('~');
      skip_space_and_comments();
      read_value();
      continue;
    }

    /* process normal entry */
    expect(':');
    expect('=');

    skip_space_and_comments();
    read_value();

    pos[1] = ftell(stream);
    dbAdd(key, pos);

    skip_space_and_comments();
  }
}



/* ---------------------------------------------------------------------- */ 

int main(int argc, char **argv)
{
  String lexFilename; 
  String dbFilename;
  
  if((argc > 3) || (argc == 1)) {
    fprintf(stderr, "syntax: %s <cdg_file> <db_file>\n", argv[0]);
    exit(1);
  }

  lexFilename = argv[1];
  dbFilename = argv[2];

  if (!dbOpen(dbFilename)) {
    fprintf(stderr, "ERROR: cannot open db file '%s'\n", dbFilename);
    exit(1);
  }
  /* open lexicon file (readonly) */
  if ((stream = fopen(lexFilename, "r")) == NULL) {
    fprintf(stderr, "ERROR: cannot open lex file '%s'\n", lexFilename);
    exit(1);
  }

  extractKeys();

  printf("%d valid keys found\n", keysFound);
  fclose(stream);
  dbClose();
  return 0;
  
}
