#!/usr/bin/env perl
# Copyright (C) 1997-2006 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
# 
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# normalise the numbers in the negra export format
# the constituents are numbered left to right and NOT depth first
#
# used to compare outputs which each other
#
# author: Klaus Dalinghaus <kdaling@nats.informatik.uni-hamburg.de>
# initial revision: 2006-04-05

use Getopt::Long;
use strict;

my $debug;
my $silent;
my $quiet;
my $schiehlen;
my $punct;
my $punctzero;
my $help;

GetOptions(
	"debug"      => \$debug,
	"silent"     => \$silent,
	"quiet"      => \$quiet,
	"schiehlen"  => \$schiehlen,
	"punct"      => \$punct,
	"punct-zero" => \$punctzero,
	"help"       => \$help);

my $infile = shift;
my $outfile = shift;
my $error = "";
my $input;
my $output;

if ($infile) {
	open(INPUT, "<$infile") or $error .= "\n\nError: can not open $infile for reading!";
	$input = *INPUT;
} else {
	$input = *STDIN;
}
if ($outfile) {
	open(OUTPUT, ">$outfile") or $error.="\n\nError: can not open $outfile for writing!";
	$output = *OUTPUT;
} else {
	$output = *STDOUT;
}

die "$error

Usage:	$0 [input-file [output-file]]

	input-file	contains negra export format
	output-file	file to write normalised output format

	-help		read this help
	-debug		debug information on STDERR
	-silent		suppress additional information (comments) in output
	-quiet		suppress the output of the sentence number on STDERR
	-schiehlen      can read output from Michael Schiehlens p2sys.pl
	-punct		normalise punctuation in output (make "." out of it)
	-punct-zero	normalise punctuation in output (set parent to 0)

" if ($help || $error);



sub printDebug {
	my $string = shift;
	print STDERR "DEBUG: $string\n" if $debug;
}



sub processSentence {
	my $nodes = shift;

	printDebug("Process Sentence");
	printDebug("renumbering nodes:");
	my $round = 1;
	my @sortednodes;
	foreach my $node (sort { $a->{id} <=> $b->{id} } values %{$nodes}) {
		if (defined $node->{word}) {
			$node->{used} = $round;
			push @sortednodes, $node;
			printDebug("found word \"$node->{word}\", id \"$node->{id}\"");
		}
	}

	my $newInnerMaxNodes = 500;
	
	while (1) {
		my @nextnodes = ();
		$round++;
		foreach my $node (@sortednodes) {
			if ($node->{used} < $round) {
				my $parentNode = $node->{parent};
				if ($parentNode) {
					if (defined $parentNode->{used}) { 
						next;
					}
					my $full = 1;
					my $childstring = "";
					foreach my $child (@{$parentNode->{children}}) {
						if (defined $child->{word}) {
							$childstring .= " " . $child->{word};
						} else {
							$childstring .= " " . $child->{nodeTag};
						}
						if (! defined $child->{used} || $child->{used} == $round) {
							$full = 0;
							last;
						}
					}
					if ($full == 1) {
						my $oldId = $parentNode->{id};
						$parentNode->{id} = $newInnerMaxNodes++;
						$parentNode->{used} = $round;
						push @nextnodes, $parentNode;
						printDebug("node $oldId ($childstring ) renamed to $parentNode->{id}");
					} else {
						push @nextnodes, $node;
					}
				}
			}
		}

		last unless @nextnodes;
		@sortednodes = @nextnodes;
	}

	printDebug("printing the sentence\n\n");
	foreach my $node (sort { $a->{id} <=> $b->{id} } values %{$nodes}) {
		my $word = $node->{word};
		my $nodeTag = $node->{nodeTag};
		my $edgeTag = $node->{edgeTag};
		my $morphTag = $node->{morphTag};
		my $parentNode = $node->{parent};
		my $parentId;
		if ($parentNode) {
			$parentId = $parentNode->{id};
		} else {
			$parentId = "0";
		}
		my $comment = $node->{comment};

		if ($nodeTag =~ /^\$/) {
			if ($punct) {
				$word = ".";
				$nodeTag = "\$.";
			}
			if ($punctzero) {
				$parentId = "0";
			}
		}
		
		if (defined $word) {
			printf $output "%-23s ", $word;
		} else {
			printf $output "#%-22d ", $node->{id};
		}
		printf $output "%-7s %-15s %-7s %-7s", $nodeTag, $morphTag, $edgeTag, $parentId;
		print $output $comment unless ($silent);
		print $output "\n";
	}

	return 1;
}


###########
# main loop
###########

my $insentence = 0;
my $linenr = 0;
my %nodes;
my $id;
my $sent_nr;
my $sent_nr2;
my $sent_nr3;
my $sent_nr4;
my $line_pattern = q{^([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)(.*)$};
if ($schiehlen) {
	$line_pattern = q{^([^\s]+)\t([^\s]+)(\t)([^\s]*)\t([^\s]+)(.*)$};
}
while (my $line = <$input>) {
	$linenr++;
	chomp $line;
	if ($line =~ /^\s*%%/ || $line =~ /^\s*$/) { # handle comment
		if ($insentence) {
			# TODO store the comment in an intelligent way
			print STDERR "TODO: store comment within sentence.\n";
		} else {
			print $output "$line\n" unless ($silent);
		}
	} elsif ($line =~ /^#BOS\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s*$/) { # start a sentence
		if ($insentence) {
			die "Error: Syntax error in input: Starting sentence without ending" .
				" the former one in line $linenr!\n\n";
		}
		$insentence = 1;
		$id = 0;
		$sent_nr = $1;
		$sent_nr2 = $2;
		$sent_nr3 = $3;
		$sent_nr4 = $4;
		unless ($quiet) {
			print STDERR "sentence $1  \r";
		}
		printDebug("");
		printDebug("found sentence no. $sent_nr");
	} elsif ($line =~ /^#EOS\s+([0-9]+)\s*$/) { # end a sentence
		if ($1 != $sent_nr) {
			die "Error: Syntax error in input: Sentence ending nr does not coincide" .
				" with sentence starting line, line $linenr!\n\n";
		}
		$insentence = 0;
		printDebug("found sentence end for no. $sent_nr");
		printDebug("generating parent relationships");
		foreach my $node (sort { $a->{id} <=> $b->{id} } values %nodes) {
			if ($node->{parent} != 0) {
				my $p = $nodes{$node->{parent}};
				$node->{parent} = $p;
				push @{$p->{children}}, $node;
			}
		}
		my $other = "";
		$other = " $sent_nr2 $sent_nr3 $sent_nr4" unless ($silent);
		print $output "#BOS $sent_nr$other\n";
		processSentence(\%nodes);
		print $output "#EOS $sent_nr\n";
		undef %nodes;
	} elsif ($line =~ /$line_pattern/ ) { # process line
		my $p = $5;
		if ($p != 0) {
			$p += 200;
		}
		my $nTag = $2;
		my $mTag = $3;
		my $eTag = $4;
		if ($schiehlen) {
			$mTag = "--";
			$nTag =~ s/([^\$])[\.\-%_+].*$/$1/g;
			$nTag =~ s/\*T[19]\*\)/\$\*LRB\*/g;
			$nTag =~ s/\*RRB\*/\*LRB\*/g;
			$nTag =~ s/^ADVNP$/ADV/g;
			$nTag =~ s/^AKTENZ$/XY/g;
			$nTag =~ s/^CRD$/CARD/g;
			$nTag =~ s/^CTIME$/CARD/g;
			$nTag =~ s/^DATE$/ADJA/g;
			$nTag =~ s/^YEAR$/CARD/g;
			$nTag =~ s/^MONEY$/CARD/g;
			$nTag =~ s/^NNE$/NE/g;
			$nTag =~ s/^PLZ$/CARD/g;
			$nTag =~ s/^SCORE$/CARD/g;
			$nTag =~ s/^TEL$/CARD/g;
			$nTag =~ s/^V[HSW]FIN$/VAFIN/g;
			$nTag =~ s/^V[HSW]INF$/VAINF/g;
			$nTag =~ s/^V[HSW]IMP$/VAIMP/g;
			$nTag =~ s/^V[HSW]PP$/VAPP/g;
			if ($nTag =~ /^\$/ && $eTag =~ /^\s*$/) {
				$eTag = "ROOT";
			}
			$eTag =~ s/\+$//g;
			$eTag =~ s/Neut\.Nom\.Sg/NK/g;
			$eTag =~ s/%m=/HD/g;
			$eTag =~ s/^$/--/g;
		}
		my $node = {
			"nodeTag" => $nTag,
			"morphTag"=> $mTag,
			"edgeTag" => $eTag,
			"parent"  => $p,
			"comment" => $6,
		};
		my $first = $1;
		if ($first =~ /^#/) {
			my $nodeid = $first;
			$nodeid =~ s/^#//;
			$node->{id} = $nodeid + 200;
		} else {
			$node->{word} = $1;
			$node->{id} = $id++;
		}
		$nodes{$node->{id}} = $node;
	} else {
		die "Error: Syntax error in input in line $linenr!\n\n";
	}
}

close($input);
close($output);

exit 0;
