# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllLexemes - manage the CDG lexicon.
##
##
## \author Michael Daum
## ----------------------------------------------------------------------------
class AllLexemes {
  inherit DataBrowser

  
  # public methods
  public method init_data {}
  public method getAllWords {}
  public method getCData {id}; ## \type TclString
  public method refreshrow {row lexeme}; ## \type TclNumber, LexiconItem

  constructor {args} {}; ## \type TclList

  # private methods
  private method _init_data {}
  private method displaybutton_action {}
  private method editbutton_action {}


  # private variables
  private variable lexemes; ## \type TclArray
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllLexemes::constructor {args} {
    set _idColumnIndex 0
    set table $itk_component(table)
    $table configure -cols 4

    itk_component add displaybutton {
	button $itk_component(childsite).display\
	    -anchor w \
	    -text "Display" \
	    -command [code $this displaybutton_action]
    } { }

    itk_component add editbutton {
	button $itk_component(childsite).edit\
	    -anchor w \
	    -text "Edit" \
	    -command [code $this editbutton_action]
    } { }

    # packing
    pack \
	$itk_component(displaybutton) \
	$itk_component(editbutton) \
	-side left -padx 3 

    # get the initial data
    _init_data
    
    # register the helpmessages
    .cdgmain help sethelpstr \
	$itk_component(table) \
	"Click selects row, Ctrl-Click selects more rows, NW cell selects all" \
        $itk_component(displaybutton) "Print out this lexeme in the shell" \
        $itk_component(editbutton) "Call editor for this lexeme" 

    eval itk_initialize $args
}



## ----------------------------------------------------------------------------
## call _init_data if necessary
## ----------------------------------------------------------------------------
body AllLexemes::init_data {} {
  global inputCurrentGrammar

  if {[InputStruct_xcdglexiconToDate_get $inputCurrentGrammar]} {
    return
  }

  .cdgmain busy compute "Init Lexemes" [code $this _init_data]
}

## ----------------------------------------------------------------------------
## get data from the cdg tool
## ----------------------------------------------------------------------------
body AllLexemes::_init_data {} {
  global inputCurrentGrammar

  set table $itk_component(table)
  $table configure -state normal
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table tag config default

  $table tag col ids 0
  $table tag configure ids -anchor w
  $table tag raise ids

  $table tag config colored -bg gray80 
  $table setCell row 0,0 { "" "file" "form" }
  $table width 0 25 1 15 2 10 3 10
    
  # prepare
  foreach id [array names lexemes] {
      unset lexemes($id)
  }

  # import data to Tcl
  set total 0
  set lexicon [InputStruct_lexicon_get $inputCurrentGrammar]
  set lexiconIterator [hashIteratorNew $lexicon]
  
  while {[set forms [hashIteratorNextValue $lexiconIterator]] != "NULL"} {
    for {set l $forms} { $l != "NULL" } { set l [ListStruct_next_get $l ] } {
      set entry [ListStruct_item_get $l]
      set id [LexiconItemStruct_description_get $entry]
      set lexemes($id) $entry
      incr total
    }
  }

  _setCount $total
  set r 1
  foreach id [lsort [array names lexemes]] {
    set lexeme $lexemes($id)
    refreshrow $r $lexeme
    incr r
    if {[expr $r % 1000] == 0} {
      .cdgmain help showstr "Init lexemes: $id"
      update idletask
    }
  }
  
  setSelection
  InputStruct_xcdglexiconToDate_set $inputCurrentGrammar 1

  $table configure -state disabled
}


## ----------------------------------------------------------------------------
## Wrapper to allow use of base class definition of refreshid
## ----------------------------------------------------------------------------
body AllLexemes::getCData {id} {
    return [getlexeme $id]
}


## ----------------------------------------------------------------------------
## fill a row with the values from a specific lexeme.
## ----------------------------------------------------------------------------
body AllLexemes::refreshrow {row lexeme} {
  set table $itk_component(table)
  # lexeme name
  set id [fixencoding [LexiconItemStruct_description_get $lexeme]]

    # base name of defining file  
    regexp "\[^/\]*$" [LexiconItemStruct_filename_get $lexeme] basename
  set basename [fixencoding $basename]
 
    # form
  set form [fixencoding [LexiconItemStruct_word_get $lexeme]]

    $itk_component(table) setCell row $row,0 \
      [list $id $basename $form ]

}


## ----------------------------------------------------------------------------
## display the lexeme in the shell
## ----------------------------------------------------------------------------
body AllLexemes::displaybutton_action {} {
    foreach id $_selection { commandEval "lexicon $id" }
}

## ----------------------------------------------------------------------------
## call editor for the declaration.
## ----------------------------------------------------------------------------
body AllLexemes::editbutton_action {} {

  # avoid flickering
  if {$_selection == ""} return
  
  # build list of filenames
  set target ""
  foreach id $_selection {

    # I suppose most editors grok "+<linenum> file" nowadays...
    set lexeme $lexemes($id)
    set linenum [LexiconItemStruct_lineNo_get $lexeme]
    set filename [LexiconItemStruct_filename_get $lexeme]
    append target "+$linenum $filename "
  }

  # build command line
  regsub -all "%f" "/bin/sh -c \"[.cdgmain cget -editor]\"" $target command

  # execute
  if {[catch {eval exec $command >/dev/null &} errMsg]} {
    ::cmd::Puts "ERROR: $errMsg"
  }

}


## ----------------------------------------------------------------------------
## Return list of all known words
## ----------------------------------------------------------------------------
body AllLexemes::getAllWords {} {
  init_data
  return [array names lexemes]
}
