# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------
## Namespace cmd
##
## this namespace defines all commands which are allowed to be evaluated
## in the cdg-script-language
##
## \todo purge this interface with the help of \c commandEval
## \todo return values of the XCDG commands are a mess. See cmd::Verify()
## and cmd::Constraint().
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## ----------------------------------------------------------------------
namespace eval ::cmd {

## ----------------------------------------------------------------------
## try to find out on what net the command will operate.
## ----------------------------------------------------------------------
proc deduceNetName {cmdline} {; ## \type TclString
  global cnMostRecentlyCreatedNet

  foreach s [split $cmdline] {

    # detect named parameters of some commands
    regsub "^net=(.*)$" $s "\1" s

    # find constraint net by that name
    set net [cnFindNet $s]
    if {$net != "NULL"} {
      return $s
    }
  }

  # assume that the command uses the most recent net
  set net $cnMostRecentlyCreatedNet
  if {$net != "NULL"} {
    return [ConstraintNetStruct_id_get $net]
  }

  return ""

}

## ----------------------------------------------------------------------
## quits the cdg-tool
## ----------------------------------------------------------------------
proc Quit {} {

  set result 1

  if {[.cdgmain cget -confirmExit]} {
    set result [.cdgmain question "Quit: Are you sure?"]
  }

  if {$result == 1} {
    wm withdraw .
    exit
  }
}

## ----------------------------------------------------------------------
## loads one or more file 
## ----------------------------------------------------------------------
proc Load {args} {; ## \type TclList

   eval .cdgmain files load $args

   return
}

## ----------------------------------------------------------------------
## loads one or more file 
## ----------------------------------------------------------------------
proc Source {file} {; ## \type TclString

   .cdgmain shell safeSource $file

   return
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Constraint {args} {; ## \type TclList
   commandEval "constraint $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Level {args} {; ## \type TclList
   commandEval "level $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Showlevel {args} {; ## \type TclList
   commandEval "showlevel $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Uselevel {args} {; ## \type TclList
   commandEval "uselevel $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Useconstraint {args} {; ## \type TclList
   commandEval "useconstraint $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Lexicon {args} {; ## \type TclList
   commandEval "lexicon $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Wordgraph {args} {; ## \type TclList
   
   commandEval "wordgraph $args"

   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Annotation {args} {; ## \type TclList
   
   commandEval "annotation $args"

   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core and add the new parse to AllParses.
## ----------------------------------------------------------------------
proc Anno2Parse {args} {; ## \type TclList
  global parseMostRecentlyCreatedParse

  commandEval "anno2parse $args"
  set p $parseMostRecentlyCreatedParse
  if {$p != "NULL"} {
   .cdgmain parses addParse $p
  }
  return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Hierarchy {args} {; ## \type TclList
   commandEval "hierarchy $args"
   return
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Section {} {
   commandEval "section"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Deactivate {args} {; ## \type TclList
   commandEval "deactivate $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Activate {args} {; ## \type TclList
   commandEval "activate $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Net {args} {; ## \type TclList
   commandEval "net $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## Additionally this function displays a busy box and adds the 
## resulting parse to AllParses
## ----------------------------------------------------------------------
proc Netsearch {args} {; ## \type TclList

  set netId [deduceNetName [join $args]]
  if {$netId == ""} {
    Printf "WARNING: netsearch called with no available netId"
    return
  }

   .cdgmain busy compute "Computing: netsearch $netId" \
      [code commandEval "netsearch $args"]


   set result [.cdgmain busy compute "Computing: parses for $netId" \
		  [code .cdgmain parses addParsesOfNet $netId]]

   .cdgmain networks refreshid $netId
   eval .cdgmain parses setSelection $result

   return $result
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc IC {args} {; ## \type TclList

   if {$args == {}} {
     Printf "WARNING: IC called with no utterance!"
     return 0
   }

   commandEval "incrementalCompletion $args"
   return 1
}

## ----------------------------------------------------------------------
## delegate the command to the C core and refresh AllNetworks.
## ----------------------------------------------------------------------
proc Netdelete {args} {; ## \type TclList
   commandEval "netdelete $args"
   .cdgmain networks init_data
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Edges {netId {startId ""} {stopId ""}} {; ## \type TclString, TclString, TclString
   commandEval "edges [concat $netId $startId $stopId]"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Writeannotation {args} {; ## \type TclList
   commandEval "writeannotation $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Writenet {args} {; ## \type TclList
   commandEval "writenet $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Status {} {
   commandEval "status"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Set {variable  {value ""}} {; ## \type TclString, TclString
  global _getValue

  set result ""
  switch $variable {
    editor {
      if {$value != ""} {
	.cdgmain configure -editor $value
      }
      set result [.cdgmain cget -editor]
    }
    grammarpath {
      if {$value != ""} {
	.cdgmain configure -grammarpath $value
      }
      set result [.cdgmain cget -grammarpath]
    }
    manywindows {
      if {$value != ""} {
	.cdgmain parses configure -manyWindows $value
      }
      set result [.cdgmain parses cget -manyWindows]
    }
    autoshow {
      if {$value != ""} {
	.cdgmain parses configure -autoshow $value
      }
      set result [.cdgmain parses cget -autoshow]
    }
    colonInTreeNames {
      if {$value != ""} {
	.cdgmain configure -colonInTreeNames $value
      }
      set result [.cdgmain cget -colonInTreeNames]
    }
    trace {
      if {$value != ""} {
	.cdgmain configure -showTrace $value
      }
      set result [.cdgmain cget -showTrace]
    }
    confirmexit {
      if {$value != ""} {
	.cdgmain configure -confirmExit $value
      }
      set result [.cdgmain cget -confirmExit]
    }
    default {
      if {$value != ""} {
	commandEval "set $variable '$value'"
      } 
      set oldHook [getHookCmd printf]
      setHookCmd printf [code _get $oldHook]
      commandEval "set $variable"
      set result $_getValue

      # save the variable in namespace CdgOption
      CdgOption::setOption $variable $result
    }
  }
  .cdgmain menu init_data
  return $result
}

## ----------------------------------------------------------------------
## callback installed by Set().
## This function captures the returnvalue of a CDG set command.
# ----------------------------------------------------------------------
proc _get {oldHook value} {; ## \type TclCommand, TclString
  global _getValue
  regsub {^[a-zA-Z]* = } $value "" _getValue
  regsub {[\n\r]+} $_getValue "" _getValue
  setHookCmd printf $oldHook
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## In addition this function refreshes AllNetworks.
## ----------------------------------------------------------------------
proc Newnet {{wordGraphId ""}} {; ## \type TclString
  global inputMostRecentlyCreatedLattice cnMostRecentlyCreatedNet

  if {$wordGraphId == ""} {
    set lattice $inputMostRecentlyCreatedLattice
    if {$lattice == "NULL"} {
      error "no wordGraphId specified"
    } else {
      set wordGraphId [LatticeStruct_id_get $lattice]
    }
  }
   
   set cnMostRecentlyCreatedNet NULL
   .cdgmain busy compute "Computing: newnet $wordGraphId" \
      [code commandEval "newnet $wordGraphId"]

   set net $cnMostRecentlyCreatedNet
   if {$net != "NULL" } {
      set netId [ConstraintNetStruct_id_get $net]
      .cdgmain networks refreshid $netId
      .cdgmain networks setSelection $netId
      return $netId
   } else {
      return ""
   }

}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Distance {netId } {; ## \type TclString
   commandEval "distance $netId"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## In addition this function refreshes AllWordgraphs.
## ----------------------------------------------------------------------
proc Inputwordgraph {word args} {; ## \type TclString, TclList
   global inputMostRecentlyCreatedLattice
   commandEval "inputwordgraph [concat $word [join $args]]"

   set lattice $inputMostRecentlyCreatedLattice
   if {$lattice != "NULL"} {
      set lattice_id [LatticeStruct_id_get $lattice]
      .cdgmain wordgraphs refreshid $lattice_id
#      .cdgmain files init_data
   } else {
      set lattice_id ""
   }
   
   # update the wordgraphs
   .cdgmain wordgraphs init_data

   return $lattice_id
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Parses2Prolog {args} {; ## \type TclList
   commandEval "parses2prolog $args"
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Annos2Prolog {args} {; ## \type TclList
   commandEval "annos2prolog $args"
   return
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## Additionally this function updates AllNetworks.
## ----------------------------------------------------------------------
proc Isearch {args} {; ## \type TclList
   global cnMostRecentlyCreatedNet
   set wgId [lindex $args 0]
   .cdgmain busy compute "Computing: isearch $wgId" \
      [code commandEval "iSearch $args"]

   set net $cnMostRecentlyCreatedNet
   if {$net != "NULL" } {
      set netId [ConstraintNetStruct_id_get $net]
      .cdgmain networks refreshid $netId
      .cdgmain networks setSelection $netId
   } else {
      set netId ""
   }

   return $netId
}


## ----------------------------------------------------------------------
## intercept the tcl command puts to print to the Shell.
## ----------------------------------------------------------------------
proc Puts {arg1 {arg2 ""}} {; ## \type TclString, TclString
   switch -- $arg1 {
      -nonewline {
	 set text $arg2
	 set newline ""
      }
      -newline {
	 set text $arg2
	 set newline "\n"
      }
      -* {
	 error "Unknown option `$mode', should be -nonewline, newline"
      }
      default {
	 if {$arg2 != ""} {
	    error "wrong #arguments to puts"
	 }
	 set text $arg1
	 set newline "\n"
      }
   }
   
   if {$text != "" || $newline != ""} {
      .cdgmain info insertMessage [fixencoding $text$newline]
      # puts $text$newline
   }
   return
}

## ----------------------------------------------------------------------
## Imitate the C command printf in tcl.
## ----------------------------------------------------------------------
proc Printf {format args} {; ## \type TclString, TclList
   set text [format $format $args]
   if {$text != "" } {
      .cdgmain shell insert $text
      .cdgmain shell setCursor end
   }
   return
}

## ----------------------------------------------------------------------
## Read a file of X resources.
## ----------------------------------------------------------------------
proc Options {filename} {; ## \type TclString

  if {[catch { option readfile $filename interactive } errMsg]} {
    puts "error in reading option file: "
    puts $errMsg
    exit 1
  }
  return
}


## ----------------------------------------------------------------------
## implement the shell command ls.
## ----------------------------------------------------------------------
proc Ls {args} {; ## \type TclList
   if {$args == ""} {
      set args "-xAF"
   }
   set allArgs ""
   set options ""


   foreach arg $args {
      if {[regexp -- "-.*" $arg]} {
	 lappend allArgs $arg
      } else {
	 lappend allArgs [glob -nocomplain -- $arg]
     }
      
   }
   set result [ \
      if {"$allArgs" == ""} {
	 exec /bin/ls 
      } else {
	 exec /bin/sh -c "/bin/ls $allArgs"
      }]
   Puts $result
   return
}

## ----------------------------------------------------------------------
## clear the shell window.
## ----------------------------------------------------------------------
proc Clear {args} {; ## \type TclList
   eval .cdgmain shell clear $args
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Hook {args} {; ## \type TclList
   commandEval "hook $args"
   return

}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Levelsort {args} {; ## \type TclList

   commandEval "levelsort $args"

   return
}

## ----------------------------------------------------------------------
## enable AllParses::showparse() in the xcdg shell.
## ----------------------------------------------------------------------
proc Showparse {args} {; ## \type TclList
   eval .cdgmain parses showparse $args
}

## ----------------------------------------------------------------------
## enable AllParses::deleteparse() in the xcdg shell.
## ----------------------------------------------------------------------
proc Deleteparse {args} {; ## \type TclList
   if {$args == {} } {
      error "missing argument `parseId'"
      return
   }
   eval .cdgmain parses deleteparse $args
   return
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Nonspeccompatible {args} {; ## \type TclList
   commandEval "nonspeccompatible $args"
   return
}

## ----------------------------------------------------------------------
## implement the extraordinary feature of directory changing.
## ----------------------------------------------------------------------
proc Cd {args} {; ## \type TclList
   eval cd $args
}

## ----------------------------------------------------------------------
## question a core reset of the system.
## ----------------------------------------------------------------------
proc Reset {args} {; ## \type TclList

  if {"$args" == "force"} {
    set result 1 
  } else {
   set result [.cdgmain question "Reset: Are you sure?\n All data will be lost!!!"]
  }
   if {$result == 1} {
     commandEval "reset"
     .cdgmain files init_data
     .cdgmain networks init_data
     .cdgmain wordgraphs init_data
     .cdgmain constraints init_data
     .cdgmain levels init_data
     .cdgmain lexemes init_data
     .cdgmain hierarchies init_data
     .cdgmain parses init_data
   }

}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## Additionally, we get a busy box and update AllParses and AllNetworks.
## ----------------------------------------------------------------------
proc Frobbing {args} {; ## \type TclList
   
  set cmdline [join $args]
  set netId [deduceNetName $cmdline]
    
  # should we allow user input during frobbing?

  # If the user specified `manual', then yes.
  # If the user gave an execute= keyword, then yes (the user
  # is reponsible for adding `q' to the program).
  # Otherwise, no.
  set interactive [regexp "execute=|manual" $cmdline]
  if {!$interactive} {
    lappend args "execute=q"
  }

  # If no input is allowed, pop up a busy box.
  if {$interactive} {
    commandEval "frobbing $args"
  } else {
    .cdgmain busy compute "Computing: frobbing $netId" commandEval "frobbing $args"
  }

  if {$netId != ""} {
    .cdgmain networks refreshid $netId
    .cdgmain parses setSelection [.cdgmain parses addParsesOfNet $netId]
  }
  return

}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## Additionally, we get a busy box and update AllParses and AllNetworks.
## ----------------------------------------------------------------------
proc Gls {args} {; ## \type TclList

  set netId [deduceNetName [join $args]]
  if {$netId == ""} {
     Printf "WARNING: gls called with no available netId"
     return
  }

  .cdgmain busy compute "Computing: gls on $netId" [code commandEval "gls $args"]

  set result [.cdgmain busy compute "Computing: parses for $netId" \
       	  [code .cdgmain parses addParsesOfNet $netId]]

  .cdgmain networks refreshid $netId
  eval .cdgmain parses setSelection $result

  return $result
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## Additionally, we get a busy box and update AllParses and AllNetworks.
## ----------------------------------------------------------------------
proc NewGls {args} {; ## \type TclList

  set netId [deduceNetName [join $args]]
  if {$netId == ""} {
     Printf "WARNING: gls called with no available netId"
     return
  }

  .cdgmain busy compute "Computing: new gls on $netId" [code commandEval "ngls $args"]

  set result [.cdgmain busy compute "Computing: parses for $netId" \
       	  [code .cdgmain parses addParsesOfNet $netId]]

  .cdgmain networks refreshid $netId
  eval .cdgmain parses setSelection $result

  return $result
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Verify {args} {; ## \type TclList
 commandEval "verify $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Compile {args} {; ## \type TclList
   commandEval "compile $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc IncrementalCompletion {args} {; ## \type TclList
   commandEval "incrementalCompletion $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Incremental {args} {; ## \type TclList
   .cdgmain busy compute "Computing: incremental" commandEval "incremental $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Beamsearch {args} {; ## \type TclList
   .cdgmain busy compute "Computing: beamsearch" commandEval "beamsearch $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc ContIS {args} {; ## \type TclList
   commandEval "contIS $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc PrintIS {args} {; ## \type TclList
   commandEval "printIS $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc PrintParse {args} {; ## \type TclList
   commandEval "printparse $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc PrintParses {args} {; ## \type TclList
   commandEval "printparses $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Renewnet {args} {; ## \type TclList
   commandEval "renewnet $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Version {args} {; ## \type TclList
   commandEval "version $args"
}


## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc WriteParses {args} {; ## \type TclList
   commandEval "writeparses $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc WriteWordgraph {args} {; ## \type TclList
   commandEval "writewordgraph $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc License {args} {; ## \type TclList
   commandEval "license $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Compareparses {args} {; ## \type TclList
   commandEval "compareparses $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Help {args} {; ## \type TclList
   commandEval "help $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Tagger {args} {; ## \type TclList
   commandEval "tagger $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Testing {args} {; ## \type TclList
   commandEval "testing $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Weight {args} {; ## \type TclList
  commandEval "weight $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Chunk {args} {; ## \type TclList
  commandEval "chunk $args"
}

## ----------------------------------------------------------------------
## delegate the command to the C core.
## ----------------------------------------------------------------------
proc Shift {args} {; ## \type TclList

  set netId [deduceNetName [join $args]]
  if {$netId == ""} {
    Printf "WARNING: shift-reduce called with no available net"
    return
  }

  commandEval "shift-reduce $args"

  .cdgmain networks refreshid $netId
  .cdgmain parses setSelection [.cdgmain parses addParsesOfNet $netId]

}

## ----------------------------------------------------------------------
## delegate command to the C core.
## ----------------------------------------------------------------------
proc Predict {args} {; ## \type TclList
   commandEval "predict $args"
   return
}



}
