# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaAnnotation - bundled information about an annotation.
## This class represents an annotation in a YadaGrammar. Scanning a YadaGrammar
## (see YadaGrammar::scan()) results in bunch of YadaAnnotation objects being 
## constructed.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaAnnotation {

  # variables ----------------------------------------------------------------

  ## name of the annotation
  public variable name ""; ## \type TclString

  ## name of the wordgraph bein annotated 
  public variable wordgraphName ""; ## \type TclString

  ## filename which this wordgraph has been extracted from
  public variable file ""; ## \type TclString

  ## linenumber in the file where this wordgraph has been defined
  public variable lineNo 0; ## \type TclNumber

  ## number of words
  public variable noWords 0; ## \type TclNumber

  ## number of annotated dependency edges
  public variable noEdges 0; ## \type TclNumber

  ## the grammar object we belong to
  private variable _grammar ""; ## \type YadaGrammar

  ## annotated structures: keys are from,to
  private variable _words; ## \type TclArray
  private variable _deps; ## \type TclArray 
  private variable _tags; ## \type TclArray 

  # methods ------------------------------------------------------------------

  # TODO access methods to the annotation data: right now we only consume
  # the noWords and noEdges properties

  constructor {grammar args} {}; ## \type YadaGrammar, TclList
  
  public method clone {args}; ## \type TclList
  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {annoNode}; ## \type domNode
}

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaAnnotation::constructor {grammar args} {
  set _grammar $grammar
  eval configure $args
}

## ----------------------------------------------------------------------------
## serializes this wordgraph into a dom document
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaAnnotation::toDom {} {
  set document [dom createDocument "annotation"]
  set rootNode [$document documentElement]

  # set the annotation's attributes
  $rootNode setAttribute "id" $name 
  $rootNode setAttribute "lattice" $wordgraphName 
  $rootNode setAttribute "nowords" $noWords 
  $rootNode setAttribute "noedges" $noEdges 

  # store the file info
  set fileNode [$document createElement "file"]
  $rootNode appendChild $fileNode
  $fileNode setAttribute "name" $file
  $fileNode setAttribute "lineNo" $lineNo

  # store the annotations
  foreach index [array names _words] {
    scan $index "%d,%d" from to

    # store the arc
    set arcNode [$document createElement "arc"]
    $rootNode appendChild $arcNode
    $arcNode setAttribute "from" $from
    $arcNode setAttribute "to" $to
    $arcNode setAttribute "word" $_words($from,$to)
      
    # store the tags
    if {[info exists _tags($from,$to)]} {
      foreach {attr value} $_tags($from,$to) {
	set tagNode [$document createElement "tag"]
	$arcNode appendChild $tagNode
	$tagNode setAttribute "$attr" "$value"
      }
    }

    # store the deps
    if {[info exists _deps($from,$to)]} {
      foreach {level label modifiee} $_deps($from,$to) {
	set depNode [$document createElement "dep"]
	$arcNode appendChild $depNode
	$depNode setAttribute "level" $level 
	$depNode setAttribute "label" $label
	$depNode setAttribute "modifiee" $modifiee
      }
    }
  }

  return $document
}

## ----------------------------------------------------------------------------
## initialize anno with dom document
## note, that only the first annotation found is used, i.e. this function is 
## pointless ;)
## @param document containing a /yada/grammar/annotation
## ----------------------------------------------------------------------------
body YadaAnnotation::fromDom {document} {
  set rootNode [$document documentElement]
  set annoNode [$rootNode selectNodes {/yada/grammar/annotation[1]}]
  fromDomNode $annoNode
}

## ----------------------------------------------------------------------------
## initialize annotation with dom node 
## @param rootNode root DOM-Node (Tag <annotation>)
## ----------------------------------------------------------------------------
body YadaAnnotation::fromDomNode {annoNode} {

  if {[$annoNode nodeName] != "annotation"} {
    error "ERROR: '$latticeNode' is not an annotation node"
  }

  set name [$annoNode getAttribute "id"]
  set noWords [$annoNode getAttribute "nowords"] 
  set noEdges [$annoNode getAttribute "noedges"] 
  set wordgraphName [$annoNode getAttribute "lattice"]
  
  set file [[$annoNode selectNodes "file"] @name]
  set lineNo [[$annoNode selectNodes "file"] @lineNo]

  # read all arcs
  foreach arcNode [$annoNode selectNodes "arc"] {
    set from [$arcNode getAttribute "from"]
    set to [$arcNode getAttribute "to"]
    set _words($from,$to) [$arcNode getAttribute "word"]

    # read all deps for this arc  
    set _deps($from,$to) {}
    foreach depNode [$arcNode selectNodes "dep"] {
      lappend _deps($from,$to) [list \
	[$depNode getAttribute "level"] \
	[$depNode getAttribute "label"] \
	[$depNode getAttribute "modifiee"] \
      ]
    }

    # read tags for this arc 
    set _tags($from,$to) {}
    foreach tagNode [$arcNode selectNodes "tag"] {
      foreach attr [$tagNode attributes] {
	set value [$tagNode getAttribute $attr]
	lappend _tags($from,$to) $attr $value
      }
    }
  }
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaAnnotation::clone {args} {

  # build the clone
  set newAnno [YadaAnnotation ::#auto $_grammar \
    -name $name \
    -wordgraphName $wordgraphName \
    -file $file \
    -lineNo $lineNo \
    -noWords $noWords \
    -noEdges $noEdges]

  # TODO

  eval $newAnno configure $args

  return $newAnno
}
