# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaDocument - embeddeable document.
## This class extends the YadaPlainDocument and offers all functionality to
## embed a document into the YadaMain widget. In parallel to this inheritance
## the YadaConfigDocument is not embeddeable into YadaMain but into YadaConfig, i.e.
## YadaGammars, YadaExperiments etc. 
## \ingroup YadaDocuments
##
## \author Michael Daum 
##
## ----------------------------------------------------------------------------
class YadaDocument {
  inherit ::itk::Widget YadaPlainDocument

  # variables ----------------------------------------------------------------

  ## link to the previous YadaDocument registered to YadaMain
  public variable prevDocument ; ## \type YadaDocument

  ## link to the next YadaDocument registered to YadaMain
  public variable nextDocument ; ## \type YadaDocument

  ## flag indicating whether the toolbar of this document is shown.
  ## This information is kept in sync with YadaMain::_toolbarFlag whenever
  ## the document is activated. \see setToggle(), activationHandle().
  private variable _toolbarFlag 1

  ## flag indicating whether the status message of this document is shown.
  ## This information is kept in sync with YadaMain::_statusFlag whenever
  ## the document is activated. \see setToggle(), activationHandle().
  private variable _statusFlag 1


  # methods ------------------------------------------------------------------
  public method backward {}
  public method forward {}
  public method setToggle {name {value ""}} ;## \virtual \type TclString, TclRef
  public method activationHandle {}; ## \virtual
  public proc newDocument {className name}; ## \type TclString, TclString
  public method updateMenus {} ; ## \virtual

  constructor {} {}
  destructor {}

  protected method askNewLoad {}

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaDocument::constructor {} {

  #
  # file menu
  #
  set fileButton [.main component fileButton]
  for {
    set i 1
    set fileMenu $fileButton.menu$i
  } {[winfo exists $fileMenu]} {incr i} {
    set fileMenu $fileButton.menu$i
  }
  itk_component add fileMenu {
     menu $fileMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  itk_component add newMenu {
     menu $itk_component(fileMenu).newMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  $itk_component(fileMenu) add command \
       	-label "Load" \
       	-underline 0 \
       	-command [code $this load] 
  $itk_component(fileMenu) add command \
       	-label "Save" \
       	-underline 0 \
       	-command [code $this save] 
  $itk_component(fileMenu) add cascade \
	-label "New" \
	-underline 0 \
	-menu $itk_component(newMenu)
  $itk_component(fileMenu) add separator
  $itk_component(fileMenu) add command \
	-label "Close" \
	-underline 0 \
	-command [code itcl::delete object $this]
  $itk_component(fileMenu) add command \
	-label "Quit" \
	-underline 0 \
	-command {.main quit}

  $itk_component(newMenu) add command \
	-label "Runner" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRunner Runner}
  $itk_component(newMenu) add command \
	-label "Difference" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaDifference Difference}
  $itk_component(newMenu) add command \
	-label "Gls Statistics" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaGlsStats GlsStatistics}
  $itk_component(newMenu) add command \
	-label "Label Statistics" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaLabelStats LabelStatistics}
  $itk_component(newMenu) add command \
	-label "Ranking" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRanking Ranking}
#  $itk_component(newMenu) add command \
#	-label "Best per Wordgraph" \
#	-underline 0 \
#	-state disabled
#  $itk_component(newMenu) add command \
#	-label "File Viewer" \
#	-underline 0 \
#	-state disabled

  #
  # edit menu
  #
  set editButton [.main component editButton]
  for {
    set i 1
    set editMenu [subst $editButton.menu$i]
  } {[winfo exists $editMenu]} {incr i} {
    set editMenu [subst $editButton.menu$i]
  }
  
  itk_component add editMenu {
     menu $editMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  #
  # view menu
  #
  set viewButton [.main component viewButton]
  for {
    set i 1
    set viewMenu [subst $viewButton.menu$i]
  } {[winfo exists $viewMenu]} {incr i} {
    set viewMenu [subst $viewButton.menu$i]
  }

  itk_component add viewMenu {
     menu $viewMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  $itk_component(viewMenu) add checkbutton \
       	-label "Toolbar" \
       	-variable [scope _toolbarFlag] \
       	-command [code $this setToggle "toolbar"]
  $itk_component(viewMenu) add checkbutton \
       	-label "Status" \
       	-variable [scope _statusFlag] \
       	-command [code $this setToggle "status"]

  #
  # toolbar
  #
  set toolbarArea [.main component toolbarArea]
  itk_component add toolbar {
    iwidgets::toolbar $toolbarArea#auto \
      -balloonfont {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
      -balloonbackground lightyellow
  } {}

  $itk_component(toolbar) add FlatButton backButton \
	-image [YadaImages::get backward] \
	-command [code $this backward] \
	-state disabled

  $itk_component(toolbar) add FlatButton forwardButton \
	-image [YadaImages::get forward] \
	-command [code $this forward] \
	-state disabled

  $itk_component(toolbar) add FlatButton loadButton \
	-image [YadaImages::get openFile] \
	-command [code $this load] \
	-balloonstr "load"

  $itk_component(toolbar) add FlatButton saveButton \
	-image [YadaImages::get floppy] \
	-command [code $this save] \
	-balloonstr "save"

  # the childsite
  itk_component add childsite {
    frame $itk_interior.childsite
  }
  
  # packing
  pack $itk_component(childsite) -fill both -expand 1 -side top -pady 2
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaDocument::destructor {} {
  ## unregister only, if document has been registered before
  if {$hasView} {
    .main unregisterDocument $this
  }
}

## ----------------------------------------------------------------------------
## backward
## ----------------------------------------------------------------------------
body YadaDocument::backward {} {
  .main activateDocument $prevDocument
}

## ----------------------------------------------------------------------------
## forward
## ----------------------------------------------------------------------------
body YadaDocument::forward {} {
  .main activateDocument $nextDocument
}

## ----------------------------------------------------------------------------
## callback for prevDocument.
## ----------------------------------------------------------------------------
configbody YadaDocument::prevDocument {
  if {$prevDocument != ""} {
    $itk_component(toolbar) itemconfigure backButton\
      -state normal \
      -balloonstr "back to [$prevDocument cget -name]"
  } else {
    $itk_component(toolbar) itemconfigure backButton\
      -state disabled \
      -balloonstr "back"
  }
}

## ----------------------------------------------------------------------------
## callback for nextDocument.
## ----------------------------------------------------------------------------
configbody YadaDocument::nextDocument {
  if {$nextDocument != ""} {
    $itk_component(toolbar) itemconfigure forwardButton\
      -state normal \
      -balloonstr "forward to [$nextDocument cget -name]"
  } else {
    $itk_component(toolbar) itemconfigure forwardButton\
      -state disabled \
      -balloonstr "forward"
  }
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaDocument::activationHandle {} {
  chain
  set _toolbarFlag [.main getToggle "toolbar"]
  set _statusFlag [.main getToggle "status"]

  set fileButton [.main component fileButton]
  set editButton [.main component editButton]
  set viewButton [.main component viewButton]
  $fileButton configure -menu $itk_component(fileMenu)
  $editButton configure -menu $itk_component(editMenu)
  $viewButton configure -menu $itk_component(viewMenu)
}

## ----------------------------------------------------------------------------
## toggle toolbar or status.
## This method switches the toolbar or the status bar on or off depending on
## its arguments.
## \param name the element to be toggled: toolbar or status
## \param value the value to which to switch to: 1 or 0
## ----------------------------------------------------------------------------
body YadaDocument::setToggle {name {value ""}} {
  switch $name {
    "toolbar" {
      set _toolbarFlag [.main setToggle "toolbar" $value]
      return $_toolbarFlag
    }
    "status" {
      set _statusFlag [.main setToggle "status" $value]
      return $_statusFlag
    }
    default {
      return [chain $name $value]
    }
  }
}

## ----------------------------------------------------------------------------
## creates a new document which becomes a component of main.
## ----------------------------------------------------------------------------
proc YadaDocument::newDocument {className name} {
  set document [.main newDocument $className $name]
  $document init
  .main activateDocument $document
  return $document
}


## ----------------------------------------------------------------------------
## Pops up a yes/no message dialog and returns the result  
## (i.e. which button has been pressed)
## ----------------------------------------------------------------------------
body YadaDocument::askNewLoad {} {
  set msg [iwidgets::messagedialog .msg \
    -text "File '[getZippedFileName]'\nhas been modified on disk. Do you want to reload?" \
    -title "Yada - Caution"  \
    -image [YadaImages::get question] \
    -modality application \
    -master $itk_interior]
  $msg buttonconfigure OK -text "Yes"
  $msg buttonconfigure Cancel -text "No"
  $msg center .
  set result [$msg activate]
  destroy $msg
  return $result
}


## ----------------------------------------------------------------------------
## Update toolbar, menubar
## ----------------------------------------------------------------------------
body YadaDocument::updateMenus {} {
  if {!$hasView || ![isVisible]} {
    return
  }

  # set the state of the "Load" action
  set loadButton [$itk_component(toolbar) component loadButton]
  if { [hasSavedState] } {
    $itk_component(fileMenu) entryconfigure 0 -state normal
    $loadButton configure -state normal
  } else {
    $itk_component(fileMenu) entryconfigure 0 -state disabled
    $loadButton configure -state disabled
  }
  
  # set the state of the "Save" action
  set saveButton [$itk_component(toolbar) component saveButton]
  if {[isModified] && [string compare $name "<none>"] != 0} {
    $itk_component(fileMenu) entryconfigure 1 -state normal   
    $saveButton configure -state normal
  } else {
    $itk_component(fileMenu) entryconfigure 1 -state disabled
    $saveButton configure -state disabled
  }
}

