# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaJob - bundled information about a yada job.
## A YadaJob is being created with the help of a YadaExperiment and organizes the
## computation of a wordgraph on a YadaMachine.
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaJob {
#  inherit CdgProcess

  ## the name of the YadaExperiment for which we created the job.
  public variable experiment ""

  ## the wordgraph which we are going to start YadaJob::experiment on
  public variable wordgraphName ""

  ## the YadaRunnableDocument that is responsible for this job.
  ## especialy all messages that want to be printed are displayed on this
  ## document.
  public variable runner ""

  ## the YadaMachine on which we run this job. 
  public variable machine ""

  ## the YadaNode on which this actually is executed
  public variable node ""

  ## the cdgp-binary to use for this Job
  ## it is set in the corresponding experiment
  public variable cdgp ""
  
  ## the file where to store the xml output of the logfile
  public variable xmlFile	""

  ## the file where to store the logfile
  public variable logFile	""

  ## state of the job.
  ## A YadaJob passes a series of states from creation time of this
  ## object til its destruction. Possible values are:
  ## - created
  ## - dequeuing
  ## - reloading
  ## - starting
  ## - running
  ## - consuming
  ## - stopping
  private variable _state "created"

  ## list of passed states. \see _state
  private variable _passedStates ""

  # methods ------------------------------------------------------------------
  public method getScript {}
  public method setState {state}; ## \type TclString
  public method getState {}
  public method getPassedStates {}
  public method hasPassedState {state}; ## \type TclString
  public method getLogFileName {}
  public method getXmlFileName {}
  public method stop {}
  public method interrupt {}
  public method term {}
  public method kill {}

  constructor {args} {}; ## \type TclList
  destructor {}
}

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaJob::constructor {args} {

  eval configure $args

  if {$runner == ""} {
    error "ERROR: no runner specified"
  }

  if {$experiment == ""} {
    error "ERROR: no experiment specified"
  }

  if {$wordgraphName == ""} {
    error "ERROR: no wordgraph specified"
  }

  # set the information
  set printCommand [code $runner print]
  set cdgp [$experiment cget -command]
  set sync 1
  set initFile "none"

  getLogFileName
  getXmlFileName
}

## ----------------------------------------------------------------------------
## stop this process and call the consumer
## ----------------------------------------------------------------------------
body YadaJob::stop {} {
  # call the consumer
  if {[catch {$runner consumeJob $this} errMsg]} {
    puts "ERROR: $errMsg\n$::errorInfo\n"
  }
}

## ----------------------------------------------------------------------------
## setStatus
## ----------------------------------------------------------------------------
body YadaJob::setState {state} {
  lappend _passedStates $_state
  set _state $state
}

## ----------------------------------------------------------------------------
## getState
## ----------------------------------------------------------------------------
body YadaJob::getState {} {
  return $_state
}

## ----------------------------------------------------------------------------
## getPassedStates
## ----------------------------------------------------------------------------
body YadaJob::getPassedStates {} {
  return $_passedStates
}

## ----------------------------------------------------------------------------
## hasPassedState
## ----------------------------------------------------------------------------
body YadaJob::hasPassedState {state} {
  return [expr [lsearch $_passedStates $state] >= 0]
}

## ----------------------------------------------------------------------------
## return the logfile which this job writes to.
## ----------------------------------------------------------------------------
body YadaJob::getLogFileName {} {
  if {$logFile == ""} {
    set logFile [$experiment getLogFileName $wordgraphName]
  }
  return $logFile
}

## ----------------------------------------------------------------------------
## return the xmlfile which this job writes to.
## ----------------------------------------------------------------------------
body YadaJob::getXmlFileName {} {
  if {$xmlFile == ""} {
    set xmlFile [$experiment getXmlFileName $wordgraphName]
  }
  return $xmlFile
}

## ----------------------------------------------------------------------------
## construct the script to run
## ----------------------------------------------------------------------------
body YadaJob::getScript {} {

  if {$runner != ""} {
    $runner markWordgraph $wordgraphName ;# unmarked in the consumer
  }
  
  # construct the script to run
  set grammarName [$experiment cget -grammarName]
  set script [$experiment cget -script]
  set dataDir [$experiment cget -dataDir]
  set grammar [.main getGrammar $grammarName]
  set tmpScript $script
  foreach file [$grammar cget -files] {
    regsub -all -- "%load" $tmpScript "load $file\n%load" tmpScript
  }
  regsub -all -- "%load" $tmpScript "" tmpScript
  regsub -all -- "%wordgraph" $tmpScript $wordgraphName tmpScript
  regsub -all -- "%path" $tmpScript $dataDir tmpScript
  
  return $tmpScript
}

## ----------------------------------------------------------------------------
## interrupt the job
## ----------------------------------------------------------------------------
body YadaJob::interrupt {} {
  $node killJob $this 2
}

## ----------------------------------------------------------------------------
## interrupt the job
## ----------------------------------------------------------------------------
body YadaJob::term {} {
  $node killJob $this 15
}

## ----------------------------------------------------------------------------
## interrupt the job
## ----------------------------------------------------------------------------
body YadaJob::kill {} {
  $node killJob $this 9
}
