#!/bin/sh
#\
exec tclsh "$0" "$@" 

package require Itcl
namespace import -force itcl::*

# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaNode - bundled information about a node of a YadaMachine
##
## A YadaNode is an object representing a host on which jobs can be processed.
## Jobs are scheduled by a YadaMachine object this node belongs to.
##
## \ingroup YadaConfiguration
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaNode {

  # variables ----------------------------------------------------------------

  ## known host name or IP
  public variable host "<undef>"; ## \type TclString

  ## number of concurrent jobs that will be started on this node.
  public variable cpus 1; ## \type Integer

  ## access manor for this node.
  ## possible values are "local" or any way of getting a shell on the nodes, i.e.
  ## \code ssh -x %node \endcode
  public variable access "<undef>"; ## \type TclString

  ## flag indicating whether we shall use this node or not
  public variable isActive 1

  ## flag indicating whether the node is connected to a cdg server
  public variable connection 0

  ## every node belongs to a YadaMachine
  private variable _machine ""

  ## list of YadaJob objects currently running on this node
  private variable _jobs {}; ## \type TclList

  ## the cdg-server to which the node is connected
  private variable server "inital_value"

  ## the cdgp-server to use
  private variable cdgserver "$env(YADA_SVR)"

  ## callback variable to inform about lost connections
  public variable connectionCallback  ""

  ## callback variable to inform about deactivation of the node
  public variable deactivationCallback  ""

  ## number of the current job to compute
  ## this number is used as a name for the job in the server
  private variable id 0
  
  # methods ------------------------------------------------------------------
  constructor {machine args} {}; ## \type YadaMachine, TclList
  destructor {}

  public method connect {}
  public method startJob {job}; ## \type YadaJob
  public method computeJob {job}
  public method stopJob {job}; ## \type YadaJob
  public method getJobs {}
  public method getLoad {}
  public method clone {}
  public method connectionCheck {}
  public method stopEvent {server}
  public method printEvent {text}
  public method changeConnection {n}
  public method killJob {job signo}
}

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaNode::constructor {machine args} {
  set _machine $machine

  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor removes this node from its machine
## ----------------------------------------------------------------------------
body YadaNode::destructor {} {
  foreach job $_jobs {
    catch {itcl::delete object $job}
  }
  $_machine removeNode $this
}

## ----------------------------------------------------------------------------
## construct a copy of this node
## ----------------------------------------------------------------------------
body YadaNode::clone {} {
  return [$_machine newNode \
    -host $host \
    -access $access \
    -cpus $cpus \
    -isActive $isActive]
}

## ----------------------------------------------------------------------------
## connect to a cdgserver
## ----------------------------------------------------------------------------
body YadaNode::connect {} {

  set pid ""
  if {$connection == 1} {
    return
  }

  if {$access == "local"} {
    set command $cdgserver
  } else {
    set command "$access $host $cdgserver"
  }

writeDebug "command=$command"
  
  # create and start a process for a cdg server over ssh
  set server [Process #auto \
    -showProgress 1 \
    -sync 1 \
    -stopCallback [code $this stopEvent]  \
    -printCommand [code $this printEvent] \
  ]
  set pid [$server start $command]
  if {$pid != 0} {
    changeConnection 1
  }

  after 5000 [code $this connectionCheck]
  
  return $pid
}

## ----------------------------------------------------------------------------
## start a job on this node
## ----------------------------------------------------------------------------
body YadaNode::startJob {job} {
  lappend _jobs $job
  incr id

  if {$connection == 0 && $isActive == 1} {
    connect
  }

  if {$connection == 1} {
    $job setState "starting"
    $job configure -node $this
    set cdgp [$job cget -cdgp]
    set xmlFile [$job getXmlFileName]
    set logFile [$job getLogFileName]

    set runner [$job cget -runner]
    set wordgraphName [$job cget -wordgraphName]
    if {$runner != ""} {
      $runner markWordgraph $wordgraphName ;# unmarked in the consumer
    }
    
    $job setState "running" 
    $server tell "NEW $job $xmlFile $logFile $cdgp"
writeDebug "NEW $job $xmlFile $logFile $cdgp"
  } else {    
    # requeue job into responsible machine
writeDebug "job $job reset to machine in node::connect"
    $_machine queueJob $job
    $_machine start
  }
}


## ----------------------------------------------------------------------------
## compute started job
## ----------------------------------------------------------------------------
body YadaNode::computeJob {job} {
  # send the script of the job to the cdgp-server
  set script [$job getScript]
  regsub -lineanchor -all -- {^} $script "TELL $job " script
  writeDebug "SCRIPT\n$script\nTPIRCS"
  $job setState "consuming" 
  $server tell $script
  $server tell "TELL $job quit"
}

## ----------------------------------------------------------------------------
## stop a job on this node
## ----------------------------------------------------------------------------
body YadaNode::stopJob {job} {

  # remove from job queue
  set index [lsearch -exact $_jobs $job]
  if {$index >= 0} {
    set _jobs [lreplace $_jobs $index $index]
    catch {itcl::delete object $job}
  }
}

## ----------------------------------------------------------------------------
## get names of all running jobs
## ----------------------------------------------------------------------------
body YadaNode::getJobs {} {
  return $_jobs
}

## ----------------------------------------------------------------------------
## get number of jobs on this node
## ----------------------------------------------------------------------------
body YadaNode::getLoad {} {
  return [llength $_jobs]
}

## ----------------------------------------------------------------------------
## check whether connection really was succssful
## if not, deactivate Node
## ----------------------------------------------------------------------------
body YadaNode::connectionCheck {} {
writeDebug connectionCheck
  if {!$connection} {
    set isActive 0
    changeConnection -1
  }
}

## ----------------------------------------------------------------------------
## listen to server whether a job has to be stopped
## ----------------------------------------------------------------------------
body YadaNode::printEvent {text} {
writeDebug $text
  # listen whether a job has to be stopped
  set text [split $text "\n"]
  foreach line $text {
    if { [regexp -- {^finished ([^\s]*)$} $line dummy jobToStop] } {
writeDebug "jobToStop: '$jobToStop'"
      $jobToStop stop
    } elseif { 
      [regexp -line -- {^Started job (.+)\.$} $line dummy job] } {
      computeJob $job
    }
  }
}

## ----------------------------------------------------------------------------
## Event if server stops to exist
## ----------------------------------------------------------------------------
body YadaNode::stopEvent {server} {
  set connection 0
  if {$connectionCallback != ""} {
    $connectionCallback $this
  }
}

## ----------------------------------------------------------------------------
## set connection of the node
## ----------------------------------------------------------------------------
body YadaNode::changeConnection {n} {
  set oldConnection $connection
  set connection $n

  # requeue the jobs if this node disconnects
  if {$connection <= 0} {
writeDebug "jobs is '$_jobs'"
    foreach job $_jobs {
writeDebug "job $job reset to machine"
      #$job setState "dequeuing"; # TODO: which state?
      $_machine queueJob $job
      $_machine start
    }
  }

writeDebug "yadaNode:: connectionCallback is '$connectionCallback'"
  if {$connectionCallback != ""} {
    $connectionCallback $this
  }
}

## ----------------------------------------------------------------------------
## kill a job with signal $signo
## ----------------------------------------------------------------------------
body YadaNode::killJob {job signo} {
  $server tell "KILL $job $signo"
}

